
#include <stdlib.h>
#include <mpi.h>
#include "az_aztec.h" 
#include <stdio.h>
#include <math.h> 
/* Common Block Declarations */

union {
    struct {
	int nx;
    } _1;
    struct {
	int n;
    } _2;
} global_;

#define global_1 (global_._1)
#define global_2 (global_._2)

/* Global variables */

static double *buf_r_xm, *buf_s_xm, *buf_s_ym, *buf_s_zm,
                  *buf_s_xp, *buf_s_yp, *buf_s_zp, *buf_r_xp, 
		  *buf_r_ym, *buf_r_yp, *buf_r_zm, *buf_r_zp;

int ikcg; /* iteration number */
double u_max ; /* max of u */

/* *************************************************************** */
/*     PROGRAM  CGMN/CARP-CG */
/* --------------------------------------------------------------- */
/*      Implementation of CGMN/CARP-CG with setup of examples 20,21,23 - 32.   

        Authors: Dan Gordon, Dept. of Computer Science, University of Haifa,   
                 Email: gordon@cs.haifa.ac.il   
                 Rachel Gordon, Dept. of Aerospace Engineering,   
                 The Technion-Israel Insitute of Technology,   
                 Email: rgordon@tx.technion.ac.il   
        Last update: April 5, 2012                                          */
/* -------------------------------------------------------------------------*/

/* Main program */ int main(int argc, char **argv)
/* Main program */ /*int MAIN(void) */
{

    double d1, d2;
    int ipr;
    static int  i_scheme, icase;
    static double mn, a2, b2, c2, gama;
    static double l_a, m_a, n_a;
    static int example_no, read_update, kcg, order, model_no, icg;
    static double freq, kn1, kn2, kn3;   
    static double c_min, c_max;
    static double xmin, ymin, xmax, ymax, zmin, zmax;
    static float  *x, *y, *z, *cxy, *ccxy;
    float  x1, y1, z1, c1;
    static double t21, t1, t2;
    static int i, j, k, ij, i1, i2, i3, n, n0, jk, ik, is, ny, nz, i_start;
    static int n_xyz, kt;
    static int ierr, iend, ntype, dim, nend, iinc, ii, jj;
    static int *index_s_xp, *index_s_xm, *index_s_yp, *index_s_ym,
           *index_s_zp, *index_s_zm, 
	   *index_r_xp, *index_r_xm, *index_r_yp, *index_r_ym,
	   *index_r_zp, *index_r_zm;
    static int nproc, me, me_new, comm_cart;
    static int n_internal, n_external, n_update_internal, n_update_external, 
               avg_nonzero_per_row, total_nz;
    int *bindx, *update_inv, *iem, *iema;
    int *data_org , *external, *update_index, *extern_index, *update;      
    static double *val, *err, *rhsm, *xnew, *ssum;
    static double *pnewl_loc, *qnewl_loc, *snewl_loc, *xnewl_loc, 
           *xnewl_loc_init, *rhsm_loc;

    static int  rank_dest_x, rank_dest_y, direction_x, direction_y, 
	    direction_z, rank_dest_z;
    static int coords[3], dims[3], disp, ntypes, kmax; 
    static int reorder, periods[3];
    static int ixp, ixm, iyp, iym, izp, izm;
    static double alpha, beta, sum, sum1, sum2, s2, s2p, s2s, pq, pqs, s2ps, 
           res0, rxmax, rmax, residmax, sumr; 
    static double sumrt, sumerr, sumerr1, sumerr2, err_l2;
    
    static double cons, omeg1;   

    static int nrow, nrow_half, nxst, nyst, nzst, nxend, nyend, nzend;
    static double *qnewl, *xnewl, *ecoef, *resid, *resid1;
    static int nlocal, nx_end, ny_end, nz_end, nxloc, nyloc, nzloc, nx_st, ny_st, nz_st;
    static int rank_source_x, rank_source_y, rank_source_z; 
    static int istart, nxproc, nyproc, nzproc;
    static int n_border, n_update, n_update_border, n_rec_length, n_neigh, n_id_neigh;
    static int n_total_send, n_send, n_send_length, n_send_list;
    int proc_config[AZ_PROC_SIZE];

    extern /* Subroutine */ int transf_2d(double *, double *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *), 
            transf_3d(double *, 
	    double *, int *, int *, int *, int *, int 
	    *, int *, int *, int *, int *, int *, int 
	    *, int *, int *, int *, int *, int *, int 
	    *, int *, int *, int *, int *, int *, int 
	    *, int *, int *, int *, int *, int *, int 
	    *);
    extern /* Subroutine */ int example_1d(int *, int *, int *, 
	    int *, int *, int *, int *, int *, int *), 
	    example_2d(int *, int *, int *, int 
	    *, int *, int *, int *, int *, int *, int 
	    *, int *, int *, int *, int *, int *, int 
	    *, int *, int *), 
            example_3d(int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *),
            example_3d_mod(int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *),
	    bound_1d(int *, int *, int *, int *, int *, 
	    int *), 
            bound_2d(int *, int *, int *, int *,
	    int *, int *, int *, int *, int *, int *),
	    bound_3d(int *, int *, int *, int *, int *
	    , int *, int *, int *, int *, int *, int *
	    , int *, int *, int *);
    extern /* Subroutine */ int example_2d_img_all(int *, int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *),
            example_2d_all(int *, int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *), 
            example_2d_img(int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *),
	    example_3d_img(int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *),
	    example_3d_img_all(int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *, int *, 
	    int *, int *, int *, int *, int *,
            int *, int *, int *, int *, int *, int*, int*);

    extern /* Subroutine */ int example20_err(double *, double *, int *, int *,
	    double *,int *), 
            example21_err(double *, double *, int *, int *,
	    double *),
            example23_err(double *, double *, int *, int *, 
	    double *, int *), 
            example24_err( int *),
            example25_err( int *), 
	    example26_err(double *, double *, int *, int *, 
	    double *, double *, double *, double *, 
	    double *, int *, int *), 
            example27_err(double *, double *, int *, int *, 
	    double *, double *, double *, double *, 
            double *, int *, int *), 
	    example28_err(double *, double *, int *, int *, 
	    double *, double *, int *, int *),
	    example29_err(double *, double *, int *, int *, int *, 
	    double *, double *, double *, double *, double *, int *, int *),
            example30_err(int *),
            example31_err_case_a(double *, double *, int *, int *,
            double *, double *, int *, int *),
            example31_err_case_b(double *, double *, int *, int *,
            double *, double *, int *, int *, double *),
            example31_err_case_c(double *, double *, int *, int *,
            double *, double *, double *, double *, int *, int *),
            example32_err_case_a(double *, double *, int *, int *,  int *,
            double *, double *, double *, int *, int *),
            example32_err_case_c(double *, double *, int *, int *,  int *,
            double *, double *, double *, double *, double *, int *, int *);

    extern /* Subroutine */ int create_matrix_row_5pt_example20(int *, int *,
	    double *, int *, int *, double *, int *, 
	    int *, int *, int *, double *), 
            create_matrix_row_5pt_example21_2rows(
	    int *, int *, double *, int *, int *, 
	    double *, int *, int *, int *, int *, 
	    double *, double *, double *),
            create_matrix_row_9pt_example23_2rows(
	    int *, int *, double *, int *, int *, 
	    double *, int *, int *, int *, int *, 
	    double *, double *, double *, int *),
            create_matrix_row_5_9_pt_example24_2rows(
            int *, int *, double *, int *, int *, double *,
            int *, int *, int *, int *, float *,
            double *, double *, double *, double *,
            int *, int *, int *, double *),
            create_matrix_row_7_27_pt_example25_2rows(
	    int *, int *, double *, int *, int *, 
	    double *, int *, int *, int *, int *, int 
	    *, int *, double *, double *, double *, 
	    double *, double *, double *, double *, 
	    double *, double *, int *, int *, int *, int *),

            create_matrix_row_9pt_example26_2rows(int *, int *,
	    double *, int *, int *, double *, int *, 
	    int *, int *, int *, double *, double *, 
	    double *, int *), 
            create_matrix_row_9pt_example27_2rows(
	    int *, int *, double *, int *, int *, 
	    double *, int *, int *, int *, int *, 
	    double *, double *, double *, int *),
            create_matrix_row_9pt_example28(int *, int *,
	    double *, int *, int *, double *, int *, 
	    int *, int *, int *, double *, int *), 
            create_matrix_row_7_27_pt_example29_2rows(
	    int *, int *, double *, int *, int *, 
	    double *, int *, int *, int *, int *, int 
	    *, int *, double *, double *, double *, int *), 
            create_matrix_row_7_27_pt_example30_2rows(
	    int *, int *, double *, int *, int *, 
	    double *, int *, int *, int *, int *, int 
	    *, int *, float *, double *, double *, 
	    double *, double *, double *, double *, 
	    int *, int *, int *, int *, double *, int *),

            create_matrix_row_9pt_example31_case_a(int *, int *,
            double *, int *, int *, double *, int *,
            int *, int *, int *, double *, int *, int *),
            create_matrix_row_9pt_example31_case_b(int *, int *,
            double *, int *, int *, double *, int *,
            int *, int *, int *, double *, int *, int *,
            double *),
            create_matrix_row_9pt_example31_case_c(int *, int *,
            double *, int *, int *, double *, int *,
            int *, int *, int *, double *, double *,  double *, int *, int *),
            create_matrix_row_7_27_pt_example32_case_a(int *, int *,
            double *, int *, int *, double *, int *, int *, int *,
            int *, int *, int *, double *, double *, int *, int *),
            create_matrix_row_7_27_pt_example32_case_c(int *, int *,
            double *, int *, int *, double *, int *, int *, int *,
            int *, int *, int *, double *, double *, double *, double *, int *, int *);

#define TRUE_ (1)
#define FALSE_ (0)
    FILE *f4;
    FILE *f7;
    FILE *f9;
    FILE *fp;

    MPI_Init ( &argc, &argv );
    MPI_Comm_size(MPI_COMM_WORLD, &nproc );
    MPI_Comm_rank(MPI_COMM_WORLD, &me );
    f7 = fopen("input.dat","r");
    if(f7==0) {printf("\n error opening input.dat"); exit;} 
    f9 = fopen("error.out1","w");
    if(f9==0) {printf("\n error opening error.out"); exit;}

/*   read input data */

    fscanf(f7, " %d %d %d ", &example_no, &global_1.nx, &read_update);
    fscanf(f7, " %d %d %d %d %d ", &nxproc, &nyproc, &nzproc, &kcg, &ierr);
    fscanf(f7, "%lf  %lf ", &residmax, &omeg1); 

    if (example_no == 20) {
        fscanf(f7, "%lf", &kn1);
    }
    if (example_no == 21) {
        fscanf(f7, "%lf  %lf %lf", &kn1, &kn2, &kn3);
    }
    if (example_no == 23) {
        fscanf(f7, "%lf  %lf %lf %d", &kn1, &kn2, &kn3, &order);
    }
    if (example_no == 26 || example_no == 27) {
        fscanf(f7, " %lf  %lf %lf %d", &kn1, &kn2, &kn3, &order);
    }
    if (example_no == 28 ) {
        fscanf(f7, "%lf  %d", &kn1, &order);
    }
    if (example_no == 29) {
        fscanf(f7,"%lf %lf %lf %d", &kn1, &kn2, &kn3, &order);
    }
    if (example_no == 31 ) {
       fscanf(f7, "%d", &icase);
       if(icase == 1)
       fscanf(f7, "%lf  %d  %d ", &mn, &order, &i_scheme);
       
       if(icase == 2) 
       fscanf(f7, "%lf  %d  %d %lf", &mn, &order, &i_scheme, &a2);   
       
       if(icase == 3) 
       fscanf(f7, "%lf   %lf  %lf  %d  %d", &a2, &b2, &c2, &order, &i_scheme );   
    }

    if (example_no == 32 ) {
       fscanf(f7, "%d", &icase);
       if(icase == 1)
       fscanf(f7, "%lf  %lf  %d  %d ", &l_a, &n_a, &order, &i_scheme);
       
       if(icase == 2) 
       fscanf(f7, "%lf  %d  %d %lf", &mn, &order, &i_scheme, &a2);   
       
       if(icase == 3) 
       fscanf(f7, "%lf  %lf  %lf  %lf %d  %d", &a2, &b2, &c2, &gama, &order, &i_scheme);   
    }
/*  Description of the input data: */
/* ------------------------------------------------------------------------- */
/*     example_no - problem number   
       NX - the number of grid points in the x direction.   
            we assume NX=NY=NZ for all test cases except examples_no=24,25,30   
       read_udate - Parameter that defines how the equations are distributed   
                    among the processors:   
                  = 1 in case that  1*1*NZproc topology of processors   
                    is used in a 3D case or 1*NYproc*1 in a 2D case.   
                  = 3 in case that NXproc*NYproc*NZproc processors   
                    topology is used in a 3D case, or NXproc*NYproc*1   
                    in a 2D case  where NXproc,NYproc,NZproc >= 1. In   
                    this case the indices of the equations of each   
                    processor are read from an external file.   
      NXproc,NYproc,NZproc - Number of sub-domains in X,Y,Z direction respectively.   
                             Each sub-domain is allocated to a different processor.   
                           (the total number of processors Np=NXproc*NYproc*NZproc)   
      KCG - the number of CARP-CG iterations.   
      ResidMax - the L2 norm of the residual required.   
      OMEG1 -  the relaxation parameter:  0 < OMEG1 < 2.   
      Kn1,Kn2,Kn3 - values of K for Helmholtz equation with three layers.   
      ierr = 0 - don't calculate the error during the iterative process
                 for examples having an analytic solution.     
           = 1 - calculate the error during the iterative process.    
*/

/* Print out the input data */

    if (me == 0) {
        printf("\n %s ","---------------------------------------------------------------");
        printf("\n    example_no= %d \n", example_no);
        printf("\n    NX= %d   ResidMax= %e   OMEG1= %e ", global_1.nx, residmax, omeg1);
        printf("\n    read_update= %d   No. of CG iterations= %d   ierr= %d \n",
                read_update,kcg,ierr);
        printf("\n    NXproc= %d    NYproc= %d     NZproc= %d",nxproc, nyproc, nzproc);

	if (example_no == 20) {
            printf("\n    Helmholtz equation:  K1= %5.1f", kn1);
	}

	if (example_no == 21) {
            printf("\n    Helmholtz equation:  K1= %5.1f    K2= %5.1f    K3=%5.1f",
               kn1, kn2, kn3);
	}

        if (example_no == 23 || example_no == 26 || example_no == 27) {
           printf("\n    Helmholtz equation:  K1= %5.1f    K2= %5.1f",kn1, kn2);
           printf( "    K3=%5.1f    scheme order= %d", kn3, order);
	}

	if (example_no == 28) {
            printf("\n    Helmholtz equation:  K1= %5.1f    scheme order= %d",
               kn1, order);
	}

/*      example 29 - 3D Helmholtz problem with an analytic solution:  */
        if (example_no == 29) {
            printf("\n  %s","    ");
            printf("\n              3D Helmholtz equation with an analytic solution\n");
            printf("\n    K1= %5.1f    K2= %5.1f    K3=%5.1f scheme order= %d", kn1, kn2, kn3, order);
            printf("\n %s ","--------------------------------------------------------------------");
        } /*  End printing of example 29 input data */

        if (example_no == 31) {
           printf("\n    case = %d ", icase);
           if( icase == 1)
           printf("\n    Helmholtz equation:  m = %10.5f    scheme order= %d, scheme of (Erlangga=1/Tsynkov=2)  %d ",
             mn, order, i_scheme);
           if( icase == 2)
           printf("\n    Helmholtz equation:  m = %10.5f    scheme order= %d, scheme of (Erlangga=1/Tsynkov=2)  %d  alpha=%10.5f",
             mn, order, i_scheme, a2);
           if( icase == 3)
           printf("\n    Helmholtz equation: a= %10.5f  b =%10.5f  c=%10.5f  scheme order= %d, scheme of (Erlangga=1/Tsynkov=2) %d",
             a2, b2, c2, order, i_scheme);
        }

        if (example_no == 32) {
           printf("\n    case = %d ", icase);
           if( icase == 1) {
           m_a = sqrt(l_a*l_a + n_a*n_a);
           printf("\n    Helmholtz equation:  l = %10.5f  m = %10.5f n= %10.5f  scheme order= %d, scheme of (Erlangga=1/Tsynkov=2)  %d ",
             l_a, m_a, n_a, order, i_scheme);
           }
           if( icase == 3)
           printf("\n    Helmholtz equation: a =%10.5f  b =%10.5f   c=%10.5f  gama=%10.5f scheme order= %d, scheme of (Erlangga=1/Tsynkov=2) %d",
             a2, b2, c2, gama, order, i_scheme);
        }
        printf("\n %s ","---------------------------------------------------------------");
    }

    if (nxproc * nyproc * nzproc != nproc) {
	printf("\n  Nproc is incorrect, Nproc= %d", nproc);
        exit (0);
    }

/*  example 24 - Marmousi problem: */
/* -------------------------------- */
    if (example_no == 24) {

        fscanf(f7, "%lf %lf %lf %lf %lf %d %d %d %d", 
        &xmin, &xmax, &ymin, &ymax, &freq, &model_no, &global_1.nx, &ny, &order); 

	nrow = 2*(global_1.nx + 4) * (ny + 4);
        nrow_half= 0.5*nrow;
        x = (float  *) calloc(nrow_half + 1, sizeof(float)); 
        y = (float  *) calloc(nrow_half + 1, sizeof(float)); 
        cxy = (float *) calloc(nrow_half + 1, sizeof(float)); 
        ccxy = (float *) calloc(nrow_half + 1, sizeof(float)); 


	if (me == 0) {
            printf("\n         Marmousi problem \n");
            printf("\n  Xmin= %f   Xmax= %f   Ymin= %f   Ymax= %f",
            xmin, xmax, ymin, ymax); 
            printf("\n  freq= %f  model_no= %d   NX= %d   NY= %d   scheme order=%d\n",
            freq, model_no, global_1.nx, ny, order);
	}
/* endif me=0 */
	if (model_no == 1) {
/*       Marmousi - 1st model problem */
/*       open(4,FILE='/home/people/rgordon/Aztec.dir/marm2D.dir/marm2D/ */
/*    1data/marm_1.dat') */
            f4 = fopen("/shared/home/gordon/CARPCG_c.dir/marm2D/data/marm_1.dat","r");
            if(f4==0) {printf("\n error opening marm_1.dat"); exit;} 
	}
	if (model_no == 2) {
/*       Marmousi - 2nd model problem */
            f4 = fopen("/shared/home/gordon/CARPCG_c.dir/marm2D/data/marm_2.dat","r");
            if(f4==0) {printf("\n error opening marm_2.dat"); exit;} 
	}
	if (model_no == 3) {
/*       Marmousi - 3rd model problem */
            f4 = fopen("/shared/home/gordon/CARPCG_c.dir/marm2D/data/marm_3.dat","r");
            if(f4==0) {printf("\n error opening marm_3.dat"); exit;} 
	}
	if (model_no == 4) {
/*       Marmousi - 4th model problem */
            f4 = fopen("/shared/home/gordon/CARPCG_c.dir/marm2D/data/marm_4.dat","r");
            if(f4==0) {printf("\n error opening marm_4.dat"); exit;} 
	}

/*  Set CXY() for each grid point(X();Y( )) */
/*  note: the value of c(x,y) is not available on the boundary and external grid */
/*        points. It will be taken as that of the adjacent grid point. */
	ij = 0;
	i1 = ny;
	for (j = 1; j <= i1; ++j) {
	    i2 = global_1.nx;
	    for (i = 1; i <= i2; ++i) {
		++ij;
		fscanf(f4," %f %f %f", &x1, &y1, &c1);
		x[ij - 1] = x1;
		y[ij - 1] = y1;
		ccxy[ij - 1]= c1;
	    }
	}
	if (me == 0) {
            printf("\n  No. of grid points of input data for c(x,y) = %d", ij);
	}

/*  The first row (boundary + external grid points): */
        cxy[0] = ccxy[0];
        cxy[1] = ccxy[0];
        is = 1;
	for (ii = 1; ii <= global_1.nx; ++ii) {
            ++is; 
	    cxy[is] = ccxy[ii-1];
	}
        ++is;
        cxy[is] = cxy[is-1];
        ++is;
        cxy[is] = cxy[is-1];

/*  define the second row:  */
        for (ii = 0; ii <= global_1.nx + 3; ++ii) {
            cxy[ii+global_1.nx+4]=cxy[ii];
        }

/*  The next NY rows: */
        is = 2*global_1.nx + 7;
	ij = 0;
	for (j = 1; j <= ny; ++j) {
	    ++is;
	    cxy[is] = ccxy[ij];
	    ++is;
            cxy[is] = cxy[is-1]; 
	    for (ii = 1; ii <= global_1.nx; ++ii) {
		++is;
		++ij;
		cxy[is] = ccxy[ij - 1];
	    }
	    ++is;
	    cxy[is] = cxy[is - 1];
	    ++is;
	    cxy[is] = cxy[is - 1];
	}

/*  the (NY+3)th row (boundary points): */
        ++is;    
        i_start=is;
        ij -= global_1.nx;
        cxy[is] = ccxy[ij];
        ++is;
        cxy[is] = cxy[is - 1];
        i2 = global_1.nx;
        for (ii = 1; ii <= i2; ++ii) {
            ++is; 
            ++ij;
            cxy[is] = ccxy[ij-1];
        }
        ++is;
        cxy[is] = cxy[is - 1];
        ++is;
        cxy[is] = cxy[is - 1];

/*  The last (NY+4)th row (external points): */

        for (ii = 0; ii <= global_1.nx + 3; ++ii) {
            ++is;
            cxy[is] = cxy[i_start + ii];
        }

/*  Calculate C_max and C_min */

	if (me == 0) {
	    c_max = cxy[0];
	    c_min = cxy[0];
	    i1 = (global_1.nx + 4) * (ny + 4) - 1;
	    for (i = 1; i <= i1; ++i) {
		if (cxy[i] > c_max) {
		    c_max = cxy[i];
		}
		if (cxy[i] < c_min) {
		    c_min = cxy[i];
		}
	    }
            printf("\n    C_max= %f    C_min= %f ", c_max, c_min);
            printf("\n %s ","---------------------------------------------------------------");
	} /*  endif me=0 */

    }  /* end of example24 Marmousi problem  definition   */


/*  example 25 - 3D Helmholtz problem:  */
/* ------------------------------------ */
    if (example_no == 25) {
        fscanf(f7,"%lf %lf %lf %lf %lf %lf %lf %lf %lf %d %d %d %d", 
        &xmin, &xmax, &ymin, &ymax, &zmin, &zmax, &kn1, &kn2, &kn3, &global_1.nx, &ny, &nz,
            &order);

        if(me == 0){
        printf("\n  %s","    ");
        printf("\n              3D Helmholtz equation \n");
        printf("\n  Xmin= %f   Xmax= %f   Ymin= %f   Ymax= %f   Zmin= %f   Zmax= %f",
        xmin, xmax, ymin, ymax, zmin, zmax); 
        printf("\n    K1= %5.1f    K2= %5.1f    K3=%5.1f", kn1, kn2, kn3);
        printf("\n    NX= %d   NY= %d    NZ= %d  scheme order= %d \n", global_1.nx, ny, nz, order);
        printf("\n %s ","---------------------------------------------------------------");
       }  /* endif: if me  = 0 printout  */ 
    } /*  End reading data of example 25 */

/*  example 30 - 3D Helmholtz problem for the SEG/EAGE salt model:  */
/* ---------------------------------------------------------------- */
    if (example_no == 30) {
        fscanf(f7,"%lf %lf %lf %lf %lf %lf %lf %d  %d %d %d %d", 
        &xmin, &xmax, &ymin, &ymax, &zmin, &zmax, &freq, &model_no, &global_1.nx, &ny, &nz,
            &order);
/*   allocate storage */
	nrow = 2*(global_1.nx + 4) * (ny + 4) * (nz + 4);
        nrow_half= 0.5*nrow;
        cxy = (float *) calloc(nrow_half + 1, sizeof(float)); 
        ccxy = (float *) calloc(nrow_half + 1, sizeof(float)); 

/* Note: nx,ny,nz are the number of inner points in each direction. */
        if(me == 0){
        printf("\n  %s","    ");
        printf("\n              3D Helmholtz eq. - the SEG/EAGE salt problem\n");
        printf("\n  Xmin= %f   Xmax= %f   Ymin= %f   Ymax= %f   Zmin= %f   Zmax= %f",
        xmin, xmax, ymin, ymax, zmin, zmax); 
        printf("\n    freq.= %5.1f    model No.= %d ", freq, model_no);
        printf("\n    NX= %d   NY= %d    NZ= %d  scheme order= %d \n", global_1.nx, ny, nz, order);
        printf("\n %s ","---------------------------------------------------------------");
       }  /* endif: if me  = 0 printout  */ 

	if (model_no == 1) {
/*  SEG/EAGE for Dx=160; 85x85x27 model problem */
            f4 = fopen("SEG_EAGE_salt.dir/salt_160.dat","r");
            if(f4==0) {printf("\n error opening marm_1.dat"); exit;} 
	}
	if (model_no == 2) {
/*  SEG/EAGE for Dx=80; 169x169x53 model problem */
            f4 = fopen("SEG_EAGE_salt.dir/salt_80.dat","r");
            if(f4==0) {printf("\n error opening marm_2.dat"); exit;} 
	}
	if (model_no == 3) {
/*  SEG/EAGE for Dx=40; 338x338x105 model problem */
            f4 = fopen("SEG_EAGE_salt.dir/salt_40.dat","r");
            if(f4==0) {printf("\n error opening marm_3.dat"); exit;} 
	}
	if (model_no == 4) {
/*  SEG/EAGE for Dx=20; 676x676x210 model problem */
            f4 = fopen("SEG_EAGE_salt.dir/salt_20.dat","r");
            if(f4==0) {printf("\n error opening marm_4.dat"); exit;} 
	}
/*  Set CXY() for each grid point(X();Y( )) */
/*  note: the value of c(x,y) is not available on the boundary and external grid */
/*        points. It will be taken as that of the adjacent grid point. */
	ij = 0;
        i3 = nz + 2;
        for (k = 1; k <= i3; ++k) {
           i1 = ny+2;
           for (j = 1; j <= i1; ++j) {
	       i2 = global_1.nx+2;
	       for (i = 1; i <= i2; ++i) {
	       	   ++ij;
		   fscanf(f4," %f %f %f %f", &x1, &y1, &z1, &c1);
/*		   x[ij - 1] = x1;
		   y[ij - 1] = y1;
		   z[ij - 1] = z1; */
		   ccxy[ij - 1]= c1;
        /*         printf("\n x1= %f  y1= %f  z1= %f  c= %f ", x1, y1, z1, c1); */
	       }
	   }
        }
	if (me == 0) {
            printf("\n  No. of grid points of input data for c(x,y,z) = %d", ij);
	}

        is = -1;
        for (k = 1; k <= nz+4; ++k) { 
           kt =k;
           if(k > 1 && k < nz+4) kt=k-1;
           if(k == nz+4) kt = nz+2;
	   i2 = global_1.nx+2;
     /*    n_xyz = (kt-1) * (nx+2) * (ny+2);  */
           n_xyz = (kt-1) * i2 * (ny+2);
        
/*  define the first row (boundary + external grid points): */
           is = is + 1;
           cxy[is] =  ccxy[n_xyz];
	   for (ii = 1; ii <= global_1.nx+2; ++ii) {
              ++is; 
	      cxy[is] = ccxy[n_xyz+ii-1];
	   } 
           ++is;
           cxy[is] = cxy[is-1];

/*  define the next ny+2 rows:  */
           ij = n_xyz;
           for (j =1; j<= ny+2; ++j) {
              ++is;
              cxy[is] = ccxy[ij];
              for (ii = 1; ii <= global_1.nx + 2; ++ii) {
                 ++is;
                 ++ij;
                 cxy[is] = ccxy[ij-1];
              }
              ++is;
              cxy[is] = cxy[is-1];
           } 
/*  define the (ny+4)th row: */
           i_start = is +1 - (global_1.nx+4);
           for (ii = 0; ii <= global_1.nx+3; ++ii) {
              ++is;
              cxy[is] = cxy[i_start+ii];
           }
        } /* enddo on k */

/*  Calculate C_max and C_min */

	if (me == 0) {
	    c_max = cxy[0];
	    c_min = cxy[0];
	    i1 = (global_1.nx + 4) * (ny + 4)* (nz + 4)  - 1;
	    for (i = 1; i <= i1; ++i) {
		if (cxy[i] > c_max) {
		    c_max = cxy[i];
		}
		if (cxy[i] < c_min) {
		    c_min = cxy[i];
                    ij=i; 
		}
	    }
            printf("\n    C_max= %f    C_min= %f i_min= %d", c_max, c_min, ij);
            printf("\n %s ","---------------------------------------------------------------");
	} /*  endif me=0 */

     free(ccxy);
    } /*  End reading data of example 30 */

/*  Set ntype - parameter which specifies the problem type (real/complex+type of boundary conditions). */

    ntype = 3;
    if (example_no == 10) {
	ntype = 1;
    }
    if (example_no == 9 || example_no == 11) {
	ntype = 2;
    }
    if (example_no == 15 || example_no == 16 || example_no == 20) {
	ntype = 2;
    } 
    if (example_no == 17 || example_no == 28 || example_no == 31) { 
	ntype = 4;
    }
    if (example_no == 21 ) {
	ntype = 5;
    }
    if (example_no == 25 || example_no == 30) {
	ntype = 6;
    }
    if (example_no == 26) {
	ntype = 7;
    }
    if (example_no == 27) {
	ntype = 8;
    }
    if (example_no == 23) {
	ntype = 9;
    }
    if (example_no == 24) {
	ntype = 10;
    }
    if (example_no == 29) {
	ntype = 11;
    }
    if (example_no == 32) {
	ntype = 12;
    }

/* Set dim - the problem dimension (1D/2D/3D) */

    if(ntype == 3 || ntype == 6 || ntype == 11 || ntype == 12) dim = 3;
    if(ntype == 2 || ntype == 4 || ntype == 5 || ntype == 7 || ntype == 8 || ntype == 9 || ntype == 10) dim = 2;
    if(ntype == 1) dim = 1;

    if (example_no == 10 && nxproc > 1) {
	if (me == 0) {
            printf("\n  %s","    ");
            printf("\n  %s","The current version of the code can not solve example 10 on more than 1 processor");
	}
        exit (0) ; 
    }

    if (example_no != 24 && example_no != 25 && example_no != 30) {
	ny = global_1.nx;
	nz = global_1.nx;
    }

    if (ntype == 12) {   /*  problem 32 */
	n =  (global_1.nx + 2) * (ny + 2) * (nz + 2);
    }

    if (ntype == 11) {   /*  problem 29 */
	n = 2 * (global_1.nx + 2) * (ny + 2) * (nz + 3);
    }

    if (ntype == 10) {   /*  problem 24: Marmousi problem */
	n = 2 * (global_1.nx + 4) * (ny + 4); 
    }
    if (ntype == 9) {   /* problem 23 */
	n = 2 * (global_1.nx + 4) * (ny + 3);
    }
    if (ntype == 8) {  /* problem 27 */
	n = 2 * (global_1.nx + 2) * (ny + 3);
    }
    if (ntype == 7) {  /* problem 26 */
	n = 2 * (global_1.nx + 3) * (ny + 2);
    }
    if (ntype == 6) {  /* problem 25 & problem 30 */
	n = 2 * (global_1.nx + 4) * (ny + 4) * (nz + 4);
    }
    if (ntype == 5) {
	n = 2 * (global_1.nx + 2) * (ny + 2);
    }
    if (ntype == 4) {
	n = (global_1.nx + 2) * (ny + 2);
    }
    if (ntype == 3) {
	n = global_1.nx * ny * nz;
    }
    if (ntype == 2) {
	n = global_1.nx * ny;
    }
    if (ntype == 1) {
	n = global_1.nx;
    }

/*     3D problems    */
/* ------------------ */
   if (dim == 3) {
      ntypes = 3;   
      dims[0] = nxproc;
      dims[1] = nyproc;
      dims[2] = nzproc;
      periods[0] = FALSE_;
      periods[1] = FALSE_;
      periods[2] = FALSE_;
      reorder = TRUE_;

/* Create a new communicator 'comm_cart' containing cartesian topology information */

      MPI_Cart_create(MPI_COMM_WORLD,ntypes,dims,periods,reorder,
         &comm_cart);


/* Get my rank in new communicator */

      MPI_Comm_rank(comm_cart, &me_new );

/*    write(NOUT,*) 'my rank in new cartesian communicator:',me_new */

/* Get my cartesian coordinates */

      MPI_Cart_get(comm_cart, ntypes, dims, periods, coords);

/* Get shifted source and destination ranks in each direction   
    for send/receive operation (i.e. get the rank of your neighbors) */

      direction_x = 0;
/* Coordinate dimension of send/rec 1-st dim */
      direction_y = 1;
      direction_z = 2;
      disp = 1;
/* displacement */
      MPI_Cart_shift( comm_cart, direction_x , disp , 
                      &rank_source_x, &rank_dest_x );
      disp = 1;
      MPI_Cart_shift( comm_cart,direction_y ,disp , 
                      &rank_source_y, &rank_dest_y );
      disp = 1;
      MPI_Cart_shift( comm_cart,direction_z ,disp , 
                      &rank_source_z, &rank_dest_z );

      printf("\n  %s","    ");
      printf("\n  me= %d  rank_source_x= %d   rank_dest_x= %d   rank_source_y= %d   rank_dest_y= %d",
               me,rank_source_x,rank_dest_x,rank_source_y,rank_dest_y);
      printf("   rank_source_z= %d   rank_dest_z= %d", rank_source_z,rank_dest_z);

    }

/*     2D problems    */
/* ------------------ */
    if (dim == 2)  {
       ntypes = 2;
       dims[0] = nxproc;
       dims[1] = nyproc;
       periods[0] = FALSE_;
       periods[1] = FALSE_;
       reorder = TRUE_;

/* Create a new communicator 'comm_cart' containing cartesian topology information */

       MPI_Cart_create(MPI_COMM_WORLD,ntypes,dims,periods,reorder,
         &comm_cart); 

/*   Get my rank in new communicator */

       MPI_Comm_rank(comm_cart, &me_new );

/*     write(NOUT,*) 'my rank in new cartesian communicator:',me_new */

/*   Get my cartesian coordinates */

       MPI_Cart_get(comm_cart, ntypes, dims, periods,  coords);  

/*  Get shifted source and destination ranks in each direction         
    for send/receive operation (i.e. get the rank of your neighbors) */

       direction_x = 0;
/* coordinate dimension of send/rec 1-st dim */
       direction_y = 1;
       disp = 1;
/* displacement */
       MPI_Cart_shift( comm_cart, direction_x ,disp ,   
                      &rank_source_x, &rank_dest_x );
       disp = 1;
       MPI_Cart_shift( comm_cart, direction_y ,disp ,    
                      &rank_source_y, &rank_dest_y );
 
       printf("\n  %s","    ");
       printf("\n  me= %d   rank_source_x= %d   rank_dest_x= %d   rank_source_y= %d   rank_dest_y= %d",
           me,rank_source_x,rank_dest_x,rank_source_y,rank_dest_y);
    }

/* Define the starting and ending grid point indices handled         
   by the current processor. */
/* --------------------------------------------------------------- */
/*   3D problems  */
/* -------------- */
    if (ntype == 3) {
       bound_3d(&global_1.nx, &ny, &nz, &nlocal, &nxst, &nxend, &nyst, &
		nyend, &nzst, &nzend, &nxproc, &nyproc, &nzproc, &me_new);
       printf("\n  me= %d   NXst= %d   NYst= %d    NZst= %d   Nlocal= %d   NXend= %d   NYend= %d   NZend= %d",
              me, nxst, nyst, nzst, nlocal, nxend, nyend, nzend);
    }
    if (ntype == 6) {  /* example25 & example30 */
       i1 = global_1.nx + 4;
       i2 = ny + 4;
       i3 = nz + 4;
       bound_3d(&i1, &i2, &i3, &nlocal, &nxst, &nxend, &nyst, &nyend,
		 &nzst, &nzend, &nxproc, &nyproc, &nzproc, &me_new);
       printf("\n  me= %d   NXst= %d   NYst= %d    NZst= %d   Nlocal= %d   NXend= %d   NYend= %d   NZend= %d",
              me, nxst, nyst, nzst, nlocal, nxend, nyend, nzend);
    }

    if (ntype == 11) {  /* example29 */
       i1 = global_1.nx + 2;
       i2 = ny + 2;
       i3 = nz + 3;
       bound_3d(&i1, &i2, &i3, &nlocal, &nxst, &nxend, &nyst, &nyend,
		 &nzst, &nzend, &nxproc, &nyproc, &nzproc, &me_new);
       printf("\n  me= %d   NXst= %d   NYst= %d    NZst= %d   Nlocal= %d   NXend= %d   NYend= %d   NZend= %d",
              me, nxst, nyst, nzst, nlocal, nxend, nyend, nzend);
    }

    if (ntype == 12) {  /* example32 */
       i1 = global_1.nx + 2;
       i2 = ny + 2;
       i3 = nz + 2;
       bound_3d(&i1, &i2, &i3, &nlocal, &nxst, &nxend, &nyst, &nyend,
		 &nzst, &nzend, &nxproc, &nyproc, &nzproc, &me_new);
       printf("\n  me= %d   NXst= %d   NYst= %d    NZst= %d   Nlocal= %d   NXend= %d   NYend= %d   NZend= %d",
              me, nxst, nyst, nzst, nlocal, nxend, nyend, nzend);
    }


/*   2D problems  */
/* -------------- */
    if (ntype == 2) {
       bound_2d(&global_1.nx, &ny, &nlocal, &nxst, &nxend, &nyst, &nyend, &
		nxproc, &nyproc, &me_new);
       printf("\n  me= %d   NXst= %d   NYst= %d   Nlocal= %d   NXend= %d   NYend= %d",
              me, nxst, nyst, nlocal, nxend, nyend);
    }
    if (ntype == 4) {
       i1 = global_1.nx + 2;
       i2 = ny + 2;
       bound_2d(&i1, &i2, &nlocal, &nxst, &nxend, &nyst, &nyend, &
		nxproc, &nyproc, &me_new);
       printf("\n  me= %d   NXst= %d   NYst= %d   Nlocal= %d   NXend= %d   NYend= %d",
              me, nxst, nyst, nlocal, nxend, nyend);
    }
    if (ntype == 5) {
       i1 = global_1.nx + 2;
       i2 = ny + 2;
       bound_2d(&i1, &i2, &nlocal, &nxst, &nxend, &nyst, &nyend, &
		nxproc, &nyproc, &me_new);
       printf("\n  me= %d   NXst= %d   NYst= %d   Nlocal= %d   NXend= %d   NYend= %d",
              me, nxst, nyst, nlocal, nxend, nyend);
    }
    if (ntype == 7) {
       i1 = global_1.nx + 3;
       i2 = ny + 2;
       bound_2d(&i1, &i2, &nlocal, &nxst, &nxend, &nyst, &nyend, &
		nxproc, &nyproc, &me_new);
       printf("\n  me= %d   NXst= %d   NYst= %d   Nlocal= %d   NXend= %d   NYend= %d",
              me, nxst, nyst, nlocal, nxend, nyend);
    }
    if (ntype == 8) {
       i1 = global_1.nx + 2;
       i2 = ny + 3;
       bound_2d(&i1, &i2, &nlocal, &nxst, &nxend, &nyst, &nyend, &
		nxproc, &nyproc, &me_new);
       printf("\n  me= %d   NXst= %d   NYst= %d   Nlocal= %d   NXend= %d   NYend= %d",
              me, nxst, nyst, nlocal, nxend, nyend);
    }
    if (ntype == 9) {
       i1 = global_1.nx + 4;
       i2 = ny + 3;
       bound_2d(&i1, &i2, &nlocal, &nxst, &nxend, &nyst, &nyend, &
		nxproc, &nyproc, &me_new);
       printf("\n  me= %d   NXst= %d   NYst= %d   Nlocal= %d   NXend= %d   NYend= %d",
              me, nxst, nyst, nlocal, nxend, nyend);
    }
    if (ntype == 10) {
       i1 = global_1.nx + 4;
       i2 = ny + 4;
       bound_2d(&i1, &i2, &nlocal, &nxst, &nxend, &nyst, &nyend, &
		nxproc, &nyproc, &me_new);
       printf("\n  me= %d   NXst= %d   NYst= %d   Nlocal= %d   NXend= %d   NYend= %d",
              me, nxst, nyst, nlocal, nxend, nyend);
    }

/*   1D problems   */
/* --------------- */
    if (ntype == 1) {
       bound_1d(&global_1.nx, &nlocal, &nxst, &nxend, &nxproc, &me_new);
       printf("\n  me= %d   NXst= %d   Nlocal= %d   NXend= %d",
              me, nxst, nlocal, nxend);
    }


    AZ_set_proc_config(proc_config, comm_cart);     
/*  AZ_set_proc_config(proc_config, MPI_COMM_WORLD);  */  
/*  me = proc_config[21];  */ 
    me = proc_config[AZ_node]; 

/*  Size of storage needed for 1D / 2D /3D problems */
/* ------------------------------------------------ */
    if (ntype == 12) {      /*  example32: 3D Helm. with  Dirichlet b.c. */
       nrow = (global_1.nx + 2 ) * (ny + 2) * (nz + 2);
       if(order == 6) avg_nonzero_per_row = 27;
       if(order == 2) avg_nonzero_per_row = 7;
    }
    if (ntype == 11) {      /*  example29: 3D Helm. with  Sommerfeld/Dirichlet b.c. */
       nrow = 2*(global_1.nx + 2 ) * (ny + 2) * (nz + 3);
       if(order == 6) avg_nonzero_per_row = 2*27;
       if(order == 2) avg_nonzero_per_row = 2*7;
    }
    if (ntype == 10) {
       nrow = 2*(global_1.nx + 4 ) * (ny + 4); /*   example24 Marmousi */
       if(order == 4 || order == 6) avg_nonzero_per_row = 2*9;
       if(order == 2 )  avg_nonzero_per_row = 2*5;
    }
    if (ntype == 9) {    /* example23 */
       nrow = 2*(global_1.nx + 4 ) * (ny + 3);
       if(order == 4 || order == 6) avg_nonzero_per_row = 2*9;
       if(order == 2 )  avg_nonzero_per_row = 2*5;
    }
    if (ntype == 8) {
       nrow = 2*(global_1.nx + 2 ) * (ny + 3);
       if(order == 4 || order == 6) avg_nonzero_per_row = 2*9;
       if(order == 2 )  avg_nonzero_per_row = 2*5;
    }
    if (ntype == 7) {
       nrow = 2*(global_1.nx + 3 ) * (ny + 2);
       if(order == 4 || order == 6) avg_nonzero_per_row = 2*9;
       if(order == 2 )  avg_nonzero_per_row = 2*5;
    }
    if (ntype == 6) {   /*  example25 & example30 */
       nrow = 2*(global_1.nx + 4 ) * (ny + 4) * (nz + 4);
       if(order == 6) avg_nonzero_per_row = 2*27;
       if(order == 2) avg_nonzero_per_row = 2*7;
    }
    if (ntype == 5) {
       nrow = 2*(global_1.nx + 2 ) * (ny + 2);
       avg_nonzero_per_row = 2*5;
    }
    if (ntype == 4) {
       nrow = (global_1.nx + 2) * (ny + 2);
       avg_nonzero_per_row = 5;
       if(example_no == 28 || example_no == 31) {
          if(order == 4 || order == 6) avg_nonzero_per_row = 9;
          if(order == 2 )  avg_nonzero_per_row = 5;
        }
    }
    if (ntype == 3) {
       nrow = global_1.nx * ny * nz;
       avg_nonzero_per_row = 7;
    }
    if (ntype == 2) {
       nrow = global_1.nx * ny;
       avg_nonzero_per_row = 5;
    }
    if (ntype == 1) {
       nrow = global_1.nx;
       avg_nonzero_per_row = 3;
    }

/* Define the equations handeled by the current processor */

    if (read_update == 1) {
	if (ntype == 1) {
	    AZ_read_update(&n_update, &update, proc_config,  nrow, 1, 0);
		}
	/* endif for ntype=1 */

	if (ntype == 2) {
	    n_update = (nxend - nxst + 1) * (nyend - nyst + 1);
            update     = (int *) AZ_allocate((n_update)*sizeof(int));
	    n0 = (nyst - 1) * global_1.nx + nxst - 1;
	    i1 = n_update - 1;
	    for (i = 0; i <= i1; ++i) {
		update[i] = i + n0;
	    }
	}
        /* endif for ntype=2 */

	if (ntype == 3) {
	    n_update = (nxend - nxst + 1) * (nyend - nyst + 1) * (nzend - 
		    nzst + 1);
            update     = (int *) AZ_allocate((n_update)*sizeof(int));
	    n0 = (nzst - 1) * ny * global_1.nx + nxst - 1;
	    i1 = n_update - 1;
	    for (i = 0; i <= i1; ++i) {
		update[i] = i + n0;
	    }
	}
	/* endif for ntype=3 */

	if (ntype == 4) {
	    n_update = (nxend - nxst + 1) * (nyend - nyst + 1);
            update     = (int *) AZ_allocate((n_update)*sizeof(int));
	    n0 = (nyst - 1) * (global_1.nx + 2) + nxst - 1;
	    i1 = n_update - 1;
	    for (i = 0; i <= i1; ++i) {
            update[i] = i + n0;
	    }
	}
	/* endif for ntype=4 */

	if (ntype == 5) {
	    n_update = 2*(nxend - nxst + 1 ) * (nyend - nyst + 1);
            update     = (int *) AZ_allocate((n_update)*sizeof(int));
	    n0 = 2*((nyst - 1) * (global_1.nx + 2) + nxst - 1);
	    i1 = n_update - 1;
	    for (i = 0; i <= i1; ++i) {
		update[i] = i + n0;
	    }
	}
	/* endif for ntype=5 */

	if (ntype == 6) {
	    n_update = 2*(nxend - nxst + 1 ) * (nyend - nyst + 1) * (
		    nzend - nzst + 1);
            update     = (int *) AZ_allocate((n_update)*sizeof(int));
	    n0 = 2*((nzst - 1) * (ny + 4) * (global_1.nx + 4) + nxst - 1);
	    i1 = n_update - 1;
	    for (i = 0; i <= i1; ++i) {
  	    update[i] = i + n0;
	    }
	} /* endif for ntype=6 */

	if (ntype == 7) {
	    n_update = 2*(nxend - nxst + 1 ) * (nyend - nyst + 1);
            update     = (int *) AZ_allocate((n_update)*sizeof(int));
	    n0 = 2*((nyst - 1) * (global_1.nx + 3) + nxst - 1);
	    i1 = n_update - 1;
	    for (i = 0; i <= i1; ++i) {
		update[i] = i + n0;
	    }
	}  /* endif for ntype=7 */

	if (ntype == 8) {
	    n_update = 2*(nxend - nxst + 1 ) * (nyend - nyst + 1);
            update     = (int *) AZ_allocate((n_update)*sizeof(int));
	    n0 = 2*((nyst - 1) * (global_1.nx + 2) + nxst - 1);
	    i1 = n_update - 1;
	    for (i = 0; i <= i1; ++i) {
	        update[i] = i + n0;
	    }
	}  /* endif for ntype=8 */

	if (ntype == 9) {
	    n_update = 2*(nxend - nxst + 1 ) * (nyend - nyst + 1);
            update     = (int *) AZ_allocate((n_update)*sizeof(int));
	    n0 = 2*((nyst - 1) * (global_1.nx + 4) + nxst - 1) ;
	    i1 = n_update - 1;
	    for (i = 0; i <= i1; ++i) {
		update[i] = i + n0;
	    }
	}
	/* endif for ntype=9 */

	if (ntype == 10) {
	    n_update = 2*(nxend - nxst + 1 ) * (nyend - nyst + 1);
            update     = (int *) AZ_allocate((n_update)*sizeof(int));
	    n0 = 2*((nyst - 1) * (global_1.nx + 4) + nxst - 1);
	    i1 = n_update - 1;
	    for (i = 0; i <= i1; ++i) {
		update[i] = i + n0;
	    }
	}  /* endif for ntype=10 */

	if (ntype == 11) {
	    n_update = 2*(nxend - nxst + 1 ) * (nyend - nyst + 1) * (
		    nzend - nzst + 1);
            update     = (int *) AZ_allocate((n_update)*sizeof(int));
	    n0 = 2*((nzst - 1) * (ny + 2) * (global_1.nx + 2) + nxst - 1);
	    i1 = n_update - 1;
	    for (i = 0; i <= i1; ++i) {
  	    update[i] = i + n0;
	    }
	}  /* endif for ntype=11 */

	if (ntype == 12) {
	    n_update = (nxend - nxst + 1 ) * (nyend - nyst + 1) * (
		    nzend - nzst + 1);
            update     = (int *) AZ_allocate((n_update)*sizeof(int));
	    n0 = (nzst - 1) * (ny + 2) * (global_1.nx + 2) + nxst - 1;
	    i1 = n_update - 1;
	    for (i = 0; i <= i1; ++i) {
  	    update[i] = i + n0;
	    }
	}  /* endif for ntype=12 */

    }  /*  endif  read_update=1 */

    if (read_update == 2) {
	AZ_read_update(&n_update, &update, proc_config,  nrow, 1, 2);
    }

    if (read_update == 3) {

/*  This corresponds to the case AZ_file of AZTEC (see page 53 of thr User's Guide Ver.2.1). 
    Each processor reads in the (global) indices of the equations assigned to it. */ 

    switch (me) {
	case 0:  goto L430;
	case 1:  goto L431;
	case 2:  goto L432;
	case 3:  goto L433;
	case 4:  goto L434;
	case 5:  goto L435;
	case 6:  goto L436;
	case 7:  goto L437;
	case 8:  goto L438;
	case 9:  goto L439;
	case 10:  goto L440;
	case 11:  goto L441;
	case 12:  goto L442;
	case 13:  goto L443;
	case 14:  goto L444;
	case 15:  goto L445;
    }
L430:    
        fp= fopen(".update_30","r");
        if(fp==0) {printf("\n error opening .update_30"); exit;} 
        goto L450;
L431:    
        fp = fopen(".update_31","r");
        if(fp==0) {printf("\n error opening .update_31"); exit;} 
        goto L450;
L432:    
        fp = fopen(".update_32","r");
        if(fp ==0) {printf("\n error opening .update_32"); exit;} 
        goto L450;
L433:    
        fp  = fopen(".update_33","r");
        if(fp ==0) {printf("\n error opening .update_33"); exit;} 
        goto L450;
L434:    
        fp  = fopen(".update_34","r");
        if(fp ==0) {printf("\n error opening .update_34"); exit;} 
        goto L450;
L435:    
        fp  = fopen(".update_35","r");
        if(fp ==0) {printf("\n error opening .update_35"); exit;} 
        goto L450;
L436:    
        fp  = fopen(".update_36","r");
        if(fp ==0) {printf("\n error opening .update_36"); exit;} 
        goto L450;
L437:    
        fp  = fopen(".update_37","r");
        if(fp ==0) {printf("\n error opening .update_37"); exit;} 
        goto L450;
L438:    
        fp  = fopen(".update_38","r");
        if(fp ==0) {printf("\n error opening .update_38"); exit;} 
        goto L450;
L439:    
        fp  = fopen(".update_39","r");
        if(fp ==0) {printf("\n error opening .update_39"); exit;} 
        goto L450;
L440:    
        fp  = fopen(".update_40","r");
        if(fp ==0) {printf("\n error opening .update_40"); exit;} 
        goto L450;
L441:    
        fp  = fopen(".update_41","r");
        if(fp ==0) {printf("\n error opening .update_41"); exit;} 
        goto L450;
L442:    
        fp  = fopen(".update_42","r");
        if(fp ==0) {printf("\n error opening .update_42"); exit;} 
        goto L450;
L443:    
        fp  = fopen(".update_43","r");
        if(fp ==0) {printf("\n error opening .update_43"); exit;} 
L444:    
        fp  = fopen(".update_44","r");
        if(fp ==0) {printf("\n error opening .update_44"); exit;} 
        goto L450;
L445:    
        fp  = fopen(".update_45","r");
        if(fp ==0) {printf("\n error opening .update_45"); exit;} 
        goto L450;
L450:   

	fscanf (fp, " %d", &n_update);
        i1 = n_update - 1;
	for (i = 0; i <= i1; ++i) {
		fscanf(fp,"%d",&update[i]);
	}
    } /*  endif   read_update =3 */
    printf("\n  me= %d   N_update= %d    nrow = %d", me, n_update, nrow);

/*  Allocate memory  */
	total_nz = n_update*avg_nonzero_per_row + 1;
        bindx = (int *) AZ_allocate(total_nz*sizeof(int));
        val   = (double *) AZ_allocate(total_nz*sizeof(double));   
        rhsm = (double *) calloc(nrow+1, sizeof(double));
        if((val == NULL) && (total_nz != 0)) {
           printf( "\n Error:  Not enough space to create matrix\n");
           printf( "\n   Try reducing the variable 'avg_nonzero_per_row'\n"); 
           exit(1);
        }
        if(dim == 3)  {
           n_send = (global_1.nx +2)*(global_1.nx +2);
           if(example_no ==29) n_send = 2*(global_1.nx +2)*(global_1.nx +2);
           if(example_no ==25 || example_no ==30)
               n_send = 2*(global_1.nx +4)*(global_1.nx +4);
        }
	if(dim == 2) {
           n_send = global_1.nx +2;
	   if(example_no == 21 )  n_send = 2*(global_1.nx+2);
           if(example_no == 23 ) n_send = 2*(global_1.nx+4);
	   if(example_no == 24 )  n_send = 2*(global_1.nx+4);
	   if(example_no == 26 || example_no ==27) n_send = 2*(global_1.nx+3);
           }

        index_s_xp  = (int *) calloc(n_send  , sizeof(int));
        index_s_xm  = (int *) calloc(n_send  , sizeof(int));
        index_s_yp  = (int *) calloc(n_send  , sizeof(int));
        index_s_ym  = (int *) calloc(n_send  , sizeof(int));
        index_r_xp  = (int *) calloc(n_send  , sizeof(int));
        index_r_xm  = (int *) calloc(n_send  , sizeof(int));
        index_r_yp  = (int *) calloc(n_send  , sizeof(int));
        index_r_ym  = (int *) calloc(n_send  , sizeof(int));

/* more allocations: */

	buf_r_xm  = (double *) calloc(n_send  , sizeof(double));
	buf_s_xm  = (double *) calloc(n_send  , sizeof(double));
    	buf_r_ym  = (double *) calloc(n_send  , sizeof(double));
        buf_s_ym  = (double *) calloc(n_send  , sizeof(double));
        buf_r_xp  = (double *) calloc(n_send  , sizeof(double));
        buf_s_xp  = (double *) calloc(n_send  , sizeof(double));
        buf_r_yp  = (double *) calloc(n_send  , sizeof(double));
        buf_s_yp  = (double *) calloc(n_send  , sizeof(double));

    if(dim == 3) {
        index_s_zp  = (int *) calloc(n_send  , sizeof(int));
        index_s_zm  = (int *) calloc(n_send  , sizeof(int));
        index_r_zp  = (int *) calloc(n_send  , sizeof(int));
        index_r_zm  = (int *) calloc(n_send  , sizeof(int));
        buf_r_zm  = (double *) calloc(n_send  , sizeof(double));
        buf_s_zm  = (double *) calloc(n_send  , sizeof(double));
        buf_r_zp  = (double *) calloc(n_send  , sizeof(double));
        buf_s_zp  = (double *) calloc(n_send  , sizeof(double));
    }

/*   Define the indices of the external boundry grid points   
     which communicate with the neigboring processors        */

    switch (ntype) {
	case 1:  goto L1;
	case 2:  goto L2;
	case 3:  goto L3;
	case 4:  goto L4;
	case 5:  goto L5;
	case 6:  goto L6;
	case 7:  goto L7;
	case 8:  goto L8;
	case 9:  goto L9;
	case 10:  goto L10;
	case 11:  goto L118;
	case 12:  goto L119;
    }
L1:
    example_1d(&nxst, &nxend, &global_1.nx, index_s_xm, index_s_xp, 
	    index_r_xm, index_r_xp, &ixp, &ixm);
    goto L60;
L2:
    example_2d(&nxst, &nxend, &nyst, &nyend, &global_1.nx, &ny, 
	    index_s_xm, index_s_xp, index_s_ym, index_s_yp, 
	    index_r_xm, index_r_xp, index_r_ym, index_r_yp, &ixp, &
	    ixm, &iyp, &iym);
    goto L60;
L3:
    example_3d(&nxst, &nxend, &nyst, &nyend, &nzst, &nzend, &global_1.nx, &
	    ny, &nz, index_s_xm, index_s_xp, index_s_ym, index_s_yp, 
	    index_s_zm, index_s_zp, index_r_xm, index_r_xp, 
	    index_r_ym, index_r_yp, index_r_zm, index_r_zp, &ixp, &
	    ixm, &iyp, &iym, &izp, &izm);
    goto L60;
L4:
    if (example_no != 28 && example_no != 31) {
	i1 = global_1.nx + 2;
	i2 = ny + 2;
	example_2d(&nxst, &nxend, &nyst, &nyend, &i1, &i2, index_s_xm,
		 index_s_xp, index_s_ym, index_s_yp, index_r_xm, 
		index_r_xp, index_r_ym, index_r_yp, &ixp, &ixm, &iyp, 
		&iym);
    }
    if (example_no == 28 || example_no == 31) {
	if (order == 2) {
	    nx_st = 1;
	    nx_end = global_1.nx + 2;
	    ny_st = 1;
	    ny_end = ny + 2;
	}
	if (order == 4 || order == 6) {
	    nx_st = 0;
	    nx_end = global_1.nx + 3;
	    ny_st = 0;
	    ny_end = ny + 3;
	}
	i1 = global_1.nx + 2;
	i2 = ny + 2;
	example_2d_all(&nxst, &nxend, &nyst, &nyend, &i1, &i2, 
		index_s_xm, index_s_xp, index_s_ym, index_s_yp, 
		index_r_xm, index_r_xp, index_r_ym, index_r_yp, &ixp, 
		&ixm, &iyp, &iym, &nx_st, &nx_end, &ny_st, 
		&ny_end);
    }
    printf("\n  me= %d   ixm= %d   ixp= %d   iym= %d   iyp= %d", 
		    me, ixm, ixp, iym, iyp);
    goto L60;
L5:
    example_2d_img(&nxst, &nxend, &nyst, &nyend, &global_1.nx, &ny, 
	    index_s_xm, index_s_xp, index_s_ym, index_s_yp, 
	    index_r_xm, index_r_xp, index_r_ym, index_r_yp, &ixp, 
	    &ixm, &iyp, &iym);
    printf("\n  me= %d   ixm= %d   ixp= %d   iym= %d   iyp= %d", 
		    me, ixm, ixp, iym, iyp);
    goto L60;
L6:
    if (order == 2) {
	nx_st = 1;
	nx_end = global_1.nx + 4;
	ny_st = 1;
	ny_end = ny + 4;
	nz_st = 1;
	nz_end = nz + 4;
    }
    if (order == 6) {
	nx_st = 0;
	nx_end = global_1.nx + 5;
	ny_st = 0;
	ny_end = ny + 5;
	nz_st = 0;
	nz_end = nz + 5;
    }
        i1=  global_1.nx + 2;
        i2 =  ny + 2;
        i3 = nz + 2;
    example_3d_img_all(&nxst, &nxend, &nyst, &nyend, &nzst, &nzend, 
	    &i1,&i2,&i3, index_s_xm, index_s_xp, index_s_ym, 
	    index_s_yp, index_s_zm, index_s_zp, index_r_xm, 
	    index_r_xp, index_r_ym, index_r_yp, index_r_zm, 
	    index_r_zp, &ixp, &ixm, &iyp, &iym, &izp, &izm,
            &nx_st,&nx_end,&ny_st,&ny_end,&nz_st,&nz_end,&me);
    printf("\n  me= %d   ixm= %d   ixp= %d   iym= %d   iyp= %d   izm= %d   izp= %d", 
		    me, ixm, ixp, iym, iyp, izm, izp);
    goto L60;
L7:
    if (order == 2) {
	nx_st = 1;
	nx_end = global_1.nx + 3;
	ny_st = 1;
	ny_end = ny + 2;
    }
    if (order == 4 || order == 6) {
	nx_st = 0;
	nx_end = global_1.nx + 4;
	ny_st = 0;
	ny_end = ny + 3;
    }
    i1 = global_1.nx + 1;
    example_2d_img_all(&nxst, &nxend, &nyst, &nyend, &i1, &ny, 
	    index_s_xm, index_s_xp, index_s_ym, index_s_yp, 
	    index_r_xm, index_r_xp, index_r_ym, index_r_yp, &ixp, &
	    ixm, &iyp, &iym, &nx_st, &nx_end, &ny_st, 
	    &ny_end);
    printf("\n  me= %d   ixm= %d   ixp= %d   iym= %d   iyp= %d", 
		    me, ixm, ixp, iym, iyp);
    goto L60;
L8:
    if (order == 2) {
	nx_st = 1;
	nx_end = global_1.nx + 2;
	ny_st = 1;
	ny_end = ny + 3;
    }
    if (order == 4 || order == 6) {
	nx_st = 0;
	nx_end = global_1.nx + 3;
	ny_st = 0;
	ny_end = ny + 4;
    }
    i1 = ny + 1;
    example_2d_img_all(&nxst, &nxend, &nyst, &nyend, &global_1.nx, &i1, 
	    index_s_xm, index_s_xp, index_s_ym, index_s_yp, 
	    index_r_xm, index_r_xp, index_r_ym, index_r_yp, &ixp, 
	    &ixm, &iyp, &iym, &nx_st, &nx_end, &ny_st, 
	    &ny_end);
    printf("\n  me= %d   ixm= %d   ixp= %d   iym= %d   iyp= %d", 
		    me, ixm, ixp, iym, iyp);
    goto L60;
L9:
    if (order == 2) {
	nx_st = 1;
	nx_end = global_1.nx + 4;
	ny_st = 1;
	ny_end = ny + 3;
    }
    if (order == 4 || order == 6) {
	nx_st = 0;
	nx_end = global_1.nx + 5;
	ny_st = 0;
	ny_end = ny + 4;
    }
    i1 = global_1.nx + 2;
    i2 = ny + 1;
    example_2d_img_all(&nxst, &nxend, &nyst, &nyend, &i1, &i2, 
	    index_s_xm, index_s_xp, index_s_ym, index_s_yp, 
	    index_r_xm, index_r_xp, index_r_ym, index_r_yp, &ixp, 
	    &ixm, &iyp, &iym, &nx_st, &nx_end, &ny_st, 
	    &ny_end);
    printf("\n  me= %d   ixm= %d   ixp= %d   iym= %d   iyp= %d", 
		    me, ixm, ixp, iym, iyp);
    goto L60;
L10:
    if (order == 2) {
	nx_st = 1;
	nx_end = global_1.nx + 4;
	ny_st = 1;
	ny_end = ny + 4;
    }
    if (order == 4 || order == 6) {
	nx_st = 0;
	nx_end = global_1.nx + 5;
	ny_st = 0;
	ny_end = ny + 5;
    }
    i1 = global_1.nx + 2;
    i2 = ny + 2;
    example_2d_img_all(&nxst, &nxend, &nyst, &nyend, &i1, &i2, 
	    index_s_xm, index_s_xp, index_s_ym, index_s_yp, 
	    index_r_xm, index_r_xp, index_r_ym, index_r_yp, &ixp, 
	    &ixm, &iyp, &iym, &nx_st, &nx_end, &ny_st, 
	    &ny_end);
    printf("\n  me= %d   ixm= %d   ixp= %d   iym= %d   iyp= %d", 
		    me, ixm, ixp, iym, iyp);
    goto L60;
L118:
    if (order == 2) {
	nx_st = 1;
	nx_end = global_1.nx + 2;
	ny_st = 1;
	ny_end = ny + 2;
	nz_st = 1;
	nz_end = nz + 3;
    }
    if (order == 6) {
	nx_st = 0;
	nx_end = global_1.nx + 3;
	ny_st = 0;
	ny_end = ny + 3;
	nz_st = 0;
	nz_end = nz + 4;
    }
     i3= nz+1;
    example_3d_img_all(&nxst, &nxend, &nyst, &nyend, &nzst, &nzend,
            &global_1.nx, &ny, &i3, index_s_xm, index_s_xp, index_s_ym,
            index_s_yp, index_s_zm, index_s_zp, index_r_xm,
            index_r_xp, index_r_ym, index_r_yp, index_r_zm,
            index_r_zp, &ixp, &ixm, &iyp, &iym, &izp, &izm,
            &nx_st,&nx_end,&ny_st,&ny_end,&nz_st,&nz_end,&me);

    printf("\n  me= %d   ixm= %d   ixp= %d   iym= %d   iyp= %d   izm= %d   izp= %d", 
		    me, ixm, ixp, iym, iyp, izm, izp);
    goto L60;
L119:
    if (order == 2) {
	nx_st = 1;
	nx_end = global_1.nx + 2;
	ny_st = 1;
	ny_end = ny + 2;
	nz_st = 1;
	nz_end = nz + 2;
    }
    if (order == 6) {
	nx_st = 0;
	nx_end = global_1.nx + 3;
	ny_st = 0;
	ny_end = ny + 3;
	nz_st = 0;
	nz_end = nz + 3;
    }
    example_3d_mod(&nxst, &nxend, &nyst, &nyend, &nzst, &nzend, &global_1.nx, 
	    &ny, &nz, index_s_xm, index_s_xp, index_s_ym, index_s_yp, 
	    index_s_zm, index_s_zp, index_r_xm, index_r_xp, 
	    index_r_ym, index_r_yp, index_r_zm, index_r_zp, &ixp, 
	    &ixm, &iyp, &iym, &izp, &izm, 
            &nx_st, &nx_end, &ny_st, &ny_end, &nz_st, &nz_end);

    printf("\n  me= %d   ixm= %d   ixp= %d   iym= %d   iyp= %d   izm= %d   izp= %d", 
		    me, ixm, ixp, iym, iyp, izm, izp);
L60:

    nxloc = nxend - nxst + 1;
    if (dim != 1) {
	nyloc = nyend - nyst + 1;
    }
    if(dim == 3) {
/*  if (ntype == 3 || ntype == 6 || ntype == 11 || ntype == 12) { */
	nzloc = nzend - nzst + 1;
    }
    bindx[0] = n_update + 1;
    nend = n_update;

/*  Define and save the matrix following 'AZTEC' sparse matrix format   */
/* -------------------------------------------------------------------- */

    if (example_no == 20) {  /*  example 20 - 2D helmholtz problem */
       i1 = nend - 1;
       for (i = 0; i <= i1; ++i) {
	create_matrix_row_5pt_example20(&update[i], &i, val, bindx, &
		kmax, rhsm, &nxst, &nyst, &nxloc, &nyloc, &kn1);
	}
	goto L290;
    }  /* endif example20 */

    if (example_no == 21) { /*  example 21 - 2D Helmholtz problem */

	i1 = n_update - 1;
	for (i = 0; i <= i1; i += 2) {
	    create_matrix_row_5pt_example21_2rows(&update[i], &i, val, 
		    bindx, &kmax, rhsm, &nxst, &nyst, &nxloc, &nyloc, &kn1, 
		    &kn2, &kn3);
	}
	goto L290;
    }   /* endif example21 */

    if (example_no == 23) { /*  example 23 - 2D Helmholtz problem */
	i1 = n_update - 1;
	for (i = 0; i <= i1; i += 2) {
	    create_matrix_row_9pt_example23_2rows(&update[i], &i, val, 
		    bindx, &kmax, rhsm, &nxst, &nyst, &nxloc, &nyloc, &kn1, 
		    &kn2, &kn3, &order);
	}
	goto L290;
    }   /* endif example23 */

    if (example_no == 24) {  /*  example 24 - Marmousi problem */
        i1 = n_update - 1;
	for (i = 0; i <= i1; i += 2) {
	    create_matrix_row_5_9_pt_example24_2rows(&update[i], &i, val, 
		    bindx, &kmax, rhsm, &nxst, &nyst, &nxloc, &nyloc, cxy, 
		    &xmin, &xmax, &ymin, &ymax, &global_1.nx, &ny,
                    &order,&freq);
	}
	goto L290;
    }   /* endif example24 */

    if (example_no == 25) { /*  example 25 - 3D Helmholtz problem */
	i1 = n_update - 1;
	for (i = 0; i <= i1; i += 2) {
	    create_matrix_row_7_27_pt_example25_2rows(&update[i], &i, val, 
		    bindx, &kmax, rhsm, &nxst, &nyst, &nzst, &nxloc, &nyloc, 
		    &nzloc, &kn1, &kn2, &kn3, &xmin, &xmax, &ymin, &ymax, 
		    &zmin, &zmax, &global_1.nx, &ny, &nz, &order);
	}
	goto L290;
    }   /* endif create matrix for example25 */

    if (example_no == 26) {
	i1 = n_update - 1;
	for (i = 0; i <= i1; i += 2) {
	    create_matrix_row_9pt_example26_2rows(&update[i], &i, val, 
		    bindx, &kmax, rhsm, &nxst, &nyst, &nxloc, &nyloc, &kn1, 
		    &kn2, &kn3, &order);
	}
	goto L290;
    }   /* endif example26 */

    if (example_no == 27) {
	i1 = n_update - 1;
	for (i = 0; i <= i1; i += 2) {
	    create_matrix_row_9pt_example27_2rows(&update[i], &i, val, 
		    bindx, &kmax, rhsm, &nxst, &nyst, &nxloc, &nyloc, &kn1, 
		    &kn2, &kn3, &order);
	}
	goto L290;
    }  /* endif example27 */

    if (example_no == 28) {  /*  example 28 - 2D helmholtz problem */
       i1 = nend - 1;
       for (i = 0; i <= i1; ++i) {
	create_matrix_row_9pt_example28(&update[i], &i, val, bindx, &
		kmax, rhsm, &nxst, &nyst, &nxloc, &nyloc, &kn1, &order);
	}
	goto L290;
    }  /* endif example28 */


    if (example_no == 29) {  /*  example 29 - 3D Helmholtz with analytic solution */
        i1 = n_update - 1;
	for (i = 0; i <= i1; i += 2) {
	    create_matrix_row_7_27_pt_example29_2rows(&update[i], &i, val, 
		    bindx, &kmax, rhsm, &nxst, &nyst,&nzst, &nxloc, &nyloc, &nzloc, 
                    &kn1, &kn2, &kn3, &order);
	}
	goto L290;
    }   /* endif example29 */

    if (example_no == 30) {  /*  example 30 - SEG/EAGE salt problem */
        i1 = n_update - 1;
	for (i = 0; i <= i1; i += 2) {
	    create_matrix_row_7_27_pt_example30_2rows(&update[i], &i, val, 
		    bindx, &kmax, rhsm, &nxst, &nyst, &nzst, &nxloc, &nyloc, &nzloc, cxy, 
		    &xmin, &xmax, &ymin, &ymax, &zmin, &zmax, &global_1.nx, &ny, &nz,
                    &order,&freq, &me);
        }
	goto L290;
    }   /* endif example30 - SEG/EAGE salt problem */

    if (example_no == 31) {  /*  example 31 - 2D helmholtz problem with variable K */
       i1 = nend - 1;
       for (i = 0; i <= i1; ++i) {
           if(icase == 1)
              create_matrix_row_9pt_example31_case_a(&update[i], &i, val, bindx, &
                kmax, rhsm, &nxst, &nyst, &nxloc, &nyloc, &mn , &order, &i_scheme);
           if(icase == 2)
              create_matrix_row_9pt_example31_case_b(&update[i], &i, val, bindx, &
                kmax, rhsm, &nxst, &nyst, &nxloc, &nyloc, &mn , &order, &i_scheme,
                &a2);
           if(icase == 3)
              create_matrix_row_9pt_example31_case_c(&update[i], &i, val, bindx, &
                kmax, rhsm, &nxst, &nyst, &nxloc, &nyloc, &a2 , &b2, &c2, &order, &i_scheme);
       }
       goto L290;
    }   /* endif example31  */

    if (example_no == 32) {  /*  example 32 - 2D helmholtz problem with variable K */
       i1 = nend - 1;
       for (i = 0; i <= i1; ++i) {
        if(icase == 1)
        create_matrix_row_7_27_pt_example32_case_a(&update[i], &i, val, bindx, &
                kmax, rhsm, &nxst, &nyst, &nzst, &nxloc, &nyloc, &nzloc, 
                &l_a, &n_a, &order, &i_scheme);
        if(icase == 3)
        create_matrix_row_7_27_pt_example32_case_c(&update[i], &i, val, bindx, &
                kmax, rhsm, &nxst, &nyst, &nzst, &nxloc, &nyloc, &nzloc, 
                &a2, &b2, &c2, &gama, &order, &i_scheme);
       }
       goto L290;
    }   /* endif example32  */


L290:
    printf("\n  me= %d   Kmax= %d   update(i)= %d   NXloc= %d   NYloc= %d", 
		    me, kmax, update[nend-1], nxloc, nyloc);

/*  Define the averaging weights */
    iem  = (int *) calloc(nrow , sizeof(int));
    iema = (int *) calloc(nrow , sizeof(int));
    ecoef = (double *) calloc(nrow , sizeof(double));

    i1 = n_update - 1;
    for (i = 0; i <= i1; ++i) {
	iem[update[i]] = 1;
	i2 = bindx[i + 1] - 1;
	for (j = bindx[i]; j <= i2; ++j) {
	    iem[bindx[j]] = 1;
	}
    }

    MPI_Allreduce(iem, iema, n, MPI_INT, MPI_SUM, MPI_COMM_WORLD );

    i1 = n - 1;   
    for (i = 0; i <= i1; ++i) {
	ecoef[i] = 1. / iema[i];
    }
/*   if(me ==0){
          i1= nrow -1;
          for (i = 0; i <= i1; ++i) { 
             printf("\n%8d  %13.7e", i,  ecoef[i]);
          }
      } */ 

/*   Tranform the matrix to a local matrix and define various variables to conform with AZTEC */


    AZ_transform(proc_config, &external, bindx, val, update, 
	    &update_index, &extern_index, &data_org, n_update, 0, 
	    0, 0, 0, AZ_MSR_MATRIX);

    n_update_internal = data_org[1];
    n_update_border = data_org[2];
    n_update_external = data_org[3];
    n_neigh = data_org[7];
    n_total_send = data_org[8];
    n_id_neigh = data_org[12];
    n_rec_length = data_org[262];
    n_send_length = data_org[512];
    n_send_list = data_org[762];

/*     write(NOUT,*)'me=',me,'  N_update_internal=',N_update_internal   
      1          ,'  N_update_border=',N_update_border   
      2          ,'  N_update_external=',N_update_external   
      3          ,'  N_neigh=',N_neigh   
      4          ,'  N_total_send=', N_total_send   
      5          ,'  N_id_neigh=',N_id_neigh   
      6          ,'  N_rec_length=',N_rec_length   
      7          ,'  N_send_length=',N_send_length   
      8          ,'  N_send_list=',N_send_list   
*/


    printf("\n  me= %d   N_update_internal= %d   N_update_border= %d",
            me, n_update_internal,n_update_border);
    printf("   N_update_external= %d   N_neigh= %d", 
            n_update_external, n_neigh);
    printf("\n  N_total_send= %d   N_id_neigh= %d   N_rec_length= %d   N_send_length= %d   N_send_list= %d", 
            n_total_send, n_id_neigh, n_rec_length, n_send_length, n_send_list);

/*  Allocate memory  */

        update_inv = (int *) calloc(nrow , sizeof(int));
        xnew = (double *) calloc(nrow , sizeof(double));
        xnewl = (double *) calloc(nrow , sizeof(double));
        err  = (double *) calloc(nrow , sizeof(double));
        qnewl= (double *) calloc(nrow , sizeof(double));
        resid= (double *) calloc(nrow , sizeof(double));
        resid1= (double *) calloc(nrow , sizeof(double));

/*      xnewl_loc = (double *) calloc(n_update + data_org[AZ_N_external] , sizeof(double));
        xnewl_loc_init = (double *) calloc(n_update + data_org[AZ_N_external] , sizeof(double));
        rhsm_loc = (double *) calloc(n_update + data_org[AZ_N_external] , sizeof(double));
        pnewl_loc = (double *) calloc(n_update + data_org[AZ_N_external] , sizeof(double));
        qnewl_loc = (double *) calloc(n_update + data_org[AZ_N_external] , sizeof(double));
        snewl_loc = (double *) calloc(n_update + data_org[AZ_N_external] , sizeof(double));  */ 
             
        xnewl_loc = (double *) calloc(n_update + 2*n_send , sizeof(double));
        xnewl_loc_init = (double *) calloc(n_update + 2*n_send , sizeof(double));
        rhsm_loc = (double *) calloc(n_update + 2*n_send , sizeof(double));
        pnewl_loc = (double *) calloc(n_update + 2*n_send , sizeof(double));
        qnewl_loc = (double *) calloc(n_update + 2*n_send , sizeof(double));
        snewl_loc = (double *) calloc(n_update + 2*n_send , sizeof(double));   
        ssum  = (double *) calloc(n_update  , sizeof(double));


/* -------------------------------------------------------------- */
/* -------------------------------------------------------------- */

/*  Calculate the initial residual - RESIDUAL(0) */

    sumr = 0.;
    i1 = n_update - 1;
    for (i = 0; i <= i1; ++i) {
	d1 = rhsm[update[i]];
	sumr += d1 * d1;
    }
    MPI_Allreduce(&sumr, &sumrt, 1, MPI_DOUBLE_PRECISION, MPI_SUM, MPI_COMM_WORLD );
    sumrt = sqrt(sumrt);
    printf("\n  RESIDUAL(0)= %.8f", sumrt);
    res0 = sumrt;

/*  Initiate vectors */

    n_internal = n_update_internal;
    n_border = n_update_border;
    n_external = n_update_external;

    i1 = n_update - 1;
    for (i = 0; i <= i1; ++i) {
	update_inv[update_index[i]] = update[i];
	rhsm_loc[update_index[i]] = rhsm[update[i]];
/*  Define xnewl_loc[i] in case that xnewl_loc # 0. */
/*	xnewl_loc[i] = 0.;*/
    }

    t1 = MPI_Wtime();
    printf("\n  %s","   ");

/*  Define a normalization factor */

    i1 = n_update - 1;
    for (i = 0; i <= i1; ++i) {
	sum = val[i] * val[i];
	i2 = bindx[i + 1] - 1;
	for (j = bindx[i]; j <= i2; ++j) {
	    d1 = val[j];
	    sum += d1 * d1;
	}
	ssum[i] = omeg1 / sum;
    }


/*  Start the initial forth and back sweeps of CGMN/CARP-CG */
/*  (required by CGMN/CARP-CG initial vectors S0( ),P0( ) ) */
/* -------------------------------------------------------------------- */
/*  Initiate the vector */
    i1 = n_update - 1;
    for (i = 0; i <= i1; ++i) {
	xnewl_loc_init[i] = xnewl_loc[i];
    }
    i1 = n_update_external - 1;
    for (i = 0; i <= i1; ++i) {
	xnewl_loc_init[extern_index[i]] = xnewl_loc[extern_index[
		i]];
    }
    for (icg = 1; icg <= 2; ++icg) {
	if (icg == 1) {
	    istart = 0;
	    iend = n_update - 1;
	    iinc = 1;
	} else {
	    istart = n_update - 1;
	    iend = 0;
	    iinc = -1;
	}
	i1 = iend;
	i2 = iinc;
	for (i = istart; i2 < 0 ? i >= i1 : i <= i1; i += i2) 
		{
	    sum = val[i] * xnewl_loc[i] - rhsm_loc[i];
	    i3 = bindx[i + 1] - 1;
	    for (j = bindx[i]; j <= i3; ++j) {
		sum += val[j] * xnewl_loc[bindx[j]];
	    }
	    cons = -sum * ssum[i];
	    xnewl_loc[i] += cons * val[i];
	    i3 = bindx[i + 1] - 1;
	    for (j = bindx[i]; j <= i3; ++j) {
		xnewl_loc[bindx[j]] += cons * val[j];
	    }
	}
/*      if(me ==0 ) {
        printf("\n ICG= %d",icg);
        i1= n_update - 1;
          for (i = 0; i <= i1; ++i) { 
             printf("\n%8d  %13.7e", i,  xnewl_loc[i]);
          }
        printf("\n ****************");
        i1= n_update_external - 1;
          for (i = 0; i <= i1; ++i) {
             printf("\n%8d  %13.7e %8d ", i,  xnewl_loc[extern_index[i]], extern_index[i]);
          }
        }    
*/ 

/*  Transfer information: */

	if (nproc > 1) {
	    i2 = n_update - 1;
	    for (i = 0; i <= i2; ++i) {
		xnewl[update_inv[i]] = xnewl_loc[i];
	    }
	    i2 = n_update_external - 1;
	    for (i = 0; i <= i2; ++i) {
		xnewl[external[i]] = xnewl_loc[extern_index[i]];
	    }


/*          if(me ==0 ) {
               printf("\n ICG= %d",icg);
               i2 = n_update_external - 1;
               for (i = 0; i <= i2; ++i) {
                  printf("\n%8d  %8d %8d", i,  external[i], extern_index[i]);
               }
               i1= nrow -1;
               for (i = 0; i <= i1; ++i) { 
                  printf("\n%8d  %13.7e ", i,  xnewl[i]);
               }
            }  */   
 
	    if (dim == 3 ) {
		transf_3d(xnewl, ecoef, update_inv, &n_update, &
			comm_cart, &me, &rank_source_x, &rank_source_y, 
			&rank_source_z, &rank_dest_x, &rank_dest_y, 
			&rank_dest_z, index_s_xm, index_s_xp, 
			index_s_ym, index_s_yp, index_s_zm, 
			index_s_zp, index_r_xm, index_r_xp, 
			index_r_ym, index_r_yp, index_r_zm, 
			index_r_zp, &ixp, &ixm, &iyp, &iym, &izp, &izm, &dim);
	    } 
            if (dim == 2 ) {
                transf_2d(xnewl, ecoef, update_inv, &n_update, &
	                comm_cart, &me, &rank_source_x, &
		        rank_source_y, &rank_dest_x, &rank_dest_y, 
		        index_s_xm, index_s_xp, index_s_ym, 
		        index_s_yp, index_r_xm, index_r_xp, 
		        index_r_ym, index_r_yp, &ixp, &ixm, &iyp, &iym);
	    }
	    i2 = n_internal + n_border - 1;
	    for (i = n_internal; i <= i2; ++i) {
		xnewl_loc[i] = xnewl[update_inv[i]];
	    }
	    i2 = n_update_external - 1;
	    for (i = 0; i <= i2; ++i) {
		xnewl_loc[extern_index[i]] = xnewl[external[i]];
	    }
/*        if(me ==0 ) {
             printf("\n IKCG= %d",icg);   
             i1= n_update ;
             for (i = n_internal; i <= i1; ++i) { 
                printf("\n%8d  %13.7e", i,  xnewl_loc[i]);
             }
             i1= n_update_external - 1;
             for (i = 0; i <= i1; ++1){
                printf("\n%8d  %13.7e %8d", i,  xnewl_loc[extern_index[i]],[extern_index[i]);
             }   
             i1= nrow -1;
             for (i = 0; i <= i1; ++i) { 
                printf("\n%8d  %13.7e", i,  xnewl[i]);
             }
           }  */  /* end of printout of  me=0 */
	}  /*  end of transfer of information for  Nproc.gt.1 */ 

    }  /*  end of the first ICG (forth and back sweeps) loop. */

/*  Start the  main forth and back sweeps of CGMN/CARP-CG: */
/* --------------------------------------------------------- */
    i2 = kcg;
    for (ikcg = 1; ikcg <= i2; ++ikcg) {
	if (ikcg == 1) {

/*  Define the initial values of P( ), S( ) and X( ): */
/* ---------------------------------------------------- */
	    s2 = 0.;
	    i1 = n_update - 1;
	    for (i = 0; i <= i1; ++i) {
		pnewl_loc[i] = xnewl_loc[i] - xnewl_loc_init[i];
		snewl_loc[i] = pnewl_loc[i];
		d1 = snewl_loc[i];
		s2 += d1 * d1;
		qnewl_loc[i] = pnewl_loc[i];
		xnewl_loc[i] = xnewl_loc_init[i];
	    }

	    i1 = n_update_external - 1;
	    for (i = 0; i <= i1; ++i) {
		pnewl_loc[extern_index[i]] = xnewl_loc[extern_index[
			i]] - xnewl_loc_init[extern_index[i]];
		snewl_loc[extern_index[i]] = pnewl_loc[extern_index[
			i]];
		qnewl_loc[extern_index[i]] = pnewl_loc[extern_index[
			i]];
		xnewl_loc[extern_index[i]] = 0.;
	    }
            MPI_Allreduce(&s2, &s2s, 1, MPI_DOUBLE_PRECISION, MPI_SUM, 
                      MPI_COMM_WORLD);

	} else {

/*   For IKCG >1:  update S( ) and P( ), initiate Q( ) and calculate S**2 and BETA:  */
/* --------------------------------------------------------------------------------- */

/* for IKCG >1 */
	    s2p = 0.;
	    i1 = n_update - 1;
	    for (i = 0; i <= i1; ++i) {
		snewl_loc[i] -= alpha * qnewl_loc[i];
		d1 = snewl_loc[i];
		s2p += d1 * d1;
	    }
	    i1 = n_update_external - 1;
	    for (i = 0; i <= i1; ++i) {
		snewl_loc[extern_index[i]] -= alpha * qnewl_loc[
			extern_index[i]];
	    }

/*   Sum up S**2 from all processors: */

            MPI_Allreduce(&s2p, &s2ps, 1, MPI_DOUBLE_PRECISION, MPI_SUM, 
                      MPI_COMM_WORLD);

	    beta = s2ps / s2s;
	    s2s = s2ps;

/*   Update P( ) and initiate Q( ) : */

	    i1 = n_update - 1;
	    for (i = 0; i <= i1; ++i) {
		pnewl_loc[i] = snewl_loc[i] + beta * pnewl_loc[i];
		qnewl_loc[i] = pnewl_loc[i];
	    }
	    i1 = n_update_external - 1;
	    for (i = 0; i <= i1; ++i) {
		pnewl_loc[extern_index[i]] = snewl_loc[extern_index[
			i]] + beta * pnewl_loc[extern_index[i]];
		qnewl_loc[extern_index[i]] = pnewl_loc[extern_index[
			i]];
	    }
	}

/* Ccalculate Q( ):  do the forth sweep */

	i1 = n_update - 1;
	for (i = 0; i <= i1; ++i) {
	    sum = val[i] * qnewl_loc[i];
	    i3 = bindx[i + 1] - 1;
	    for (j = bindx[i]; j <= i3; ++j) {
		sum += val[j] * qnewl_loc[bindx[j]];
	    }
	    cons = -sum * ssum[i];
	    qnewl_loc[i] += cons * val[i];
	    i3 = bindx[i + 1] - 1;
	    for (j = bindx[i]; j <= i3; ++j) {
		qnewl_loc[bindx[j]] += cons * val[j];
	    }
	}

/*  Transfer information (for Nproc>1): */

	if (nproc > 1) {
	    i1 = n_update - 1;
	    for (i = 0; i <= i1; ++i) {
		qnewl[update_inv[i]] = qnewl_loc[i];
	    }
	    i1 = n_update_external - 1;
	    for (i = 0; i <= i1; ++i) {
		qnewl[external[i]] = qnewl_loc[extern_index[i]];
	    }
	    if (dim == 3 ) {
		transf_3d(qnewl, ecoef, update_inv, &n_update, &
			comm_cart, &me, &rank_source_x, &rank_source_y, 
			&rank_source_z, &rank_dest_x, &rank_dest_y, &
			rank_dest_z, index_s_xm, index_s_xp, 
			index_s_ym, index_s_yp, index_s_zm, 
			index_s_zp, index_r_xm, index_r_xp, 
			index_r_ym, index_r_yp, index_r_zm, 
			index_r_zp, &ixp, &ixm, &iyp, &iym, &izp, &izm, &
			dim);
	    } 
	    if (dim == 2 ) {
		transf_2d(qnewl, ecoef, update_inv, &n_update, &
			comm_cart, &me, &rank_source_x, &
		        rank_source_y, &rank_dest_x, &rank_dest_y, 
			index_s_xm, index_s_xp, index_s_ym, 
			index_s_yp, index_r_xm, index_r_xp, 
			index_r_ym, index_r_yp, &ixp, &ixm, &iyp, &iym);
	    }
	    i1 = n_update - 1;
	    for (i = n_internal; i <= i1; ++i) {
		qnewl_loc[i] = qnewl[update_inv[i]];
	    }
	    i1 = n_update_external - 1;
	    for (i = 0; i <= i1; ++i) {
		qnewl_loc[extern_index[i]] = qnewl[external[i]];
	    }
	}

/*  Do the back sweep */

	for (i = n_update - 1; i >= 0; --i) {
	    sum = val[i] * qnewl_loc[i];
	    i1 = bindx[i + 1] - 1;
	    for (j = bindx[i]; j <= i1; ++j) {
		sum += val[j] * qnewl_loc[bindx[j]];
	    }
	    cons = -sum * ssum[i];
	    qnewl_loc[i] += cons * val[i];
	    i1 = bindx[i + 1] - 1;
	    for (j = bindx[i]; j <= i1; ++j) {
		qnewl_loc[bindx[j]] += cons * val[j];
	    }
	}

/*  Transfer information (for Nproc>1): */

	if (nproc > 1) {
	    i1 = n_update - 1;
	    for (i = 0; i <= i1; ++i) {
		qnewl[update_inv[i]] = qnewl_loc[i];
	    }
	    i1 = n_update_external - 1;
	    for (i = 0; i <= i1; ++i) {
		qnewl[external[i]] = qnewl_loc[extern_index[i]];
	    }
            if (dim == 3 ) {
		transf_3d(qnewl, ecoef, update_inv, &n_update, &
			comm_cart, &me, &rank_source_x, &rank_source_y, 
			&rank_source_z, &rank_dest_x, &rank_dest_y, &
			rank_dest_z, index_s_xm, index_s_xp, 
			index_s_ym, index_s_yp, index_s_zm, 
			index_s_zp, index_r_xm, index_r_xp, 
			index_r_ym, index_r_yp, index_r_zm, 
			index_r_zp, &ixp, &ixm, &iyp, &iym, &izp, &izm, &dim);
	    } 
	    if (dim == 2 ) {
		transf_2d(qnewl, ecoef, update_inv, &n_update, &
		        comm_cart, &me, &rank_source_x, &
			rank_source_y, &rank_dest_x, &rank_dest_y, 
			index_s_xm, index_s_xp, index_s_ym, 
			index_s_yp, index_r_xm, index_r_xp, 
			index_r_ym, index_r_yp, &ixp, &ixm, &iyp, &iym);
	    }
	    i1 = n_update - 1;
	    for (i = n_internal; i <= i1; ++i) {
		qnewl_loc[i] = qnewl[update_inv[i]];
	    }
	    i1 = n_update_external - 1;
	    for (i = 0; i <= i1; ++i) {
		qnewl_loc[extern_index[i]] = qnewl[external[i]];
	    }
	}  /* end of information transfer for nproc > 1  */

	i1 = n_update - 1;
	for (i = 0; i <= i1; ++i) {
	    qnewl_loc[i] = pnewl_loc[i] - qnewl_loc[i];
	}
	i1 = n_update_external - 1;
	for (i = 0; i <= i1; ++i) {
	    qnewl_loc[extern_index[i]] = pnewl_loc[extern_index[i]
		    ] - qnewl_loc[extern_index[i]];
	}

/*   Calculate ALPHA: */

	pq = 0.;
	i1 = n_update - 1;
	for (i = 0; i <= i1; ++i) {
	    pq += pnewl_loc[i] * qnewl_loc[i];
	}
        MPI_Allreduce(&pq, &pqs, 1, MPI_DOUBLE_PRECISION, MPI_SUM, 
                      MPI_COMM_WORLD);
	alpha = s2s / pqs;

/*  Update  X( ): */

	i1 = n_update - 1;
	for (i = 0; i <= i1; ++i) {
	    xnewl_loc[i] += alpha * pnewl_loc[i];
	}
	i1 = n_update_external - 1;
	for (i = 0; i <= i1; ++i) {
	    xnewl_loc[extern_index[i]] += alpha * pnewl_loc[
		    extern_index[i]];
	}


/*  Calculate the residual: */
/* ---------------------------- */
/*      if (ikcg / 1 * 1 == ikcg) {  */ 
/*	if (ikcg / 5 * 5 == ikcg) {  */
  	if (ikcg / 10 * 10 == ikcg) { 
	    sumr = 0.;
	    sumr = 0.;
	    i1 = n_update - 1;
	    for (i = 0; i <= i1; ++i) {
		sum = val[i] * xnewl_loc[i] - rhsm_loc[i];
		i3 = bindx[i + 1] - 1;
		for (j = bindx[i]; j <= i3; ++j) {
		    sum += val[j] * xnewl_loc[bindx[j]];
		}
		d1 = sum;
		sumr += d1 * d1;
	    }
            MPI_Allreduce(&sumr, &sumrt, 1, MPI_DOUBLE_PRECISION, MPI_SUM, 
                      MPI_COMM_WORLD);
	    sumrt = sqrt(sumrt);
	    if (me == 0) {
                printf("\n  IKCG= %d   RESIDUAL= %.9g     RESIDUAL/Res0= %.9g",
                ikcg, sumrt, d1 = sumrt / res0);
	    }
            if(ierr == 1 ) {   /* Calculate the l2 error and errmax  */

/*  Get XNEW( ) - the global vector of solution */
	    i1 = n - 1;
	    for (i = 0; i <= i1; ++i) {
		xnewl[i] = 0.;
	    }
	    i1 = n_update - 1;
	    for (i = 0; i <= i1; ++i) {
		xnewl[update[i]] = xnewl_loc[update_index[i]];
	    }

            MPI_Reduce(xnewl, xnew, n, MPI_DOUBLE_PRECISION, MPI_SUM, 
                      0, MPI_COMM_WORLD);

	    if (me == 0) {
		ipr=0;    
		switch (example_no) {
		    case 20:  goto L330;
		    case 26:  goto L336;
		    case 27:  goto L337;
		    case 28:  goto L338;
		    case 29:  goto L339;
                    case 31:  goto L340;
                    case 32:  goto L341;
                    default : goto L380;
		}
L330:
		example20_err(xnew, err, &global_1.nx, &ny, &sumerr, &ipr);
		if (ikcg == 5 || ikcg == 10) {
                    fprintf(f9,"\n    Results for example 20:");
		    fprintf(f9,"\n    from Erlangga, Vuik, Oosterlee, 2004 paper: \n");
		}
		goto L370;
L336:
		example26_err(xnew, err, &global_1.nx, &ny, &sumerr1, &
			sumerr2, &kn1, &kn2, &kn3, &order, &ipr);
		if (ikcg == 5 || ikcg == 10) {
                    fprintf(f9,"\n    Results for example 26:");
		    fprintf(f9,"\n    from Erlangga & Turkel, 2011 paper \n");
		}
		goto L370;
L337:
		example27_err(xnew, err, &global_1.nx, &ny, &sumerr1, &
			sumerr2, &kn1, &kn2, &kn3, &order, &ipr);
		if (ikcg == 5 || ikcg == 10) {
                    fprintf(f9,"\n    Results for example 27:");
		    fprintf(f9,"\n    from Erlangga & Turkel, 2011 paper \n");
		}
		goto L370;
L338:
		example28_err(xnew, err, &global_1.nx, &ny, &sumerr, &kn1, &order, &ipr);
		if (ikcg == 5 || ikcg == 10) {
                    fprintf(f9,"\n    Results for example 28:\n");
		}
		goto L370;
L339:
		example29_err(xnew, err, &global_1.nx, &ny, &nz, &sumerr1, &sumerr2, 
                &kn1, &kn2, &kn3, &order, &ipr);
		if (ikcg == 5 || ikcg == 10) {
                    fprintf(f9,"\n    Results for example 29:\n");
                }
                goto L370;
L340:
                if(icase == 1)
                example31_err_case_a(xnew, err, &global_1.nx, &ny, &sumerr, &mn, &order, &ipr);
                if(icase == 2)
                example31_err_case_b(xnew, err, &global_1.nx, &ny, &sumerr, &mn, &order, &ipr, &a2);
                if(icase == 3)
                example31_err_case_c(xnew, err, &global_1.nx, &ny, &sumerr, &a2, &b2, &c2, &order, &ipr);
                if (ikcg == 5 || ikcg == 10) {
                      fprintf(f9,"\n    Results for example 31:\n");
                }
                goto L370;
L341:
                if(icase == 1)
                example32_err_case_a(xnew, err, &global_1.nx, &ny, &nz, &sumerr, &l_a, &n_a, &order, &ipr);
                if(icase == 3)
                example32_err_case_c(xnew, err, &global_1.nx, &ny, &nz, &sumerr, &a2, &b2, &c2,
                      &gama, &order, &ipr);
                if (ikcg == 5 || ikcg == 10) {
                      fprintf(f9,"\n    Results for example 32:\n");
                }
L370:

/*  Calculate ERR_L2 */

		sum = 0.;
		i1 = n - 1;
		for (i = 0; i <= i1; ++i) {
		    d1 = err[i];
		    sum += d1 * d1;
		}
		if (example_no != 26 && example_no != 27 && example_no != 29) {
		    err_l2 = sqrt(sum) / sumerr;
		}
		if (example_no == 26 || example_no == 27 || example_no == 29) {
		    d1 = sumerr1;
		    d2 = sumerr2;
		    err_l2 = sqrt(sum) / sqrt(d1 * d1 + d2 * d2);
		}

/*  Calculate the maximum error */

		jk = 0;
		rmax = fabs(err[0]);
		i1 = n - 1;
		for (i = 1; i <= i1; ++i) {
		    if ((d1 = err[i], fabs(d1)) > rmax) {
			jk = i;
			rmax = (d1 = err[i], fabs(d1));
		    }
		}
  		fprintf(f9," I= %d    ERR_L2= %.9g    Err_max= %.9g \n", ikcg, err_l2, rmax);    

	     }  /* endif me = 0   */
            }   /* endif ierr=1   */
L380:
	    if (sumrt < residmax) {
		goto L150;
	    }
	}
    }
/*  End of CGMN/CARP-CG main loop */

L150:


/*  Get XNEW( ) - the global vector of solution */

    if(example_no == 30) goto L151;   

    i2 = n - 1;
    for (i = 0; i <= i2; ++i) {
	xnewl[i] = 0.;
    }
    i2 = n_update - 1;
    for (i = 0; i <= i2; ++i) {
	xnewl[update[i]] = xnewl_loc[update_index[i]];
    }

    MPI_Reduce(xnewl, xnew,  n, MPI_DOUBLE_PRECISION, MPI_SUM, 
                      0, MPI_COMM_WORLD);

/*  print the solution - xnew on file xnew.dat */
/*  if(me ==0) {
        fp= fopen("xnew.dat","w");
        if(fp==0) {
         printf("\n error opening xnew.dat"); exit;
        } 

        printf("\n");   
        for (ii=0; ii<=n-1; ii+=4) 
	   { for (jj=ii; jj<=ii+3 & jj<= n-1; jj++)
	        fprintf(fp,"%10d  %13.6e", jj, xnew[jj]) ;    
		fprintf(fp,"\n") ;
	   }
    } */  /* endif print me=0 */

/*    get the runtime */

L151:
    t2 = MPI_Wtime();
    t21 = t2 - t1;
/*  printf("\n  %s","   "); */
    printf("\n  me= %d   TIME= %.4f   T1= %.4f   T2= %.4f", me, t21, t1, t2);

/*  Get the final L2 norm of the residual: */
    if(example_no ==30) goto L615;

    i2 = n - 1;
    for (i = 0; i <= i2; ++i) {
	resid1[i] = 0.;
    }
    sumr = 0.;
    i2 = n_update - 1;
    for (i = 0; i <= i2; ++i) {
	sum = val[i] * xnewl_loc[i] - rhsm_loc[i];
	i1 = bindx[i + 1] - 1;
	for (j = bindx[i]; j <= i1; ++j) {
	    sum += val[j] * xnewl_loc[bindx[j]];
	}
	d1 = sum;
	sumr += d1 * d1;
	resid1[update_inv[i]] = -sum;
    }
    MPI_Allreduce(&sumr, &sumrt, 1, MPI_DOUBLE_PRECISION, MPI_SUM, 
                      MPI_COMM_WORLD);

/*  Get RESID( ) - the global residual vector: */

    MPI_Reduce(resid1, resid,  n, MPI_DOUBLE_PRECISION, MPI_SUM, 
                      0, MPI_COMM_WORLD);
    sumrt = sqrt(sumrt);


/*  Calculate the L2 norm of the error */

    if (me == 0) {

/*  Calculate the maximum residual: */
	ik = 0;
	rxmax = resid[0];
	i2 = n - 1;
	for (i = 1; i <= i2; ++i) {
	    if (resid[i] > rxmax) {
		ik = i;
		rxmax = resid[i];
	    }
	}

/*  Calculate the L2 norm of the error and the maximum error when */
/*  the analytic solution is known, and print out information */
/*  about the test case: */
        ipr= 1;
	switch (example_no) {
	    case 20:  goto L30;
	    case 21:  goto L131;
	    case 23:  goto L133;
	    case 24:  goto L134;
	    case 25:  goto L135;
	    case 26:  goto L136;
	    case 27:  goto L137;
	    case 28:  goto L138;
	    case 29:  goto L139;
	    case 30:  goto L140;
            case 31:  goto L141;
            case 32:  goto L142;
            default : goto L70;
	}
L30:
	example20_err(xnew, err, &global_1.nx, &ny, &sumerr,&ipr);
	goto L70;
L131:
	example21_err(xnew, err, &global_1.nx, &ny, &sumerr);
	goto L70;
L133:
	example23_err(xnew, err, &global_1.nx, &ny, &sumerr, &order);
	goto L70;
L134:
	example24_err(&order);
	goto L70;
L135:
	example25_err(&order);
	goto L70;
L136:
	example26_err(xnew, err, &global_1.nx, &ny, &sumerr1, &sumerr2, &
		kn1, &kn2, &kn3, &order, &ipr);
	goto L70;
L137:
	example27_err(xnew, err, &global_1.nx, &ny, &sumerr1, &sumerr2, &
		kn1, &kn2, &kn3, &order, &ipr);
	goto L70;
L138:
	example28_err(xnew, err, &global_1.nx, &ny, &sumerr, &kn1, &order, &ipr);
	goto L70;
L139:
	example29_err(xnew, err, &global_1.nx, &ny, &nz, &sumerr1, &sumerr2, 
        &kn1, &kn2, &kn3, &order, &ipr);
	goto L70;
L140:
	example30_err(&order);
	goto L70;
L141:
        if(icase == 1)
        example31_err_case_a(xnew, err, &global_1.nx, &ny, &sumerr, &mn, &order, &ipr);
        if(icase == 2)
        example31_err_case_b(xnew, err, &global_1.nx, &ny, &sumerr, &mn, &order, &ipr, &a2);
        if(icase == 3)
        example31_err_case_c(xnew, err, &global_1.nx, &ny, &sumerr, &a2, &b2, &c2, &order, &ipr);
	goto L70;
L142:
        if(icase == 1)
        example32_err_case_a(xnew, err, &global_1.nx, &ny, &nz, &sumerr, &l_a, &n_a, &order, &ipr);
        if(icase == 3)
        example32_err_case_c(xnew, err, &global_1.nx, &ny, &nz, &sumerr, &a2, &b2, &c2,
              &gama, &order, &ipr);

/* Note: the true (analytic) solution is unknown for  examples 16,17,18,19,21,22,23,24,25,30 */ 

L70:

/*  Calculate the L2 norm of the error and the maximum error */
/*  when the true solution is known */

	if (example_no < 16 || example_no == 20 || example_no == 26 || 
		example_no == 27 || example_no == 28|| example_no == 29 ||
                  example_no == 31 || example_no == 32) {
	    if (example_no == 26 || example_no == 27||example_no == 29) {
		goto L605;
	    }
	    sum = 0.;
	    i2 = n - 1;
	    for (i = 0; i <= i2; ++i) {
		d1 = err[i];
		sum += d1 * d1;
	    }
	    err_l2 = sqrt(sum) / sumerr;
	    goto L610;
L605:
	    sum1 = 0.;
	    sum2 = 0.;
	    i2 = n - 1;
	    for (i = 0; i <= i2; i += 2) {
		d1 = err[i];
		sum1 += d1 * d1;
		d1 = err[i + 1];
		sum2 += d1 * d1;
	    }
	    d1 = sumerr1;
	    d2 = sumerr2;
	    err_l2 = sqrt(sum1 + sum2) / sqrt(d1 * d1 + d2 * d2);
	    printf("\n  ERR_L2_real = %.8f   ERR_L2_imag = %.8f", sqrt(sum1)/sumerr1, sqrt(sum2)/sumerr2);
L610:

/*  Calculate the maximum error */

	    jk = 0;
	    rmax = fabs(err[0]);
	    i2 = n - 1;
	    for (i = 1; i <= i2; ++i) {
		if ((d1 = err[i], fabs(d1)) > rmax) {
		    jk = i;
		    rmax = (d1 = err[i], fabs(d1));
		}
	    }
	} /* end of error calculation and print */

	printf("\n  RESIDUAL= %.10g ", sumrt);
	if (example_no < 16 || example_no == 20 || example_no == 26 || 
		example_no == 27 || example_no == 28|| example_no == 29 ||
                   example_no == 31 || example_no == 32) {
	    printf("\n  ERR_L2= %.10g ", err_l2);
	}
	printf("\n  Res_max= %.10g   Res_max_Index= %d", rxmax, ik);
	if (example_no < 16 || example_no == 20 || example_no == 26 || 
		example_no == 27 || example_no == 28 || example_no == 29 ||
                   example_no == 31 || example_no == 32) {
            printf("\n  Err_max= %.10g   Err_max_Index= %d", rmax, jk);
	}
    }
/* endif  me=0 */
L615: 
    MPI_Finalize(); 
    exit (0);
    return 0; 
} /* MAIN */

/* ******************************************************************** */
/* ******************************************************************** */

/* Subroutine */ int bound_3d(int *nx, int *ny, int *nz, 
	int *nlocal, int *nxst, int *nxend, int *nyst, 
	int *nyend, int *nzst, int *nzend, int *nxproc, 
	int *nyproc, int *nzproc, int *me)
{
    static int n, i1, i2, i3, i4, nremx, nremy, nremz, nlocalx, nlocaly, 
	    nlocalz, nyzproc;

/*  Purpose: Define the starting and ending grid point indices handled   
    by the current processor, for a 3D problem. */

    n = *nx * *ny * *nz;
    nlocalx = *nx / *nxproc;
    nlocaly = *ny / *nyproc;
    nlocalz = *nz / *nzproc;
    nyzproc = *nyproc * *nzproc;
    i1 = *me / nyzproc;
    i2 = *me % nyzproc;
    i3 = i2 / *nzproc;
    i4 = i2 % *nzproc;
    *nxst = nlocalx * i1 + 1;
    *nxend = *nxst + nlocalx - 1;
    *nyst = nlocaly * i3 + 1;
    *nyend = *nyst + nlocaly - 1;
    *nzst = nlocalz * i4 + 1;
    *nzend = *nzst + nlocalz - 1;
    nremx = *nx % *nxproc;
    if (nremx != 0) {
	*nxst += min(nremx,i1);
	*nxend = *nxst + nlocalx - 1;
	if (i1 < nremx) {
	    ++(*nxend);
	}
    }
    nremy = *ny % *nyproc;
    if (nremy != 0) {
	*nyst += min(nremy,i3);
	*nyend = *nyst + nlocaly - 1;
	if (i3 < nremy) {
	    ++(*nyend);
	}
    }
    nremz = *nz % *nzproc;
    if (nremz != 0) {
	*nzst += min(nremz,i4);
	*nzend = *nzst + nlocalz - 1;
	if (i4 < nremz) {
	    ++(*nzend);
	}
    }
    *nlocal = (*nzend - *nzst + 1) * (*nyend - *nyst + 1) * (*nxend - *nxst + 
	    1);
    return 0;
} /* bound_3d */

/* ******************************************************************** */
/* ******************************************************************** */

/* Subroutine */ int bound_2d(int *nx, int *ny, int *nlocal, 
	int *nxst, int *nxend, int *nyst, int *nyend, int 
	*nxproc, int *nyproc, int *me)
{
    static int n, i1, i2, nremx, nremy, nlocalx, nlocaly;

/*  Purpose: Define the starting and ending grid point indices handled   
    by the current processor, for a 2D problem. */

    n = *nx * *ny;
    nlocalx = *nx / *nxproc;
    nlocaly = *ny / *nyproc;
    i1 = *me / *nyproc;
    i2 = *me % *nyproc;
    *nxst = nlocalx * i1 + 1;
    *nxend = *nxst + nlocalx - 1;
    *nyst = nlocaly * i2 + 1;
    *nyend = *nyst + nlocaly - 1;
    nremx = *nx % *nxproc;
    if (nremx != 0) {
	*nxst += min(nremx,i1);
	*nxend = *nxst + nlocalx - 1;
	if (i1 < nremx) {
	    ++(*nxend);
	}
    }
    nremy = *ny % *nyproc;
    if (nremy != 0) {
	*nyst += min(nremy,i2);
	*nyend = *nyst + nlocaly - 1;
	if (i2 < nremy) {
	    ++(*nyend);
	}
    }
    *nlocal = (*nyend - *nyst + 1) * (*nxend - *nxst + 1);
    return 0;
} /* bound_2d */

/* ******************************************************************** */
/* ******************************************************************** */
/* Subroutine */ int bound_1d(int *nx, int *nlocal, int *nxst, 
	int *nxend, int *nxproc, int *me)
{
    static int n, i1, nremx, nlocalx;

/*  Purpose: Define the starting and ending grid point indices handled   
    by the current processor, for a 1D problem. */

    n = *nx;
    nlocalx = *nx / *nxproc;
    i1 = *me;
    *nxst = nlocalx * i1 + 1;
    *nxend = *nxst + nlocalx - 1;
    nremx = *nx % *nxproc;
    if (nremx != 0) {
	*nxst += min(nremx,i1);
	*nxend = *nxst + nlocalx - 1;
	if (i1 < nremx) {
	    ++(*nxend);
	}
    }
    *nlocal = *nxend - *nxst + 1;
    return 0;
} /* bound_1d */

/* ******************************************************************** */
/* ******************************************************************** */

/* Subroutine */ int example_3d(int *ist, int *iend, int *jst, 
	int *jend, int *kst, int *kend, int *nx, int *ny, 
	int *nz, int *index_s_xm, int *index_s_xp, int *
	index_s_ym, int *index_s_yp, int *index_s_zm, int *
	index_s_zp, int *index_r_xm, int *index_r_xp, int *
	index_r_ym, int *index_r_yp, int *index_r_zm, int *
	index_r_zp, int *ixp, int *ixm, int *iyp, int *iym, 
	int *izp, int *izm)
{
    int i1, i2, i3;
    static int nxylocal, i, j, k, ii1, ii2, ii3, ii4, jcoef_ii_1, 
	    jcoef_ii_2, jcoef_ii_3, jcoef_ii_4, jcoef_ii_5, 
	    jcoef_ii_6, jcoef_ii_7, ii, igl, nxy, nxlocal, nylocal;

/*  Purpose:  Determine the indices of the external boundary grid points   
    comunicating with the neighboring  subdomains (3D, 7pt problems) */
/* ----------------------------------------------------------------------- */

    /* Parameter adjustments */
    --index_r_zp;
    --index_r_zm;
    --index_r_yp;
    --index_r_ym;
    --index_r_xp;
    --index_r_xm;
    --index_s_zp;
    --index_s_zm;
    --index_s_yp;
    --index_s_ym;
    --index_s_xp;
    --index_s_xm;

    /* Function Body */
    nxy = *nx * *ny;
    nxlocal = *iend - *ist + 1;
    nylocal = *jend - *jst + 1;
    nxylocal = nxlocal * nylocal;
    *ixm = 0;
    *ixp = 0;
    *iym = 0;
    *iyp = 0;
    *izm = 0;
    *izp = 0;
    ii = 0;
    i1 = *kend;
    for (k = *kst; k <= i1; ++k) {
	i2 = *jend;
	for (j = *jst; j <= i2; ++j) {
	    i3 = *iend;
	    for (i = *ist; i <= i3; ++i) {
		++ii;
		ii1 = (ii - 1) / nxylocal;
		ii2 = (ii - 1) % nxylocal;
		ii3 = ii2 / nxlocal;
		ii4 = ii2 % nxlocal;
		igl = (*kst - 1 + ii1) * nxy + *nx * (*jst - 1 + ii3) + *ist + 
			ii4;
		jcoef_ii_1 = igl;
		jcoef_ii_2 = igl + 1;
		jcoef_ii_3 = igl + *nx;
		jcoef_ii_4 = igl - 1;
		jcoef_ii_5 = igl - *nx;
		jcoef_ii_6 = igl + nxy;
		jcoef_ii_7 = igl - nxy;

		if (i == *ist && i != 1) {
		    ++(*ixm);
		    index_s_xm[*ixm] = jcoef_ii_4;
		    index_r_xm[*ixm] = jcoef_ii_1;
		}
		if (i == *iend && i != *nx) {
		    ++(*ixp);
		    index_s_xp[*ixp] = jcoef_ii_2;
		    index_r_xp[*ixp] = jcoef_ii_1;
		}
		if (j == *jst && j != 1) {
		    ++(*iym);
		    index_s_ym[*iym] = jcoef_ii_5;
		    index_r_ym[*iym] = jcoef_ii_1;
		}
		if (j == *jend && j != *ny) {
		    ++(*iyp);
		    index_s_yp[*iyp] = jcoef_ii_3;
		    index_r_yp[*iyp] = jcoef_ii_1;
		}
		if (k == *kst && k != 1) {
		    ++(*izm);
		    index_s_zm[*izm] = jcoef_ii_7;
		    index_r_zm[*izm] = jcoef_ii_1;
		}
		if (k == *kend && k != *nz) {
		    ++(*izp);
		    index_s_zp[*izp] = jcoef_ii_6;
		    index_r_zp[*izp] = jcoef_ii_1;
		}
	    }
	}
    }
    return 0;
} /* example_3d */

/* ********************************************************************* */
/* ******************************************************************** */

/* Subroutine */ int example_3d_mod(int *ist, int *iend, int *jst, 
	int *jend, int *kst, int *kend, int *nx, int *ny, 
	int *nz, int *index_s_xm, int *index_s_xp, int *
	index_s_ym, int *index_s_yp, int *index_s_zm, int *
	index_s_zp, int *index_r_xm, int *index_r_xp, int *
	index_r_ym, int *index_r_yp, int *index_r_zm, int *
	index_r_zp, int *ixp, int *ixm, int *iyp, int *iym, 
	int *izp, int *izm, int *nx_st, int *nx_end, int *ny_st, int *ny_end,
        int *nz_st, int *nz_end)
{
    int i1, i2, i3;
    static int nxylocal, i, j, k, ii1, ii2, ii3, ii4, jcoef_ii_1, 
	    jcoef_ii_2, jcoef_ii_3, jcoef_ii_4, jcoef_ii_5, 
	    jcoef_ii_6, jcoef_ii_7, ii, igl, nxy, nxlocal, nylocal;

/*  Purpose:  Determine the indices of the external boundary grid points   
    comunicating with the neighboring  subdomains (3D, 7pt problems) */
/* ----------------------------------------------------------------------- */

    /* Parameter adjustments */
    --index_r_zp;
    --index_r_zm;
    --index_r_yp;
    --index_r_ym;
    --index_r_xp;
    --index_r_xm;
    --index_s_zp;
    --index_s_zm;
    --index_s_yp;
    --index_s_ym;
    --index_s_xp;
    --index_s_xm;

    /* Function Body */
/*  nxy = *nx * *ny; */
    nxy = (*nx + 2) * (*ny + 2);
    nxlocal = *iend - *ist + 1;
    nylocal = *jend - *jst + 1;
    nxylocal = nxlocal * nylocal;
    *ixm = 0;
    *ixp = 0;
    *iym = 0;
    *iyp = 0;
    *izm = 0;
    *izp = 0;
    ii = 0;
    i1 = *kend;
    for (k = *kst; k <= i1; ++k) {
	i2 = *jend;
	for (j = *jst; j <= i2; ++j) {
	    i3 = *iend;
	    for (i = *ist; i <= i3; ++i) {
		++ii;
		ii1 = (ii - 1) / nxylocal;
		ii2 = (ii - 1) % nxylocal;
		ii3 = ii2 / nxlocal;
		ii4 = ii2 % nxlocal;
/*		igl = (*kst - 1 + ii1) * nxy + *nx * (*jst - 1 + ii3) + *ist + ii4; */
		igl = (*kst - 1 + ii1) * nxy + (*nx + 2) * (*jst - 1 + ii3) + *ist + ii4;
		jcoef_ii_1 = igl;
		jcoef_ii_2 = igl + 1;
/*		jcoef_ii_3 = igl + *nx;  */
		jcoef_ii_3 = igl + (*nx + 2);
		jcoef_ii_4 = igl - 1;
/*		jcoef_ii_5 = igl - *nx; */
		jcoef_ii_5 = igl - (*nx + 2);
		jcoef_ii_6 = igl + nxy;
		jcoef_ii_7 = igl - nxy;

		if (i == *ist && i > 1  && (j > *ny_st && j < *ny_end) && (k > *nz_st && k < *nz_end)) {
		    ++(*ixm);
		    index_s_xm[*ixm] = jcoef_ii_4;
		    index_r_xm[*ixm] = jcoef_ii_1;
		}
		if (i == *iend && i <  *nx + 2  && (j > *ny_st && j < *ny_end) && (k > *nz_st && k < *nz_end)) {
		    ++(*ixp);
		    index_s_xp[*ixp] = jcoef_ii_2;
		    index_r_xp[*ixp] = jcoef_ii_1;
		}
		if (j == *jst && j > 1 && (i > *nx_st && i < *nx_end) && (k > *nz_st && k < *nz_end)) {
		    ++(*iym);
		    index_s_ym[*iym] = jcoef_ii_5;
		    index_r_ym[*iym] = jcoef_ii_1;
		}
		if (j == *jend && j < *ny + 2  && (i > *nx_st && i < *nx_end) && (k > *nz_st && k < *nz_end)) {
		    ++(*iyp);
		    index_s_yp[*iyp] = jcoef_ii_3;
		    index_r_yp[*iyp] = jcoef_ii_1;
		}
		if (k == *kst && k > 1 && (i > *nx_st && i < *nx_end) && (j >*ny_st && j < *ny_end)) {
		    ++(*izm);
		    index_s_zm[*izm] = jcoef_ii_7;
		    index_r_zm[*izm] = jcoef_ii_1;
		}
		if (k == *kend && k < *nz + 2 && (i > *nx_st && i < *nx_end) && (j >*ny_st && j < *ny_end)) {
		    ++(*izp);
		    index_s_zp[*izp] = jcoef_ii_6;
		    index_r_zp[*izp] = jcoef_ii_1;
		}
	    }
	}
    }
    return 0;
} /* example_3d_mod */

/* ********************************************************************* */
/* ******************************************************************** */

/* Subroutine */ int example_2d(int *ist, int *iend, int *jst, 
	int *jend, int *nx, int *ny, int *index_s_xm, 
	int *index_s_xp, int *index_s_ym, int *index_s_yp, 
	int *index_r_xm, int *index_r_xp, int *index_r_ym, 
	int *index_r_yp, int *ixp, int *ixm, int *iyp, 
	int *iym)
{
    int i1, i2;
    static int nxylocal, i, j, ii1, ii2, ii3, ii4, jcoef_ii_1, 
	    jcoef_ii_2, jcoef_ii_3, jcoef_ii_4, jcoef_ii_5, ii, igl, 
	    nxy, nxlocal, nylocal;

/* Purpose: Determine the indices of the external boundary grid points   
   comunicating with the neighboring  subdomains (2D, 5pt problems) */
/* ----------------------------------------------------------------------- */

    /* Parameter adjustments */
    --index_r_yp;
    --index_r_ym;
    --index_r_xp;
    --index_r_xm;
    --index_s_yp;
    --index_s_ym;
    --index_s_xp;
    --index_s_xm;

    /* Function Body */
    nxy = *nx * *ny;
    nxlocal = *iend - *ist + 1;
    nylocal = *jend - *jst + 1;
    nxylocal = nxlocal * nylocal;
    *ixm = 0;
    *ixp = 0;
    *iym = 0;
    *iyp = 0;
    ii = 0;
    i1 = *jend;
    for (j = *jst; j <= i1; ++j) {
	i2 = *iend;
	for (i = *ist; i <= i2; ++i) {
	    ++ii;
	    ii1 = (ii - 1) / nxylocal;
	    ii2 = (ii - 1) % nxylocal;
	    ii3 = ii2 / nxlocal;
	    ii4 = ii2 % nxlocal;
	    igl = *nx * (*jst - 1 + ii3) + *ist + ii4;
	    jcoef_ii_1 = igl;
	    jcoef_ii_2 = igl + 1;
	    jcoef_ii_3 = igl + *nx;
	    jcoef_ii_4 = igl - 1;
	    jcoef_ii_5 = igl - *nx;

	    if (i == *ist && i != 1) {
		++(*ixm);
		index_s_xm[*ixm] = jcoef_ii_4;
		index_r_xm[*ixm] = jcoef_ii_1;
	    }
	    if (i == *iend && i != *nx) {
		++(*ixp);
		index_s_xp[*ixp] = jcoef_ii_2;
		index_r_xp[*ixp] = jcoef_ii_1;
	    }
	    if (j == *jst && j != 1) {
		++(*iym);
		index_s_ym[*iym] = jcoef_ii_5;
		index_r_ym[*iym] = jcoef_ii_1;
	    }
	    if (j == *jend && j != *ny) {
		++(*iyp);
		index_s_yp[*iyp] = jcoef_ii_3;
		index_r_yp[*iyp] = jcoef_ii_1;
	    }
	}
    }
    return 0;
} /* example_2d */

/* ******************************************************************** */
/* ******************************************************************** */

/* Subroutine */ int example_2d_all(int *ist, int *iend, int *
	jst, int *jend, int *nx, int *ny, int *index_s_xm, 
	int *index_s_xp, int *index_s_ym, int *index_s_yp, 
	int *index_r_xm, int *index_r_xp, int *index_r_ym, 
	int *index_r_yp, int *ixp, int *ixm, int *iyp, 
	int *iym, int *nx_st, int *nx_end, 
	int *ny_st, int *ny_end)
{
    int i1, i2;
    static int nxylocal, i, j, ii1, ii2, ii3, ii4, jcoef_ii_1, 
	    jcoef_ii_2, jcoef_ii_3, jcoef_ii_4, jcoef_ii_5, ii, igl, 
	    nxy, nxlocal, nylocal;

/* Purpose: Determine the indices of the external boundary grid points   
   comunicating with the neighboring  subdomains (2D, 5pt problems) */
/* ----------------------------------------------------------------------- */

    /* Parameter adjustments */
    --index_r_yp;
    --index_r_ym;
    --index_r_xp;
    --index_r_xm;
    --index_s_yp;
    --index_s_ym;
    --index_s_xp;
    --index_s_xm;

    /* Function Body */
    nxy = *nx * *ny;
    nxlocal = *iend - *ist + 1;
    nylocal = *jend - *jst + 1;
    nxylocal = nxlocal * nylocal;
    *ixm = 0;
    *ixp = 0;
    *iym = 0;
    *iyp = 0;
    ii = 0;
    i1 = *jend;
    for (j = *jst; j <= i1; ++j) {
	i2 = *iend;
	for (i = *ist; i <= i2; ++i) {
	    ++ii;
	    ii1 = (ii - 1) / nxylocal;
	    ii2 = (ii - 1) % nxylocal;
	    ii3 = ii2 / nxlocal;
	    ii4 = ii2 % nxlocal;
	    igl = *nx * (*jst - 1 + ii3) + *ist + ii4;
	    jcoef_ii_1 = igl;
	    jcoef_ii_2 = igl + 1;
	    jcoef_ii_3 = igl + *nx;
	    jcoef_ii_4 = igl - 1;
	    jcoef_ii_5 = igl - *nx;

	    if (i == *ist && i != 1 && (j > *ny_st && j < *ny_end)) {
		++(*ixm);
		index_s_xm[*ixm] = jcoef_ii_4;
		index_r_xm[*ixm] = jcoef_ii_1;
	    }
	    if (i == *iend && i != *nx && (j > *ny_st && j < *ny_end))
		     {
		++(*ixp);
		index_s_xp[*ixp] = jcoef_ii_2;
		index_r_xp[*ixp] = jcoef_ii_1;
	    }
	    if (j == *jst && j != 1 && (i > *nx_st && i < *nx_end)) {
		++(*iym);
		index_s_ym[*iym] = jcoef_ii_5;
		index_r_ym[*iym] = jcoef_ii_1;
	    }
	    if (j == *jend && j != *ny && (i > *nx_st && i < *nx_end))
		     {
		++(*iyp);
		index_s_yp[*iyp] = jcoef_ii_3;
		index_r_yp[*iyp] = jcoef_ii_1;
	    }
	}
    }
    return 0;
} /* example_2d_all */

/* ******************************************************************** */
/* ******************************************************************** */

/* Subroutine */ int example_3d_img_all(int *ist, int *iend, int *
	jst, int *jend, int *kst, int *kend, int *nx, int 
	*ny, int *nz, int *index_s_xm, int *index_s_xp, 
	int *index_s_ym, int *index_s_yp, int *index_s_zm, 
	int *index_s_zp, int *index_r_xm, int *index_r_xp, 
	int *index_r_ym, int *index_r_yp, int *index_r_zm, 
	int *index_r_zp, int *ixp, int *ixm, int *iyp, 
	int *iym, int *izp, int *izm,
        int *nx_st, int *nx_end, int *ny_st, int *ny_end, int *nz_st, int *nz_end ,int *me)
{
    int i1, i2, i3;
    static int nxylocal, i, j, k, ii1, ii2, ii3, ii4, jcoef_ii_1, 
	    jcoef_ii_2, jcoef_ii_3, jcoef_ii_4, jcoef_ii_5, 
	    jcoef_ii_6, jcoef_ii_7, ii, igl, nxy, nxy2, nxlocal, nylocal;

/*  Purpose:  Determine the indices of the external boundary grid points   
    comunicating with the neighboring subdomains (3D, 7 or 27 point schemes) */
/* ------------------------------------------------------------------------- */

    /* Parameter adjustments */
    --index_r_zp;
    --index_r_zm;
    --index_r_yp;
    --index_r_ym;
    --index_r_xp;
    --index_r_xm;
    --index_s_zp;
    --index_s_zm;
    --index_s_yp;
    --index_s_ym;
    --index_s_xp;
    --index_s_xm;

    /* Function Body */
    nxy = (*nx + 2) * (*ny + 2);
    nxy2 = 2*nxy;
    nxlocal = *iend - *ist + 1;
    nylocal = *jend - *jst + 1;
    nxylocal = nxlocal * nylocal;
/*  if(*me ==0) printf("\n nx+2= %8d  ny+2= %8d  nxy2= %8d", *nx+2, *ny+2, &nxy2);
    if(*me ==0) printf("\n ist=  %8d  iend=  %8d  jst=  %8d jend= %8d kst= %8d kst= %8d", *ist, *iend, *jst, *jend, *kst, *kend); */
    *ixm = 0;
    *ixp = 0;
    *iym = 0;
    *iyp = 0;
    *izm = 0;
    *izp = 0;
    ii = 0;
    i1 = *kend;
    for (k = *kst; k <= i1; ++k) {
	i2 = *jend;
	for (j = *jst; j <= i2; ++j) {
	    i3 = *iend;
	    for (i = *ist; i <= i3; ++i) {
		++ii;
		ii1 = (ii - 1) / nxylocal;
		ii2 = (ii - 1) % nxylocal;
		ii3 = ii2 / nxlocal;
		ii4 = ii2 % nxlocal;
		igl = 2*((*kst - 1 + ii1) * nxy + (*nx + 2) * (*jst - 1 + ii3) + 
			*ist + ii4 ) - 1;
		jcoef_ii_1 = igl;
		jcoef_ii_2 = igl + 2;
		jcoef_ii_3 = igl + 2*(*nx + 2);
		jcoef_ii_4 = igl - 2;
		jcoef_ii_5 = igl - 2*(*nx + 2);
		jcoef_ii_6 = igl + nxy2;
		jcoef_ii_7 = igl - nxy2;

/*		if (i == *ist && i > 1 && (j > 1 && j < *ny + 2) && (k > 
			1 && k < *nz + 2)) {  */
		if ((i == *ist && i > 1) && (j > *ny_st && j < *ny_end) && (k > 
			*nz_st && k < *nz_end)) {
		    ++(*ixm);
		    index_s_xm[*ixm] = jcoef_ii_4;
		    index_r_xm[*ixm] = jcoef_ii_1;
		    ++(*ixm);
		    index_s_xm[*ixm] = jcoef_ii_4 + 1;
		    index_r_xm[*ixm] = jcoef_ii_1 + 1;
		}
/*		if (i == *iend && i < *nx + 2 && (j > 1 && j < *ny + 2) &&
			 (k > 1 && k < *nz + 2)) {   */
		if ((i == *iend && i < *nx + 2) && (j > *ny_st && j < *ny_end ) &&
			 (k > *nz_st && k < *nz_end)) {
		    ++(*ixp);
		    index_s_xp[*ixp] = jcoef_ii_2;
		    index_r_xp[*ixp] = jcoef_ii_1;
		    ++(*ixp);
		    index_s_xp[*ixp] = jcoef_ii_2 + 1;
		    index_r_xp[*ixp] = jcoef_ii_1 + 1;
		}
/*		if (j == *jst && j > 1 && (i > 1 && i < *nx + 2) && (k > 
			1 && k < *nz + 2)) { */
		if ((j == *jst && j > 1) && (i > *nx_st && i < *nx_end ) && (k > 
			*nz_st && k < *nz_end)) {
		    ++(*iym);
		    index_s_ym[*iym] = jcoef_ii_5;
		    index_r_ym[*iym] = jcoef_ii_1;
		    ++(*iym);
		    index_s_ym[*iym] = jcoef_ii_5 + 1;
		    index_r_ym[*iym] = jcoef_ii_1 + 1;
		}
/*		if (j == *jend && j < *ny + 2 && (i > 1 && i < *nx + 2) &&
			 (k > 1 && k < *nz + 2)) {  */
		if ((j == *jend && j < *ny + 2) && (i > *nx_st && i < *nx_end) &&
			 (k > *nz_st && k < *nz_end)) {
		    ++(*iyp);
		    index_s_yp[*iyp] = jcoef_ii_3;
		    index_r_yp[*iyp] = jcoef_ii_1;
		    ++(*iyp);
		    index_s_yp[*iyp] = jcoef_ii_3 + 1;
		    index_r_yp[*iyp] = jcoef_ii_1 + 1;
		}
/*		if (k == *kst && k > 1 && (i > 1 && i < *nx + 2) && (j > 
			1 && j < *ny + 2)) {  */
		if ((k == *kst && k > 1) && (i > *nx_st && i < *nx_end) && (j > 
			*ny_st && j < *ny_end)) {
		    ++(*izm);
		    index_s_zm[*izm] = jcoef_ii_7;
		    index_r_zm[*izm] = jcoef_ii_1;
		    ++(*izm);
		    index_s_zm[*izm] = jcoef_ii_7 + 1;
		    index_r_zm[*izm] = jcoef_ii_1 + 1;
		}
/*		if (k == *kend && k < *nz && (i > 1 && i < *nx + 2) && (j 
			> 1 && j < *ny + 2)) {  */
		if ((k == *kend && k < *nz+2) && (i > *nx_st && i < *nx_end) && (j 
			> *ny_st && j < *ny_end)) {
		    ++(*izp);
		    index_s_zp[*izp] = jcoef_ii_6;
		    index_r_zp[*izp] = jcoef_ii_1;
       /*         if(*me ==0) {
                    printf("\n izp=  %8d  %8d %8d ", *izp, index_s_zp[*izp], index_r_zp[*izp]); } */
		    ++(*izp);
		    index_s_zp[*izp] = jcoef_ii_6 + 1;
		    index_r_zp[*izp] = jcoef_ii_1 + 1;
      /*          if(*me ==0) {
                    printf("\n izp=  %8d  %8d %8d ", *izp, index_s_zp[*izp], index_r_zp[*izp]); } */
		}
	    }
	}
    }
    return 0;
} /* example_3d_img_all */

/* ******************************************************************** */
/* ******************************************************************** */

/* Subroutine */ int example_3d_img(int *ist, int *iend, int *
	jst, int *jend, int *kst, int *kend, int *nx, int 
	*ny, int *nz, int *index_s_xm, int *index_s_xp, 
	int *index_s_ym, int *index_s_yp, int *index_s_zm, 
	int *index_s_zp, int *index_r_xm, int *index_r_xp, 
	int *index_r_ym, int *index_r_yp, int *index_r_zm, 
	int *index_r_zp, int *ixp, int *ixm, int *iyp, 
	int *iym, int *izp, int *izm)
{
    int i1, i2, i3;
    static int nxylocal, i, j, k, ii1, ii2, ii3, ii4, jcoef_ii_1, 
	    jcoef_ii_2, jcoef_ii_3, jcoef_ii_4, jcoef_ii_5, 
	    jcoef_ii_6, jcoef_ii_7, ii, igl, nxy, nxy2, nxlocal, nylocal;

/*  Purpose:  Determine the indices of the external boundary grid points   
    comunicating with the neighboring subdomains (3D, 7pt problems) */
/* ----------------------------------------------------------------------- */

    /* Parameter adjustments */
    --index_r_zp;
    --index_r_zm;
    --index_r_yp;
    --index_r_ym;
    --index_r_xp;
    --index_r_xm;
    --index_s_zp;
    --index_s_zm;
    --index_s_yp;
    --index_s_ym;
    --index_s_xp;
    --index_s_xm;

    /* Function Body */
    nxy = (*nx + 2) * (*ny + 2);
    nxy2 = 2*nxy;
    nxlocal = *iend - *ist + 1;
    nylocal = *jend - *jst + 1;
    nxylocal = nxlocal * nylocal;
    *ixm = 0;
    *ixp = 0;
    *iym = 0;
    *iyp = 0;
    *izm = 0;
    *izp = 0;
    ii = 0;
    i1 = *kend;
    for (k = *kst; k <= i1; ++k) {
	i2 = *jend;
	for (j = *jst; j <= i2; ++j) {
	    i3 = *iend;
	    for (i = *ist; i <= i3; ++i) {
		++ii;
		ii1 = (ii - 1) / nxylocal;
		ii2 = (ii - 1) % nxylocal;
		ii3 = ii2 / nxlocal;
		ii4 = ii2 % nxlocal;
		igl = 2*((*kst - 1 + ii1) * nxy + (*nx + 2) * (*jst - 1 + ii3) + *
			ist + ii4) - 1;
		jcoef_ii_1 = igl;
		jcoef_ii_2 = igl + 2;
		jcoef_ii_3 = igl + 2*(*nx + 2);
		jcoef_ii_4 = igl - 2;
		jcoef_ii_5 = igl - 2*(*nx + 2);
		jcoef_ii_6 = igl + nxy2;
		jcoef_ii_7 = igl - nxy2;

		if (i == *ist && i > 1 && (j > 1 && j < *ny + 2) && (k > 
			1 && k < *nz + 2)) {
		    ++(*ixm);
		    index_s_xm[*ixm] = jcoef_ii_4;
		    index_r_xm[*ixm] = jcoef_ii_1;
		    ++(*ixm);
		    index_s_xm[*ixm] = jcoef_ii_4 + 1;
		    index_r_xm[*ixm] = jcoef_ii_1 + 1;
		}
		if (i == *iend && i < *nx + 2 && (j > 1 && j < *ny + 2) &&
			 (k > 1 && k < *nz + 2)) {
		    ++(*ixp);
		    index_s_xp[*ixp] = jcoef_ii_2;
		    index_r_xp[*ixp] = jcoef_ii_1;
		    ++(*ixp);
		    index_s_xp[*ixp] = jcoef_ii_2 + 1;
		    index_r_xp[*ixp] = jcoef_ii_1 + 1;
		}
		if (j == *jst && j > 1 && (i > 1 && i < *nx + 2) && (k > 
			1 && k < *nz + 2)) {
		    ++(*iym);
		    index_s_ym[*iym] = jcoef_ii_5;
		    index_r_ym[*iym] = jcoef_ii_1;
		    ++(*iym);
		    index_s_ym[*iym] = jcoef_ii_5 + 1;
		    index_r_ym[*iym] = jcoef_ii_1 + 1;
		}
		if (j == *jend && j < *ny + 2 && (i > 1 && i < *nx + 2) &&
			 (k > 1 && k < *nz + 2)) {
		    ++(*iyp);
		    index_s_yp[*iyp] = jcoef_ii_3;
		    index_r_yp[*iyp] = jcoef_ii_1;
		    ++(*iyp);
		    index_s_yp[*iyp] = jcoef_ii_3 + 1;
		    index_r_yp[*iyp] = jcoef_ii_1 + 1;
		}
		if (k == *kst && k > 1 && (i > 1 && i < *nx + 2) && (j > 
			1 && j < *ny + 2)) {
		    ++(*izm);
		    index_s_zm[*izm] = jcoef_ii_7;
		    index_r_zm[*izm] = jcoef_ii_1;
		    ++(*izm);
		    index_s_zm[*izm] = jcoef_ii_7 + 1;
		    index_r_zm[*izm] = jcoef_ii_1 + 1;
		}
		if (k == *kend && k < *nz && (i > 1 && i < *nx + 2) && (j 
			> 1 && j < *ny + 2)) {
		    ++(*izp);
		    index_s_zp[*izp] = jcoef_ii_6;
		    index_r_zp[*izp] = jcoef_ii_1;
		    ++(*izp);
		    index_s_zp[*izp] = jcoef_ii_6 + 1;
		    index_r_zp[*izp] = jcoef_ii_1 + 1;
		}
	    }
	}
    }
    return 0;
} /* example_3d_img */

/* ********************************************************************* */
/* ******************************************************************** */

/* Subroutine */ int example_2d_img(int *ist, int *iend, int *
	jst, int *jend, int *nx, int *ny, int *index_s_xm, 
	int *index_s_xp, int *index_s_ym, int *index_s_yp, 
	int *index_r_xm, int *index_r_xp, int *index_r_ym, 
	int *index_r_yp, int *ixp, int *ixm, int *iyp, 
	int *iym)
{
    int i1, i2;
    static int nxylocal, i, j, ii2, ii3, ii4, jcoef_ii_1, jcoef_ii_2, 
	    jcoef_ii_3, jcoef_ii_4, jcoef_ii_5, ii, igl, nxlocal, 
	    nylocal;

/* Purpose: determine the indices of the external boundary grid points   
   comunicating with the neighboring subdomains (2D, 5pt problems)   
   for problems with an imaginary solution.   
   This routine follows Heroux paper using: (x1;y1, x2:y2,...,   ,xn;yn) */
/* ----------------------------------------------------------------------- */

    /* Parameter adjustments */
    --index_r_yp;
    --index_r_ym;
    --index_r_xp;
    --index_r_xm;
    --index_s_yp;
    --index_s_ym;
    --index_s_xp;
    --index_s_xm;

    /* Function Body */
    nxlocal = *iend - *ist + 1;
    nylocal = *jend - *jst + 1;
    nxylocal = nxlocal * nylocal;
    *ixm = 0;
    *ixp = 0;
    *iym = 0;
    *iyp = 0;
    ii = 0;
    i1 = *jend;
    for (j = *jst; j <= i1; ++j) {
	i2 = *iend;
	for (i = *ist; i <= i2; ++i) {
	    ++ii;
	    ii2 = ii - 1;
	    ii3 = ii2 / nxlocal;
	    ii4 = ii2 % nxlocal;
	    igl = 2*((*nx + 2) * (*jst - 1 + ii3) + *ist + ii4) - 1;
	    jcoef_ii_1 = igl;
	    jcoef_ii_2 = igl + 2;
	    jcoef_ii_3 = igl + 2*(*nx + 2);
	    jcoef_ii_4 = igl - 2;
	    jcoef_ii_5 = igl - 2*(*nx + 2);

	    if (i == *ist && i > 1 && (j > 1 && j < *ny + 2)) {
		++(*ixm);
		index_s_xm[*ixm] = jcoef_ii_4;
		index_r_xm[*ixm] = jcoef_ii_1;
		++(*ixm);
		index_s_xm[*ixm] = jcoef_ii_4 + 1;
		index_r_xm[*ixm] = jcoef_ii_1 + 1;
	    }
	    if (i == *iend && i < *nx + 2 && (j > 1 && j < *ny + 2)) {
		++(*ixp);
		index_s_xp[*ixp] = jcoef_ii_2;
		index_r_xp[*ixp] = jcoef_ii_1;
		++(*ixp);
		index_s_xp[*ixp] = jcoef_ii_2 + 1;
		index_r_xp[*ixp] = jcoef_ii_1 + 1;
	    }
	    if (j == *jst && j > 1 && (i > 1 && i < *nx + 2)) {
		++(*iym);
		index_s_ym[*iym] = jcoef_ii_5;
		index_r_ym[*iym] = jcoef_ii_1;
		++(*iym);
		index_s_ym[*iym] = jcoef_ii_5 + 1;
		index_r_ym[*iym] = jcoef_ii_1 + 1;
	    }
	    if (j == *jend && j < *ny + 2 && (i > 1 && i < *nx + 2)) {
		++(*iyp);
		index_s_yp[*iyp] = jcoef_ii_3;
		index_r_yp[*iyp] = jcoef_ii_1;
		++(*iyp);
		index_s_yp[*iyp] = jcoef_ii_3 + 1;
		index_r_yp[*iyp] = jcoef_ii_1 + 1;
	    }
	}
    }
    return 0;
} /* example_2d_img */

/* ********************************************************************* */
/* ******************************************************************** */

/* Subroutine */ int example_2d_img_all(int *ist, int *iend, 
	int *jst, int *jend, int *nx, int *ny, int *
	index_s_xm, int *index_s_xp, int *index_s_ym, int *
	index_s_yp, int *index_r_xm, int *index_r_xp, int *
	index_r_ym, int *index_r_yp, int *ixp, int *ixm, 
	int *iyp, int *iym, int *nx_st, int *
	nx_end, int *ny_st, int *ny_end)
{
    int i1, i2;
    static int nxylocal, i, j, ii2, ii3, ii4, jcoef_ii_1, jcoef_ii_2, 
	    jcoef_ii_3, jcoef_ii_4, jcoef_ii_5, ii, igl, nxlocal, 
	    nylocal;

/* Purpose:  Determine the indices of the external boundary grid points   
   comunicating with the neighboring subdomains (2D, 5 or 9 point schemes)   
   for problems with an imaginary solution.   
   This routine follows Heroux paper using: (x1;y1, x2:y2,...,   ,xn;yn) */
/* ----------------------------------------------------------------------- */

    /* Parameter adjustments */
    --index_r_yp;
    --index_r_ym;
    --index_r_xp;
    --index_r_xm;
    --index_s_yp;
    --index_s_ym;
    --index_s_xp;
    --index_s_xm;

    /* Function Body */
    nxlocal = *iend - *ist + 1;
    nylocal = *jend - *jst + 1;
    nxylocal = nxlocal * nylocal;
    *ixm = 0;
    *ixp = 0;
    *iym = 0;
    *iyp = 0;
    ii = 0;
    i1 = *jend;
    for (j = *jst; j <= i1; ++j) {
	i2 = *iend;
	for (i = *ist; i <= i2; ++i) {
	    ++ii;
	    ii2 = ii - 1;
	    ii3 = ii2 / nxlocal;
	    ii4 = ii2 % nxlocal;
	    igl = 2*((*nx + 2) * (*jst - 1 + ii3) + *ist + ii4) - 1;
	    jcoef_ii_1 = igl;
	    jcoef_ii_2 = igl + 2;
	    jcoef_ii_3 = igl + 2*(*nx + 2);
	    jcoef_ii_4 = igl - 2;
	    jcoef_ii_5 = igl - 2*(*nx + 2);

	    if (i == *ist && i > 1 && (j > *ny_st && j < *ny_end)) {
		++(*ixm);
		index_s_xm[*ixm] = jcoef_ii_4;
		index_r_xm[*ixm] = jcoef_ii_1;
		++(*ixm);
		index_s_xm[*ixm] = jcoef_ii_4 + 1;
		index_r_xm[*ixm] = jcoef_ii_1 + 1;
	    }
	    if (i == *iend && i < *nx + 2 && (j > *ny_st && j < *
		    ny_end)) {
		++(*ixp);
		index_s_xp[*ixp] = jcoef_ii_2;
		index_r_xp[*ixp] = jcoef_ii_1;
		++(*ixp);
		index_s_xp[*ixp] = jcoef_ii_2 + 1;
		index_r_xp[*ixp] = jcoef_ii_1 + 1;
	    }
	    if (j == *jst && j > 1 && (i > *nx_st && i < *nx_end)) {
		++(*iym);
		index_s_ym[*iym] = jcoef_ii_5;
		index_r_ym[*iym] = jcoef_ii_1;
		++(*iym);
		index_s_ym[*iym] = jcoef_ii_5 + 1;
		index_r_ym[*iym] = jcoef_ii_1 + 1;
	    }
	    if (j == *jend && j < *ny + 2 && (i > *nx_st && i < *
		    nx_end)) {
		++(*iyp);
		index_s_yp[*iyp] = jcoef_ii_3;
		index_r_yp[*iyp] = jcoef_ii_1;
		++(*iyp);
		index_s_yp[*iyp] = jcoef_ii_3 + 1;
		index_r_yp[*iyp] = jcoef_ii_1 + 1;
	    }
	}
    }
    return 0;
} /* example_2d_img_all */

/* ********************************************************************* */
/* ******************************************************************** */

/* Subroutine */ int example_1d(int *ist, int *iend, int *nx, 
	int *index_s_xm, int *index_s_xp, int *index_r_xm, 
	int *index_r_xp, int *ixp, int *ixm)
{
    int i1;
    static int i, ii4, jcoef_ii_1, jcoef_ii_2, jcoef_ii_4, ii, igl,
	     nxlocal;

/*   Purpose: Determine the indices of the external boundary grid points   
            comunicating with the neighboring subdomains (1D, 3pt problems) */
/* -------------------------------------------------------------------------- */

    /* Parameter adjustments */
    --index_r_xp;
    --index_r_xm;
    --index_s_xp;
    --index_s_xm;

    /* Function Body */
    nxlocal = *iend - *ist + 1;
    *ixm = 0;
    *ixp = 0;
    ii = 0;
    i1 = *iend;
    for (i = *ist; i <= i1; ++i) {
	++ii;
	ii4 = (ii - 1) % nxlocal;
	igl = *ist + ii4;
	jcoef_ii_1 = igl;
	jcoef_ii_2 = igl + 1;
	jcoef_ii_4 = igl - 1;
	if (i == *ist && i != 1) {
	    ++(*ixm);
	    index_s_xm[*ixm] = jcoef_ii_4;
	    index_r_xm[*ixm] = jcoef_ii_1;
	}
	if (i == *iend && i != *nx) {
	    ++(*ixp);
	    index_s_xp[*ixp] = jcoef_ii_2;
	    index_r_xp[*ixp] = jcoef_ii_1;
	}
    }
    return 0;
} /* example_1d */

/* ******************************************************************** */
/* ******************************************************************** */
/* Subroutine */ int example20_err(double *xnew, double *err, 
	int *nx, int *ny, double *sumerr,int *ipr)
{
    int i1, i2;
    static double h;
    static int i, j;
    static double u, x, y;
    static int ii;
    static double pi, sumerr1;

/*  Purpose: calculate the error for the 2D Helmholtz problem - Example20.   
    Example20 - taken from Erlangga, Vuik, Oosterlee, 2004 paper, 
    applied Numerical Mathematics. Example #1.   

    Domain: [0 1]*[0 1]   

        Uxx + Uyy + K**2*U = (K**2 -5*PI**2)*sin(PI*X)*sin(2*PI*Y)   

    With Dirichlet boundary conditions:  u=0 on the boundaries.   

    The analytic solution is:   
      U=sin(PI*x)*sin(2*PI*y)   on the inner grid points, */
/* ----------------------------------------------------------------------- */
    if(*ipr == 0) goto L100;
    printf("\n  %s", "    ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Results for example 20: ");
    printf("\n  %s","   Similar to the example in Erlangga, Vuik, Oosterlee 2004 paper.");
    printf("\n  %s", "    ");
    printf("\n  %s","   Domain: [0 1]*[0 1]  ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Uxx + Uyy + K**2*U = (K**2-5*PI**2)*sin(PI*X)*sin(2*PI*Y) ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Dirichlet boundary condition: U=0 on on all boundaries ");
    printf("\n  %s", "    ");
    printf("\n  %s","   The analytic solution is: U=sin(PI*x)*sin(2*PI*y)");
    printf("\n  %s","--------------------------------------------------------------" );
    printf("\n  %s","   ");
L100:
/*  pi = 3.14159265359793; */
    pi = atan(1.) * 4.;
    h = 1. / (double) (*nx + 1);
    sumerr1 = 0.;
    ii = 0;
    i1 = *ny;
    for (j = 1; j <= i1; ++j) {
	i2 = *nx;
	for (i = 1; i <= i2; ++i) {
	    x = (double) i * h;
	    y = (double) j * h;
	    u = sin(pi * x) * sin(pi * 2. * y);
	    sumerr1 += u * u;
	    err[ii] = xnew[ii] - u;
	    ++ii;
	}
    }
    *sumerr = sqrt(sumerr1);

    return 0;
} /* example20_err */

/* ********************************************************************* */
/* ******************************************************************** */
/* Subroutine */ int example21_err(double *xnew, double *err, 
	int *nx, int *ny, double *sumerr)
{

/*  Purpose: print out information abuot the 2D Helmholtz problem which was solved - 
    Example21.  
    Example21 is similar to example #2 from Erlangga,   
    Vuik, Oosterlee, 2004 paper, in applied Numerical Mathematics.   

    Domain: [0 1]*[0 1];   

         Uxx + Uyy + K**2*U = f   

    Where: f=0   

    With Dirichlet boundary condition on y=0: U=0 on y=0, except at   
    the center point:   
            U=1   for x=0.5; y=0   
            U=0   for x#0.5; y=0   

    and Smmerfeld b.c. DU/Dn -i*K*U=0 on all other boundaries. */
/* ----------------------------------------------------------------------- */

    printf("\n  %s", "    ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Results for example 21: ");
    printf("\n  %s","   Similar to example #2 in Erlangga, Vuik, Oosterlee 2004 paper.");
    printf("\n  %s", "    ");
    printf("\n  %s","   Domain: [0 1]*[0 1]  ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Uxx + Uyy + K**2*U = f  ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Where:  f=0.           ");
    printf("\n  %s", "    ");
    printf("\n  %s","   With Dirichlet boundary condition on y=0: U=0 on y=0, ");
    printf("\n  %s","   except:  U=1 at (x=0.5 , y=0).               ");
    printf("\n  %s","   Sommerfeld b.c. DU/Dn -i*K*U=0 on all other boundaries");
    printf("\n  %s", "    ");
    printf("\n  %s","--------------------------------------------------------------------" );
    printf("\n  %s","   ");

    return 0;
} /* example21_err */

/* ********************************************************************* */
/* ******************************************************************** */
/* Subroutine */ int example23_err(double *xnew, double *err, 
	int *nx, int *ny, double *sumerr, int *order)
{

/*  Purpose: print out information abuot the 2D Helmholtz problem which was solved - 
    Example23.   
    Example23 is similar to example #2 from Erlangga,   
    Vuik, Oosterlee, 2004 paper, in applied Numerical Mathematics.   
    The equations are solved using a 2nd/4th/6th  order difference scheme.   

    Domain: [0 1]*[0 1];   

            Uxx + Uyy + K**2*U = f   

     where: f=0   

     With Dirichlet boundary condition on y=0: U=0 on y=0, except   
     at the center point:   
              U=1   for x=0.5; y=0   
              U=0   for x#0.5; y=0   

     and Sommerfeld b.c. DU/Dn -i*K*U=0 on all other boundaries. */
/* ----------------------------------------------------------------------- */
    printf("\n  %s", "    ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Results for example 23:");
    printf("\n  %s","   Similar to example #2 in Erlangga, Vuik & Oosterlee 2004 paper.");
    printf("\n  %s", "    ");
    printf("\n  %s","   Domain: [0 1]*[0 1]  ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Uxx + Uyy + K**2*U = f  ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Where:  f=0.           ");
    printf("\n  %s", "    ");
    printf("\n  %s","   With Dirichlet boundary condition on y=0: U=0 on y=0, except: ");
    printf("\n  %s","   U=1. at (x=0.5, y=0.)");
    printf("\n  %s","   Sommerfeld b.c.:  DU/Dn -i*K*U=0 on all other boundaries.");
    printf("\n  %s","    ");
    printf("\n     Scheme order= %d \n",*order);
    printf("\n  %s","-------------------------------------------------------------------" );
    printf("\n  %s","   ");

    return 0;
} /* example23_err */

/* ******************************************************************** */
/* ******************************************************************** */
/* Subroutine */ int example24_err(int *order) 
{

/*  Purpose: print out information abuot the 2D Helmholtz problem which was solved -
    Example24.
    Example24 - Marmousi problem, was taken from O. Schenk  2009 paper, 
    SIAM J. Scientific Computing.  
    Marmpusi problem:   

    Domain: [Xmin-hx  Xmax+hx]*[Ymin-hy Ymax+hy];   

            Uxx + Uyy + K**2*U = f   
            K=2*PI*freq/C   
    Sommerfeld b.c. DU/Dn-i*K*U=gi on all  boundaries.   
    and:   
              f=1   for x=0.5, y=0   
              f=0    else */
/* ----------------------------------------------------------------------- */

    printf("\n  %s", "    "); 
    printf("\n  %s", "    ");
    printf("\n  %s","   Results for example 24:");
    printf("\n  %s","   Marmousi problem from O. Schenk 2009 paper. ");
    printf("\n  %s","    ");
    printf("\n  %s","   Uxx + Uyy + K**2*U = f");
    printf("\n  %s","    ");
    printf("\n  %s","   Where:  f=1/(hx*Hy)   for: x=0.5(Xmin+Xmax), ");
    printf("\n  %s","           f=0           else");
    printf("\n  %s","    ");
    printf("\n  %s","   Sommerfeld b.c.: DU/Dn-i*K*U=0 on all boundaries");
    printf("\n  %s","    ");
    printf("\n     Scheme order= %d \n",*order);
    printf("\n  %s","  ----------------------------------------------------");
    printf("\n  %s", "    ");

    return 0;
} /* example24_err */

/* ************************************************************************ */
/* ******************************************************************** */
/* Subroutine */ int example25_err(int *order) 
{

/*  Purpose: print out information abuot the 3D problem which was solved - Example25.
    Example25  is similar to the 3D problem in  O. Schenk  2009 paper,
    SIAM J. Scientific Computing.   

    Domain: [Xmin-hx  Xmax+hx]*[Ymin-hy Ymax+hy]*[Zmin-hz Zmax+hz];   

            Uxx + Uyy + Uzz + K**2*U = f   
            K=2*PI*freq/C   
    Sommerfeld b.c. DU/Dn-i*K*U=gi on all  boundaries.   
    and:   
              f=1   for x=0.5*(Xmin+Xmax), y=0, z=0.5*(Zmin+Zmax)   
              f=0   else */
/* ----------------------------------------------------------------------- */
    printf("\n  %s", "    ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Results for example 25:");
    printf("\n  %s","   3D Helmholtz problem from O. Schenk 2009 paper.");
    printf("\n  %s","    ");
    printf("\n  %s","   Uxx + Uyy + Uzz + K**2*U = f");
    printf("\n  %s","    ");
    printf("\n  %s","   Where:   ");
    printf("\n  %s","   f = 1. for: x=0.5*(Xmin+Xmax), y=0, z=0.5*(Zmin+Zmax); ");
/*  printf("\n  %s","   f = 1/(hx*Hy*Hz) for: x=0.5*(Xmin+Xmax), y=0, z=0.5*(Zmin+Zmax); "); */
    printf("\n  %s","   f = 0. else      ");
    printf("\n  %s","  "); 
    printf("\n  %s","   Sommerfeld b.c.:  DU/Dn-i*K*U=0 on all boundaries.");
    printf("\n     Scheme order= %d \n",*order);
/*  printf("\n  %s","   Scheme order=","\n  %d \n",order); */
    printf("\n  %s","   ------------------------------------------------------------------ ");
    printf("\n  %s","  "); 

    return 0;
} /* end of example25_err */

/* ************************************************************************ */
/* ************************************************************************ */

/* Subroutine */ int example26_err(double *xnew, double *err, 
	int *nx, int *ny, double *sumerr1, double *sumerr2, 
	double *kn1, double *kn2, double *kn3, int *order, 
        int *ipr)
{
    int i1, i2;
    double d1, d2;
    static double h;
    static int i, j;
    static double x, y;
    static int ii;
    static double kn, pi, beta, u_imag, u_real;

/*  Purpose: calculate the error for the 2D Helmholtz problem -Example26.   
    Example26 is similar to the example of Erlangga and E. Turkel,   
    in the 2012 paper.   
    The equations are solved using a 2nd/4th/6th order difference scheme.   

    Domain: [0 1]*[-1/2 1/2]   

          Uxx + Uyy + K**2*U = f   

    Where: f=0   
    With Dirichlet boundary conditions:   
    U=0 on y=-1/2,    U=0 on y=1/2;   
    U=cos(PI*y) on x=0.   
    and Sommerfeld b.c. on x=1: DU/Dn-i*beta*U=0 on x=1.,   
    Where: beta**2+PI**2=K**2   

    The analytic solution is: U=cos(PI*y)*exp(i*beta*x)   
                               =cos(PI*y)*[cos(beta*x)+ i*sin(beta*x)]     */
/* ----------------------------------------------------------------------- */

    if(*ipr == 0) goto L80;
    printf("\n  %s", "    ");
    printf("\n  %s", "    ");
    printf("\n  %s","  Results for example 26:");
    printf("\n  %s","  Similar to the example in Erlangga and Turkel, 2011 paper.");
    printf("\n  %s", "    ");
    printf("\n  %s","  Domain: [0 1]*[-1/2 1/2]  ");
    printf("\n  %s", "    ");
    printf("\n  %s","  Uxx + Uyy + K**2*U = f  ");
    printf("\n  %s", "    ");
    printf("\n  %s","  Where:  f=0.           ");
    printf("\n  %s", "    ");
    printf("\n  %s","  With Dirichlet boundary conditions:      ");
    printf("\n  %s","     U=0 on y=-1/2 ; U=0 on  y=1/2; U=cos(PI*y) on  x=0");
    printf("\n  %s","  Sommerfeld b.c.:  DU/Dn-i*beta*U=0 on x=1.;           ");
    printf("\n  %s","  Where: beta**2+PI**2= K**2 "                         );
    printf("\n  %s","   " );
    printf("\n  %s","   The analytic solution is: U=cos(PI*y)*exp(i*beta*x) " );
    printf("\n     Scheme order= %d \n",*order);
    printf("\n  %s","------------------------------------------------------------------" );
    printf("\n  %s","   ");
    printf("\n  %s","   ");
L80:

    pi = atan(1.) * 4.;
    h = 1. / (double) (*nx + 1);
    *sumerr1 = 0.;
    *sumerr2 = 0.;
    ii = 0;
    i1 = *ny + 1;
    for (j = 0; j <= i1; ++j) {
	y = (double) j * h - .5;

	if (y <= (1./3.)-0.5 ) kn = *kn1;
	if (y > (1./3.)-0.5 && y <= (2./3.)-0.5 ) kn = *kn2;
	if (y > (2./3.)-0.5 ) kn = *kn3;
	d1 = kn;
	d2 = pi;
	beta = sqrt(d1 * d1 - d2 * d2);
	i2 = *nx + 2;
	for (i = 0; i <= i2; ++i) {
	    x = (double) i * h;
	    u_real = cos(pi * y) * cos(beta * x);
	    u_imag = cos(pi * y) * sin(beta * x);
	    d1 = u_real;
	    *sumerr1 += d1 * d1;
	    d1 = u_imag;
	    *sumerr2 += d1 * d1;
	    err[ii] = xnew[ii] - u_real;
	    ++ii;
	    err[ii] = xnew[ii] - u_imag;
	    ++ii;
	}
    }
    *sumerr1 = sqrt(*sumerr1);
    *sumerr2 = sqrt(*sumerr2);
    if(*ipr == 1)
    printf("\n  SUMERR1= %f    SUMERR2= %f", *sumerr1, *sumerr2);
    return 0;
} /* example26_err */

/* ******************************************************************** */
/* ******************************************************************** */

/* Subroutine */ int example27_err(double *xnew, double *err, 
	int *nx, int *ny, double *sumerr1, double *sumerr2, 
	double *kn1, double *kn2, double *kn3, int *order,
         int *ipr )
{
    int i1, i2;
    double d1, d2;
    static double h;
    static int i, j;
    static double x, y;
    static int ii;
    static double kn, pi, beta, u_imag, u_real;

/*  Purpose: calculate the error for the 2D Helmholtz problem - Example27.   
    Example27 is the example of Erlangga and Turkel,  2012 paper.
    The equations are solved using a 2nd/4th/6th order difference scheme.   

    Domain: [-1/2 1/2]*[0 1]   

           Uxx + Uyy + K**2*U = f   

    Where: f=0   

    With Dirichlet boundary conditions:   
    On the left side: U=0 on x=-1/2,  On the right side:  u=0 on x=1/2;   
    On y=0:  U=cos(PI*x) on y=0.   
    Sommerfld b.c. on y=1: DU/Dy -i*beta*U=0 on y=1.,   
    where: beta**2+PI**2=K**2   

    The analytic solution is: u=cos(PI*x)*exp(i*beta*y)   
                               =cos(PI*x)*[cos(beta*y)+ i*sin(beta*y)] */
/* ----------------------------------------------------------------------- */

    if(*ipr == 0) goto L80;
    printf("\n  %s", "    ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Results for example 27:");
    printf("\n  %s","   Similar to the example in Erlangga and Turkel, 2012 paper.");
    printf("\n  %s", "    ");
    printf("\n  %s","   Domain: [-1/2 1/2]*[0 1]  ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Uxx + Uyy + K**2*U = f  ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Where:  f=0.           ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Dirichlet boundary conditions:      ");
    printf("\n  %s","      U=0 on x=-1/2 ; U=0 on  x=1/2; U=cos(PI*x) on  y=0");
    printf("\n  %s","   Sommerfeld b.c.: DU/Dy-i*beta*U=0 on y=1.;           ");
    printf("\n  %s","   Where: beta**2+PI**2= K**2 "                         );
    printf("\n  %s","   " );                                                
    printf("\n  %s","   The analytic solution is: U=cos(PI*x)*exp(i*beta*y) "    );
    printf("\n     Scheme order= %d \n",*order);
    printf("\n  %s","---------------------------------------------------------------" );
    printf("\n  %s","   ");                                                
    printf("\n  %s","   ");                                                
L80:
    pi = atan(1.) * 4.;
    h = 1. / (double) (*nx + 1);
    *sumerr1 = 0.;
    *sumerr2 = 0.;
    ii = 0;
    i1 = *ny + 2;
    for (j = 0; j <= i1; ++j) {
	y = (double) j * h;

	if (y <= 1./3.) kn = *kn1;
	if (y > 1./3. && y <= 2./3.) kn = *kn2;
	if (y > 2./3.) kn = *kn3;
	 
	d1 = kn;
	d2 = pi;
	beta = sqrt(d1 * d1 - d2 * d2);
	i2 = *nx + 1;
	for (i = 0; i <= i2; ++i) {
	    x = (double) i * h - .5;
	    u_real = cos(pi * x) * cos(beta * y);
	    u_imag = cos(pi * x) * sin(beta * y);
	    d1 = u_real;
	    *sumerr1 += d1 * d1;
	    d1 = u_imag;
	    *sumerr2 += d1 * d1;
	    err[ii] = xnew[ii] - u_real;
	    ++ii;
	    err[ii] = xnew[ii] - u_imag;
	    ++ii;
	}
    }
    *sumerr1 = sqrt(*sumerr1);
    *sumerr2 = sqrt(*sumerr2);
    if(*ipr == 1)
    printf("\n  SUMERR1= %f    SUMERR2= %f", *sumerr1, *sumerr2);
    return 0;
} /* example27_err */

/* ******************************************************************** */
/* ******************************************************************** */

/* Subroutine */ int example28_err(double *xnew, double *err, 
	int *nx, int *ny, double *sumerr, double *kn, int 
	*order, int *ipr)
{
    int i1, i2;
    double d1, d2;
    static double h;
    static int i, j;
    static double u, x, y;
    static int ii;
    static double pi, beta;

/*  Purpose: calculate the error for the 2D Helmholtz problem - Example28.   
    Example28 - is our example.  
    The equations are solved using a 2nd/4th/6th order difference scheme.   

    Domain: [0 1]*[0 1]   

          Uxx + Uyy + K**2*U = f   

    Where: f=0   

    Case A:   
    With Dirichlet boundary conditions:  U=0 on X=0.;    
    U=0 on x=1.0;  U=0. on y=0.   
    U=sin(PI*X)*sin(beta) on y=1.   

    The analytic solution is:   
            U(x,y)= sin(PI*X)*sin(beta*Y)  with: PI**2+beta**2=K**2       */
/* ---------------------------------------------------------------------- */

    if(*ipr == 0) goto L80;
    printf("\n  %s", "    ");
    printf("\n  %s", "    ");
/*  printf("\n  %s","   Results for example 28, case A:  "); */
/*  printf("\n  %s","   Results for example 28, case B:  "); */
    printf("\n  %s","   Results for example 28, case C:  ");   
    printf("\n  %s", "    ");
    printf("\n  %s","   Domain: [0 1]*[0 1]  ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Uxx + Uyy + K**2*U = f  ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Where:  f=0.           ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Dirichlet boundary condition on all boundaries:      ");
    printf("\n  %s", "    ");
/*  printf("\n  %s","   Case A: Analytic solution: U=sin(PI*x)*sin(beta*y)");
    printf("\n  %s","   Where: PI**2+beta**2=K**2                             "); */
    printf("\n  %s","   Case C: Analytic solution: U=sin(PI*x)*cos(beta*y)");
    printf("\n  %s","   Where: PI**2+beta**2=K**2                             ");   
    printf("\n     Scheme order= %d \n",*order);
    printf("\n  %s","-------------------------------------------------------" );
    printf("\n  %s", "    ");
L80:
    pi = atan(1.) * 4.;
    h = 1. / (double) (*nx + 1);
    d1 = *kn;
    d2 = pi;
    beta = sqrt(d1 * d1 - d2 * d2);
    *sumerr = 0.;
    ii = 0;
    i1 = *ny + 1;
    for (j = 0; j <= i1; ++j) {
	y = (double) j * h;
	i2 = *nx + 1;
	for (i = 0; i <= i2; ++i) {
	    x = (double) i * h;
/*  case A: */
/* --------- */
/*          u=sin(pi * x)*sin(beta * y);  */ 
/*  case B: */
/* --------- */
/*          const=d1/sqrt(2.) ;*/
/*          U=sin(const * x)*cos(const * y) */
/*  case C: */
/* --------- */
            u = sin(pi * x) * cos(beta * y);   
	    *sumerr += u * u;
	    err[ii] = xnew[ii] - u;
	    ++ii;
	}
    }
    *sumerr = sqrt(*sumerr);
    if(*ipr == 1)
    printf("\n  SUMERR= %f", *sumerr);

    return 0;
} /* example28_err */

/* ******************************************************************** */
/* ******************************************************************** */

/* Subroutine */ int example29_err(double *xnew, double *err, 
	int *nx, int *ny, int *nz, double *sumerr1, double *sumerr2, 
	double *kn1, double *kn2,  double *kn3, int *order, int *ipr)
/*
    Purpose: calculate the error for a 3D Helmholtz problem, which has an
    analytic solution.
    Example29 is  our own built example for checking the 6th vs. the 2nd
    order schemes.
    The equations are solved using a 2nd/6th order difference scheme.
    The 6th order finite difference schemes follows G. Sutmann in the J. of
    Computational and Applied Mathematics, pp. 15-31, 2007,
    - eq. 17 in that paper.
    The expression for the normal derivatives in the b.c are taken from 
    Turkel and erlangga 2012 paper.
 
    Domain: [0 1]*[0 1]*[0 1]
 
           Uxx + Uyy + + Uzz + K**2*U = f
 
    where: f=0
 
    With Dirichlet boundary conditions on all boundaries except on the x=0
    plane where Sommerfeld condition Du/Dn-iKU=0 is prescribed.
 
    The analytic solution is:
 
         U=cos(PI*x)*sin(PI*y)*(alpha*cos(beta1*z)- i*sin(beta1*z))
 
     where:  beta1**2= K**2-2*PI**2
       and   alpha=beta1/K
 ----------------------------------------------------------------------- */

{
    int i1, i2, k1;
    double d1, d2, d3;
    static int i, j, kk;
    static double kn, h,  x, y, z, u_real, u_imag, alpha, beta1, pi;
    static double  z13, z23;
    static int ii;


    if(*ipr == 0) goto L80;
    printf("\n  %s", "    ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Results for example 29:");
    printf("\n  %s", "    ");
    printf("\n  %s","   Domain: [0 1]*[0 1]*[0 1]  ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Uxx + Uyy + Uzz + K**2*U = f  ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Where:  f=0.           ");
    printf("\n  %s", "    ");
    printf("\n  %s","   With Dirichlet boundary condition on all the boundaries,      ");
    printf("\n  %s","   except on the x=0 plane - Sommerfeld b.c.              ");
    printf("\n  %s", "    ");
    printf("\n  %s","   The analytic solution:");
    printf("\n  %s","   U=cos(PI*x)*sin(PI*y)*(alpha*cos(beta1*z)-i*sin(beta1*z))");
    printf("\n  %s", "    ");
    printf("\n  %s","   Where: beta1**2= K**2-2*PI**2  and:  alpha=beta1/K   ");
    printf("\n     Scheme order= %d \n",*order);
    printf("\n  %s","------------------------------------------------------------" );
    printf("\n  %s", "    ");
L80:
    
    pi = atan(1.) * 4.;
    h = 1. / (double) (*nx + 1);
    *sumerr1 = 0.;
    *sumerr2 = 0.;
    ii = 0;
    k1 = *nz + 2;
    for (kk = 0; kk <= k1; ++kk) {
       z = -h + (double) kk*h;
       z13= (1./3.) ;
       z23= (2./3.) ;
    
       if (z <  z13) {
         kn = *kn1;
       }
       if (z >=  z13 && z <  z23 ) {
         kn = *kn2;
       }
       if (z >=  z23 ) {
         kn = *kn3;
       }
       i1 = *ny + 1;
       for (j = 0; j <= i1; ++j) {
           y = (double) j * h;
           /* d1 = *kn; */
           d1 = kn;
           beta1=sqrt(d1 * d1 - 2.*pi*pi);
           alpha=beta1/d1;
           i2 = *nx + 1;
           for (i = 0; i <= i2; ++i) {
	       x = (double) i * h;
               u_real = alpha*cos(beta1*z)*sin(pi*y)*cos(pi*x);
               u_imag = -sin(beta1*z)*sin(pi*y)*cos(pi*x);
	       d2 = u_real;
	       d3 = u_imag;
	       *sumerr1 += d2 * d2;
	       *sumerr2 += d3 * d3;
	       err[ii] = xnew[ii] - u_real;
	       ++ii;
               err[ii] = xnew[ii] - u_imag;
	       ++ii;
           }
       }
    }
    *sumerr1 = sqrt(*sumerr1);
    *sumerr2 = sqrt(*sumerr2);
    if(*ipr == 1)
    printf("\n  SUMERR= %f  SUMERR2= %f", *sumerr1,*sumerr2);

    return 0;
} /* end of example29_err */

/* ******************************************************************** */
/* ******************************************************************** */

/* Subroutine */ int example30_err(int *order) 
{
/*  Purpose: print out information abuot the 3D problem which was solved - Example30.
    Example30 - is similar to the 3D 3 layer problem in  O. Schenk  2009 paper, in   
    SIAM J. Scientific Computing, but with variable K, applied to the SEG/EAGE salt model.

    Domain: [Xmin-hx  Xmax+hx]*[Ymin-hy Ymax+hy]*[Zmin-hz Zmax+hz];   

            Uxx + Uyy +Uzz + K**2*U = F   
            K=2*PI*freq/C   
    With Sommerfeld b.c. DU/Dn-i*K*U=gi on all  the boundaries.   
    and:   
              f=1   for x=0.5*(Xmin+Xmax), y=0, z=0.5*(Zmin+Zmax)   
              f=0    else */
/* ----------------------------------------------------------------------- */
    printf("\n  %s", "    ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Results for example 30:");
    printf("\n  %s","   3D Helmholtz problem for the SEG/EAGE salt model");
    printf("\n  %s","    ");
    printf("\n  %s","   Uxx + Uyy + Uzz + K**2*U = f");
    printf("\n  %s","    ");
    printf("\n  %s","   Where:   ");
    printf("\n  %s","   f = 1. for: x=0.5*(Xmin+Xmax), y=0.5*(Ymin+Ymax), z=0.; ");
/*  printf("\n  %s","   f = 1/(hx*Hy*Hz) for: x=0.5*(Xmin+Xmax), y=0.5*(Ymin+Ymax) z=0.; "); */
    printf("\n  %s","   f = 0. else      ");
    printf("\n  %s","  "); 
    printf("\n  %s","   Sommerfeld b.c.:  DU/Dn-i*K*U=0 on all the boundaries.");
    printf("\n     Scheme order= %d \n",*order);
    printf("\n  %s","   ------------------------------------------------------------------ ");
    printf("\n  %s","  "); 

    return 0;
} /* end of example30_err */

/* ******************************************************************** */
/* ******************************************************************** */

/* Subroutine */ int example31_err_case_a(double *xnew, double *err, 
	int *nx, int *ny, double *sumerr, double *m, int *order, int *ipr) 
{
    int i1, i2;
    static double h;
    static int i, j;
    static double u, x, y, sinhmy;
    static int ii;
    static double pi;

/*  Purpose: calculate the error for a 2D Helmholtz problem with variable K.   
    Example31_case_a - is the example of E. Turkel, Gordon R&D and Tsynkov 2012,  
    but it is not included in the paper.  
    The equations are solved using a 2nd/4th/6th order difference scheme.   

    Domain: [0 PI]*[0 PI]   

          Uxx + Uyy + K**2*U = f   

    With Dirichlet boundary conditions on all the boundaries.               

    The analytic solution is:   
            U(x,y)= sin(sin(m*x)*sinh(n*y))  
    
    with m=n  so that f=0.

    and:  K**2 = m**2*cos(m*x)**2*sinh(n*y)**2 + n**2*sin(m*x)**2*cosh(n*y)**2   */ 
/* ----------------------------------------------------------------------------- */

    if(*ipr == 0) goto L80;
    printf("\n  %s", "    ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Results for example 31 - case A with RHS=0");   
    printf("\n  %s", "    ");
    printf("\n  %s","   Domain: [0 pi]*[0 pi]  ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Uxx + Uyy + K**2*U = f  ");
    printf("\n  %s", "    ");
    printf("\n  %s","   With Dirichlet boundary condition on all the boundaries:      ");
    printf("\n  %s", "    ");
    printf("\n  %s","   The analytic solution: U=sin(sin(m*x)*sinh(n*y)) ");
    printf("\n  %s", "    ");
    printf("\n  %s","   With: m = n  ==> f=0            ");
    printf("\n  %s", "    ");
    printf("\n  %s","   K**2 = m**2*(cos(m*x)**2*sinh(n*y)**2 + sin(m*x)**2*cosh(n*y)**2) ");
    printf("\n     Scheme order= %d     m=n=%5.2f \n",*order, *m);
    printf("\n  %s","---------------------------------------------------------------------" );
    printf("\n  %s", "    ");
L80:
    
    pi = atan(1.) * 4.;
    h = pi / (double) (*nx + 1);
    *sumerr = 0.;
    ii = 0;
    i1 = *ny + 1;
    for (j = 0; j <= i1; ++j) {
	y = (double) j * h;
        sinhmy = sinh(*m * y);
	i2 = *nx + 1;
	for (i = 0; i <= i2; ++i) {
	    x = (double) i * h;
            u=sin(sin(*m * x) * sinhmy);   
	    *sumerr += u * u;
	    err[ii] = xnew[ii] - u;
/*      printf("\n  ii=  %d  x= %10.5f  y= %10.5f  u=  %10.5f xnew[ii]= %10.5f err(ii)= %10.5f",
               ii,x,y,u, xnew[ii],err[ii]); */
	    ++ii;
	}
    }
    *sumerr = sqrt(*sumerr);
    if(*ipr == 1)
    printf("\n  SUMERR= %f", *sumerr);
    return 0;
} /* example31_err_case_a */

/* ******************************************************************** */
/* ******************************************************************** */

/* Subroutine */ int example31_err_case_b(double *xnew, double *err, 
	int *nx, int *ny, double *sumerr, double *m, int *order, int *ipr, 
        double *alpha) 
{
    int i1, i2;
    static double h;
    static int i, j;
    static double u, x, y, expmy;
    static int ii;
    static double pi;

/*  Purpose: calculate the error for a 2D Helmholtz problem with variable K.   
    Example31_case_b - is the example of E. Turkel, Gordon R&D and Tsynkov 2012,
    but it is not included in the paper.  
    The equations are solved using a 2nd/4th/6th order difference scheme.   

    Domain: [0 PI]*[0 PI]   

          Uxx + Uyy + K**2*U = f   

    With Dirichlet boundary conditions on all the boundaries.                   

    The analytic solution is:   
            U(x,y)= sin(alpha*sin(m*x)*exp(n*y))
    
    with  m = n  ==> f=0
    and:    K**2 = m**2*(alpha*exp(m*y))**2                                   */   
/* -------------------------------------------------------------------------- */

    if(*ipr == 0) goto L80;
    printf("\n  %s", "    ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Results for example 31 - case B with RHS=0");   
    printf("\n  %s", "    ");
    printf("\n  %s","   Domain: [0 pi]*[0 pi]  ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Uxx + Uyy + K**2*U = f  ");
    printf("\n  %s", "    ");
    printf("\n  %s","   With Dirichlet boundary condition on all the boundaries:      ");
    printf("\n  %s", "    ");
    printf("\n  %s","   The analytic solution: U=sin(alpha*sin(m*x)*exp(n*y)) ");
    printf("\n  %s", "    ");
    printf("\n  %s","   With:   m=n ==> f=0      ");
    printf("\n  %s", "    ");
    printf("\n  %s","   K**2 = m**2*(alpha*exp(m*y))**2 ");
    printf("\n  %s", "    ");
    printf("\n     Scheme order= %d     m=n=%5.2f  alpha=%7.2f \n",*order, *m, *alpha);
    printf("\n  %s","---------------------------------------------------------------------" );
    printf("\n  %s", "    ");
L80:
    pi = atan(1.) * 4.;
    h = pi / (double) (*nx + 1);
    *sumerr = 0.;
    ii = 0;
    i1 = *ny + 1;
    for (j = 0; j <= i1; ++j) {
	y = (double) j * h;
        expmy = *alpha * exp(*m * y);
	i2 = *nx + 1;
	for (i = 0; i <= i2; ++i) {
	    x = (double) i * h;
            u=sin(sin(*m * x) * expmy);   
	    *sumerr += u * u;
	    err[ii] = xnew[ii] - u;
/*      printf("\n ii=  %d  x= %10.5f  y= %10.5f  u=  %10.5f xnew[ii]= %10.5f err(ii)= %10.5f", 
        ii,x,y,u, xnew[ii],err[ii]); */
	    ++ii;
	}
    }
    *sumerr = sqrt(*sumerr);
    if(*ipr == 1)
    printf("\n  SUMERR= %f", *sumerr);
    return 0;
} /* example31_err_case_b */

/* ************************************************************************ */
/* ******************************************************************** */

/* Subroutine */ int example31_err_case_c(double *xnew, double *err, 
	int *nx, int *ny, double *sumerr, double *a, double *b, double *c, int *order, int *ipr) 
{
    int i1, i2;
    static double h;
    static int i, j;
    static double u, x, y, siny, kx;
    static int ii;
    static double pi;

/*  Purpose: calculate the error for a 2D Helmholtz problem with variable K.   
    Example31_case_c - is the example of E. Turkel, Gordon R&D and Tsynkov 2012 paper.  
    The equations are solved using a 2nd/4th/6th order difference scheme.   

    Domain: [0 PI]*[0 PI]   

          Uxx + Uyy + K**2*U = f   

    With Dirichlet boundary conditions calculated from the analytic solution.   

    The analytic solution is:   
            U(x,y)= exp(-K(x)/c)*sin(sqrt(a**2+b**2)*y)   for general a, b and c. 

    with: K = a - b*sin(c*x); 
    if  c=-2.*a  ==> f=0   
*/
/* -------------------------------------------------------------------------- */

    if(*ipr == 0) goto L80;
    printf("\n  %s", "    ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Results for example 31 - case C ");   
    printf("\n  %s", "    ");
    printf("\n  %s","   Domain: [0 pi]*[0 pi]  ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Uxx + Uyy + K**2*U = f  ");
    printf("\n  %s", "    ");
    printf("\n  %s","   With Dirichlet boundary condition on all the boundaries:      ");
    printf("\n  %s", "    ");
    printf("\n  %s","   The analytic solution: U=exp(-K(x)/c)*sin(sqrt(a**2+b**2)*y) ");
    printf("\n  %s", "    ");
    printf("\n  %s","   K= a - b*sin(c*x) ");
    printf("\n     Scheme order= %d     a=%5.2f  b=%7.2f  c=%7.2f \n",*order, *a, *b, *c);
    printf("\n  %s","---------------------------------------------------------------------" );
    printf("\n  %s", "    ");

L80:
    pi = atan(1.) * 4.;
    h = pi / (double) (*nx + 1);
    if (ikcg==10 || *ipr == 1) { *sumerr = 0.; u_max=0; }
    ii = 0;
    i1 = *ny + 1;
    for (j = 0; j <= i1; ++j) {
	y = (double) j * h;
        siny = sin(sqrt(*a * *a + *b * *b)*y);
	i2 = *nx + 1;
	for (i = 0; i <= i2; ++i) {
	    x = (double) i * h;

            kx = *a - *b * sin( *c * x);
            u = exp(-kx / *c) * siny;   
            if (ikcg==10 || *ipr == 1)
               { if(fabs(u) > u_max) u_max= fabs(u);
                 *sumerr += u * u;
               }
	    err[ii] = xnew[ii] - u;
/*      printf("\n ii=  %d  x= %10.5f  y= %10.5f  u=  %10.5f xnew[ii]= %10.5f err(ii)= %10.5f", 
        ii,x,y,u, xnew[ii],err[ii]); */
	    ++ii;
	}
    }
    if (ikcg==10 || *ipr == 1) *sumerr = sqrt(*sumerr);
    if(*ipr == 1)
    {
    printf("\n  SUMERR= %f", *sumerr);
    printf("\n  u_max= %15.9f", u_max);
    }
    return 0;
} /* example31_err_case_c */

/* ******************************************************************** */
/* ******************************************************************** */

/* Subroutine */ int example32_err_case_a(double *xnew, double *err, 
	int *nx, int *ny, int *nz, double *sumerr, double *l_a, double *n_a, 
        int *order, int *ipr) 
{
    int i1, i2, k1;
    static double h;
    static int i, j;
    static double u, x, y, z, la, ma, na;
    static int ii, kk;
    static double pi;

/*  Purpose: calculate the error for a 3D Helmholtz problem with variable K.   
    Example32_case_a - is the example of E. Turkel, Gordon R&D and Tsynkov 2012,  
    but it is not included in the paper.
    The equations are solved using a 2nd/6th order difference scheme.   

    Domain: [0 PI]*[0 PI]*[0 PI]   

          Uxx + Uyy + Uzz + K**2*U = f   

    With Dirichlet boundary conditions on all the boundaries.                

    The analytic solution is:   
            U(x,y)= sin(sin(l*x)*sinh(m*y)*sin(nz))  

    with: K**2 = l**2*sin(n*z)**2*(sin(l*x)**2 + sinh(m*y)**2) + 
                 n**2*sin(l*x)**2*(sinh(m*y)**2 + sin(n*z)**2)   

    and:  m**2 = l**2 + n**2  so that  f=0.                                  */
/* ------------------------------------------------------------------------- */

    if(*ipr == 0) goto L80;
     la = *l_a;
     na = *n_a;
     ma = sqrt(la*la + na*na); 
    printf("\n  %s", "    ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Results for example 32 - case A  with RHS=0");   
    printf("\n  %s", "    ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Domain: [0 pi]*[0 pi]]*[0 pi]  ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Uxx + Uyy + Uzz + K**2*U= f  ");
    printf("\n  %s", "    ");
    printf("\n  %s","   With Dirichlet boundary condition on all the boundaries.  ");
    printf("\n  %s", "    ");
    printf("\n  %s","   The analytic solution: U=sin(sin(l*x)*sinh(m*y)*sin(nz)) ");
    printf("\n  %s", "    ");
    printf("\n  %s","   With:  m**2 = l**2 + n**2 ==> f=0.           ");
    printf("\n  %s", "    ");
    printf("\n  %s","   K**2 = l**2*sin(n*z)**2*(sin(l*x)**2 + sinh(m*y)**2) +  ");
    printf("\n  %s","          n**2*sin(l*x)**2*(sinh(m*y)**2 + sin(n*z)**2)  ");
    printf("\n  %s", "    ");
    printf("\n     l= %6.2f  n= %6.2f m= %9.5f ", la, na, ma);
    printf("\n     Scheme order= %d \n",*order);
    printf("\n  %s","---------------------------------------------------------------------" );
    printf("\n  %s", "    ");

L80:
     la = *l_a;
     na = *n_a;
     ma = sqrt(la*la + na*na); 
     pi = atan(1.) * 4.;
     h = pi / (double) (*nx + 1);
     *sumerr = 0.;
     ii = 0;
     k1 =  *nz + 1;
     for (kk = 0; kk <= k1; ++kk) {
        z = (double) kk * h;
        i1 = *ny + 1;
        for (j = 0; j <= i1; ++j) {
           y = (double) j * h;
           i2 = *nx + 1;
           for (i = 0; i <= i2; ++i) {
              x = (double) i * h;
              u = sin(sin(la*x)*sinh(ma*y)*sin(na*z));
	      *sumerr += u * u;
	      err[ii] = xnew[ii] - u;
/*      printf("\n  ii=  %d  x= %10.5f  y= %10.5f  u=  %10.5f xnew[ii]= %10.5f err(ii)= %10.5f",
                ii,x,y,u, xnew[ii],err[ii]); */
	      ++ii;
           }
	}
    }
    *sumerr = sqrt(*sumerr);
    if(*ipr == 1)
    printf("\n  SUMERR= %f", *sumerr);
    return 0;
} /* example32_err_case_a */

/* ******************************************************************** */
/* ******************************************************************** */

/* Subroutine */ int example32_err_case_c(double *xnew, double *err, 
	int *nx, int *ny, int *nz, double *sumerr, double *a2, double *b2, 
        double *c2, double *gam, int *order, int *ipr) 
{
    int i1, i2, k1;
    static double h, a, b, c, beta, gama;
    static int i, j;
    static double u, x, y, z, kx;
    static int ii, kk;
    static double pi;

/*  Purpose: calculate the error for a 3D Helmholtz problem with variable K.   
    Example32_case_c - is the example of E. Turkel, Gordon R&D and Tsynkov 2012 paper.  
    The equations are solved using a 2nd/6th order difference scheme.   

    Domain: [0 PI]*[0 PI]*[0 PI]   

          Uxx + Uyy + Uzz + K**2*U = f   

    With Dirichlet boundary conditions on all the boundaries.

    The analytic solution is:   
         U(x,y)= exp(-kx/c)*sin(beta*y)*sin(gama*z)   for general a, b, c  and gama. 
    with: K =  a - b*sin(c*x) 
    For: c=-2.*a ==> f=0                              
                                                                                       */
/* ----------------------------------------------------------------------------------- */

    a = *a2;
    b = *b2;
    c = *c2;
    gama = *gam;
    beta = sqrt(a*a + b*b - gama*gama);
    if(*ipr == 0) goto L80;
    printf("\n  %s", "    ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Results for example 32 - case C");   
    printf("\n  %s", "    ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Domain: [0 pi]*[0 pi]]*[0 pi]  ");
    printf("\n  %s", "    ");
    printf("\n  %s","   Uxx + Uyy + Uzz + K**2*U= f  ");
    printf("\n  %s", "    ");
    printf("\n  %s","   With Dirichlet boundary condition on all the boundaries:      ");
    printf("\n  %s", "    ");
    printf("\n  %s","   The analytic solution: U=exp(-k(x)/c)*sin(beta*y)*sin(gama*z) ");
    printf("\n  %s", "    ");
    printf("\n  %s","   K =  a - b*sin(c*x)                                ");
    printf("\n  %s", "    ");
    printf("\n     a= %6.2f  b= %6.2f c= %6.2f beta= %6.2f gama= %6.2f", a, b, c, beta, gama);
    printf("\n     Scheme order= %d \n",*order);
    printf("\n  %s","---------------------------------------------------------------------" );
    printf("\n  %s", "    ");

L80:
     pi = atan(1.) * 4.;
     h = pi / (double) (*nx + 1);
     if (ikcg==10 || *ipr == 1) { *sumerr = 0.; u_max=0; } 
     ii = 0;
     k1 =  *nz + 1;
     for (kk = 0; kk <= k1; ++kk) {
        z = (double) kk * h;
        i1 = *ny + 1;
        for (j = 0; j <= i1; ++j) {
           y = (double) j * h;
           i2 = *nx + 1;
           for (i = 0; i <= i2; ++i) {
              x = (double) i * h;
              kx = a - b * sin(c * x);
              u = exp(-kx/c) * sin(beta * y) * sin(gama * z);
              if (ikcg==10 || *ipr == 1) 
	 	 { if(fabs(u) > u_max) u_max=fabs(u);
		   *sumerr += u * u;
		 }
	      err[ii] = xnew[ii] - u;
/*      printf("\n  ii=  %d  x= %10.5f  y= %10.5f  u=  %10.5f xnew[ii]= %10.5f err(ii)= %10.5f", 
               ii,x,y,u, xnew[ii],err[ii]); */
	      ++ii;
           }
	}
    }
	
	if (ikcg==10 || *ipr == 1) *sumerr = sqrt(*sumerr);
        if(*ipr == 1) 
    	  { printf("\n  SUMERR= %f", *sumerr);
    	    printf("\n  u_max= %15.9f", u_max);
	  }
    return 0;
} /* example32_err_case_c */

/* ******************************************************************** */
/* ******************************************************************** */

/* Subroutine */ int create_matrix_row_5pt_example20(int *row, int *
	location, double *val, int *bindx, int *kmax, double *
	rhs, int *nxst, int *nyst, int *nxloc, int *nyloc, 
	double *kn)
{
    int i1;
    double d1;
    static double f, h;
    static int i, j, k;
    static double x, y, pi;
    static int icn;
    static double sum, sum1;

/*  Purpose:   
    Add one row to an MSR matrix corresponding to a 5pt discrete   
    approximation to the 2D EXAMPLE20 operator on an n x n  square.   

    Example20  taken from Erlangga, Vuik, Oosterlee, 2004 paper, in   
    applied Numerical Mathematics. Example #1 for Helemholtz eq..   

    Domain: [0 1]*[0 1]   

      Uxx + Uyy + K**2*U = (K**2 -5*PI**2)*sin(PI*X)*sin(2*PI*Y)   

    Dirichlet boundary conditions:  u=0 on the boundaries.   

    The analytic solution is:   
       u=sin(PI*X)*sin(2*PI*Y)   on the inner grid points. */
/* ----------------------------------------------------------------------- */
/*  Parameters:   
       row          == global row number of the new row to be added.   
       location     == local row where diagonal of the new row will be stored.   
       val,bindx    == (see user's guide). On output, val[] and bindx[]   
                       are appended such that the new row has been added. */

    pi = atan(1.) * 4.;
    h = 1. / (double) (global_2.n + 1);
    i = *location % *nxloc;
    j = *location / *nxloc % *nyloc;
    x = (*nxst + i) * h;
    y = (*nyst + j) * h;

    icn = 0;
    sum = 0.;
    k = bindx[*location];
    bindx[k] = *row + 1;
    if (*row % global_2.n != global_2.n - 1) {
	val[k] = 1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
    }

    bindx[k] = *row - 1;
    if (*row % global_2.n != 0) {
	val[k] = 1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
    }

    bindx[k] = *row + global_2.n;
    if (*row / global_2.n % global_2.n != global_2.n - 1) {
	val[k] = 1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
    }

    bindx[k] = *row - global_2.n;
    if (*row / global_2.n % global_2.n != 0) {
	val[k] = 1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
    }

    *kmax = k;
    bindx[*location + 1] = k;
    val[*location] = *kn * *kn * h * h - 4.;
    f = (*kn * *kn - pi * 5. * pi) * sin(pi * x) * sin(pi * 2. * y);
    rhs[*row] = f * h * h;

/*  return 0; */
/*  Normalize the equations: */

    d1 = val[*location];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[*row] /= sum1;
    val[*location] /= sum1;
    i1 = icn;
    for (i = 1; i <= i1; ++i) {
	val[*kmax - i] /= sum1;
    }

    return 0;
} /* create_matrix_row_5pt_example20 */

/* ********************************************************************* */
/* ************************************************************************ */

/* Subroutine */ int create_matrix_row_5pt_example21_2rows(int *row, 
	int *location, double *val, int *bindx, int *kmax, 
	double *rhs, int *nxst, int *nyst, int *nxloc, 
	int *nyloc, double *kn1, double *kn2, double *kn3)
{
    int i1;
    double d1;
    static double f, h;
    static int i, j, k;
    static double x, y;
    static int ii;
    static double kn;
    static int icn;
    static double eps, sum, sum1;
    static int loca, rowa;

/*  Purpose:   
    Add one row to an MSR matrix corresponding to a 5pt discrete   
    approximation to the 2D EXAMPLE_21 operator on an (n+2)x(n+2) square.   

    Example 21 = similar to example #2 of Y.A. Erlangga, C. Vuik C.W. Oosterlee   
    2004 paper, in Applied Numerical Mathematics.   

    Domain: [0 1]*[0 1];   

            Uxx + Uyy + K**2*U = F   

    Where: F=0   

    With Dirichlet boundary condition on the bottom: U=0 on y=0, except at   
    the center point:   
              U=1   for x=0.5; y=0   
              U=0   for x#0.5; y=0   

    and Sommerfeld b.c. DU/Dn -i*K*U=0 on all other boundaries. */
/* ----------------------------------------------------------------------- */
/*  Parameters:   
       row          == global row number of the new row to be added.   
       location     == local row where diagonal of the new row will be stored.   
       val,bindx    == (see user's guide). On output, val[] and bindx[]   
                       are appended such that the new row has been added. */

    eps = 1.e-5;
    h = 1. / (double) (global_2.n + 1);
    i = *location / 2 % *nxloc;
    j = *location / 2 / *nxloc;
    x = (*nxst + i - 1) * h;
    y = (*nyst + j - 1) * h;

/*  Note:  The results of the following single and double precision variable definitions may differ. 
    Though, the single def. in fortran = the single idef. in c. */

    if (y <= (1.f/3.f) ) kn = *kn1;
    if ((y > (1.f/3.f)) && (y <= (2.f/3.f))) kn= *kn2;
    if (y > (2.f/3.f) ) kn= *kn3;

/*  if (y <= (1./3.) ) kn = *kn1;
    if (y > (1./3.) && y <= (2./3.) ) kn = *kn2;
    if (y > (2./3.) ) kn = *kn3; */ 

    f = 0.;
/*  For a perturbation at the bottom center: */
    d1=x - .5;
    if (*nyst + j - 1 == 0 && fabs(d1) <= h * .5 + eps)  f = 1.;

/*  Set the  boundary conditions equations   
    on the upper boundary:  du/dn -i*Kn*u=0:  
    where du/dn is the outward derivative. */

    if (j + *nyst - 1 == global_2.n + 1) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	bindx[k] = *row - 2*(global_2.n + 2);
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
	bindx[k] = *row + 1;
	val[k] = kn * h;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	*kmax = k;
	bindx[*location + 1] = k;
/*  the diagonal element: */
	val[*location] = 1.;
	goto L100;
    }

/*     the lower boundary (Dirichlet b.c.): */

    if (j + *nyst - 1 == 0) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	*kmax = k;
	bindx[*location + 1] = k;
	val[*location] = 1.;
	goto L100;
    }

/*  the left hand-side boundary (Sommerfeld b.c.) */
    if (i + *nxst - 1 == 0 && (j + *nyst - 1 > 0 && j + *nyst - 1 < 
	    global_2.n + 1)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	bindx[k] = *row + 2;
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	bindx[k] = *row + 1;
	val[k] = kn * h;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	*kmax = k;
	bindx[*location + 1] = k;
	val[*location] = 1.;
	goto L100;
    }

/*     the right hand-side boundary  (Sommerfeld b.c.) */
    if (i + *nxst - 1 == global_2.n + 1 && (j + *nyst - 1 > 0 && j + *nyst 
	    - 1 < global_2.n + 1)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	bindx[k] = *row - 2;
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	bindx[k] = *row + 1;
	val[k] = kn * h;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	*kmax = k;
	bindx[*location + 1] = k;
	val[*location] = 1.;
	goto L100;
    }
/* --------    end of the b.c. equations setting -------- */

/*  Set the eq. for the inner grid points.   
    In the present example there are no imaginary terms   
    for the inner grid points. */

    icn = 0;
    sum = 0.;
    k = bindx[*location];
    bindx[k] = *row + 2;
    val[k] = 1.;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row - 2;
    val[k] = 1.;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row + 2*(global_2.n + 2);
    val[k] = 1.;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row - 2*(global_2.n + 2);
    val[k] = 1.;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    *kmax = k;
    bindx[*location + 1] = k;
    val[*location] = kn * kn * h * h - 4.;
L100:
    rhs[*row] = f;

/*  return 0; */
/*  Normalize the equations: */

    d1 = val[*location];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[*row] /= sum1;
    val[*location] /= sum1;
/*  printf("\n%7d %13.10f %13.10f", *row, rhs[*row], val[*location]); */
/*  for (ii=1; ii<=i1; ++ii) printf(" %13.10f", val[*kmax - ii]) ; */
    i1 = icn;
    for (ii = 1; ii <= i1; ++ii) {
	val[*kmax - ii] /= sum1;
    }

/* ----------------------------------- */
/* ----------------------------------- */
/*  Store the imaginary equation coefficients (B; A) */

    loca = *location + 1;
    rowa = *row + 1;
    f = 0.;

/*  Set the  b.c. equations: dU/dn -i*Kn*U=0: */

/*  The upper boundary: */
    if (j + *nyst - 1 == global_2.n + 1) {
/*  the real part: */
	icn = 0;
	sum = 0.;
	k = bindx[loca];
	bindx[k] = rowa - 2*(global_2.n + 2);
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
	bindx[k] = rowa - 1;
	val[k] = -kn * h;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	*kmax = k;
	bindx[loca + 1] = k;
/*  the diagonal element: */
	val[loca] = 1.;
	goto L200;
    }

/*  The lower boundary (Dirichlet b.c.): */

    if (j + *nyst - 1 == 0) {
	icn = 0;
	sum = 0.;
	k = bindx[loca];
	*kmax = k;
	bindx[loca + 1] = k;
	val[loca] = 1.;
	goto L200;
    }

/*  The left hand-side boundary (Sommerfeld b.c.): */
    if (i + *nxst - 1 == 0 && (j + *nyst - 1 > 0 && j + *nyst - 1 < 
	    global_2.n + 1)) {
/*  the real part: */
	icn = 0;
	sum = 0.;
	k = bindx[loca];
	bindx[k] = rowa + 2;
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
	bindx[k] = rowa - 1;
	val[k] = -kn * h;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	*kmax = k;
	bindx[loca + 1] = k;
/*  the diagonal element: */
	val[loca] = 1.;
	goto L200;
    }

/*  The right hand-side boundary  (Sommerfeld b.c.): */
    if (i + *nxst - 1 == global_2.n + 1 && (j + *nyst - 1 > 0 && j + *nyst 
	    - 1 < global_2.n + 1)) {
/*  the real part: */
	icn = 0;
	sum = 0.;
	k = bindx[loca];
	bindx[k] = rowa - 2;
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
	bindx[k] = rowa - 1;
	val[k] = -kn * h;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	*kmax = k;
	bindx[loca + 1] = k;
/*  the diagonal element: */
	val[loca] = 1.;
	goto L200;
    }
/* --------    end of the b.c. equation settings -------- */


/*  Set the eq. for the inner grid points.   
    In the present example there are no imaginary terms   
    for the inner grid points. */

    icn = 0;
    sum = 0.;
    k = bindx[loca];
    bindx[k] = rowa + 2;
    val[k] = 1.;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = rowa - 2;
    val[k] = 1.;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = rowa + 2*(global_2.n + 2);
    val[k] = 1.;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = rowa - 2*(global_2.n + 2);
    val[k] = 1.;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    *kmax = k;
    bindx[loca + 1] = k;
    val[loca] = kn * kn * h * h - 4.;
L200:
    rhs[rowa] = f;

/*  return */
/*  Normalize the equations: */

    d1 = val[loca];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[rowa] /= sum1;
    val[loca] /= sum1;
    i1 = icn;
    for (ii = 1; ii <= i1; ++ii) {
	val[*kmax - ii] /= sum1;
    }

    return 0;
} /* create_matrix_row_5pt_example21_2rows */

/* ********************************************************************* */
/* ************************************************************************ */

/* Subroutine */ int create_matrix_row_9pt_example23_2rows(int *row, 
	int *location, double *val, int *bindx, int *kmax, 
	double *rhs, int *nxst, int *nyst, int *nxloc, 
	int *nyloc, double *kn1, double *kn2, double *kn3, 
	int *order)
{
    int i1;
    double d1, d2;
    static double f, h;
    static int i, j, k;
    static double x, y, a0, ac;
    static int ii;
    static double as, kn, del;
    static int icn;
    static double eps, sum, hkn2, sum1, beta;
    static int loca, rowa;

/*  Purpose:   
    Add one row to an MSR matrix corresponding to a 9pt discrete   
    approximation to the 2D EXAMPLE_23 operator on an (n+3)x(n+4) mesh.   

    Example 23 = similar to example #2 of Y.A. Erlangga, C. Vuik and   
    C.W. Oosterlee 2004 paper, in Applied Numerical Mathematics.   
    The variable Order defines the scheme order: 2nd / 4th/ 6th order scheme.   
    The 4th/6th order finite difference schemes follow Singer and   
    Turkel 2006, Erlanga and Turkel  2011? and Harrari and Turkel 1995.   
    with gamma=14/5, and DEL= -1.;0. or. 2. for the 6th order scheme.   

    Domain: [0 1]*[0 1];   

            Uxx + Uyy + K**2*U = F   

    Where: F=0   

    With Dirichlet boundary condition on the bottom: U=0 on y=0, except at   
    the center point:   
             U=1   for x=0.5; y=0   
             U=0   for x#0.5; y=0   

    and Sommerfeld b.c. DU/Dn -i*K*U=0 on all other boundaries. */
/* ----------------------------------------------------------------------- */
/*  Parameters:   
       row          == global row number of the new row to be added.   
       location     == local row where diagonal of the new row will be stored.   
       val,bindx    == (see user's guide). On output, val[] and bindx[]   
                       are appended such that the new row has been added. */
/*  eps = 1.e-5; */
    h = 1./ (double) (global_2.n + 1);
    i = *location / 2 % *nxloc;
    j = *location / 2 / *nxloc;
    x = (*nxst + i - 1) * h;
    y = (*nyst + j - 1) * h;

    if (y <= (1.f/3.f)) kn = *kn1;
    if (y > (1.f/3.f) && y <= (2.f/3.f) ) kn = *kn2;
    if (y > (2.f/3.f) ) kn = *kn3;

    beta = -kn;
    f = 0.;
    d1=x - .5;
/*  if (*nyst + j - 1 == 0 && d1  <= h * .5 + eps && d1 >= -(.5*h + eps)) f = 1.; */
/*  if (*nyst + j - 1 == 0 && d1  <= h * .5 + eps && d1 >= -(.5*h + eps)) f = h*h; */ 
/*  if (*nyst + j - 1 == 0 && fabs(d1) <= h * .5 + eps) f = 1.;   */
    if (*nxst + i - 1 == (global_2.n + 3) / 2 )  f=1.; 
/*  if (*nyst + j - 1 == 0 && fabs(d1) <= h * .5 + eps) f = h*h; */

/*  Set the  boundary conditions equations */
/*  On the upper boundary:  du/dn -i*Kn*u=0: */
/*  where du/dn is the outward derivative. */

    if (j + *nyst - 1 == global_2.n + 2) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	bindx[k] = *row - 4*(global_2.n + 4);
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
	bindx[k] = *row - 2*(global_2.n + 4) + 1;
	if (*order == 4 || *order == 6) {
	    d1 = beta * h;
	    d2 = beta * h, d2 *= d2;
	    val[k] = beta * -2. * h * (1. - d1 * d1 / 6. + d2 * d2 
		    / 120.);
	}
	if (*order == 2) {
	    val[k] = beta * -2. * h;
	}
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	*kmax = k;
	bindx[*location + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
	val[*location] = 1.;
	goto L100;
    }

/*  The lower boundary (Dirichlet b.c.) */

    if (j + *nyst - 1 == 0) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	*kmax = k;
	bindx[*location + 1] = k;
	val[*location] = 1.;
	goto L100;
    }

/*  The left hand-side boundary (Sommerfeld b.c.: dU/dn -i*Kn*U=0) */
    if (i + *nxst - 1 == 0 && (j + *nyst - 1 > 0 && j + *nyst - 1 <= 
	    global_2.n + 1)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	bindx[k] = *row + 4;
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part */
	bindx[k] = *row + 3;
	if (*order == 4 || *order == 6) {
	    d1 = beta * h;
/* Computing 4th power */
	    d2 = beta * h, d2 *= d2;
	    val[k] = beta * -2. * h * (1. - d1 * d1 / 6. + d2 * d2 
		    / 120.);
	}
	if (*order == 2) {
	    val[k] = beta * -2. * h;
	}
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

	if (*order == 2) {
	    goto L35;
	}
	bindx[k] = *row + 2*(global_2.n + 4);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part */
	bindx[k] = *row + 2*(global_2.n + 4) + 1;
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the real part */
	bindx[k] = *row - 2*(global_2.n + 4);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginay part */
	bindx[k] = *row - 2*(global_2.n + 4) + 1;
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
L35:
	*kmax = k;
	bindx[*location + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
	val[*location] = 1.;
	goto L100;
    }

/*  The right hand-side boundary  (Sommerfeld b.c.: dU/dn -i*Kn*U=0 ) */
    if (i + *nxst - 1 == global_2.n + 3 && (j + *nyst - 1 > 0 && j + *nyst 
	    - 1 <= global_2.n + 1)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	bindx[k] = *row - 4;
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginay part */
	bindx[k] = *row - 1;
	if (*order == 4 || *order == 6) {
	    d1 = beta * h;
/* Computing 4th power */
	    d2 = beta * h, d2 *= d2;
	    val[k] = beta * -2. * h * (1. - d1 * d1 / 6. + d2 * d2 
		    / 120.);
	}
	if (*order == 2) {
	    val[k] = beta * -2. * h;
	}
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	if (*order == 2) {
	    goto L45;
	}
	bindx[k] = *row + 2*(global_2.n + 4);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part */
	bindx[k] = *row + 2*(global_2.n + 4) + 1;
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the real part */
	bindx[k] = *row - 2*(global_2.n + 4);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginay part */
	bindx[k] = *row - 2*(global_2.n + 4) + 1;
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
L45:
	*kmax = k;
	bindx[*location + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
	val[*location] = 1.;
	goto L100;
    }
/* --------    end of the b.c. equations setting -------- */

/*  Set the eq. for the inner grid points.   
    In the present example there are no imaginary terms   
    for the inner grid points. */

    hkn2 = kn * kn * h * h;
    if (*order == 2) {
	a0 = hkn2 - 4.;
	as = 1.;
	ac = 0.;
    }
    if (*order == 4) {
	a0 = hkn2 * 67. / 90. - 10./3.;
	as = hkn2 * 2. / 45. + 2./3.;
	ac = hkn2 * 7. / 360. + 1./6.;
    }
    if (*order == 6) {
/*      del= 2. */
	del = -1.;
/*      del= 0. */
	d1 = hkn2;
	a0 = hkn2 * 67. / 90. - 10./3. + d1 * d1 * (del - 3.)  / 180.;
	d1 = hkn2;
	as = hkn2 * 2. / 45. + 2./3. + d1 * d1 * (3. - del *  2.) / 720.;
	d1 = hkn2;
	ac = hkn2 * 7. / 360. + 1./6. + d1 * d1 * del / 720.;
    }
    icn = 0;
    sum = 0.;
    k = bindx[*location];
    bindx[k] = *row + 2;
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row - 2;
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row + 2*(global_2.n + 4);
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    if (*order == 2) {
	goto L50;
    }

    bindx[k] = *row + 2*(global_2.n + 4) + 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row + 2*(global_2.n + 4) - 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

L50:

    bindx[k] = *row - 2*(global_2.n + 4);
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    if (*order == 2) {
	goto L55;
    }
    bindx[k] = *row - 2*(global_2.n + 4) + 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row - 2*(global_2.n + 4) - 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

L55:

    *kmax = k;
    bindx[*location + 1] = k;
    val[*location] = a0;
L100:
    rhs[*row] = f;

/*  return 0; */
/*  Normalize the equations: */

    d1 = val[*location];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[*row] /= sum1;
    val[*location] /= sum1;
/*  printf("\n%7d %13.10f %13.10f", *row, rhs[*row], val[*location]);*/  
    i1 = icn;
    for (ii = 1; ii <= i1; ++ii) {
	val[k - ii] /= sum1;
    }

/* --------------- */
/* --------------- */
/*  Store the imaginary equation coefficients (B; A) */

    loca = *location + 1;
    rowa = *row + 1;
    f = 0.;

/*  Set the  b.c. equations: du/dn -i*Kn*u=0: */
/*  The upper boundary: */

    if (j + *nyst - 1 == global_2.n + 2) {
/*  the real part: */
	icn = 0;
	sum = 0.;
	k = bindx[loca];
	bindx[k] = rowa - 4*(global_2.n + 4);
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
	bindx[k] = rowa - 2*(global_2.n + 4) - 1;
	if (*order == 4 || *order == 6) {
	    d1 = beta * h;
/* Computing 4th power */
	    d2 = beta * h, d2 *= d2;
	    val[k] = beta * 2. * h * (1. - d1 * d1 / 6. + d2 * 
		    d2 / 120.);
	}
	if (*order == 2) {
	    val[k] = beta * 2 * h;
	}
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	*kmax = k;
	bindx[loca + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
	val[loca] = 1.;
	goto L200;
    }

/*  The lower boundary (Dirichlet b.c.) */

    if (j + *nyst - 1 == 0) {
	icn = 0;
	sum = 0.;
	k = bindx[loca];
	*kmax = k;
	bindx[loca + 1] = k;
	val[loca] = 1.;
	goto L200;
    }

/*  The left hand-side boundary (Sommerfeld b.c.) */
    if (i + *nxst - 1 == 0 && (j + *nyst - 1 > 0 && j + *nyst - 1 <= 
	    global_2.n + 1)) {
/*  the real part: */
	icn = 0;
	sum = 0.;
	k = bindx[loca];
	bindx[k] = rowa + 4;
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
	bindx[k] = rowa + 1;
	if (*order == 4 || *order == 6) {
	    d1 = beta * h;
/* Computing 4th power */
	    d2 = beta * h, d2 *= d2;
	    val[k] = beta * 2. * h * (1. - d1 * d1 / 6. + d2 * 
		    d2 / 120.);
	}
	if (*order == 2) {
	    val[k] = beta * 2. * h;
	}
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	if (*order == 2) {
	    goto L65;
	}

	bindx[k] = rowa + 2*(global_2.n + 4);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	bindx[k] = rowa + 2*(global_2.n + 4) - 1;
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	bindx[k] = rowa - 2*(global_2.n + 4);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

	bindx[k] = rowa - 2*(global_2.n + 4) - 1;
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

L65:
	*kmax = k;
	bindx[loca + 1] = k;
/*  the diagonal element: */
	val[loca] = 1.;
	goto L200;
    }

/*  The right hand-side boundary  (Sommerfeld b.c.) */
    if (i + *nxst - 1 == global_2.n + 3 && (j + *nyst - 1 > 0 && j + *nyst 
	    - 1 <= global_2.n + 1)) {
/*  the real part: */
	icn = 0;
	sum = 0.;
	k = bindx[loca];
	bindx[k] = rowa - 4;
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
	bindx[k] = rowa - 3;
	if (*order == 4 || *order == 6) {
	    d1 = beta * h;
/* Computing 4th power */
	    d2 = beta * h, d2 *= d2;
	    val[k] = beta * 2. * h * (1. - d1 * d1 / 6. + d2 *  d2 / 120.);
	}
	if (*order == 2) {
	    val[k] = beta * 2. * h;
	}
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	if (*order == 2) {
	    goto L75;
	}

	bindx[k] = rowa + 2*(global_2.n + 4);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	bindx[k] = rowa + 2*(global_2.n + 4) - 1;
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	bindx[k] = rowa - 2*(global_2.n + 4);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

	bindx[k] = rowa - 2*(global_2.n + 4) - 1;
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

L75:
	*kmax = k;
	bindx[loca + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
	val[loca] = 1.;
	goto L200;
    }
/* --------    end of the b.c. equation setting -------- */

/*  Set the eq. for the inner grid points.   
    In the present example there are no imaginary terms   
    for the inner grid points. */

    icn = 0;
    sum = 0.;
    k = bindx[loca];
    bindx[k] = rowa + 2;
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = rowa - 2;
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = rowa + 2*(global_2.n + 4);
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    if (*order == 2) {
	goto L150;
    }

    bindx[k] = rowa + 2*(global_2.n + 4) + 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = rowa + 2*(global_2.n + 4) - 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

L150:

    bindx[k] = rowa - 2*(global_2.n + 4);
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    if (*order == 2) {
	goto L155;
    }

    bindx[k] = rowa - 2*(global_2.n + 4) + 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = rowa - 2*(global_2.n + 4) - 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

L155:

    *kmax = k;
    bindx[loca + 1] = k;
    val[loca] = a0;
L200:
    rhs[rowa] = f;

/*  return */
/*  Normalize the equations: */
    d1 = val[loca];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[rowa] /= sum1;
    val[loca] /= sum1;
/*  printf("\n%7d %13.10f %13.10f", rowa, rhs[rowa], val[loca]); */    
    i1 = icn;
    for (ii = 1; ii <= i1; ++ii) {
	val[k - ii] /= sum1;
    }

    return 0;
} /* create_matrix_row_9pt_example23_2rows */
/* ********************************************************************** */
/* ********************************************************************** */
/* Subroutine */ int create_matrix_row_5_9_pt_example24_2rows(int *row, 
	int *location, double *val, int *bindx, int *kmax, 
	double *rhs, int *nxst, int *nyst, int *nxloc, 
	int *nyloc, float *cxy, double *xmin, double *xmax, 
	double *ymin, double *ymax, int *nx, int *ny, int 
	*order, double *freq)
{
    int i1;
    double d1, d2;
    static double f, h;
    static int i, j, k;
    static double ac_im1_jm1, ac_ip1_jm1, ac_im1_jp1, ac_ip1_jp1, 
	    x, y, kn_im1_jm1, kn_im1_jp1, kn_ip1_jm1, kn_ip1_jp1, a0, 
	    ac;
    static int ii;
    static double as, pi, hx, hy, del;
    static int icn;
    static double  sum, sum1, beta;
    static int loca, rowa;
    static double cons;
    static double kn_i_j, as_im1_j, as_i_jm1, as_ip1_j, 
	    as_i_jp1, kn_im1_j, kn_i_jm1, kn_ip1_j, kn_i_jp1;


/*  Purpose:    
    Add one row to an MSR matrix corresponding to a 5 / 9 pt discrete   
    approximation to the 2D Marmousi operator on an (nx+4)x(ny+4) domain.   

    Example 24 = example #2 of O. Schenk.  2009   
    paper, SIAM J. Scientific Computing.   

    Domain: [Xmin-hx  Xmax+hx]*[Ymin-hy Ymax+hy];   

            Uxx + Uyy + K**2*U = F   

            K=2*PI*freq/C   

     Sommerfeld b.c. DU/Dn -i*K*U=gi on all  boundaries.   

    and:   
              f=1   for x=0.5, y=0   
              f=0    else   
*/
/* ----------------------------------------------------------------------- */

/*  Parameters: */
/*     row          == global row number of the new row to be added. */
/*     location     == local row where diagonal of the new row will be stored. */
/*     val,bindx    == (see user's guide). On output, val[] and bindx[] */
/*                     are appended such that the new row has been added. */


    hx = (*xmax - *xmin) / (*nx - 1);
    hy = (*ymax - *ymin) / (*ny - 1);
    i = *location / 2 % *nxloc;
    j = *location / 2 / *nxloc;
    x = *xmin - hx + (*nxst + i - 1) * hx;
    y = *ymin - hy + (*nyst + j - 1) * hy;
/* here we assume:  H = Hx = Hy ; */
    h = hx;

    pi = atan(1.) * 4.;
    cons = pi * 2. * *freq;
    kn_i_j = cons / cxy[*row / 2];
/* note: on the grid points outside the domain cxy() is defined as that of the nearest */
/*       grid point of the domain for which the finite difference scheme is applied. */
    if (*order > 2 && i + *nxst - 1 > 0 && i + *nxst - 1 < *nx + 3 && j + 
	    *nyst - 1 > 0 && j + *nyst - 1 < *ny + 3) {
	kn_im1_j = cons / cxy[*row / 2 - 1];
	kn_ip1_j = cons / cxy[*row / 2 + 1];
	kn_ip1_jp1 = cons / cxy[*row / 2 + (*nx + 4) + 1];
	kn_i_jp1 = cons / cxy[*row / 2 + (*nx + 4)];
	kn_im1_jp1 = cons / cxy[*row / 2 + (*nx + 4) - 1];
	kn_ip1_jm1 = cons / cxy[*row / 2 - (*nx + 4) + 1];
	kn_i_jm1 = cons / cxy[*row / 2 - (*nx + 4)];
	kn_im1_jm1 = cons / cxy[*row / 2 - (*nx + 4) - 1];
    }
/*  store the real equation coefficients (ai;-bi) */
    f = 0.;
/*  for a perturbation at the bottom center: */
/*  after the addition of the two external grid points at: (Xmin-h;*) and (Xmax+h;*): */
    if (*nyst + j - 1 == 1 && *nxst + i - 1 == (*nx + 3) / 2) {
	f = 1.;
    }

/*  Set the  boundary conditions equations: */
/* ----------------------------------------- */
/*  on the upper boundary:  du/dn -i*Kn*u=0: */
/*  where du/dn is the outward derivative. */

    if (j + *nyst - 1 == *ny + 3) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
/*  Note: the b.c. equation is set on the point on the boundary (not */
/*  the additional external point) */
	bindx[k] = *row - 4*(*nx + 4);
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
	bindx[k] = *row - 2*(*nx + 4) + 1;
	beta = -kn_i_j;
	if (*order == 4 || *order == 6) {
	    d1 = beta * hy;
	    d2 = beta * hy, d2 *= d2;
	    val[k] = beta * -2 * hy * (1. - d1 * d1 / 6. + d2 * d2 /
		     120.);
	}
	if (*order == 2) {
	    val[k] = beta * -2. * hy;
	}
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	*kmax = k;
	bindx[*location + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
	val[*location] = 1.;
	goto L100;
    }

/*     the lower boundary (Sommerfeld b.c.) */

    if (j + *nyst - 1 == 0) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	bindx[k] = *row + 4*(*nx + 4);
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
	bindx[k] = *row + 2*(*nx + 4) + 1;
	beta = -kn_i_j;
	if (*order == 4 || *order == 6) {
	    d1 = beta * hy;
	    d2 = beta * hy, d2 *= d2;
	    val[k] = beta * -2. * hy * (1. - d1 * d1 / 6. + d2 * 
		    d2 / 120.);
	}
	if (*order == 2) {
	    val[k] = beta * -2. * hy;
	}
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	*kmax = k;
	bindx[*location + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
	val[*location] = 1.;
	goto L100;
    }

/*     the left hand-side boundary (Sommerfeld b.c.) */
    if (i + *nxst - 1 == 0 && (j + *nyst - 1 > 0 && j + *nyst - 1 < *ny + 3)
	    ) {
	icn = 0;
	sum = 0.f;
	k = bindx[*location];
	bindx[k] = *row + 4;
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
	bindx[k] = *row + 3;
	beta = -kn_i_j;
	if (*order == 4 || *order == 6) {
	    d1 = beta * hx;
	    d2 = beta * hx, d2 *= d2;
	    val[k] = beta * -2. * hx * (1. - d1 * d1 / 6. + d2 * 
		    d2 / 120.);
	}
	if (*order == 2) {
	    val[k] = beta * -2. * hx;
	}
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	if (*order == 2) {
	    goto L35;
	}
/*  set zero value for the point exactly above the current (i=0) grid point: */
/*  the real part */
	bindx[k] = *row + 2*(*nx + 4);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part */
	bindx[k] = *row + 2*(*nx + 4) + 1;
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  set zero value for the point exactly below the grid point */
/*  the real part */
	bindx[k] = *row - 2*(*nx + 4);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginay part */
	bindx[k] = *row - 2*(*nx + 4) + 1;
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
L35:
	*kmax = k;
	bindx[*location + 1] = k;
/*  the diagonal element (the grid point outside the computational domain): */
	val[*location] = 1.;
	goto L100;
    }

/*     the right hand-side boundary  (Sommerfeld b.c.) */
    if (i + *nxst - 1 == *nx + 3 && (j + *nyst - 1 > 0 && j + *nyst - 1 < *
	    ny + 3)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	bindx[k] = *row - 4;
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
	bindx[k] = *row - 1;
	beta = -kn_i_j;
	if (*order == 4 || *order == 6) {
	    d1 = beta * hx;
	    d2 = beta * hx, d2 *= d2;
	    val[k] = beta * -2. * hx * (1. - d1 * d1 / 6. + d2 * 
		    d2 / 120.);
	}
	if (*order == 2) {
	    val[k] = beta * -2. * hx;
	}
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	if (*order == 2) {
	    goto L45;
	}
/*  set zero value for the point exactly above the grid point */
/*  the real part */
	bindx[k] = *row + 2*(*nx + 4);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part */
	bindx[k] = *row + 2*(*nx + 4) + 1;
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  set zero value for the point exactly below the grid point */
/*  the real part */
	bindx[k] = *row - 2*(*nx + 4);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginay part */
	bindx[k] = *row - 2*(*nx + 4) + 1;
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
L45:
	*kmax = k;
	bindx[*location + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
	val[*location] = 1.;
	goto L100;
    }
/* --------    end of the b.c. equations setting -------- */

/*   set the eq. for the inner grid points. */
/*   In the present example there are no imaginary terms */
/*   for the inner grid points */

/* here we assume:  H = Hx = Hy ; */
    if (*order == 2) {
	a0 = kn_i_j * kn_i_j * h * h - 4.;
	as = 1.;
	ac = 0.;
	as_i_jm1 = as;
	as_i_jp1 = as;
	as_im1_j = as;
	as_ip1_j = as;
	ac_im1_jm1 = ac;
	ac_ip1_jm1 = ac;
	ac_im1_jp1 = ac;
	ac_ip1_jp1 = ac;
    }
    if (*order == 4) {
/*        DEL= 2. */
/*        DEL=-1. */
/*        DEL= 2.8 !for this value the coefficiens are:  67./90.; 2./45. and 7./360. */
	del = 0.;
/*        A0  = -10.D0/3.D0 + HKN2*67.D0/90.D0 */
	d1 = kn_i_j * h;
	a0 = d1 * d1 * (del + 24.) / 36. - 10./3.;
/*  symmetric points: */
	d1 = kn_i_jm1 * h;
	as_i_jm1 = d1 * d1 * (6. - del) / 72. + 2./3.;
	d1 = kn_i_jp1 * h;
	as_i_jp1 = d1 * d1 * (6. - del) / 72. + 2./3.;
	d1 = kn_im1_j * h;
	as_im1_j = d1 * d1 * (6. - del) / 72. + 2./3.;
	d1 = kn_ip1_j * h;
	as_ip1_j = d1 * d1 * (6. - del) / 72. + 2./3.;
/*  corner points: */
	d1 = kn_im1_jm1 * h;
	ac_im1_jm1 = d1 * d1 * del / 144. + 1./6.;
	d1 = kn_ip1_jm1 * h;
	ac_ip1_jm1 = d1 * d1 * del / 144. + 1./6.;
	d1 = kn_im1_jp1 * h;
	ac_im1_jp1 = d1 * d1 * del / 144. + 1./6.;
	d1 = kn_ip1_jp1 * h;
	ac_ip1_jp1 = d1 * d1 * del / 144. + 1./6.;
    }
/* endif 4th order */
    if (*order == 6) {
/*        DEL= 2. */
	del = -1.;
/*        DEL= 0. */
	d1 = kn_i_j * h;
	d2 = kn_i_j * h, d2 *= d2;
	a0 = d1 * d1 * 67. / 90. - 10./3. + d2 * d2 * (
		del - 3.) / 180.;

/*  symmetric points: */
	d1 = kn_i_jm1 * h;
	d2 = kn_i_jm1 * h, d2 *= d2;
	as_i_jm1 = d1 * d1 * 2. / 45. + 2./3. + d2 * 
		d2 * (3. - del * 2.) / 720.;
	d1 = kn_i_jp1 * h;
	d2 = kn_i_jp1 * h, d2 *= d2;
	as_i_jp1 = d1 * d1 * 2. / 45. + 2./3. + d2 * 
		d2 * (3. - del * 2.) / 720.;
	d1 = kn_im1_j * h;
	d2 = kn_im1_j * h, d2 *= d2;
	as_im1_j = d1 * d1 * 2. / 45. + 2./3. + d2 * 
		d2 * (3. - del * 2.) / 720.;
	d1 = kn_ip1_j * h;
	d2 = kn_ip1_j * h, d2 *= d2;
	as_ip1_j = d1 * d1 * 2. / 45. + 2./3. + d2 * 
		d2 * (3. - del * 2.) / 720.;
/*   corner points: */
	d1 = kn_im1_jm1 * h;
	d2 = kn_im1_jm1 * h, d2 *= d2;
	ac_im1_jm1 = d1 * d1 * 7. / 360. + 1./6. + d2 * 
		d2 * del / 720.;
	d1 = kn_ip1_jm1 * h;
	d2 = kn_ip1_jm1 * h, d2 *= d2;
	ac_ip1_jm1 = d1 * d1 * 7. / 360. + 1./6. + d2 * 
		d2 * del / 720.;
	d1 = kn_im1_jp1 * h;
	d2 = kn_im1_jp1 * h, d2 *= d2;
	ac_im1_jp1 = d1 * d1 * 7. / 360. + 1./6. + d2 * 
		d2 * del / 720.;
	d1 = kn_ip1_jp1 * h;
	d2 = kn_ip1_jp1 * h, d2 *= d2;
	ac_ip1_jp1 = d1 * d1 * 7. / 360. + 1./6. + d2 * 
		d2 * del / 720.;
    }
/* endif 6th order */
    icn = 0;
    sum = 0.;
    k = bindx[*location];
    bindx[k] = *row + 2;
    val[k] = as_ip1_j;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row - 2;
    val[k] = as_im1_j;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row + 2*(*nx + 4);
    val[k] = as_i_jp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row - 2*(*nx + 4);
    val[k] = as_i_jm1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    if (*order == 2) {
	goto L50;
    }

    bindx[k] = *row + 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row + 2*(*nx + 4) - 2;
    val[k] = ac_im1_jp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row - 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jm1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row - 2*(*nx + 4) - 2;
    val[k] = ac_im1_jm1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
L50:

    *kmax = k;
    bindx[*location + 1] = k;
    val[*location] = a0;
L100:
    rhs[*row] = f;

/*     return */
/*  normalize the equations: */
    d1 = val[*location];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[*row] /= sum1;
    val[*location] /= sum1;
    i1 = icn;
    for (ii = 1; ii <= i1; ++ii) {
	val[*kmax - ii] /= sum1;
    }

/* --------------- */
/* --------------- */
/*    store the imaginary equation coefficients (bi; ai) */

    loca = *location + 1;
    rowa = *row + 1;
    f = 0.;

/*    Set the  b.c. equations: du/dn -i*Kn*u=0: */

/*     the upper boundary: */

    if (j + *nyst - 1 == *ny + 3) {
/*  the real part: */
	icn = 0;
	sum = 0.;
	k = bindx[loca];
	bindx[k] = rowa - 4*(*nx + 4);
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
	bindx[k] = rowa - 2*(*nx + 4) - 1;
	beta = -kn_i_j;
	if (*order == 4 || *order == 6) {
	    d1 = beta * hy;
	    d2 = beta * hy, d2 *= d2;
	    val[k] = beta * 2. * hy * (1. - d1 * d1 / 6. + d2 * d2 
		    / 120.);
	}
	if (*order == 2) {
	    val[k] = beta * 2 * hy;
	}
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	*kmax = k;
	bindx[loca + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
	val[loca] = 1.;
	goto L200;
    }

/*     the lower boundary: du/dn -i*Kn*u=0 */

    if (j + *nyst - 1 == 0) {
/*  the real part: */
	icn = 0;
	sum = 0.;
	k = bindx[loca];
	bindx[k] = rowa + 4*(*nx + 4);
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
	bindx[k] = rowa + 2*(*nx + 4) - 1;
	beta = -kn_i_j;
	if (*order == 4 || *order == 6) {
	    d1 = beta * hy;
	    d2 = beta * hy, d2 *= d2;
	    val[k] = beta * 2. * hy * (1. - d1 * d1 / 6. + d2 * d2 
		    / 120.);
	}
	if (*order == 2) {
	    val[k] = beta * 2. * hy;
	}
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	*kmax = k;
	bindx[loca + 1] = k;
/*  the diagonal element: */
/* the additional point - outside the computational domain: */
	val[loca] = 1.;
	goto L200;
    }

/*     the left hand-side boundary (Sommerfeld b.c.) */
    if (i + *nxst - 1 == 0 && (j + *nyst - 1 > 0 && j + *nyst - 1 < *ny + 3)
	    ) {
/*  the real part: */
	icn = 0;
	sum = 0.;
	k = bindx[loca];
	bindx[k] = rowa + 4;
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
/*        bindx(k)  = rowA - 1 */
	bindx[k] = rowa + 1;
	beta = -kn_i_j;
	if (*order == 4 || *order == 6) {
	    d1 = beta * hx;
	    d2 = beta * hx, d2 *= d2;
	    val[k] = beta * 2. * hx * (1. - d1 * d1 / 6. + d2 * d2 
		    / 120.);
	}
	if (*order == 2) {
	    val[k] = beta * 2. * hx;
	}
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

	if (*order == 2) {
	    goto L65;
	}

	bindx[k] = rowa + 2*(*nx + 4);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

	bindx[k] = rowa + 2*(*nx + 4) - 1;
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

	bindx[k] = rowa - 2*(*nx + 4);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

	bindx[k] = rowa - 2*(*nx + 4) - 1;
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
L65:
	*kmax = k;
	bindx[loca + 1] = k;
/*  the diagonal element: */
	val[loca] = 1.;
	goto L200;
    }

/*     the right hand-side boundary  (Sommerfeld b.c.) */
    if (i + *nxst - 1 == *nx + 3 && (j + *nyst - 1 > 0 && j + *nyst - 1 < *
	    ny + 3)) {
/*  the real part: */
	icn = 0;
	sum = 0.;
	k = bindx[loca];
	bindx[k] = rowa - 4;
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
	bindx[k] = rowa - 3;
	beta = -kn_i_j;
	if (*order == 4 || *order == 6) {
	    d1 = beta * hx;
	    d2 = beta * hx, d2 *= d2;
	    val[k] = beta * 2. * hx * (1. - d1 * d1 / 6. + d2 * d2 
		    / 120.);
	}
	if (*order == 2) {
	    val[k] = beta * 2. * hx;
	}
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	if (*order == 2) {
	    goto L75;
	}

	bindx[k] = rowa + 2*(*nx + 4);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

	bindx[k] = rowa + 2*(*nx + 4) - 1;
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

	bindx[k] = rowa - 2*(*nx + 4);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

	bindx[k] = rowa - 2*(*nx + 4) - 1;
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
L75:

	*kmax = k;
	bindx[loca + 1] = k;
/* the diagonal element: */
/* the additional point - outside the computational domain: */
	val[loca] = 1.;
	goto L200;
    }
/* --------    end of the b.c. equation settings -------- */


/*   set the eq. for the inner grid points. */
/*   In the present example there are no imaginary terms */
/*   for the inner grid points */

    icn = 0;
    sum = 0.;
    k = bindx[loca];
    bindx[k] = rowa + 2;
    val[k] = as_ip1_j;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = rowa - 2;
    val[k] = as_im1_j;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = rowa + 2*(*nx + 4);
    val[k] = as_i_jp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = rowa - 2*(*nx + 4);
    val[k] = as_i_jm1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    if (*order == 2) {
	goto L150;
    }

    bindx[k] = rowa + 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = rowa + 2*(*nx + 4) - 2;
    val[k] = ac_im1_jp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = rowa - 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jm1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = rowa - 2*(*nx + 4) - 2;
    val[k] = ac_im1_jm1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
L150:
    *kmax = k;
    bindx[loca + 1] = k;
    val[loca] = a0;
L200:
    rhs[rowa] = f;

/*     return */
/*  normalize the equations: */

    d1 = val[loca];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[rowa] /= sum1;
    val[loca] /= sum1;
    i1 = icn;
    for (ii = 1; ii <= i1; ++ii) {
	val[*kmax - ii] /= sum1;
    }

    return 0;
} /* create_matrix_row_5_9_pt_example24_2rows */

/* ************************************************************************ */
/* ************************************************************************ */

/* Subroutine */ int create_matrix_row_7_27_pt_example25_2rows(int *row, 
	int *location, double *val, int *bindx, int *kmax, 
	double *rhs, int *nxst, int *nyst, int *nzst, int 
	*nxloc, int *nyloc, int *nzloc, double *kn1, double *
	kn2, double *kn3, double *xmin, double *xmax, double *
	ymin, double *ymax, double *zmin, double *zmax, int *
	nx, int *ny, int *nz, int *order)
{
    int i1;
    double d1, d2, d3;
    static double f;
    static int i, j, k;
    static double x, y, z;
    static int ii, kk;
    static double kn, hx, hy, hz, h, beta;
    static int icn;
    static double sum, sum1;
    static int loca;
    static int rowa;
    static double ztot;   
    double z13,z23;
    double as, ac, ad, h2;
    static double ac_i_j_k,ac_ip1_j_k,ac_im1_j_k,ac_i_jp1_k,ac_i_jm1_k,
       ac_i_j_kp1,ac_i_j_km1;
    static double ac_ip1_jp1_k,ac_im1_jp1_k,ac_ip1_jm1_k,ac_im1_jm1_k;
    static double ac_ip1_j_kp1,ac_im1_j_kp1,ac_i_jp1_kp1,ac_i_jm1_kp1,
       ac_ip1_jp1_kp1,ac_im1_jp1_kp1,ac_ip1_jm1_kp1,ac_im1_jm1_kp1;
    static double ac_ip1_j_km1,ac_im1_j_km1,ac_i_jp1_km1,ac_i_jm1_km1,
       ac_ip1_jp1_km1,ac_im1_jp1_km1,ac_ip1_jm1_km1,ac_im1_jm1_km1;


/*  Purpose:   
    Add one row to an MSR matrix corresponding to a 7 /27 pt discrete   
    approximation to the 3D operator on an (nx+4)x(ny+4)x(nz+4) domain.   

    Example 25 is similar to example #3 of O. Schenk.  2009   
    paper, in SIAM J. Scientific Computing, (3D, 3 layer case).   

    Domain: [Xmin-hx  Xmax+hx]*[Ymin-hy Ymax+hy]*[Zmin-hz Zmax+hz];   

            Uxx + Uyy + Uzz + K**2*U = F   

            K=K1; K2; K3 for the 3 different layers.   

    Sommerfeld b.c. DU/Dn -i*K*U = gi  on all boundaries.   

    and:      gi=0   
              f=1   for x=0.5, y=0.5, z=0.   
              f=0    else   
    Order -  the scheme order: 2nd / 6th order scheme.
    The 6th order finite difference scheme follows G. Sutmann, J. of
    Computational and Applied Mathematics, pp. 15-31, 2007,
    - eq. 17 in that paper. A 3D 4th order scheme is not implemented.
    The expression for the b.c are taken from Turkel and Erlangga 2012 paper. */   
/* -------------------------------------------------------------------------- */

/*  Parameters:   
       row          == global row number of the new row to be added.   
       location     == local row where diagonal of the new row will be stored.   
       val,bindx    == (see user's guide). On output, val[] and bindx[]   
                       are appended such that the new row has been added. */
    hx = (*xmax - *xmin) / (*nx + 1);
    hy = (*ymax - *ymin) / (*ny + 1);
    hz = (*zmax - *zmin) / (*nz + 1);
    h = hx;
    i = *location / 2 % *nxloc;
    j = *location / 2 / *nxloc % *nyloc;
    kk = *location / 2 / (*nxloc * *nyloc) % *nzloc;
    x = *xmin - hx + (*nxst + i - 1) * hx;
    y = *ymin - hy + (*nyst + j - 1) * hy;
    z = *zmin - hz + (*nzst + kk - 1) * hz;
    ztot = *zmax - *zmin;   
    z13= (1./3.)*ztot ;
    z23= (2./3.)*ztot ;
    if (z < *zmin + z13) {
	kn = *kn1;
    }
    if (z >= *zmin + z13 && z < *zmin + z23 ) {
	kn = *kn2;
    }
    if (z >= *zmin + z23 ) {
	kn = *kn3;
    }
/*  Store the real equation coefficients (ai;-bi) */
    f = 0.;
/*     for a perturbation at the bottom center: */
/*  if (*nyst + j - 1 == 1 && *nxst + i - 1 == (*nx + 4) / 2 && *nzst + kk 
	    - 1 == (*nz + 4) / 2) {
	f = 1. / (hx * hy * hz);
    } */
/*   1  (NZst+KK-1.eq.(NZ+1)/2)) F = +1. */
/*     for a perturbation at the bottom center: */
    if (*nzst + kk - 1 == 1 && *nxst + i - 1 == (*nx + 4) / 2 && *nyst + j 
	    - 1 == (*ny + 4) / 2) {
  	f = 1. / (hx * hy * hz);   
/*	f = 1. ;   */
/*	f = 1. / (hx * hy * hz); */
    }
/*  Set Sommerfeld b.c. on the upper boundary: y=Ymax: dU/dn -i*Kn*U=0:   
    where dU/dn is the outward derivative. */
    if (j + *nyst - 1 == *ny + 3) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        if (*order == 2) {
           goto L28;
        }
        if (i + *nxst - 1 == *nx + 3) {
           goto L25;
        }
        bindx[k] = *row + 2;
        val[k] = 0.;
        d1 = val[k]; 
	sum += d1 * d1;
	++icn;
	++k;
L25:
        if (i + *nxst - 1 == 0) {
           goto L26;
        }
        bindx[k] = *row - 2;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L26:
        if (kk + *nzst - 1 == *nz + 3) {
           goto L27;
        }
        bindx[k] = *row + 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L27:
        if (kk + *nzst - 1 == 0) {
            goto L28;
        }
        bindx[k] = *row - 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L28:
        bindx[k] = *row - 4*(*nx + 4);
        val[k] = -1.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
/*  the imaginary part: */
        bindx[k] = *row - 2*(*nx + 4) + 1;
        beta = -kn;
        if (*order == 4 || *order == 6) {
        d1 = beta * hy;
/*Computing 4th power */
        d2 = beta * hy, d2 *= d2;
            val[k] = beta * -2. * hy * (1. - d1 * d1 / 6. + d2 *
                    d2 / 120.);
        }
        if (*order == 2) {
            val[k] = beta * -2. * hy;
        }
        d1 = val[k];
        sum += d1 * d1; 
        ++icn;
        ++k;
        *kmax = k;
        bindx[*location + 1] = k;
/*  the diagonal element: */
        val[*location] = 1.;
        goto L100;
    }
/*  Set Sommerfeld b.c. on the lower (y=0) boundary: */

    if (j + *nyst - 1 == 0) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        if (*order == 2) {
            goto L18;
        }
        if (i + *nxst - 1 == *nx + 3) {
            goto L15;
        }
        bindx[k] = *row + 2;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L15:
        if (i + *nxst - 1 == 0) {
            goto L16;
        }
        bindx[k] = *row - 2;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L16:
        if (kk + *nzst - 1 == *nz + 3) {
            goto L17;
        }
        bindx[k] = *row + 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L17:
        if (kk + *nzst - 1 == 0) {
            goto L18;
        }
        bindx[k] = *row - 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L18:
        bindx[k] = *row + 4*(*nx + 4);
        val[k] = -1.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
/*  the imaginary part: */
        bindx[k] = *row + 2*(*nx + 4) + 1;
        beta = -kn;
        if (*order == 6) {
            d1 = beta * hy;
            d2 = beta * hy, d2 *= d2;
            val[k] = beta * -2. * hy * (1. - d1 * d1 / 6. + d2 *
                    d2 / 120.);
        }
        if (*order == 2) {
            val[k] = beta * -2. * hy;
        }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[*location + 1] = k;
/*  the diagonal element: */
        val[*location] = 1.;
        goto L100;
    }
/*  The left hand-side boundary x=0 Sommerfeld b.c): */
    if (i + *nxst - 1 == 0 && (j + *nyst - 1 > 0 && j + *nyst - 1 < *ny + 3)
	     && (kk + *nzst - 1 > 0 && kk + *nzst - 1 < *nz + 3)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        if (*order == 2) {
            goto L20;
        }
        if (j + *nyst - 1 == *ny + 3) {
            goto L21;
        }
        bindx[k] = *row + 2*(*nx + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L21:
        if (j + *nyst - 1 == 0) {
            goto L22;
        }
        bindx[k] = *row - 2*(*nx + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L22:
        if (kk + *nzst - 1 == *nz + 3) {
            goto L23;
        }
        bindx[k] = *row + 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L23:
        if (kk + *nzst - 1 == 0) {
            goto L20;
        }
        bindx[k] = *row - 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L20:
        bindx[k] = *row + 4;
        val[k] = -1.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
/*  the imaginary part: */
        bindx[k] = *row + 3;
        beta = -kn;
        if (*order == 4 || *order == 6) {
            d1 = beta * hx;
            d2 = beta * hx, d2 *= d2;
            val[k] = beta * -2. * hx * (1. - d1 * d1 / 6. + d2 *
                    d2 / 120.);
        }
        if (*order == 2) {
            val[k] = beta * -2. * hx;
        }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[*location + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
        val[*location] = 1.;
        goto L100;
    }
/*  Set Sommerfeld b.c. on right hand-side boundary (x = Xmax) : */
    if (i + *nxst - 1 == *nx + 3 && (j + *nyst - 1 > 0 && j + *nyst - 1 < *
	    ny + 3) && (kk + *nzst - 1 > 0 && kk + *nzst - 1 < *nz + 3)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        if (*order == 2) {
            goto L10;
        }
        if (j + *nyst - 1 == *ny + 3) {
            goto L11;
        }
        bindx[k] = *row + 2*(*nx + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L11:
        if (j + *nyst - 1 == 0) {
            goto L12;
        }
        bindx[k] = *row - 2*(*nx + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L12:
        if (kk + *nzst - 1 == *nz + 3) {
            goto L13;
        }
        bindx[k] = *row + 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L13:
        if (kk + *nzst - 1 == 0) {
            goto L10;
        }
        bindx[k] = *row - 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L10:
        bindx[k] = *row - 4;
        val[k] = -1.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
/*  the imaginary part: */
        bindx[k] = *row - 1;
        beta = -kn;
        if (*order == 4 || *order == 6) {
            d1 = beta * hx;
            d2 = beta * hx, d2 *= d2;
            val[k] = beta * -2. * hx * (1. - d1 * d1 / 6. + d2 *
                    d2 / 120.);
        }
        if (*order == 2) {
            val[k] = beta * -2. * hx;
        }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[*location + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
        val[*location] = 1.;
        goto L100;
    }
/*  On the back side boundary z=0 Sommerfeld b.c.: */
    if (kk + *nzst - 1 == 0 && (i + *nxst - 1 >= 0 && i + *nxst - 1 <= *
	    nx + 3) && (j + *nyst - 1 > 0 && j + *nyst - 1 < *ny + 3)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        bindx[k] = *row + 4*(*nx + 4) * (*ny + 4);
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
        bindx[k] = *row + 2*(*nx + 4) * (*ny + 4) + 1;
        beta = -kn;
        if (*order == 4 || *order == 6) {
            d1 = beta * hz;
            d2 = beta * hz, d2 *= d2;
            val[k] = beta * -2. * hz * (1. - d1 * d1 / 6. + d2 *
                    d2 / 120.);
        }
        if (*order == 2) {
            val[k] = beta * -2. * hz;
        }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[*location + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
        val[*location] = 1.;
        goto L100;
    }
/*  On the front side boundary Z=Zmax Sommerfeld b.c.: */
    if (kk + *nzst - 1 == *nz + 3 && (i + *nxst - 1 >= 0 && i + *nxst - 1 
	    <= *nx + 3) && (j + *nyst - 1 > 0 && j + *nyst - 1 < *ny + 3)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        bindx[k] = *row - 4*(*nx + 4) * (*ny + 4);
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
        bindx[k] = *row - 2*(*nx + 4) * (*ny + 4) + 1;
        beta = -kn;
        if (*order == 4 || *order == 6) {
            d1 = beta * hz;
            d2 = beta * hz, d2 *= d2;
            val[k] = beta * -2. * hz * (1. - d1 * d1 / 6. + d2 *
                    d2 / 120.);
        }
        if (*order == 2) {
            val[k] = beta * -2. * hz;
        }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[*location + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain */
        val[*location] = 1.;
        goto L100;
    }
/* --------    end of the b.c. equations setting -------- */
/*  Set the eq. for the inner grid points. */
/*  here we assume:  H = Hx = Hy = Hz ; */
    h2 = h * h;
    if (*order == 2) {
        ac_i_j_k = (kn * kn * h2 - 6.) / h2;
        ac = 1. / h2;
        ac_i_jm1_k = ac;
        ac_i_jp1_k = ac;
        ac_im1_j_k = ac;
        ac_ip1_j_k = ac;
        ac_i_j_kp1 = ac;
        ac_i_j_km1 = ac;
    }
/* endif Order=2 */
    if (*order == 6) {
/*  k plane: */
/* ------------- */
/*  The following expresions follows Sutmann 2007 paper in JCAM - */
/*  - using the L6c scheme eq. (17) in this paper. */

        d1 = kn * h;
/* Computing 4th power */
        d2 = kn * h, d2 *= d2;
/* Computing 6th power */
        d3 = kn * h, d3 *= d3;
        ac_i_j_k = -64. / 15. / h2 * (1. - d1 * d1 * .25 +
                5. / 256. * d2 * d2  - 1./ 1536. * d3 * (d3 * d3)); 
        as = 7./ 15. / h2 * (1. - d1 * d1 / 21.);
        ac = .1 / h2 * (1. + d1 * d1 / 18.);
        ad = 1./ 30. / h2;

/*  symmetric points */
        ac_i_jm1_k = as;
        ac_i_jp1_k = as;
        ac_im1_j_k = as;
        ac_ip1_j_k = as;
/*  corner points: */
        ac_im1_jm1_k = ac;
        ac_ip1_jm1_k = ac;
        ac_im1_jp1_k = ac;
        ac_ip1_jp1_k = ac;
/*  k-1 plane: */
/* -------------- */
        ac_i_j_km1 = as;
/*  symmetric points : */
        ac_i_jm1_km1 = ac;
        ac_i_jp1_km1 = ac;
        ac_im1_j_km1 = ac;
        ac_ip1_j_km1 = ac;
/*  corner points: */
        ac_im1_jm1_km1 = ad;
        ac_ip1_jm1_km1 = ad;
        ac_im1_jp1_km1 = ad;
        ac_ip1_jp1_km1 = ad;
/*  k+1 plane: */
/* -------------- */
        ac_i_j_kp1 = as;
/*  symmetric points */
        ac_i_jm1_kp1 = ac;
        ac_i_jp1_kp1 = ac;
        ac_im1_j_kp1 = ac;
        ac_ip1_j_kp1 = ac;
/*  corner points: */
        ac_im1_jm1_kp1 = ad;
        ac_ip1_jm1_kp1 = ad;
        ac_im1_jp1_kp1 = ad;
        ac_ip1_jp1_kp1 = ad;
    }
/* endif 6th order */
    icn = 0;
    sum = 0.;
    k = bindx[*location];
    bindx[k] = *row + 2;
    val[k] =  ac_ip1_j_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2;
    val[k] = ac_im1_j_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(*nx + 4);
    val[k] = ac_i_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(*nx + 4);
    val[k] = ac_i_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(*nx + 4) * (*ny + 4);
    val[k] = ac_i_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(*nx + 4) * (*ny + 4);
    val[k] = ac_i_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    if (*order == 2) {
        goto L50;
    }
    bindx[k] = *row + 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(*nx + 4) - 2;
    val[k] = ac_im1_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(*nx + 4) - 2;
    val[k] = ac_im1_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* ----- k+1 plane: */
    bindx[k] = *row + 2*(*nx + 4) * (*ny + 4) + 2;
    val[k] = ac_ip1_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(*nx + 4) * (*ny + 4) - 2;
    val[k] = ac_im1_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4);
    val[k] = ac_i_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4);
    val[k] = ac_i_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* -- */
    bindx[k] = *row + 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4) - 2;
    val[k] = ac_im1_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4) - 2;
    val[k] = ac_im1_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* ----- k-1 plane: */
    bindx[k] = *row - 2*(*nx + 4) * (*ny + 4) + 2;
    val[k] = ac_ip1_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(*nx + 4) * (*ny + 4) - 2;
    val[k] = ac_im1_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4);
    val[k] = ac_i_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4);
    val[k] = ac_i_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* -- */
    bindx[k] = *row - 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4) - 2;
    val[k] = ac_im1_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4) - 2;
    val[k] = ac_im1_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
L50:
    *kmax = k;
    bindx[*location + 1] = k;
    val[*location] = ac_i_j_k;

L100:
    rhs[*row] = f;
/*  return */
/*  normalize the equations: */
    d1 = val[*location];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[*row] /= sum1;
    val[*location] /= sum1;
    i1 = icn;
    for (ii = 1; ii <= i1; ++ii) {
        val[*kmax - ii] /= sum1;
    }
/*  printf("\n%7d %13.10f %13.10f", *row, rhs[*row], val[*location]);
    printf("\n ---\n  ");
    for (ii=1; ii<=i1; ++ii) printf(" %13.10f", val[*kmax - ii]) ;  */ 
/* -------------------------------- */
/* -------------------------------- */
/*  Store the imaginary equation coefficients (bi; ai) */

    loca = *location + 1;
    rowa = *row + 1;
    f = 0.;
/*  Set Sommerfeld b.c. on all boundaries: dU/dn -i*Kn*U=0: */

/*  On the upper boundary y=Ymax Sommefeld b.c.: */
    if (j + *nyst - 1 == *ny + 3) {
/*  the real part: */
	icn = 0;
	sum = 0.;
	k = bindx[loca];
        if (*order == 2) {
            goto L134;
        }
        if (i + *nxst - 1 == *nx + 3) {
            goto L131;
        }
        bindx[k] = rowa + 2;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L131:
        if (i + *nxst - 1 == 0) {
            goto L132;
        }
        bindx[k] = rowa - 2;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L132:
        if (kk + *nzst - 1 == *nz + 3) {
            goto L133;
        }
        bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L133:
        if (kk + *nzst - 1 == 0) {
            goto L134;
        }
        bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L134:
        bindx[k] = rowa - 4*(*nx + 4);
        val[k] = -1.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
/*  the imaginary part: */
        bindx[k] = rowa - 2*(*nx + 4) - 1;
        beta = -kn;
        if (*order == 4 || *order == 6) {
            d1 = beta * hy;
            d2 = beta * hy, d2 *= d2;
            val[k] = beta * 2. * hy * (1. - d1 * d1 / 6. + d2 * d2
                    / 120.);
        }
        if (*order == 2) {
            val[k] = beta * 2 * hy;
        }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[loca + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
        val[loca] = 1.;
        goto L200;
    }
/*  On the lower boundary y=Ymin: dU/dn -i*Kn*U=0 */
    if (j + *nyst - 1 == 0) {
/*  the real part: */
        icn = 0;
        sum = 0.;
        k = bindx[loca];
        if (*order == 2) {
            goto L104;
        }
        if (i + *nxst - 1 == *nx + 3) {
            goto L101;
        }
        bindx[k] = rowa + 2;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L101:
        if (i + *nxst - 1 == 0) {
            goto L102;
        }
        bindx[k] = rowa - 2;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L102:
        if (kk + *nzst - 1 == *nz + 3) {
            goto L103;
        }
        bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L103:
        if (kk + *nzst - 1 == 0) {
            goto L104;
        }
        bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L104:
        bindx[k] = rowa + 4*(*nx + 4);
        val[k] = -1.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
/*  the imaginary part: */
        bindx[k] = rowa + 2*(*nx + 4) - 1;
        beta = -kn;
        if (*order == 4 || *order == 6) {
            d1 = beta * hy;
            d2 = beta * hy, d2 *= d2;
            val[k] = beta * 2. * hy * (1. - d1 * d1 / 6. + d2 * d2
                    / 120.);
        }
        if (*order == 2) {
            val[k] = beta * 2. * hy;
        }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[loca + 1] = k;
/*  the diagonal element: */
/* the additional point - outside the computational domain: */
        val[loca] = 1.;
        goto L200;
    }
/*  On the left boundary Sommerfeld b.c.: */
    if (i + *nxst - 1 == 0 && (j + *nyst - 1 > 0 && j + *nyst - 1 < *ny + 3)
             && (kk + *nzst - 1 > 0 && kk + *nzst - 1 < *nz + 3)) {
/*  the real part: */
        icn = 0;
        sum = 0.;
        k = bindx[loca];
        if (*order == 2) {
            goto L124;
        }
        if (j + *nyst - 1 == *ny + 3) {
            goto L121;
        }
        bindx[k] = rowa + 2*(*nx + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L121:
        if (j + *nyst - 1 == 0) {
            goto L122;
        }
        bindx[k] = rowa - 2*(*nx + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L122:
        if (kk + *nzst - 1 == *nz + 3) {
            goto L123;
        }
        bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L123:
        if (kk + *nzst - 1 == 0) {
            goto L124;
        }
        bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L124:
        bindx[k] = rowa + 4;
        val[k] = -1.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
/*  the imaginary part: */
        bindx[k] = rowa + 1;
        beta = -kn;
        if (*order == 4 || *order == 6) {
            d1 = beta * hx;
            d2 = beta * hx, d2 *= d2;
            val[k] = beta * 2. * hx * (1. - d1 * d1 / 6. + d2 * d2
                    / 120.);
        }
        if (*order == 2) {
            val[k] = beta * 2. * hx;
        }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[loca + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
        val[loca] = 1.;
        goto L200;
    }
/*  On the right boundary  X=Xmax Sommerfeld b.c.: */
    if (i + *nxst - 1 == *nx + 3 && (j + *nyst - 1 > 0 && j + *nyst - 1 < *
            ny + 3) && (kk + *nzst - 1 > 0 && kk + *nzst - 1 < *nz + 3)) {
/*  the real part: */
        icn = 0;
        sum = 0.;
        k = bindx[loca];

        if (*order == 2) {
            goto L114;
        }
        if (j + *nyst - 1 == *ny + 3) {
            goto L111;
        }
        bindx[k] = rowa + 2*(*nx + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L111:
        if (j + *nyst - 1 == 0) {
            goto L112;
        }
        bindx[k] = rowa - 2*(*nx + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L112:
        if (kk + *nzst - 1 == *nz + 3) {
            goto L113;
        }
        bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L113:
        if (kk + *nzst - 1 == 0) {
            goto L114;
        }
        bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L114:
        bindx[k] = rowa - 4;
        val[k] = -1.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
/*  the imaginary part: */
        bindx[k] = rowa - 3;
        beta = -kn;
        if (*order == 4 || *order == 6) {
            d1 = beta * hx;
            d2 = beta * hx, d2 *= d2;
            val[k] = beta * 2. * hx * (1. - d1 * d1 / 6. + d2 * d2
                    / 120.); }
        if (*order == 2) {
            val[k] = beta * 2. * hx; }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[loca + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
        val[loca] = 1.;
        goto L200;
    }
/*  On the back side boundary Sommerfeld b.c.: */
    if (kk + *nzst - 1 == 0 && (i + *nxst - 1 >= 0 && i + *nxst - 1 <= *
            nx + 3) && (j + *nyst - 1 > 0 && j + *nyst - 1 < *ny + 3)) {
/*  the real part: */
        icn = 0;
        sum = 0.;
        k = bindx[loca];
        bindx[k] = rowa + 4*(*nx + 4) * (*ny + 4);
        val[k] = -1.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
/*  the imaginary part: */
        bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4) - 1;
        beta = -kn;
        if (*order == 4 || *order == 6) {
            d1 = beta * hz;
            d2 = beta * hz, d2 *= d2;
            val[k] = beta * 2. * hz * (1. - d1 * d1 / 6. + d2 * d2
                    / 120.); }
        if (*order == 2) {
            val[k] = beta * 2. * hz; }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[loca + 1] = k;
/*  the diagonal element: */
        val[loca] = 1.;
        goto L200;
    }
/*  On the front boundary iZ=Zmax Sommerfeld b.c.: */
    if (kk + *nzst - 1 == *nz + 3 && (i + *nxst - 1 >= 0 && i + *nxst - 1
            <= *nx + 3) && (j + *nyst - 1 > 0 && j + *nyst - 1 < *ny + 3)) {
        icn = 0;
        sum = 0.;
        k = bindx[loca];
        bindx[k] = rowa - 4*(*nx + 4) * (*ny + 4);
        val[k] = -1.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
/*  the imaginary part: */
        bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4) - 1;
        beta = -kn;
        if (*order == 4 || *order == 6) {
            d1 = beta * hz;
            d2 = beta * hz, d2 *= d2;
            val[k] = beta * 2. * hz * (1. - d1 * d1 / 6. + d2 * d2
                    / 120.); }
        if (*order == 2) {
            val[k] = beta * 2. * hz; }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[loca + 1] = k;
/*  the diagonal element: */
        val[loca] = 1.;
        goto L200;
    }
/* --------    end of the b.c. equation settings -------- */

/*  Set the eq. for the inner grid points. */
    icn = 0;
    sum = 0.;
    k = bindx[loca];
    bindx[k] = rowa + 2;
    val[k] = ac_ip1_j_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2;
    val[k] = ac_im1_j_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(*nx + 4);
    val[k] = ac_i_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(*nx + 4);
    val[k] = ac_i_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4);
    val[k] = ac_i_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4);
    val[k] = ac_i_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    if (*order == 2) {
        goto L150; }
    bindx[k] = rowa + 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(*nx + 4) - 2;
    val[k] = ac_im1_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(*nx + 4) - 2;
    val[k] = ac_im1_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* ----- k+1 plane: */
    bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4) + 2;
    val[k] = ac_ip1_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4) - 2;
    val[k] = ac_im1_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4);
    val[k] = ac_i_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4);
    val[k] = ac_i_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* -- */
    bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4) - 2;
    val[k] = ac_im1_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4) - 2;
    val[k] = ac_im1_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* ----- k-1 plane: */
    bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4) + 2;
    val[k] = ac_ip1_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4) - 2;
    val[k] = ac_im1_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4);
    val[k] = ac_i_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4);
    val[k] = ac_i_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* -- */
    bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4) - 2;
    val[k] = ac_im1_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4) - 2;
    val[k] = ac_im1_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
L150:
    *kmax = k;
    bindx[loca + 1] = k;
    val[loca] = ac_i_j_k;
L200:
    rhs[rowa] = f;

/*  return */
/*  normalize the equations: */
    d1 = val[loca];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[rowa] /= sum1;
    val[loca] /= sum1;
    i1 = icn;
    for (ii = 1; ii <= i1; ++ii) {
        val[*kmax - ii] /= sum1;
    }
/*  printf("\n%7d %13.10f %13.10f", rowa, rhs[rowa], val[loca]);
    for (ii=1; ii<=i1; ++ii) printf(" %13.10f", val[*kmax - ii]) ; */   
    return 0;
} /* create_matrix_row_7pt_example25_2rows */

/* ************************************************************************ */
/* ************************************************************************ */

/* Subroutine */ int create_matrix_row_9pt_example26_2rows(int *row, 
	int *location, double *val, int *bindx, int *kmax, 
	double *rhs, int *nxst, int *nyst, int *nxloc, 
	int *nyloc, double *kn1, double *kn2, double *kn3, 
	int *order)
{
    int i1;
    double d1, d2;
    static double f, h;
    static int i, j, k;
    static double x, y, a0, ac;
    static int ii;
    static double as, pi, kn, del;
    static int icn;
    static double sum, hkn2, sum1, beta;
    static int loca, rowa;

/*  Purpose:   
    Add one row to an MSR matrix corresponding to a 5 / 9 point discrete   
    approximation to the 2D EXAMPLE_26 operator on an (n+3)*(n+2) rectangle.   
    Note: An external line of points is added to implement a high order scheme   
    of the boundary condition on the right side of the domain.   

    Example 26 = similar to the example  of Y.A. Erlangga,  and   
    E. Turkel 2012 paper.   
    We use a 4th/6th order finite difference scheme following Singer and   
    Turkel 2006, Erlanga and Turkel  2012 and Harrari and Turkel 1995.        
    The 6th order schem uses gamma=14/5 and DEl=-1. / 2./ 0.   

    Domain: [0 1]*[-0.5 0.5];   

            Uxx + Uyy + K**2*U = F   

    where: F=0   

    With Dirichlet boundary condition on the bottom: U=0 on y=-0.5; on   
    the top:  U=0 on y=0.5, and on the left side: U=cos(PI*y) on x=0.   
    On the right side, Sommerfeld boundary condition: DU/Dn-i*beta*U=0 on x=1.   

    The analytic solution of this problem is:   
        U(x,y)=cos(PI*y)*exp(i*beta*x)   where: PI**2+beta**2=K**2         */
/* ----------------------------------------------------------------------- */
/*  Parameters:   
       row          == global row number of the new row to be added.   
       location     == local row where diagonal of the new row will be stored.   
       val,bindx    == (see user's guide). On output, val[] and bindx[]   
                       are appended such that the new row has been added. */

    pi = atan(1.) * 4.;
    h = 1. / (double) (global_2.n + 1);
    i = *location / 2 % *nxloc;
    j = *location / 2 / *nxloc;
    x = (*nxst + i - 1) * h;
    y = (*nyst + j - 1) * h - .5;

    if (y <= 1.f/3.f-0.5f ) kn = *kn1;
    if (y > 1.f/3.f-0.5f && y <= 2.f/3.f-0.5f ) kn = *kn2;
    if (y > 2.f/3.f-0.5f ) kn = *kn3;

    d1 = kn;
    d2 = pi;
    beta = sqrt(d1 * d1 - d2 * d2);
    f = 0.;
    if (i + *nxst - 1 == 0 && (j + *nyst - 1 > 0 && j + *nyst - 1 < 
	    global_2.n + 1)) f = cos(pi * y);

/*  Set the  boundary conditions equations */

/*  On the upper boundary (Dirichlet b.c.): u=0 */
    if (j + *nyst - 1 == global_2.n + 1) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	*kmax = k;
	bindx[*location + 1] = k;
/*  the diagonal element: */
	val[*location] = 1.;
	goto L100;
    }

/*  On the lower boundary (Dirichlet b.c.): u=0 */
    if (j + *nyst - 1 == 0) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	*kmax = k;
	bindx[*location + 1] = k;
	val[*location] = 1.;
	goto L100;
    }

/*  The left hand-side boundary (Dirichlet b.c.): u=cos(y) */
    if (i + *nxst - 1 == 0 && (j + *nyst - 1 > 0 && j + *nyst - 1 < 
	    global_2.n + 1)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	*kmax = k;
	bindx[*location + 1] = k;
	val[*location] = 1.;
	goto L100;
    }

/*  The right hand-side boundary  (Sommerfeld b.c.): DU/Dn-i*beta*U=0 */
    if (i + *nxst - 1 == global_2.n + 2 && (j + *nyst - 1 > 0 && j + *nyst 
	    - 1 < global_2.n + 1)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	bindx[k] = *row - 4;
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginay part */
	bindx[k] = *row - 1;
	if (*order == 4 || *order == 6) {
	    d1 = beta * h;
/* Computing 4th power */
	    d2 = beta * h, d2 *= d2;
	    val[k] = beta * 2. * h * (1. - d1 * d1 / 6. + d2 * d2 
		    / 120.);
	}
	if (*order == 2) {
	    val[k] = beta * 2. * h;
	}
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	*kmax = k;
	bindx[*location + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
	val[*location] = 1.;
	goto L100;
    }
/* --------    end of the b.c. equations setting -------- */

/*  Set the eq. for the inner grid points.   
    In the present example there are no imaginary terms   
    for the inner grid points */

    hkn2 = kn * kn * h * h;
    if (*order == 2) {
	a0 = hkn2 - 4.;
	as = 1.;
	ac = 0.;
    }
    if (*order == 4) {
	a0 = hkn2 * 67. / 90. - 10./3.;
	as = hkn2 * 2. / 45. + 2./3.;
	ac = hkn2 * 7. / 360. + 1./6.;
    }
    if (*order == 6) {
/*      del= 2. */
	del = -1.;
/*      del= 0. */
	d1 = hkn2;
	a0 = hkn2 * 67. / 90. - 10./3. + d1 * d1 * (del - 3.) / 180.; 
	d1 = hkn2;
	as = hkn2 * 2. / 45. + 2./3. + d1 * d1 * (3. - del * 2.) / 720.;
	d1 = hkn2;
	ac = hkn2 * 7. / 360. + 1./6. + d1 * d1 * del / 720.;
    }
    icn = 0;
    sum = 0.;
    k = bindx[*location];
    bindx[k] = *row + 2;
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row - 2;
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row + 2*(global_2.n + 3);
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    if (*order == 2) {
	goto L50;
    }

    bindx[k] = *row + 2*(global_2.n + 3) + 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row + 2*(global_2.n + 3) - 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

L50:

    bindx[k] = *row - 2*(global_2.n + 3);
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    if (*order == 2) {
	goto L55;
    }
    bindx[k] = *row - 2*(global_2.n + 3) + 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row - 2*(global_2.n + 3) - 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

L55:

    *kmax = k;
    bindx[*location + 1] = k;
    val[*location] = a0;
L100:
    rhs[*row] = f;

/*  return */
/*  Normalize the equations: */

    d1 = val[*location];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[*row] /= sum1;
    val[*location] /= sum1;
    i1 = icn;
    for (ii = 1; ii <= i1; ++ii) {
	val[k - ii] /= sum1;
    }

/* --------------- */
/* --------------- */
/*  Store the imaginary equation coefficients (B; A) */

    loca = *location + 1;
    rowa = *row + 1;
    f = 0.;

/*  Set the  b.c. equations: */

/*  The upper boundary u=0: */
    if (j + *nyst - 1 == global_2.n + 1) {
/*  the real part: */
	icn = 0;
	sum = 0.;
	k = bindx[loca];
	*kmax = k;
	bindx[loca + 1] = k;
/*  the diagonal element: */
	val[loca] = 1.;
	goto L200;
    }

/*  The lower boundary (Dirichlet b.c.) u=0: */
    if (j + *nyst - 1 == 0) {
	icn = 0;
	sum = 0.;
	k = bindx[loca];
	*kmax = k;
	bindx[loca + 1] = k;
	val[loca] = 1.;
	goto L200;
    }

/*  The left hand-side boundary (Dirichlet b.c.): u=cos(PI*y) */
    if (i + *nxst - 1 == 0 && (j + *nyst - 1 > 0 && j + *nyst - 1 < 
	    global_2.n + 1)) {
/*  the real part: */
	icn = 0;
	sum = 0.;
	k = bindx[loca];
	*kmax = k;
	bindx[loca + 1] = k;
/*  the diagonal element: */
	val[loca] = 1.;
	goto L200;
    }

/*  The right hand-side boundary  (Sommerfeld b.c.) */
    if (i + *nxst - 1 == global_2.n + 2 && (j + *nyst - 1 > 0 && j + *nyst 
	    - 1 < global_2.n + 1)) {
/*  the real part: */
	icn = 0;
	sum = 0.;
	k = bindx[loca];
	bindx[k] = rowa - 4;
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
	bindx[k] = rowa - 3;
	if (*order == 4 || *order == 6) {
	    d1 = beta * h;
/* Computing 4th power */
	    d2 = beta * h, d2 *= d2;
	    val[k] = beta *-2. * h * (1. - d1 * d1 / 6. + d2 * d2 / 120.);
	}
	if (*order == 2) {
	    val[k] = beta *-2. * h;
	}
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	*kmax = k;
	bindx[loca + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
	val[loca] = 1.;
	goto L200;
    }
/* --------    end of the b.c. equation setting -------- */

/*  Set the eq. for the inner grid points.   
    In the present example there are no imaginary terms   
    for the inner grid points. */

    icn = 0;
    sum = 0.;
    k = bindx[loca];
    bindx[k] = rowa + 2;
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = rowa - 2;
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = rowa + 2*(global_2.n + 3);
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    if (*order == 2) {
	goto L150;
    }

    bindx[k] = rowa + 2*(global_2.n + 3) + 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = rowa + 2*(global_2.n + 3) - 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

L150:

    bindx[k] = rowa - 2*(global_2.n + 3);
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    if (*order == 2) {
	goto L155;
    }

    bindx[k] = rowa - 2*(global_2.n + 3) + 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = rowa - 2*(global_2.n + 3) - 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

L155:

    *kmax = k;
    bindx[loca + 1] = k;
    val[loca] = a0;
L200:
    rhs[rowa] = f;

/*  return 0; */
/*  Normalize the equations: */

    d1 = val[loca];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[rowa] /= sum1;
    val[loca] /= sum1;
    i1 = icn;
    for (ii = 1; ii <= i1; ++ii) {
	val[k - ii] /= sum1;
    }

    return 0;
} /* create_matrix_row_9pt_example26_2rows */

/* ************************************************************************ */
/* ******************************************************************** */

/* Subroutine */ int create_matrix_row_9pt_example27_2rows(int *row, 
	int *location, double *val, int *bindx, int *kmax, 
	double *rhs, int *nxst, int *nyst, int *nxloc, 
	int *nyloc, double *kn1, double *kn2, double *kn3, 
	int *order)
{
    int i1;
    double d1, d2;
    static double f, h;
    static int i, j, k;
    static double x, y, a0, ac;
    static int ii;
    static double as, pi, kn, del;
    static int icn;
    static double sum, hkn2, sum1, beta;
    static int loca, rowa;

/*  Purpose:   
    Add one row to an MSR matrix corresponding to a 5 / 9 point discrete   
    approximation to the 2D EXAMPLE_27 operator on an (n+2)*(n+3) rectangle.   
    Note: An external line of points is added to implement a high order scheme   
    of the boundary condition on the right side of the domain.   

    Example 27 = similar to the example  of Y.A. Erlangga,  and   
    E. Turkel 2012 paper.   
    The variable Order defines the scheme order: 2nd / 4th/ 6th order scheme.   
    The 4th/6th order finite difference schemes follow Singer and   
    Turkel 2006, Erlanga and Turkel 2012 and Harrari and Turkel 1995.   
    The 6th order schem uses gamma=14/5 and DEl=-1. / 2./ 0.   

    Domain: [-0.5 0.5]*[0 1];   

           Uxx + Uyy + K**2*U = F   

    Where: F=0   

    With Dirichlet boundary condition on the left side: U=0 on x=-0.5; on   
    the right side:  U=0 on x=0.5, and on the bottom: U=cos(PI*x) on y=0.   
    On the top Sommerfeld boundary condition: DU/Dy - i*beta*U=0 on y=1.   

    The analytic solution of this problem is:   
         U(x,y)=cos(PI*x)*exp(i*beta*y)   where: PI**2+beta**2=K**2        */
/* ----------------------------------------------------------------------- */
/*  Parameters:   
       row          == global row number of the new row to be added.   
       location     == local row where diagonal of the new row will be stored.   
       val,bindx    == (see user's guide). On output, val[] and bindx[]   
                       are appended such that the new row has been added. 
       Order        == scheme order */

    pi = atan(1.) * 4.;
    h = 1. / (double) (global_2.n + 1);
    i = *location / 2 % *nxloc;
    j = *location / 2 / *nxloc;
    x = (*nxst + i - 1) * h - .5;
    y = (*nyst + j - 1) * h;

    if (x <= 1.f/3.f-0.5f) kn = *kn1;
    if (x > 1.f/3.f-0.5 && y <= 2.f/3.f-0.5f) kn = *kn2;
    if (x > 2.f/3.f-0.5f ) kn = *kn3;

    d1 = kn;
    d2 = pi;
    beta = sqrt(d1 * d1 - d2 * d2);

    f = 0.;
    if (j + *nyst - 1 == 0 && (i + *nxst - 1 > 0 && i + *nxst - 1 < 
	    global_2.n + 1)) f = cos(pi * x);

/*  Set the  boundary conditions equations */

/*  On the right side boundary (Dirichlet b.c.): u=0 */

    if (i + *nxst - 1 == global_2.n + 1) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	if (*order == 2) {
	    goto L35;
	}
	if ((*order == 4 || *order == 6) && (j + *nyst - 1 == 0 || j + *nyst 
		- 1 == global_2.n + 2)) {
	    goto L35;
	}
	bindx[k] = *row + 2*(global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

	bindx[k] = *row - 2*(global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
L35:
	*kmax = k;
	bindx[*location + 1] = k;
/*  the diagonal element: */
	val[*location] = 1.;
	goto L100;
    }

/*  On the left side boundary (Dirichlet b.c.): u=0 */

    if (i + *nxst - 1 == 0) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	if (*order == 2) {
	    goto L45;
	}
	if ((*order == 4 || *order == 6) && (j + *nyst - 1 == 0 || j + *nyst 
		- 1 == global_2.n + 2)) {
	    goto L45;
	}
	bindx[k] = *row + 2*(global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

	bindx[k] = *row - 2*(global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
L45:
	*kmax = k;
	bindx[*location + 1] = k;
	val[*location] = 1.;
	goto L100;
    }

/*  The  bottom boundary (Dirichlet b.c.): U=cos(Pi*x) */
    if (j + *nyst - 1 == 0 && (i + *nxst - 1 > 0 && i + *nxst - 1 < 
	    global_2.n + 1)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	*kmax = k;
	bindx[*location + 1] = k;
	val[*location] = 1.;
	goto L100;
    }

/*  The top boundary  (Sommerfeld b.c.): DU/Dy-i*beta*U=0 */
    if (j + *nyst - 1 == global_2.n + 2 && (i + *nxst - 1 > 0 && i + *
	    nxst - 1 < global_2.n + 1)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	bindx[k] = *row - 4*(global_2.n + 2);
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginay part */
	bindx[k] = *row - 2*(global_2.n + 2) + 1;
	if (*order == 4 || *order == 6) {
	    d1 = beta * h;
/* Computing 4th power */
	    d2 = beta * h, d2 *= d2;
	    val[k] = beta * 2. * h * (1. - d1 * d1 / 6. + d2 * d2 / 120.); 
	}
	if (*order == 2) {
	    val[k] = beta * 2. * h;
	}
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	*kmax = k;
	bindx[*location + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
	val[*location] = 1.;
	goto L100;
    }
/* --------    end of the b.c. equations setting -------- */

/*  Set the eq. for the inner grid points.   
    In the present example there are no imaginary terms   
    for the inner grid points. */

    hkn2 = kn * kn * h * h;
    if (*order == 2) {
	a0 = hkn2 - 4.;
	as = 1.;
	ac = 0.;
    }
    if (*order == 4) {
	a0 = hkn2 * 67. / 90. - 10./3.;
	as = hkn2 * 2. / 45. + 2./3.;
	ac = hkn2 * 7. / 360. + 1./6.;
    }
    if (*order == 6) {
/*      del= 2. */
	del = -1.;
/*      del= 0. */
	d1 = hkn2;
	a0 = hkn2 * 67. / 90. - 10./3. + d1 * d1 * (del - 3.) / 180.;
	d1 = hkn2;
	as = hkn2 * 2. / 45. + 2./3. + d1 * d1 * (3. - del *  2.) / 720.;
	d1 = hkn2;
	ac = hkn2 * 7. / 360. + 1./6. + d1 * d1 * del / 720.;
    }
    icn = 0;
    sum = 0.;
    k = bindx[*location];
    bindx[k] = *row + 2;
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row - 2;
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row + 2*(global_2.n + 2);
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    if (*order == 2) {
	goto L50;
    }

    bindx[k] = *row + 2*(global_2.n + 2) + 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row + 2*(global_2.n + 2) - 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

L50:

    bindx[k] = *row - 2*(global_2.n + 2);
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    if (*order == 2) {
	goto L55;
    }
    bindx[k] = *row - 2*(global_2.n + 2) + 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row - 2*(global_2.n + 2) - 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

L55:

    *kmax = k;
    bindx[*location + 1] = k;
    val[*location] = a0;
L100:
    rhs[*row] = f;

/*  return 0; */
/*  Normalize the equations: */

    d1 = val[*location];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[*row] /= sum1;
    val[*location] /= sum1;
    i1 = icn;
    for (ii = 1; ii <= i1; ++ii) {
	val[k - ii] /= sum1;
    }

/* ------------------------------------ */
/* ------------------------------------ */
/*  Store the imaginary equation coefficients (B; A) */

    loca = *location + 1;
    rowa = *row + 1;
    f = 0.;

/*  Set the  b.c. equations: */

/*  The right side boundary u=0: */

    if (i + *nxst - 1 == global_2.n + 1) {
/*  the real part: */
	icn = 0;
	sum = 0.;
	k = bindx[loca];
	if (*order == 2) {
	    goto L56;
	}
	if ((*order == 4 || *order == 6) && (j + *nyst - 1 == 0 || j + *nyst 
		- 1 == global_2.n + 2)) {
	    goto L56;
	}
	bindx[k] = rowa + 2*(global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

	bindx[k] = rowa - 2*(global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
L56:
	*kmax = k;
	bindx[loca + 1] = k;
/*  the diagonal element: */
	val[loca] = 1.;
	goto L200;
    }

/*  The left side boundary (Dirichlet b.c.) u=0: */

    if (i + *nxst - 1 == 0) {
	icn = 0;
	sum = 0.;
	k = bindx[loca];
	if (*order == 2) {
	    goto L58;
	}
	if ((*order == 4 || *order == 6) && (j + *nyst - 1 == 0 || j + *nyst 
		- 1 == global_2.n + 2)) {
	    goto L58;
	}
	bindx[k] = rowa + 2*(global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

	bindx[k] = rowa - 2*(global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
L58:
	*kmax = k;
	bindx[loca + 1] = k;
	val[loca] = 1.;
	goto L200;
    }

/*  The bottom boundary (Dirichlet b.c.): u=cos(Pi*x) */
    if (j + *nyst - 1 == 0 && (i + *nxst - 1 > 0 && i + *nxst - 1 < 
	    global_2.n + 1)) {
/*  the real part: */
	icn = 0;
	sum = 0.;
	k = bindx[loca];
	*kmax = k;
	bindx[loca + 1] = k;
/*  the diagonal element: */
	val[loca] = 1.;
	goto L200;
    }

/*  The top boundary  (Sommerfeld b.c.): */
    if (j + *nyst - 1 == global_2.n + 2 && (i + *nxst - 1 > 0 && i + *
	    nxst - 1 < global_2.n + 1)) {
/*  the real part: */
	icn = 0;
	sum = 0.;
	k = bindx[loca];
	bindx[k] = rowa - 4*(global_2.n + 2);
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
	bindx[k] = rowa - 2*(global_2.n + 2) - 1;
	if (*order == 4 || *order == 6) {
	    d1 = beta * h;
	    d2 = beta * h, d2 *= d2;
	    val[k] = beta *-2. * h * (1. - d1 * d1 / 6. + d2 * d2 / 120.);
	}
	if (*order == 2) {
	    val[k] = beta *-2. * h;
	}
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
	*kmax = k;
	bindx[loca + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
	val[loca] = 1.;
	goto L200;
    }
/* --------    end of the b.c. equation setting -------- */


/*  Set the eq. for the inner grid points.   
    In the present example there are no imaginary terms   
    for the inner grid points */

    icn = 0;
    sum = 0.;
    k = bindx[loca];
    bindx[k] = rowa + 2;
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = rowa - 2;
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = rowa + 2*(global_2.n + 2);
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    if (*order == 2) {
	goto L150;
    }

    bindx[k] = rowa + 2*(global_2.n + 2) + 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = rowa + 2*(global_2.n + 2) - 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

L150:

    bindx[k] = rowa - 2*(global_2.n + 2);
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    if (*order == 2) {
	goto L155;
    }

    bindx[k] = rowa - 2*(global_2.n + 2) + 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = rowa - 2*(global_2.n + 2) - 2;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

L155:

    *kmax = k;
    bindx[loca + 1] = k;
    val[loca] = a0;
L200:
    rhs[rowa] = f;

/*  return 0; */
/*  Normalize the equations: */

    d1 = val[loca];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[rowa] /= sum1;
    val[loca] /= sum1;
    i1 = icn;
    for (ii = 1; ii <= i1; ++ii) {
	val[k - ii] /= sum1;
    }

    return 0;
} /* create_matrix_row_9pt_example27_2rows */

/* ************************************************************************ */
/* ******************************************************************** */
/* Subroutine */ int create_matrix_row_9pt_example28(int *row, int *
	location, double *val, int *bindx, int *kmax, double *
	rhs, int *nxst, int *nyst, int *nxloc, int *nyloc, 
	double *kn, int *order)
{
    int i1;
    double d1, d2;
    static double f, h;
    static int i, j, k;
    static double x, y, a0, ac;
    static int ii;
    static double as, pi, del;
    static int icn;
    static double sum, hkn2, sum1, beta;

/*  Purpose:   
    Add one row to an MSR matrix corresponding to a 5 / 9 point discrete   
    approximation to the 2D EXAMPLE_28 operator on an (n+2)*(n+2) rectangle.   
    Note: An external line of points is added to implement a high order scheme   
    of the boundary condition on the right side of the domain.   

    Example 28 - is our example.  

    The variable Order defines the scheme order: 2nd / 4th/ 6th order scheme.   
    The 4th/6th order finite difference schemes follow Singer and   
    Turkel 2006, Erlanga and Turkel 2012 and Harrari and Turkel 1995.        
    The 6th order scheme uses gamma=14/5, and DEl=-1. / 2./ 0.   

    Domain: [0 1]*[0 1];   

            Uxx + Uyy + K**2*U = F   

    Where: F=0   
    Case A:   
    With Dirichlet boundary condition on the left side: u=0 on x=0.; on   
    the right side:  u=0 on x=1.0, and on the bottom: u=0. on y=0.   
    On the top: u=sin(PI*x)*sin(beta) on y=1.   

    The analytic solution is:   
          u(x,y)= sin(PI*x)*sin(beta*y)  with: PI**2+beta**2=K**2          */
/* ----------------------------------------------------------------------- */
/*   Parameters:   
       row          == global row number of the new row to be added.   
       location     == local row where diagonal of the new row will be stored.   
       val,bindx    == (see user's guide). On output, val[] and bindx[]   
                       are appended such that the new row has been added.   
       Order        == scheme order */


    pi = atan(1.) * 4.;
    h = 1. / (double) (global_2.n + 1);
    i = *location % *nxloc;
    j = *location / *nxloc % *nyloc;
    x = (*nxst + i - 1) * h;
    y = (*nyst + j - 1) * h;
    d1 = *kn;
    d2 = pi;
    beta = sqrt(d1 * d1 - d2 * d2);

    f = 0.;
/*  CASE A: */
/* --------- */
/*   if (J+NYst-1.eq.0.and.(I+NXst-1.gt.0.and.I+NXst-1.lt.n+1)) */
/*  1    F=0.D0 */
/*   if (J+NYst-1.eq.n+1.and.(I+NXst-1.gt.0.and.I+NXst-1.lt.n+1)) */
/*  1    F=dsin(PI*X)*dsin(beta) */
/*   if (I+NXst-1.eq.n+1) F=0.D0 */
/*   if (I+NXst-1.eq.0) F=0.D0 */

/*  if (j + *nyst - 1 == 0 && (i + *nxst - 1 > 0 && i + *nxst - 1 <
            global_2.n + 1)) f = 0.;

    if (j + *nyst - 1 == global_2.n + 1 && (i + *nxst - 1 > 0 && i + *
            nxst - 1 < global_2.n + 1)) f = sin(pi * x) * sin(beta);

    if (i + *nxst - 1 == global_2.n + 1) f = 0.;
    if (i + *nxst - 1 == 0) f = 0.;
*/
/*  CASE B: */
/* --------- */
/*   const=Kn/dsqrt(2.D0) */
/*   if (J+NYst-1.eq.0.and.(I+NXst-1.gt.0.and.I+NXst-1.lt.n+1)) */
/*  1    F=dsin(const*X) */
/*   if (J+NYst-1.eq.n+1.and.(I+NXst-1.gt.0.and.I+NXst-1.lt.n+1)) */
/*  1    F=dsin(const*X)*dcos(const) */
/*   if (I+NXst-1.eq.n+1) F=0.D0 */
/*   if (I+NXst-1.eq.0) F=dsin(const)*dcos(const*Y) */
/*  CASE C: */
/* --------- */
    if (j + *nyst - 1 == 0 && (i + *nxst - 1 > 0 && i + *nxst - 1 < 
	    global_2.n + 1)) f = sin(pi * x);

    if (j + *nyst - 1 == global_2.n + 1 && (i + *nxst - 1 > 0 && i + *
	    nxst - 1 < global_2.n + 1)) f = sin(pi * x) * cos(beta);

    if (i + *nxst - 1 == global_2.n + 1) f = 0.;
    if (i + *nxst - 1 == 0) f = 0.;   
/* --------- */

/*  Set the  boundary conditions equations */

/*  On the right side boundary (Dirichlet b.c.): u=0 */

    if (i + *nxst - 1 == global_2.n + 1) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	if (*order == 2) {
	    goto L35;
	}
	if ((*order == 4 || *order == 6) && (j + *nyst - 1 == 0 || j + *nyst 
		- 1 == global_2.n + 1)) {
	    goto L35;
	}
	bindx[k] = *row + (global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

	bindx[k] = *row - (global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
L35:
	*kmax = k;
	bindx[*location + 1] = k;
/*  the diagonal element: */
	val[*location] = 1.;
	goto L100;
    }

/*  On the left side boundary (Dirichlet b.c.): u=0 */

    if (i + *nxst - 1 == 0) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];

	if (*order == 2) {
	    goto L45;
	}
	if ((*order == 4 || *order == 6) && (j + *nyst - 1 == 0 || j + *nyst 
		- 1 == global_2.n + 1)) {
	    goto L45;
	}
	bindx[k] = *row + (global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

	bindx[k] = *row - (global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
L45:
	*kmax = k;
	bindx[*location + 1] = k;
	val[*location] = 1.;
	goto L100;
    }

/*  The bottom boundary (Dirichlet b.c.): */
    if (j + *nyst - 1 == 0 && (i + *nxst - 1 > 0 && i + *nxst - 1 < 
	    global_2.n + 1)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	*kmax = k;
	bindx[*location + 1] = k;
	val[*location] = 1.;
	goto L100;
    }

/*  The top boundary: */
    if (j + *nyst - 1 == global_2.n + 1 && (i + *nxst - 1 > 0 && i + *
	    nxst - 1 < global_2.n + 1)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	*kmax = k;
	bindx[*location + 1] = k;
	val[*location] = 1.;
	goto L100;
    }
/* --------    end of the b.c. equations setting -------- */

/*  Set the eq. for the inner grid points.   
    In the present example there are no imaginary terms   
    for the inner grid points. */

    hkn2 = *kn * *kn * h * h;
    if (*order == 2) {
	a0 = hkn2 - 4.;
	as = 1.;
	ac = 0.;
    }
    if (*order == 4) {
	a0 = hkn2 * 67. / 90. - 10./3.;
	as = hkn2 * 2. / 45. + 2./3. ;
	ac = hkn2 * 7. / 360. + 1./6.;
    }
    if (*order == 6) {
/*      del= 2. */
	del = -1.;
/*      del= 0. */
	d1 = hkn2;
	a0 = hkn2 * 67. / 90. - 10./3.  + d1 * d1 * (del - 3.) 
		/ 180.;
	d1 = hkn2;
	as = hkn2 * 2. / 45. + 2./3.  + d1 * d1 * (3. - del * 
		2.) / 720.;
	d1 = hkn2;
	ac = hkn2 * 7. / 360. + 1./6.  + d1 * d1 * del / 720.;
    }
    icn = 0;
    sum = 0.;
    k = bindx[*location];
    bindx[k] = *row + 1;
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row - 1;
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row + (global_2.n + 2);
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    if (*order == 2) {
	goto L50;
    }

    bindx[k] = *row + (global_2.n + 2) + 1;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row + (global_2.n + 2) - 1;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

L50:

    bindx[k] = *row - (global_2.n + 2);
    val[k] = as;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    if (*order == 2) {
	goto L55;
    }
    bindx[k] = *row - (global_2.n + 2) + 1;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row - (global_2.n + 2) - 1;
    val[k] = ac;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

L55:

    *kmax = k;
    bindx[*location + 1] = k;
    val[*location] = a0;
L100:
    rhs[*row] = f;

/*  return 0; */
/*  Normalize the equations: */

    d1 = val[*location];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[*row] /= sum1;
    val[*location] /= sum1;
    i1 = icn;
    for (ii = 1; ii <= i1; ++ii) {
	val[k - ii] /= sum1;
    }

    return 0;
} /* create_matrix_row_9pt_example28 */

/* ************************************************************************ */
/* ******************************************************************** */

/* Subroutine */ int create_matrix_row_7_27_pt_example29_2rows(int *row, int *
	location, double *val, int *bindx, int *kmax, double *
	rhs, int *nxst, int *nyst, int *nzst, int *nxloc, int *nyloc, int *nzloc, 
	double *kn1, double *kn2, double *kn3, int *order)

{
    int i1;
    double d1, d2;
    static double f;
    static int i, j, kk, ii, k;
    static double x, y, z;
    static double kn, hx, hy, hz, h, alpha, beta, beta1, f1, g1, pi;
    static int nx, ny, nz, icn;
    static double sum, sum1;
    static int loca;
    static int rowa;
    static double z13,z23;   
    static double as, ac, ad, h2;
    static double ac_i_j_k,ac_ip1_j_k,ac_im1_j_k,ac_i_jp1_k,ac_i_jm1_k,
       ac_i_j_kp1,ac_i_j_km1;
    static double ac_ip1_jp1_k,ac_im1_jp1_k,ac_ip1_jm1_k,ac_im1_jm1_k;
    static double ac_ip1_j_kp1,ac_im1_j_kp1,ac_i_jp1_kp1,ac_i_jm1_kp1,
       ac_ip1_jp1_kp1,ac_im1_jp1_kp1,ac_ip1_jm1_kp1,ac_im1_jm1_kp1;
    static double ac_ip1_j_km1,ac_im1_j_km1,ac_i_jp1_km1,ac_i_jm1_km1,
       ac_ip1_jp1_km1,ac_im1_jp1_km1,ac_ip1_jm1_km1,ac_im1_jm1_km1;

/* Purpose:   
   Add one row to an MSR matrix corresponding to a 7 / 27 point discrete
   approximation to the 3D EXAMPLE_29 operator on a 2*(NX+2)*(NY+2)*(NZ+3) cube.
   Example 29 = is our own built example for checking the error of the 6th vs.
   the 2nd order schemes.
   Order - is the scheme order: 2nd / 6th order scheme.
   The 6th order finite difference schemes follows G. Sutmann in the J. of
   Computational and Applied Mathematics, pp. 15-31, 2007.
   the 6th order scheme - eq. 17 in that paper.
   For this 3D case no 4th order scheme is implemented.
   The expression for the b.c are taken from Turkel and erlangga 2012 paper.
 
   Domain: [0 1]*[0 1]*[0 1];
 
           Uxx + Uyy + Uzz + K**2*U = F
 
    where: F=0
 
    With Dirichlet boundary condition on all boundaries except the z=0. plane.
    On the z=0. plane Sommerfeld condition: Du/Dn-i*K*U=0  where n
    is directed outward.
    The analytic solution is:
           U(x,y)=cos(PI*x)*sin(PI*y)*(alpha*cos(beta1*z)- i*sin(beta1*z)) 
     where: beta1**2= K**2-2*PI**2
       and  alpha=beta1/K                                                     */
/*----------------------------------------------------------------------------*/
/*  Parameters:   
       row          == global row number of the new row to be added.   
       location     == local row where diagonal of the new row will be stored.   
       val,bindx    == (see user's guide). On output, val[] and bindx[]   
                       are appended such that the new row has been added.     */
    pi = atan(1.) * 4.;
    h = 1. / (double) (global_2.n + 1);
    hx = h;
    hy = h;
    hz = h;
    nx = global_2.n;
    ny = global_2.n;
    nz = global_2.n;
    i = *location / 2 % *nxloc;
    j = *location / 2 / *nxloc % *nyloc;
    kk = *location / 2 / (*nxloc * *nyloc) % *nzloc;
    x = (*nxst + i - 1) * hx;
    y = (*nyst + j - 1) * hy;
    z = -hz + (*nzst + kk - 1) * hz;

    z13= 1./3. ;
    z23= 2./3. ;
    if (z <  z13) {
	kn = *kn1;
    }
    if (z >=  z13 && z <  z23 ) {
	kn = *kn2;
    }
    if (z >=  z23 ) {
	kn = *kn3;
    }
    beta1 = sqrt(kn*kn - 2.*pi*pi);
    alpha = beta1/kn;

/*  Store the real equation coefficients (ai;-bi) */
/*----------------------------------------------- */
    f = 0.;
    f1 = alpha*cos(beta1*z)*sin(pi*y)*cos(pi*x);
    g1 =  -sin(beta1*z)*sin(pi*y)*cos(pi*x);

/*  Set Dirichlet b.c. on the x=0 plane: */
    if (i + *nxst - 1 == 0 && (j + *nyst - 1 > 0 && j + *nyst - 1 < ny + 1)
	     && (kk + *nzst - 1 >= 0 && kk + *nzst - 1 <= nz + 2)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        if (*order == 2) {
            goto L20;
        }
        if (j + *nyst - 1 == ny + 1) {
            goto L21;
        }
        bindx[k] = *row + 2*(nx + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L21:
        if (j + *nyst - 1 == 0) {
            goto L22;
        }
        bindx[k] = *row - 2*(nx + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L22:
        if (kk + *nzst - 1 == nz + 2) {
            goto L23;
        }
        bindx[k] = *row + 2*(nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L23:
        if (kk + *nzst - 1 == 0) {
            goto L20;
        }
        bindx[k] = *row - 2*(nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L20:
        f = f1;
        *kmax = k;
        bindx[*location + 1] = k;
        val[*location] = 1.;
        goto L100;
    }

/*  Set Dirichlet b.c. on the x=1. plane */ 
    if (i + *nxst - 1 == nx + 1 && (j + *nyst - 1 > 0 && j + *nyst - 1 < 
	    ny + 1) && (kk + *nzst - 1 >= 0 && kk + *nzst - 1 <= nz + 2)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        if (*order == 2) {
           goto L10;
        }
        if (j + *nyst - 1 == ny + 1) {
           goto L11;
        }
        bindx[k] = *row + 2*(nx + 2);
        val[k] = 0.;
        d1 = val[k]; 
	sum += d1 * d1;
	++icn;
	++k;
L11:
        if (j + *nyst - 1 == 0) {
           goto L12;
        }
        bindx[k] = *row - 2*(nx + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L12:
        if (kk + *nzst - 1 == nz + 2) {
           goto L13;
        }
        bindx[k] = *row + 2*(nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L13:
        if (kk + *nzst - 1 == 0) {
            goto L10;
        }
        bindx[k] = *row - 2*(nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L10:
        f = f1;
        *kmax = k;
        bindx[*location + 1] = k;
        val[*location] = 1.;
        goto L100;
    }

/*  Set Dirichlet b.c. on the y=0 plane: */
    if (j + *nyst - 1 == 0) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        if (*order == 2) {
            goto L18;
        }
        if (i + *nxst - 1 == nx + 1) {
            goto L15;
        }
        bindx[k] = *row + 2;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L15:
        if (i + *nxst - 1 == 0) {
            goto L16;
        }
        bindx[k] = *row - 2;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L16:
        if (kk + *nzst - 1 == nz + 2) {
            goto L17;
        }
        bindx[k] = *row + 2*(nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L17:
        if (kk + *nzst - 1 == 0) {
            goto L18;
        }
        bindx[k] = *row - 2*(nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L18:
        f = f1;
        *kmax = k;
        bindx[*location + 1] = k;
        val[*location] = 1.;
        goto L100;
    }

/*  Set Dirichlet b.c. on the y=1 plane*/            
    if (j + *nyst - 1 == ny + 1) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        if (*order == 2) {
            goto L28;
        }
        if (i + *nxst - 1 == nx + 1) {
            goto L25;
        }
        bindx[k] = *row + 2;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L25:
        if (i + *nxst - 1 == 0) {
            goto L26;
        }
        bindx[k] = *row - 2;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L26:
        if (kk + *nzst - 1 == nz + 2) {
            goto L27;
        }
        bindx[k] = *row + 2*(nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L27:
        if (kk + *nzst - 1 == 0) {
            goto L28;
        }
        bindx[k] = *row - 2*(nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L28:
        f = f1;
        *kmax = k;
        bindx[*location + 1] = k;
        val[*location] = 1.;
        goto L100;
    }

/*  Set Sommerfeld b.c. on the boundary z = 0 : Du/Dn - K*u =0 */
/*  where n is directed outward, thus: Du/DZ +iKu=0 */
    if (kk + *nzst - 1 == 0 && (i + *nxst - 1 > 0 && i + *nxst - 1 < 
	    nx + 1) && (j + *nyst - 1 > 0 && j + *nyst - 1 < ny + 1)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        bindx[k] = *row +  4*(nx+2)*(ny+2);
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
        bindx[k] = *row + 2*(nx+2)*(ny+2)+1 ;
        beta = -kn;
        if (*order == 6) {
            d1 = beta * hz;
            d2 = beta * hz, d2 *= d2;
            val[k] = beta * -2. * hz * (1. - d1 * d1 / 6. + d2 *
                    d2 / 120.);
        }
        if (*order == 2) {
            val[k] = beta * -2. * hz;
        }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[*location + 1] = k;
        val[*location] = 1.;
        goto L100;
    }
/*  Set Dirichlet b.c. on the z=1. plane */
    if (kk + *nzst - 1 == nz + 2 && (i + *nxst - 1 > 0 && i + *nxst - 1 
	    < nx + 1) && (j + *nyst - 1 > 0 && j + *nyst - 1 < ny + 1)) {
	icn = 0;
	sum = 0.;
        f = f1;
	k = bindx[*location];
        *kmax = k;
        bindx[*location + 1] = k;
        val[*location] = 1.;
        goto L100;
    }
/* --------    end of the b.c. equations setting -------- */
/*  Set the eq. for the inner grid points. */
/*  We assume:  H = Hx = Hy = Hz ; */
    h2 = h * h;
    if (*order == 2) {
        ac_i_j_k = (kn * kn * h2 - 6.) / h2;
        ac = 1. / h2;
        ac_i_jm1_k = ac;
        ac_i_jp1_k = ac;
        ac_im1_j_k = ac;
        ac_ip1_j_k = ac;
        ac_i_j_kp1 = ac;
        ac_i_j_km1 = ac;
    } /* endif Order=2 */

    if (*order == 6) {
/*  k plane: */
/* ------------- */
/*  The following expresions follows Sutmann 2007 paper in JCAM - */
/*  - using the L6c scheme eq. (17) in this paper. */

        d1 = kn * h;
/* Computing 4th power */
        d2 = kn * h, d2 *= d2;
/* Computing 6th power */
/*      d3 = kn * h, d3 *= d3; */
        ac_i_j_k = -64. / 15. / h2 * (1. - d1 * d1 * .25 +
                5. / 256. * d2 * d2  - 1./ 1536. * d2 * (d2 * d2)); 
        as = 7./ 15. / h2 * (1. - d1 * d1 / 21.);
        ac = .1 / h2 * (1. + d1 * d1 / 18.);
        ad = 1./ 30. / h2;

/*  symmetric points */
        ac_i_jm1_k = as;
        ac_i_jp1_k = as;
        ac_im1_j_k = as;
        ac_ip1_j_k = as;
/*  corner points: */
        ac_im1_jm1_k = ac;
        ac_ip1_jm1_k = ac;
        ac_im1_jp1_k = ac;
        ac_ip1_jp1_k = ac;
/*  k-1 plane: */
/* -------------- */
        ac_i_j_km1 = as;
/*  symmetric points : */
        ac_i_jm1_km1 = ac;
        ac_i_jp1_km1 = ac;
        ac_im1_j_km1 = ac;
        ac_ip1_j_km1 = ac;
/*  corner points: */
        ac_im1_jm1_km1 = ad;
        ac_ip1_jm1_km1 = ad;
        ac_im1_jp1_km1 = ad;
        ac_ip1_jp1_km1 = ad;
/*  k+1 plane: */
/* -------------- */
        ac_i_j_kp1 = as;
/*  symmetric points */
        ac_i_jm1_kp1 = ac;
        ac_i_jp1_kp1 = ac;
        ac_im1_j_kp1 = ac;
        ac_ip1_j_kp1 = ac;
/*  corner points: */
        ac_im1_jm1_kp1 = ad;
        ac_ip1_jm1_kp1 = ad;
        ac_im1_jp1_kp1 = ad;
        ac_ip1_jp1_kp1 = ad;
    }
/* endif 6th order */
    icn = 0;
    sum = 0.;
    k = bindx[*location];
    bindx[k] = *row + 2;
    val[k] =  ac_ip1_j_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2;
    val[k] = ac_im1_j_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(nx + 2);
    val[k] = ac_i_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(nx + 2);
    val[k] = ac_i_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(nx + 2) * (ny + 2);
    val[k] = ac_i_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(nx + 2) * (ny + 2);
    val[k] = ac_i_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    if (*order == 2) {
        goto L50;
    }
    bindx[k] = *row + 2*(nx + 2) + 2;
    val[k] = ac_ip1_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(nx + 2) - 2;
    val[k] = ac_im1_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(nx + 2) + 2;
    val[k] = ac_ip1_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(nx + 2) - 2;
    val[k] = ac_im1_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* ----- k+1 plane: */
    bindx[k] = *row + 2*(nx + 2) * (ny + 2) + 2;
    val[k] = ac_ip1_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(nx + 2) * (ny + 2) - 2;
    val[k] = ac_im1_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(nx + 2) * (ny + 2) + 2*(nx + 2);
    val[k] = ac_i_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(nx + 2) * (ny + 2) - 2*(nx + 2);
    val[k] = ac_i_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* -- */
    bindx[k] = *row + 2*(nx + 2) * (ny + 2) + 2*(nx + 2) + 2;
    val[k] = ac_ip1_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(nx + 2) * (ny + 2) + 2*(nx + 2) - 2;
    val[k] = ac_im1_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(nx + 2) * (ny + 2) - 2*(nx + 2) + 2;
    val[k] = ac_ip1_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(nx + 2) * (ny + 2) - 2*(nx + 2) - 2;
    val[k] = ac_im1_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* ----- k-1 plane: */
    bindx[k] = *row - 2*(nx + 2) * (ny + 2) + 2;
    val[k] = ac_ip1_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(nx + 2) * (ny + 2) - 2;
    val[k] = ac_im1_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(nx + 2) * (ny + 2) + 2*(nx + 2);
    val[k] = ac_i_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(nx + 2) * (ny + 2) - 2*(nx + 2);
    val[k] = ac_i_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* -- */
    bindx[k] = *row - 2*(nx + 2) * (ny + 2) + 2*(nx + 2) + 2;
    val[k] = ac_ip1_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(nx + 2) * (ny + 2) + 2*(nx + 2) - 2;
    val[k] = ac_im1_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(nx + 2) * (ny + 2) - 2*(nx + 2) + 2;
    val[k] = ac_ip1_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(nx + 2) * (ny + 2) - 2*(nx + 2) - 2;
    val[k] = ac_im1_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
L50:
    *kmax = k;
    bindx[*location + 1] = k;
    val[*location] = ac_i_j_k;
L100:
    rhs[*row] = f;
/*  return */
/*  normalize the equations: */
    d1 = val[*location];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[*row] /= sum1;
    val[*location] /= sum1;
    i1 = icn;
    for (ii = 1; ii <= i1; ++ii) {
        val[*kmax - ii] /= sum1;
    }
/*  printf("\n%7d %13.10f %13.10f", *row, rhs[*row], val[*location]);
    for (ii=1; ii<=i1; ++ii) printf(" %13.10f", val[*kmax - ii]) ;  */ 
/* -------------------------------- */
/* -------------------------------- */
/*  Store the imaginary equation coefficients (bi; ai) */

    loca = *location + 1;
    rowa = *row + 1;
    f = 0.;

/*  Set Dirichlet b.c. on the x=0 plane: */
    if (i + *nxst - 1 == 0 && (j + *nyst - 1 > 0 && j + *nyst - 1 < ny + 1)
             && (kk + *nzst - 1 >= 0 && kk + *nzst - 1 <= nz + 2)) {
        icn = 0;
        sum = 0.;
        k = bindx[loca];
        if (*order == 2) {
            goto L124;
        }
        if (j + *nyst - 1 == ny + 1) {
            goto L121;
        }
        bindx[k] = rowa + 2*(nx + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L121:
        if (j + *nyst - 1 == 0) {
            goto L122;
        }
        bindx[k] = rowa - 2*(nx + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L122:
        if (kk + *nzst - 1 == nz + 2) {
            goto L123;
        }
        bindx[k] = rowa + 2*(nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L123:
        if (kk + *nzst - 1 == 0) {
            goto L124;
        }
        bindx[k] = rowa - 2*(nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L124:
        f = g1;
        *kmax = k;
        bindx[loca + 1] = k;
        val[loca] = 1.;
        goto L200;
    }

/*  Set Dirichlet b.c. on the plane x=1.: */
    if (i + *nxst - 1 == nx + 1 && (j + *nyst - 1 > 0 && j + *nyst - 1 < 
            ny + 1) && (kk + *nzst - 1 >= 0 && kk + *nzst - 1 <= nz + 2)) {
        icn = 0;
        sum = 0.;
        k = bindx[loca];

        if (*order == 2) {
            goto L114;
        }
        if (j + *nyst - 1 == ny + 1) {
            goto L111;
        }
        bindx[k] = rowa + 2*(nx + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L111:
        if (j + *nyst - 1 == 0) {
            goto L112;
        }
        bindx[k] = rowa - 2*(nx + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L112:
        if (kk + *nzst - 1 == nz + 2) {
            goto L113;
        }
        bindx[k] = rowa + 2*(nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L113:
        if (kk + *nzst - 1 == 0) {
            goto L114;
        }
        bindx[k] = rowa - 2*(nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L114:
        f = g1;
        *kmax = k;
        bindx[loca + 1] = k;
        val[loca] = 1.;
        goto L200;
    }

/*  Set Dirichlet b.c. on the y=0. plane */
    if (j + *nyst - 1 == 0) {
        icn = 0;
        sum = 0.;
        k = bindx[loca];
        if (*order == 2) {
            goto L104;
        }
        if (i + *nxst - 1 == nx + 1) {
            goto L101;
        }
        bindx[k] = rowa + 2;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L101:
        if (i + *nxst - 1 == 0) {
            goto L102;
        }
        bindx[k] = rowa - 2;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L102:
        if (kk + *nzst - 1 == nz + 2) {
            goto L103;
        }
        bindx[k] = rowa + 2*(nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L103:
        if (kk + *nzst - 1 == 0) {
            goto L104;
        }
        bindx[k] = rowa - 2*(nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L104:
        f = g1;
        *kmax = k;
        bindx[loca + 1] = k;
        val[loca] = 1.;
        goto L200;
    }

/*  Set Dirichlet b.c. on the y=1. plane */
    if (j + *nyst - 1 == ny + 1) {
	icn = 0;
	sum = 0.;
	k = bindx[loca];
        f = g1;
        if (*order == 2) {
            goto L134;
        }
        if (i + *nxst - 1 == nx + 1) {
            goto L131;
        }
        bindx[k] = rowa + 2;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L131:
        if (i + *nxst - 1 == 0) {
            goto L132;
        }
        bindx[k] = rowa - 2;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L132:
        if (kk + *nzst - 1 == nz + 2) {
            goto L133;
        }
        bindx[k] = rowa + 2*(nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L133:
        if (kk + *nzst - 1 == 0) {
            goto L134;
        }
        bindx[k] = rowa - 2*(nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L134:
        *kmax = k;
        bindx[loca + 1] = k;
/*  the diagonal element: */
        val[loca] = 1.;
        goto L200;
    }

/*  Set Sommerfeld b.c. on plane z=0: */
    if (kk + *nzst - 1 == 0 && (i + *nxst - 1 > 0 && i + *nxst - 1 <
            nx + 1) && (j + *nyst - 1 > 0 && j + *nyst - 1 < ny + 1)) {
/*  the real part: */
        icn = 0;
        sum = 0.;
        k = bindx[loca];
        bindx[k] = rowa + 4*(nx + 2) * (ny + 2);
        val[k] = -1.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
/*  the imaginary part: */
        bindx[k] = rowa + 2*(nx + 2) * (ny + 2) - 1;
        beta = -kn;
        if (*order == 6) {
            d1 = beta * hz;
            d2 = beta * hz, d2 *= d2;
            val[k] = beta * 2. * hz * (1. - d1 * d1 / 6. + d2 * d2
                    / 120.); }
        if (*order == 2) {
            val[k] = beta * 2. * hz; }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[loca + 1] = k;
/*  the diagonal element: */
        val[loca] = 1.;
        goto L200;
    }

/*  Set Dirichlet b.c. on the z=1. plane */
    if (kk + *nzst - 1 == nz + 2 && (i + *nxst - 1 > 0 && i + *nxst - 1
            < nx + 1) && (j + *nyst - 1 > 0 && j + *nyst - 1 < ny + 1)) {
        icn = 0;
        sum = 0.;
        f = g1;
        k = bindx[loca];
        *kmax = k;
        bindx[loca + 1] = k;
        val[loca] = 1.;
        goto L200;
    }
/* --------    end of the b.c. equation settings -------- */

/*  Set the eq. for the inner grid points. */
    icn = 0;
    sum = 0.;
    k = bindx[loca];
    bindx[k] = rowa + 2;
    val[k] = ac_ip1_j_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2;
    val[k] = ac_im1_j_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(nx + 2);
    val[k] = ac_i_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(nx + 2);
    val[k] = ac_i_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(nx + 2) * (ny + 2);
    val[k] = ac_i_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(nx + 2) * (ny + 2);
    val[k] = ac_i_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    if (*order == 2) {
        goto L150; }
    bindx[k] = rowa + 2*(nx + 2) + 2;
    val[k] = ac_ip1_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(nx + 2) - 2;
    val[k] = ac_im1_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(nx + 2) + 2;
    val[k] = ac_ip1_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(nx + 2) - 2;
    val[k] = ac_im1_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* ----- k+1 plane: */
    bindx[k] = rowa + 2*(nx + 2) * (ny + 2) + 2;
    val[k] = ac_ip1_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(nx + 2) * (ny + 2) - 2;
    val[k] = ac_im1_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(nx + 2) * (ny + 2) + 2*(nx + 2);
    val[k] = ac_i_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(nx + 2) * (ny + 2) - 2*(nx + 2);
    val[k] = ac_i_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* -- */
    bindx[k] = rowa + 2*(nx + 2) * (ny + 2) + 2*(nx + 2) + 2;
    val[k] = ac_ip1_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(nx + 2) * (ny + 2) + 2*(nx + 2) - 2;
    val[k] = ac_im1_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(nx + 2) * (ny + 2) - 2*(nx + 2) + 2;
    val[k] = ac_ip1_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(nx + 2) * (ny + 2) - 2*(nx + 2) - 2;
    val[k] = ac_im1_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* ----- k-1 plane: */
    bindx[k] = rowa - 2*(nx + 2) * (ny + 2) + 2;
    val[k] = ac_ip1_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(nx + 2) * (ny + 2) - 2;
    val[k] = ac_im1_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(nx + 2) * (ny + 2) + 2*(nx + 2);
    val[k] = ac_i_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(nx + 2) * (ny + 2) - 2*(nx + 2);
    val[k] = ac_i_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* -- */
    bindx[k] = rowa - 2*(nx + 2) * (ny + 2) + 2*(nx + 2) + 2;
    val[k] = ac_ip1_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(nx + 2) * (ny + 2) + 2*(nx + 2) - 2;
    val[k] = ac_im1_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(nx + 2) * (ny + 2) - 2*(nx + 2) + 2;
    val[k] = ac_ip1_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(nx + 2) * (ny + 2) - 2*(nx + 2) - 2;
    val[k] = ac_im1_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
L150:
    *kmax = k;
    bindx[loca + 1] = k;
    val[loca] = ac_i_j_k;
L200:
    rhs[rowa] = f;

/*  return */
/*  normalize the equations: */
    d1 = val[loca];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[rowa] /= sum1;
    val[loca] /= sum1;
    i1 = icn;
    for (ii = 1; ii <= i1; ++ii) {
        val[*kmax - ii] /= sum1;
    }
/*  printf("\n%7d %13.10f %13.10f", rowa, rhs[rowa], val[loca]);
    for (ii=1; ii<=i1; ++ii) printf(" %13.10f", val[*kmax - ii]) ;  */
    return 0;
} /* create_matrix_row_7_27pt_example29_2rows */

/* ************************************************************************ */
/* ************************************************************************ */

/* Subroutine */ int create_matrix_row_7_27_pt_example30_2rows(int *row, 
	int *location, double *val, int *bindx, int *kmax, 
	double *rhs, int *nxst, int *nyst, int *nzst, int 
	*nxloc, int *nyloc, int *nzloc, float *cxy, 
	double *xmin, double *xmax, double *ymin, double *ymax,
	double *zmin, double *zmax, int *nx, int *ny, int *nz,
        int *order, double *freq, int  *me)
{
    int i1;
    double d1, d2;
    static double f;
    static int i, j, k, ii, kk;
    static double x, y, z;
    static double kn, hx, hy, hz, h, beta, cons, pi;
    static int icn;
    static double sum, sum1;
    static int loca;
    static int rowa;
    static double as, ac, ad, h2;
    static double ac_i_j_k,ac_ip1_j_k,ac_im1_j_k,ac_i_jp1_k,ac_i_jm1_k,
       ac_i_j_kp1,ac_i_j_km1;
    static double ac_ip1_jp1_k,ac_im1_jp1_k,ac_ip1_jm1_k,ac_im1_jm1_k;
    static double ac_ip1_j_kp1,ac_im1_j_kp1,ac_i_jp1_kp1,ac_i_jm1_kp1,
       ac_ip1_jp1_kp1,ac_im1_jp1_kp1,ac_ip1_jm1_kp1,ac_im1_jm1_kp1;
    static double ac_ip1_j_km1,ac_im1_j_km1,ac_i_jp1_km1,ac_i_jm1_km1,
       ac_ip1_jp1_km1,ac_im1_jp1_km1,ac_ip1_jm1_km1,ac_im1_jm1_km1;


/*  Purpose:   
    Add one row to an MSR matrix corresponding to a 7 /27 pt discrete   
    approximation to the 3D operator on an (nx+4)x(ny+4)x(nz+4) domain.   

    Example 30 is implementation of the SEG/EAGE salt model problem. The problem 
    definition and boundary conditions are  similar to example #3 of O. Schenk
    2009 paper, in SIAM J. Scientific Computing, (3D, 3 layer case), but
    with variable K given by the salt model.  

    Domain: [Xmin-hx  Xmax+hx]*[Ymin-hy Ymax+hy]*[Zmin-hz Zmax+hz];   

            Uxx + Uyy + Uzz + K**2*U = f  

            K=K1; K2; K3 for the 3 different layers.   

    Sommerfeld b.c. DU/Dn -i*K*U = gi  on all boundaries.   

    and:      gi=0   
              f=1   for x=0.5, y=0.5, z=0.   
              f=0    else 
    Order - is the scheme order: 2nd / 6th order scheme.
    The 6th order finite difference scheme follows G. Sutmann, in the J. of
    Computational and Applied Mathematics, pp. 15-31, 2007,
    - eq. 17 in that paper. Note: this 6th order scheme is applicable only for 
    constant K. A 4th order scheme is not implemented here.
    The expression for the b.c are taken from Turkel and Erlangga 2012  paper. */
/* --------------------------------------------------------------------------- */

/*  Parameters:   
       row          == global row number of the new row to be added.   
       location     == local row where diagonal of the new row will be stored.   
       val,bindx    == (see user's guide). On output, val[] and bindx[]   
                       are appended such that the new row has been added.     */

    hx = (*xmax - *xmin) / (*nx + 1);
    hy = (*ymax - *ymin) / (*ny + 1);
    hz = (*zmax - *zmin) / (*nz + 1);
    h  =  hx;
    i = *location / 2 % *nxloc;
    j = *location / 2 / *nxloc % *nyloc;
    kk = *location / 2 / (*nxloc * *nyloc) % *nzloc;
    x = *xmin - hx + (*nxst + i - 1) * hx;
    y = *ymin - hy + (*nyst + j - 1) * hy;
    z = *zmin - hz + (*nzst + kk - 1) * hz;
    pi = atan(1.) * 4.;
    cons = pi * 2. * *freq;
    kn = cons / cxy[*row / 2 ];

/*  Store the real equation coefficients (ai;-bi) */
    f = 0.;
/*     for a perturbation at the bottom center (on y=0): */
/*  if (*nyst + j - 1 == 1 && *nxst + i - 1 == (*nx + 4) / 2 && *nzst + kk 
	    - 1 == (*nz + 4) / 2) {
	f = 1. / (hx * hy * hz); } */

/*   1  (NZst+KK-1.eq.(NZ+1)/2)) F = +1. */
/*     for a perturbation at the center of the back plane (on z=0): */
    if (*nzst + kk - 1 == 1 && *nxst + i - 1 == (*nx + 4) / 2 && *nyst + j 
	    - 1 == (*ny + 4) / 2) {
  	f = 1. ; }   
/*	f = 1. / (hx * hy * hz); } */

/*  Set Sommerfeld b.c. on the x=Xmin plane: */
    if (i + *nxst - 1 == 0 && (j + *nyst - 1 > 0 && j + *nyst - 1 < *ny + 3)
	     && (kk + *nzst - 1 > 0 && kk + *nzst - 1 < *nz + 3)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        if (*order == 2) {
            goto L20;
        }
        if (j + *nyst - 1 == *ny + 3) {
            goto L21;
        }
        bindx[k] = *row + 2*(*nx + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L21:
        if (j + *nyst - 1 == 0) {
            goto L22;
        }
        bindx[k] = *row - 2*(*nx + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L22:
        if (kk + *nzst - 1 == *nz + 3) {
            goto L23;
        }
        bindx[k] = *row + 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L23:
        if (kk + *nzst - 1 == 0) {
            goto L20;
        }
        bindx[k] = *row - 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L20:
        bindx[k] = *row + 4;
        val[k] = -1.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
/*  the imaginary part: */
        bindx[k] = *row + 3;
        beta = -kn;
        if (*order == 6) {
            d1 = beta * hx;
            d2 = beta * hx, d2 *= d2;
            val[k] = beta * -2. * hx * (1. - d1 * d1 / 6. + d2 *
                    d2 / 120.);
        }
        if (*order == 2) {
            val[k] = beta * -2. * hx;
        }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[*location + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
        val[*location] = 1. ;
        goto L100;
    }
/*  Set Sommerfeld b.c. on right boundary x = Xmax : */
    if (i + *nxst - 1 == *nx + 3 && (j + *nyst - 1 > 0 && j + *nyst - 1 < 
	    *ny + 3) && (kk + *nzst - 1 > 0 && kk + *nzst - 1 < *nz + 3)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        if (*order == 2) {
            goto L10;
        }
        if (j + *nyst - 1 == *ny + 3) {
            goto L11;
        }
        bindx[k] = *row + 2*(*nx + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L11:
        if (j + *nyst - 1 == 0) {
            goto L12;
        }
        bindx[k] = *row - 2*(*nx + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L12:
        if (kk + *nzst - 1 == *nz + 3) {
            goto L13;
        }
        bindx[k] = *row + 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L13:
        if (kk + *nzst - 1 == 0) {
            goto L10;
        }
        bindx[k] = *row - 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L10:
        bindx[k] = *row - 4;
        val[k] = -1.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
/*  the imaginary part: */
        bindx[k] = *row - 1;
        beta = -kn;
        if (*order == 6) {
            d1 = beta * hx;
            d2 = beta * hx, d2 *= d2;
            val[k] = beta * -2. * hx * (1. - d1 * d1 / 6. + d2 *
                    d2 / 120.);
        }
        if (*order == 2) {
            val[k] = beta * -2. * hx;
        }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[*location + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
        val[*location] = 1.;
        goto L100;
    }

/*  Set Sommerfeld b.c. on the y=0 boundary: */

    if (j + *nyst - 1 == 0) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        if (*order == 2) {
            goto L18;
        }
        if (i + *nxst - 1 == *nx + 3) {
            goto L15;
        }
        bindx[k] = *row + 2;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L15:
        if (i + *nxst - 1 == 0) {
            goto L16;
        }
        bindx[k] = *row - 2;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L16:
        if (kk + *nzst - 1 == *nz + 3) {
            goto L17;
        }
        bindx[k] = *row + 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L17:
        if (kk + *nzst - 1 == 0) {
            goto L18;
        }
        bindx[k] = *row - 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L18:
        bindx[k] = *row + 4*(*nx + 4);
        val[k] = -1.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
/*  the imaginary part: */
        bindx[k] = *row + 2*(*nx + 4) + 1;
        beta = -kn;
        if (*order == 6) {
            d1 = beta * hy;
            d2 = beta * hy, d2 *= d2;
            val[k] = beta * -2. * hy * (1. - d1 * d1 / 6. + d2 *
                    d2 / 120.);
        }
        if (*order == 2) {
            val[k] = beta * -2. * hy;
        }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[*location + 1] = k;
/*  the diagonal element: */
        val[*location] = 1.;
        goto L100;
    }

/*  Set Sommerfeld b.c. on the upper boundary y=Ymax: dU/dn -i*Kn*U=0:   
    where dU/dn is the outward derivative. */
    if (j + *nyst - 1 == *ny + 3) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        if (*order == 2) {
           goto L28;
        }
        if (i + *nxst - 1 == *nx + 3) {
           goto L25;
        }
        bindx[k] = *row + 2;
        val[k] = 0.;
        d1 = val[k]; 
	sum += d1 * d1;
	++icn;
	++k;
L25:
        if (i + *nxst - 1 == 0) {
           goto L26;
        }
        bindx[k] = *row - 2;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L26:
        if (kk + *nzst - 1 == *nz + 3) {
           goto L27;
        }
        bindx[k] = *row + 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L27:
        if (kk + *nzst - 1 == 0) {
            goto L28;
        }
        bindx[k] = *row - 2*(*nx + 4 ) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L28:
        bindx[k] = *row - 4*(*nx + 4);
        val[k] = -1.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
/*  the imaginary part: */
        bindx[k] = *row - 2*(*nx + 4) + 1;
        beta = -kn;
        if (*order == 6) {
            d1 = beta * hy;
            d2 = beta * hy, d2 *= d2;
            val[k] = beta * -2. * hy * (1. - d1 * d1 / 6. + d2 *
                    d2 / 120.);
        }
        if (*order == 2) {
            val[k] = beta * -2. * hy;
        }
        d1 = val[k];
        sum += d1 * d1; 
        ++icn;
        ++k;
        *kmax = k;
        bindx[*location + 1] = k;
/*  the diagonal element: */
        val[*location] = 1.;
        goto L100;
    }

/*  Set Sommerfeld b.c. on the z=Zmin plane: */
    if (kk + *nzst - 1 == 0 && (i + *nxst - 1 >= 0 && i + *nxst - 1 <= 
	    *nx + 3) && (j + *nyst - 1 > 0 && j + *nyst - 1 < *ny + 3)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        bindx[k] = *row + 4*(*nx + 4) * (*ny + 4);
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
        bindx[k] = *row + 2*(*nx + 4) * (*ny + 4) + 1;
        beta = -kn;
        if (*order == 6) {
            d1 = beta * hz;
            d2 = beta * hz, d2 *= d2;
            val[k] = beta * -2. * hz * (1. - d1 * d1 / 6. + d2 *
                    d2 / 120.);
        }
        if (*order == 2) {
            val[k] = beta * -2. * hz;
        }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[*location + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
        val[*location] = 1.;
        goto L100;
    }
/*  Set Sommerfeld b.c. on the Z=Zmax plane: */
    if (kk + *nzst - 1 == *nz + 3 && (i + *nxst - 1 >= 0 && i + *nxst - 1 
	    <= *nx + 3) && (j + *nyst - 1 > 0 && j + *nyst - 1 < *ny + 3)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        bindx[k] = *row - 4*(*nx + 4) * (*ny + 4);
	val[k] = -1.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
/*  the imaginary part: */
        bindx[k] = *row - 2*(*nx + 4) * (*ny + 4) + 1;
        beta = -kn;
        if (*order == 6) {
            d1 = beta * hz;
            d2 = beta * hz, d2 *= d2;
            val[k] = beta * -2. * hz * (1. - d1 * d1 / 6. + d2 *
                    d2 / 120.);
        }
        if (*order == 2) {
            val[k] = beta * -2. * hz;
        }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[*location + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain */
        val[*location] = 1.;
        goto L100;
    }
/* --------    end of the b.c. equations setting -------- */

/*  Set the eq. for the inner grid points. */
/*  We assume:  H = Hx = Hy = Hz ; */
    h2 = h * h;
    if (*order == 2) {
        ac_i_j_k = (kn * kn * h2 - 6.) / h2;
        ac = 1. / h2;
        ac_i_jm1_k = ac;
        ac_i_jp1_k = ac;
        ac_im1_j_k = ac;
        ac_ip1_j_k = ac;
        ac_i_j_kp1 = ac;
        ac_i_j_km1 = ac;
    } /* endif Order=2 */

    if (*order == 6) {
/*  k plane: */
/* ------------- */
/*  The following expresions follows Sutmann 2007 paper in JCAM - */
/*  - using the L6c scheme eq. (17) in this paper. */

        d1 = kn * h;
/* Computing 4th power */
        d2 = kn * h, d2 *= d2;
/* Computing 6th power */
/*      d3 = kn * h, d3 *= d3; */
        ac_i_j_k = -64. / 15. / h2 * (1. - d1 * d1 * .25 +
                5. / 256. * d2 * d2  - 1./ 1536. * d2 * (d2 * d2)); 
        as = 7./ 15. / h2 * (1. - d1 * d1 / 21.);
        ac = .1 / h2 * (1. + d1 * d1 / 18.);
        ad = 1./ 30. / h2;

/*  symmetric points */
        ac_i_jm1_k = as;
        ac_i_jp1_k = as;
        ac_im1_j_k = as;
        ac_ip1_j_k = as;
/*  corner points: */
        ac_im1_jm1_k = ac;
        ac_ip1_jm1_k = ac;
        ac_im1_jp1_k = ac;
        ac_ip1_jp1_k = ac;
/*  k-1 plane: */
/* -------------- */
        ac_i_j_km1 = as;
/*  symmetric points : */
        ac_i_jm1_km1 = ac;
        ac_i_jp1_km1 = ac;
        ac_im1_j_km1 = ac;
        ac_ip1_j_km1 = ac;
/*  corner points: */
        ac_im1_jm1_km1 = ad;
        ac_ip1_jm1_km1 = ad;
        ac_im1_jp1_km1 = ad;
        ac_ip1_jp1_km1 = ad;
/*  k+1 plane: */
/* -------------- */
        ac_i_j_kp1 = as;
/*  symmetric points */
        ac_i_jm1_kp1 = ac;
        ac_i_jp1_kp1 = ac;
        ac_im1_j_kp1 = ac;
        ac_ip1_j_kp1 = ac;
/*  corner points: */
        ac_im1_jm1_kp1 = ad;
        ac_ip1_jm1_kp1 = ad;
        ac_im1_jp1_kp1 = ad;
        ac_ip1_jp1_kp1 = ad;
    }
/* endif 6th order */
    icn = 0;
    sum = 0.;
    k = bindx[*location];
    bindx[k] = *row + 2;
    val[k] =  ac_ip1_j_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2;
    val[k] = ac_im1_j_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(*nx + 4);
    val[k] = ac_i_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(*nx + 4);
    val[k] = ac_i_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(*nx + 4) * (*ny + 4);
    val[k] = ac_i_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(*nx + 4) * (*ny + 4);
    val[k] = ac_i_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    if (*order == 2) {
        goto L50;
    }
    bindx[k] = *row + 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(*nx + 4) - 2;
    val[k] = ac_im1_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(*nx + 4) - 2;
    val[k] = ac_im1_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* ----- k+1 plane: */
    bindx[k] = *row + 2*(*nx + 4) * (*ny + 4) + 2;
    val[k] = ac_ip1_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(*nx + 4) * (*ny + 4) - 2;
    val[k] = ac_im1_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4);
    val[k] = ac_i_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4);
    val[k] = ac_i_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* -- */
    bindx[k] = *row + 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4) - 2;
    val[k] = ac_im1_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4) - 2;
    val[k] = ac_im1_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* ----- k-1 plane: */
    bindx[k] = *row - 2*(*nx + 4) * (*ny + 4) + 2;
    val[k] = ac_ip1_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(*nx + 4) * (*ny + 4) - 2;
    val[k] = ac_im1_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4);
    val[k] = ac_i_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4);
    val[k] = ac_i_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* -- */
    bindx[k] = *row - 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4) - 2;
    val[k] = ac_im1_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4) - 2;
    val[k] = ac_im1_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
L50:
    *kmax = k;
    bindx[*location + 1] = k;
    val[*location] = ac_i_j_k;

L100:
    rhs[*row] = f;
/*  return */
/*  normalize the equations: */
    d1 = val[*location];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[*row] /= sum1;
    val[*location] /= sum1;
    i1 = icn;
    for (ii = 1; ii <= i1; ++ii) {
        val[*kmax - ii] /= sum1;
    }
/*    if( *me == 1) {    */ 
/*    if( *me == 0 &&  ( *row >=45414 && *row <= 90826)) { */
/*    if( *me == 0 &&  ( *row >=60552 && *row <= 121102)) { */  
/*  printf("\n%7d %13.10f %13.10f", *row, rhs[*row], val[*location]);
    for (ii=1; ii<=i1; ++ii) printf(" %13.10f", val[*kmax - ii]) ;   
    for (ii=1; ii<=i1; ++ii) printf(" %8d", bindx[*kmax - ii]) ;   
    }*/ 
/* -------------------------------- */
/* -------------------------------- */
/*  Store the imaginary equation coefficients (bi; ai) */

    loca = *location + 1;
    rowa = *row + 1;
    f = 0.;

/*  Set Sommerfeld b.c. on all boundaries: dU/dn -i*Kn*U=0: */

/*  Set Sommerfeld b.c. on the  x=Xmin plane: */
    if (i + *nxst - 1 == 0 && (j + *nyst - 1 > 0 && j + *nyst - 1 < *ny + 3)
             && (kk + *nzst - 1 > 0 && kk + *nzst - 1 < *nz + 3)) {
/*  the real part: */
        icn = 0;
        sum = 0.;
        k = bindx[loca];
        if (*order == 2) {
            goto L124;
        }
        if (j + *nyst - 1 == *ny + 3) {
            goto L121;
        }
        bindx[k] = rowa + 2*(*nx + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L121:
        if (j + *nyst - 1 == 0) {
            goto L122;
        }
        bindx[k] = rowa - 2*(*nx + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L122:
        if (kk + *nzst - 1 == *nz + 3) {
            goto L123;
        }
        bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L123:
        if (kk + *nzst - 1 == 0) {
            goto L124;
        }
        bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L124:
        bindx[k] = rowa + 4;
        val[k] = -1.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
/*  the imaginary part: */
        bindx[k] = rowa + 1;
        beta = -kn;
        if (*order == 6) {
            d1 = beta * hx;
            d2 = beta * hx, d2 *= d2;
            val[k] = beta * 2. * hx * (1. - d1 * d1 / 6. + d2 * d2
                    / 120.);
        }
        if (*order == 2) {
            val[k] = beta * 2. * hx;
        }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[loca + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
        val[loca] = 1.;
        goto L200;
    }

/*  Set Sommerfeld b.c. on the  x=Xmax plane: */
    if (i + *nxst - 1 == *nx + 3 && (j + *nyst - 1 > 0 && j + *nyst - 1 < *
            ny + 3) && (kk + *nzst - 1 > 0 && kk + *nzst - 1 < *nz + 3)) {
/*  the real part: */
        icn = 0;
        sum = 0.;
        k = bindx[loca];

        if (*order == 2) {
            goto L114;
        }
        if (j + *nyst - 1 == *ny + 3) {
            goto L111;
        }
        bindx[k] = rowa + 2*(*nx + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L111:
        if (j + *nyst - 1 == 0) {
            goto L112;
        }
        bindx[k] = rowa - 2*(*nx + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L112:
        if (kk + *nzst - 1 == *nz + 3) {
            goto L113;
        }
        bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L113:
        if (kk + *nzst - 1 == 0) {
            goto L114;
        }
        bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L114:
        bindx[k] = rowa - 4;
        val[k] = -1.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
/*  the imaginary part: */
        bindx[k] = rowa - 3;
        beta = -kn;
        if (*order == 6) {
            d1 = beta * hx;
            d2 = beta * hx, d2 *= d2;
            val[k] = beta * 2. * hx * (1. - d1 * d1 / 6. + d2 * d2
                    / 120.); }
        if (*order == 2) {
            val[k] = beta * 2. * hx; }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[loca + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
        val[loca] = 1.;
        goto L200;
    }

/*  Set Sommerfeld b.c. on the  y=Ymin plane: */
    if (j + *nyst - 1 == 0) {
/*  the real part: */
        icn = 0;
        sum = 0.;
        k = bindx[loca];
        if (*order == 2) {
            goto L104;
        }
        if (i + *nxst - 1 == *nx + 3) {
            goto L101;
        }
        bindx[k] = rowa + 2;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L101:
        if (i + *nxst - 1 == 0) {
            goto L102;
        }
        bindx[k] = rowa - 2;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L102:
        if (kk + *nzst - 1 == *nz + 3) {
            goto L103;
        }
        bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L103:
        if (kk + *nzst - 1 == 0) {
            goto L104;
        }
        bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L104:
        bindx[k] = rowa + 4*(*nx + 4);
        val[k] = -1.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
/*  the imaginary part: */
        bindx[k] = rowa + 2*(*nx + 4) - 1;
        beta = -kn;
        if (*order == 6) {
            d1 = beta * hy;
            d2 = beta * hy, d2 *= d2;
            val[k] = beta * 2. * hy * (1. - d1 * d1 / 6. + d2 * d2
                    / 120.);
        }
        if (*order == 2) {
            val[k] = beta * 2. * hy;
        }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[loca + 1] = k;
/*  the diagonal element: */
/* the additional point - outside the computational domain: */
        val[loca] = 1.;
        goto L200;
    }

/*  Set Sommerfeld b.c. on the  y=Ymax plane: */
    if (j + *nyst - 1 == *ny + 3) {
/*  the real part: */
	icn = 0;
	sum = 0.;
	k = bindx[loca];
        if (*order == 2) {
            goto L134;
        }
        if (i + *nxst - 1 == *nx + 3) {
            goto L131;
        }
        bindx[k] = rowa + 2;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L131:
        if (i + *nxst - 1 == 0) {
            goto L132;
        }
        bindx[k] = rowa - 2;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L132:
        if (kk + *nzst - 1 == *nz + 3) {
            goto L133;
        }
        bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L133:
        if (kk + *nzst - 1 == 0) {
            goto L134;
        }
        bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L134:
        bindx[k] = rowa - 4*(*nx + 4);
        val[k] = -1.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
/*  the imaginary part: */
        bindx[k] = rowa - 2*(*nx + 4) - 1;
        beta = -kn;
        if (*order == 6) {
            d1 = beta * hy;
            d2 = beta * hy, d2 *= d2;
            val[k] = beta * 2. * hy * (1. - d1 * d1 / 6. + d2 * d2
                    / 120.);
        }
        if (*order == 2) {
            val[k] = beta * 2 * hy;
        }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[loca + 1] = k;
/*  the diagonal element: */
/*  the additional point - outside the computational domain: */
        val[loca] = 1.;
        goto L200;
    }


/*  Set Sommerfeld b.c. on the  z=Zmin plane: */
    if (kk + *nzst - 1 == 0 && (i + *nxst - 1 >= 0 && i + *nxst - 1 <= 
            *nx + 3) && (j + *nyst - 1 > 0 && j + *nyst - 1 < *ny + 3)) {
/*  the real part: */
        icn = 0;
        sum = 0.;
        k = bindx[loca];
        bindx[k] = rowa + 4*(*nx + 4) * (*ny + 4);
        val[k] = -1.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
/*  the imaginary part: */
        bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4) - 1;
        beta = -kn;
        if (*order == 6) {
            d1 = beta * hz;
            d2 = beta * hz, d2 *= d2;
            val[k] = beta * 2. * hz * (1. - d1 * d1 / 6. + d2 * d2
                    / 120.); }
        if (*order == 2) {
            val[k] = beta * 2. * hz; }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[loca + 1] = k;
/*  the diagonal element: */
        val[loca] = 1.;
        goto L200;
    }
/*  Set Sommerfeld b.c. on the  z=Zmax plane: */
    if (kk + *nzst - 1 == *nz + 3 && (i + *nxst - 1 >= 0 && i + *nxst - 1
            <= *nx + 3) && (j + *nyst - 1 > 0 && j + *nyst - 1 < *ny + 3)) {
        icn = 0;
        sum = 0.;
        k = bindx[loca];
        bindx[k] = rowa - 4*(*nx + 4) * (*ny + 4);
        val[k] = -1.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
/*  the imaginary part: */
        bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4) - 1;
        beta = -kn;
        if (*order == 6) {
            d1 = beta * hz;
            d2 = beta * hz, d2 *= d2;
            val[k] = beta * 2. * hz * (1. - d1 * d1 / 6. + d2 * d2
                    / 120.); }
        if (*order == 2) {
            val[k] = beta * 2. * hz; }
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
        *kmax = k;
        bindx[loca + 1] = k;
/*  the diagonal element: */
        val[loca] = 1.;
        goto L200;
    }
/* --------    end of the b.c. equation settings -------- */

/*  Set the eq. for the inner grid points. */
    icn = 0;
    sum = 0.;
    k = bindx[loca];
    bindx[k] = rowa + 2;
    val[k] = ac_ip1_j_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2;
    val[k] = ac_im1_j_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(*nx + 4);
    val[k] = ac_i_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(*nx + 4);
    val[k] = ac_i_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4);
    val[k] = ac_i_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4);
    val[k] = ac_i_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    if (*order == 2) {
        goto L150; }
    bindx[k] = rowa + 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(*nx + 4) - 2;
    val[k] = ac_im1_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(*nx + 4) - 2;
    val[k] = ac_im1_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* ----- k+1 plane: */
    bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4) + 2;
    val[k] = ac_ip1_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4) - 2;
    val[k] = ac_im1_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4);
    val[k] = ac_i_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4);
    val[k] = ac_i_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* -- */
    bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4) - 2;
    val[k] = ac_im1_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa + 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4) - 2;
    val[k] = ac_im1_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* ----- k-1 plane: */
    bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4) + 2;
    val[k] = ac_ip1_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4) - 2;
    val[k] = ac_im1_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4);
    val[k] = ac_i_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4);
    val[k] = ac_i_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* -- */
    bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4) + 2*(*nx + 4) - 2;
    val[k] = ac_im1_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4) + 2;
    val[k] = ac_ip1_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = rowa - 2*(*nx + 4) * (*ny + 4) - 2*(*nx + 4) - 2;
    val[k] = ac_im1_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
L150:
    *kmax = k;
    bindx[loca + 1] = k;
    val[loca] = ac_i_j_k;
L200:
    rhs[rowa] = f;

/*  return */
/*  normalize the equations: */
    d1 = val[loca];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[rowa] /= sum1;
    val[loca] /= sum1;
    i1 = icn;
    for (ii = 1; ii <= i1; ++ii) {
        val[*kmax - ii] /= sum1;
    }
/*    if( *me == 1) {    */ 
/*    if( *me == 0 &&  ( *row >=45414 && *row <= 90826)) {  */ 
/*      if( *me == 0 &&  ( *row >=60552 && *row <= 121102)) {  */   
/*  printf("\n%7d %13.10f %13.10f", rowa, rhs[rowa], val[loca]);
    for (ii=1; ii<=i1; ++ii) printf(" %13.10f", val[*kmax - ii]) ;  
    for (ii=1; ii<=i1; ++ii) printf(" %8d", bindx[*kmax - ii]) ;   
    }*/ 
    return 0;
} /* create_matrix_row_27_pt_example30_2rows */
/* ******************************************************************** */
/* ******************************************************************** */
/* Subroutine */ int create_matrix_row_9pt_example31_case_a(int *row, int *
	location, double *val, int *bindx, int *kmax, double *
	rhs, int *nxst, int *nyst, int *nxloc, int *nyloc, double *m,
        int *order, int *i_scheme) 
{
    int i1;
    int m_2, m3;
    double  d1, d2, d3, d4, d5, d6, d7, d8;
    static double f, h;
    static int i, j, k;
    static double x, y, a1, a2, a3, a4, a5, a6, a7, a8, a9;
    static double kn1, kn2, kn3, kn4, kn5, kn6, kn7, kn8, kn9;
    static double  xmh, xph, ymh, yph, hkn_2, h_3, kx2, ky2, del_k2; 
    static int ii;
    static double  pi, del;
    static int icn;
    static double sum, hkn2, sum1;

/* Purpose:   
   Add one row to an MSR matrix corresponding to a 5 / 9 point discrete   
   approximation to the 2D EXAMPLE_31 operator on an (n+2)*(n+2) rectangle.   
   Example 31_case_a - is the example of Turkel, R&D and Tsynkov 2012, 
   but it is not included in the paper.

   Order - is the scheme order: 2nd / 4th/ 6th order scheme.
   i_scheme = 1  for the 2nd order scheme, the 4th or the 6th order schemes of
                 Turkel & Harrari 1995, Turkel & Erlangga 2012.
   i_scheme = 2  for the 6th order scheme of Turkel, R&D and Tsynkov 2012.

   Domain: [0 PI]*[0 PI]    

           Uxx + Uyy + K**2*U = f   

   With Dirichlet boundary condition on all sides. The value of f on the boundaries 
   is obtained from the analytic solution.

   The analytic solution is:   
         U(x,y)= sin(sin(m*x)*sinh(n*y))                                 

   with: m=n   so that f=0
   and:  K**2 = m**2*cos(m*x)**2*sinh(n*y)**2 + n**2*sin(m*x)**2*cosh(n*y)**2     
                                                                                   */ 
/* --------------------------------------------------------------------------------*/
/*   Parameters:   
       row          == global row number of the new row to be added.   
       location     == local row where diagonal of the new row will be stored.   
       val,bindx    == (see user's guide). On output, val[] and bindx[]   
                       are appended such that the new row has been added.   
       Order        == scheme order                                                */

    pi = atan(1.) * 4.;
    h = pi / (double) (global_2.n + 1);
    i = *location % *nxloc;
    j = *location / *nxloc % *nyloc;
    x = (*nxst + i - 1) * h;
    y = (*nyst + j - 1) * h;

    f = 0.;
    if (j + *nyst - 1 == 0 && (i + *nxst - 1 > 0 && i + *nxst - 1 <
            global_2.n + 1)) f = sin(sin(*m * x) * sinh(*m * y));

    if (j + *nyst - 1 == global_2.n + 1 && (i + *nxst - 1 > 0 && i + *
            nxst - 1 < global_2.n + 1)) f = sin(sin(*m * x) * sinh(*m * y));

    if (i + *nxst - 1 == global_2.n + 1) f = sin(sin(*m * x) * sinh(*m * y));
    if (i + *nxst - 1 == 0) f = sin(sin(*m * x) * sinh(*m * y));


/*  Set the  boundary conditions equations */
/*-----------------------------------------*/
/*  On the right side boundary (Dirichlet b.c.): u=0 */

    if (i + *nxst - 1 == global_2.n + 1) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	if (*order == 2) {
	    goto L35;
	}
	if ((*order == 4 || *order == 6) && (j + *nyst - 1 == 0 || j + *nyst 
		- 1 == global_2.n + 1)) {
	    goto L35;
	}
	bindx[k] = *row + (global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

	bindx[k] = *row - (global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
L35:
	*kmax = k;
	bindx[*location + 1] = k;
/*  the diagonal element: */
	val[*location] = 1.;
	goto L100;
    }

/*  On the left side boundary (Dirichlet b.c.): u=0 */

    if (i + *nxst - 1 == 0) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];

	if (*order == 2) {
	    goto L45;
	}
	if ((*order == 4 || *order == 6) && (j + *nyst - 1 == 0 || j + *nyst 
		- 1 == global_2.n + 1)) {
	    goto L45;
	}
	bindx[k] = *row + (global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

	bindx[k] = *row - (global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
L45:
	*kmax = k;
	bindx[*location + 1] = k;
	val[*location] = 1.;
	goto L100;
    }

/*  The bottom boundary (Dirichlet b.c.): */
    if (j + *nyst - 1 == 0 && (i + *nxst - 1 > 0 && i + *nxst - 1 < 
	    global_2.n + 1)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	*kmax = k;
	bindx[*location + 1] = k;
	val[*location] = 1.;
	goto L100;
    }

/*  The top boundary: */
    if (j + *nyst - 1 == global_2.n + 1 && (i + *nxst - 1 > 0 && i + *
	    nxst - 1 < global_2.n + 1)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	*kmax = k;
	bindx[*location + 1] = k;
	val[*location] = 1.;
	goto L100;
    }
/* --------    end of the b.c. equations setting -------- */

/*  Set the eq. for the inner grid points.  */ 
    if (*order == 2) {
        m_2 = *m * *m;
        d1 = cos(*m * x);
        d2 = d1 * d1;
        d3 = sinh(*m * y);
        d4 = d3 * d3;
        d5 = sin(*m * x);
        d6 = d5 * d5;
        d7 = cosh(*m * y);
        d8 = d7 * d7;
        kn2= m_2 *(d2 * d4 + d6 * d8);
        hkn2 = kn2 * h * h;
	a5 = hkn2 - 4.;
	a2 = 1.;
	a4 = 1.;
	a6 = 1.;
	a8 = 1.;
	a1 = 0.;
        a3 = 0.;
        a7 = 0.;
        a9 = 0.;
	goto L40;
    }

/*  definition of the coefficients following Turkel & Harrari 1995, Turkel & Erlangga 2012:  */
    if(*i_scheme == 1 && (*order == 4 || *order == 6)) {
       xph = x + h;
       xmh = x - h;
       yph = y + h;
       ymh = y - h;
       m_2 = *m * *m;
/*     del= 2.; */
/*     del = -1.; */
       del= 0.; 
       d1 = cos(*m * xph);
       d2 = d1 * d1;
       d3 = sinh(*m * yph);
       d4 = d3 * d3;
       d5 = sin(*m * xph);
       d6 = d5 * d5;
       d7 = cosh(*m * yph);
       d8 = d7 * d7;
       kn1= m_2 *(d2 * d4 + d6 * d8);
       hkn_2 = kn1 * h * h;
/*     if(*order == 4) a1 = hkn_2 * 7. / 360. + 1./6.; */
       if(*order == 4) a1 = hkn_2 * del / 144. + 1./6.;
       if(*order == 6) a1 = hkn_2 * 7. / 360. + 1./6.  + hkn_2 * hkn_2 * del / 720.;

       d1 = cos(*m * x);
       d2 = d1 * d1;
       d3 = sinh(*m * yph);
       d4 = d3 * d3;
       d5 = sin(*m * x);
       d6 = d5 * d5;
       d7 = cosh(*m * yph);
       d8 = d7 * d7;
       kn2= m_2 *(d2 * d4 + d6 * d8);
       hkn_2 = kn2 * h * h;
/*     if(*order == 4) a2 = hkn_2 * 2. / 45. + 2./3. ; */
       if(*order == 4) a2 = hkn_2 * (6. - del)/ 72. + 2./3. ;
       if(*order == 6) a2 = hkn_2 * 2. / 45. + 2./3.  + hkn_2 * hkn_2 * (3. - del * 2.) / 720.;

       d1 = cos(*m * xmh);
       d2 = d1 * d1;
       d3 = sinh(*m * yph);
       d4 = d3 * d3;
       d5 = sin(*m * xmh);
       d6 = d5 * d5;
       d7 = cosh(*m * yph);
       d8 = d7 * d7;
       kn3= m_2 *(d2 * d4 + d6 * d8);
       hkn_2 = kn3 * h * h;
/*     if(*order == 4) a3 = hkn_2 * 7. / 360. + 1./6.; */
       if(*order == 4) a3 = hkn_2 * del / 144. + 1./6.;
       if(*order == 6) a3 = hkn_2 * 7. / 360. + 1./6.  + hkn_2 * hkn_2 * del / 720.;

       d1 = cos(*m * xph);
       d2 = d1 * d1;
       d3 = sinh(*m * y);
       d4 = d3 * d3;
       d5 = sin(*m * xph);
       d6 = d5 * d5;
       d7 = cosh(*m * y);
       d8 = d7 * d7;
       kn4= m_2 *(d2 * d4 + d6 * d8);
       hkn_2 = kn4 * h * h;
/*     if(*order == 4) a4 = hkn_2 * 2. / 45. + 2./3. ; */
       if(*order == 4) a4 = hkn_2 * (6. - del)/ 72. + 2./3. ;
       if(*order == 6) a4 = hkn_2 * 2. / 45. + 2./3.  + hkn_2 * hkn_2 * (3. - del * 2.) / 720.;

       d1 = cos(*m * x);
       d2 = d1 * d1;
       d3 = sinh(*m * y);
       d4 = d3 * d3;
       d5 = sin(*m * x);
       d6 = d5 * d5;
       d7 = cosh(*m * y);
       d8 = d7 * d7;
       kn5= m_2 *(d2 * d4 + d6 * d8);
       hkn_2 = kn5 * h * h;
/*     if(*order == 4) a5 = hkn_2 * 67. / 90. - 10./3.; */
       if(*order == 4) a5 = hkn_2 * (24. + del) / 36. - 10./3.;
       if(*order == 6) a5 = hkn_2 * 67. / 90. - 10./3.  + hkn_2 * hkn_2 * (del - 3.); 

       d1 = cos(*m * xmh);
       d2 = d1 * d1;
       d3 = sinh(*m * y);
       d4 = d3 * d3;
       d5 = sin(*m * xmh);
       d6 = d5 * d5;
       d7 = cosh(*m * y);
       d8 = d7 * d7;
       kn6= m_2 *(d2 * d4 + d6 * d8);
       hkn_2 = kn6 * h * h;
/*     if(*order == 4) a6 = hkn_2 * 2. / 45. + 2./3. ; */
       if(*order == 4) a6 = hkn_2 * (6. - del)/ 72. + 2./3. ;
       if(*order == 6) a6 = hkn_2 * 2. / 45. + 2./3.  + hkn_2 * hkn_2 * (3. - del * 2.) / 720.;

       d1 = cos(*m * xph);
       d2 = d1 * d1;
       d3 = sinh(*m * ymh);
       d4 = d3 * d3;
       d5 = sin(*m * xph);
       d6 = d5 * d5;
       d7 = cosh(*m * ymh);
       d8 = d7 * d7;
       kn7= m_2 *(d2 * d4 + d6 * d8);
       hkn_2 = kn7 * h * h;
/*     if(*order == 4) a7 = hkn_2 * 7. / 360. + 1./6.; */
       if(*order == 4) a7 = hkn_2 * del / 144. + 1./6.;
       if(*order == 6) a7 = hkn_2 * 7. / 360. + 1./6.  + hkn_2 * hkn_2 * del / 720.;

       d1 = cos(*m * x);
       d2 = d1 * d1;
       d3 = sinh(*m * ymh);
       d4 = d3 * d3;
       d5 = sin(*m * x);
       d6 = d5 * d5;
       d7 = cosh(*m * ymh);
       d8 = d7 * d7;
       kn8= m_2 *(d2 * d4 + d6 * d8);
       hkn_2 = kn8 * h * h;
/*     if(*order == 4) a8 = hkn_2 * 2. / 45. + 2./3. ; */
       if(*order == 4) a8 = hkn_2 * (6. - del)/ 72. + 2./3. ;
       if(*order == 6) a8 = hkn_2 * 2. / 45. + 2./3.  + hkn_2 * hkn_2 * (3. - del * 2.) / 720.;

       d1 = cos(*m * xmh);
       d2 = d1 * d1;
       d3 = sinh(*m * ymh);
       d4 = d3 * d3;
       d5 = sin(*m * xmh);
       d6 = d5 * d5;
       d7 = cosh(*m * ymh);
       d8 = d7 * d7;
       kn9= m_2 *(d2 * d4 + d6 * d8);
       hkn_2 = kn9 * h * h;
/*     if(*order == 4) a9 = hkn_2 * 7. / 360. + 1./6.; */
       if(*order == 4) a9 = hkn_2 * del / 144. + 1./6.;
       if(*order == 6) a9 = hkn_2 * 7. / 360. + 1./6.  + hkn_2 * hkn_2 * del / 720.;
       goto L40;
    }

/*---------------------------------------*/

/*  definition of the coefficients following Turkel, R&D and Tsynkov 2012:  */
    if(*i_scheme == 2 &&  *order == 6) {
       xph = x + h;
       xmh = x - h;
       yph = y + h;
       ymh = y - h;
       h_3 = h*h*h;
       m_2 = *m * *m;

/*     The derivatives of k**2 are calculated at the center point   
       kx2 - the first derivative of k**2 with respect to x
       ky2 - the first derivative of k**2 with respect to y */

       m3 =  *m * m_2;
       kx2 = m3 * sin(2.* *m * x);
       ky2 = m3 * sinh(2.* *m *y);

       d1 = cos(*m * xph);
       d2 = d1 * d1;
       d3 = sinh(*m * yph);
       d4 = d3 * d3;
       d5 = sin(*m * xph);
       d6 = d5 * d5;
       d7 = cosh(*m * yph);
       d8 = d7 * d7;
       kn1= m_2 *(d2 * d4 + d6 * d8);
       hkn_2 = kn1 * h * h;
       a1 = 1./6. + hkn_2/90. + h_3/120.*(kx2 + ky2);

       d1 = cos(*m * x);
       d2 = d1 * d1;
       d3 = sinh(*m * yph);
       d4 = d3 * d3;
       d5 = sin(*m * x);
       d6 = d5 * d5;
       d7 = cosh(*m * yph);
       d8 = d7 * d7;
       kn2= m_2 *(d2 * d4 + d6 * d8);
       hkn_2 = kn2 * h * h;
       a2 = 2./3. + hkn_2/90. + h_3*ky2*(2./3. + hkn_2/6.)/20.; 

       d1 = cos(*m * xmh);
       d2 = d1 * d1;
       d3 = sinh(*m * yph);
       d4 = d3 * d3;
       d5 = sin(*m * xmh);
       d6 = d5 * d5;
       d7 = cosh(*m * yph);
       d8 = d7 * d7;
       kn3= m_2 *(d2 * d4 + d6 * d8);
       hkn_2 = kn3 * h * h;
       a3 = 1./6. + hkn_2/90. + h_3/120.*(-kx2 + ky2);

       d1 = cos(*m * xph);
       d2 = d1 * d1;
       d3 = sinh(*m * y);
       d4 = d3 * d3;
       d5 = sin(*m * xph);
       d6 = d5 * d5;
       d7 = cosh(*m * y);
       d8 = d7 * d7;
       kn4= m_2 *(d2 * d4 + d6 * d8);
       hkn_2 = kn4 * h * h;
       a4 = 2./3. + hkn_2/90. + h_3*kx2*(2./3. + hkn_2/6.)/20.; 

       d1 = cos(*m * x);
       d2 = d1 * d1;
       d3 = sinh(*m * y);
       d4 = d3 * d3;
       d5 = sin(*m * x);
       d6 = d5 * d5;
       d7 = cosh(*m * y);
       d8 = d7 * d7;
       kn5= m_2 *(d2 * d4 + d6 * d8);
       hkn_2 = kn5 * h * h;
       del_k2 = 2. * *m * m3 * ( cos(2.* *m * x) + 2.*cosh(2.* *m *y));
       a5 = -10./3. +41./45.*hkn_2 + h*h_3*(del_k2 - kn5*kn5)/20. ; 

       d1 = cos(*m * xmh);
       d2 = d1 * d1;
       d3 = sinh(*m * y);
       d4 = d3 * d3;
       d5 = sin(*m * xmh);
       d6 = d5 * d5;
       d7 = cosh(*m * y);
       d8 = d7 * d7;
       kn6= m_2 *(d2 * d4 + d6 * d8);
       hkn_2 = kn6 * h * h;
       a6 = 2./3. + hkn_2/90. - h_3*kx2*(2./3. + hkn_2/6.)/20.; 

       d1 = cos(*m * xph);
       d2 = d1 * d1;
       d3 = sinh(*m * ymh);
       d4 = d3 * d3;
       d5 = sin(*m * xph);
       d6 = d5 * d5;
       d7 = cosh(*m * ymh);
       d8 = d7 * d7;
       kn7= m_2 *(d2 * d4 + d6 * d8);
       hkn_2 = kn7 * h * h;
       a7 = 1./6. + hkn_2/90. + h_3/120.*(kx2 - ky2);

       d1 = cos(*m * x);
       d2 = d1 * d1;
       d3 = sinh(*m * ymh);
       d4 = d3 * d3;
       d5 = sin(*m * x);
       d6 = d5 * d5;
       d7 = cosh(*m * ymh);
       d8 = d7 * d7;
       kn8= m_2 *(d2 * d4 + d6 * d8);
       hkn_2 = kn8 * h * h;
       a8 = 2./3. + hkn_2/90. - h_3*ky2*(2./3. + hkn_2/6.)/20.; 

       d1 = cos(*m * xmh);
       d2 = d1 * d1;
       d3 = sinh(*m * ymh);
       d4 = d3 * d3;
       d5 = sin(*m * xmh);
       d6 = d5 * d5;
       d7 = cosh(*m * ymh);
       d8 = d7 * d7;
       kn9= m_2 *(d2 * d4 + d6 * d8);
       hkn_2 = kn9 * h * h;
       a9 = 1./6. + hkn_2/90. - h_3/120.*(kx2 + ky2);
    }
L40: 
    icn = 0;
    sum = 0.;
    k = bindx[*location];
    bindx[k] = *row + 1;
    val[k] = a4;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row - 1;
    val[k] = a6;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row + (global_2.n + 2);
    val[k] = a2;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    if (*order == 2) {
	goto L50;
    }

    bindx[k] = *row + (global_2.n + 2) + 1;
    val[k] = a1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row + (global_2.n + 2) - 1;
    val[k] = a3;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

L50:

    bindx[k] = *row - (global_2.n + 2);
    val[k] = a8;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    if (*order == 2) {
	goto L55;
    }
    bindx[k] = *row - (global_2.n + 2) + 1;
    val[k] = a7;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row - (global_2.n + 2) - 1;
    val[k] = a9;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

L55:

    *kmax = k;
    bindx[*location + 1] = k;
    val[*location] = a5;
L100:
    rhs[*row] = f;

/*  return 0; */
/*  Normalize the equations: */

    d1 = val[*location];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[*row] /= sum1;
    val[*location] /= sum1;
    i1 = icn;
    for (ii = 1; ii <= i1; ++ii) {
	val[k - ii] /= sum1;
    }

    return 0;
} /* create_matrix_row_9pt_example31_case_a */

/* ******************************************************************** */
/* ******************************************************************** */
/* Subroutine */ int create_matrix_row_9pt_example31_case_b(int *row, int *
	location, double *val, int *bindx, int *kmax, double *
	rhs, int *nxst, int *nyst, int *nxloc, int *nyloc, double *m,
        int *order, int *i_scheme, double *alpha) 
{
    int i1;
    int m_2, m3;
    double  d1, d2, d3;
    static double f, h;
    static int i, j, k;
    static double x, y, a1, a2, a3, a4, a5, a6, a7, a8, a9;
    static double kn1, kn2, kn3, kn4, kn5, kn6, kn7, kn8, kn9;
    static double  xmh, xph, ymh, yph, hkn_2, h_3, kx2, ky2, del_k2; 
    static int ii;
    static double  pi, del;
    static int icn;
    static double sum, hkn2, sum1;

/* Purpose:   
   Add one row to an MSR matrix corresponding to a 5 / 9 point discrete   
   approximation to the 2D EXAMPLE_31 operator on an (n+2)*(n+2) rectangle.   
   Example 31_case_b - is the example of Turkel, R&D and Tsynkov 2012,
   but it is not included in the paper. 

   Order - is the scheme order: 2nd / 4th/ 6th order scheme.   
   i_scheme = 1  for the 2nd order scheme, the 4th or the 6th order schemes of 
                 Turkel & Harrari 1995, Turkel & Erlangga 2012.
   i_scheme = 2  for the 6th order scheme of Turkel, R&D and Tsynkov 2012.

   Domain: [0 PI]*[0 PI]   

         Uxx + Uyy + K**2*U = f   

   With Dirichlet boundary conditions on all the boundaries.

   The analytic solution is:   
          U(x,y)= sin(alpha*sin(m*x)*exp(n*y))
            
   with:  m=n    so that f=0

   and:   K**2 = m**2*(alpha*exp(m*y))**2 
                                                                               */      
/* ----------------------------------------------------------------------------*/
/*   Parameters:   
       row          == global row number of the new row to be added.   
       location     == local row where diagonal of the new row will be stored.   
       val,bindx    == (see user's guide). On output, val[] and bindx[]   
                       are appended such that the new row has been added.   
       Order        == scheme order                                            */

    pi = atan(1.) * 4.;
    h = pi / (double) (global_2.n + 1);
    i = *location % *nxloc;
    j = *location / *nxloc % *nyloc;
    x = (*nxst + i - 1) * h;
    y = (*nyst + j - 1) * h;

    f = 0.;
    if (j + *nyst - 1 == 0 && (i + *nxst - 1 > 0 && i + *nxst - 1 <
            global_2.n + 1)) f = sin(*alpha*sin(*m * x));

    if (j + *nyst - 1 == global_2.n + 1 && (i + *nxst - 1 > 0 && i + *
            nxst - 1 < global_2.n + 1)) f = sin(*alpha*sin(*m * x) * exp(*m * pi));

    if (i + *nxst - 1 == global_2.n + 1) f = 0.;
    if (i + *nxst - 1 == 0) f = 0.;


/*  Set the  boundary conditions equations */

/*  On the right side boundary (Dirichlet b.c.): u=0 */

    if (i + *nxst - 1 == global_2.n + 1) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	if (*order == 2) {
	    goto L35;
	}
	if ((*order == 4 || *order == 6) && (j + *nyst - 1 == 0 || j + *nyst 
		- 1 == global_2.n + 1)) {
	    goto L35;
	}
	bindx[k] = *row + (global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

	bindx[k] = *row - (global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
L35:
	*kmax = k;
	bindx[*location + 1] = k;
/*  the diagonal element: */
	val[*location] = 1.;
	goto L100;
    }

/*  On the left side boundary (Dirichlet b.c.): u=0 */

    if (i + *nxst - 1 == 0) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];

	if (*order == 2) {
	    goto L45;
	}
	if ((*order == 4 || *order == 6) && (j + *nyst - 1 == 0 || j + *nyst 
		- 1 == global_2.n + 1)) {
	    goto L45;
	}
	bindx[k] = *row + (global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

	bindx[k] = *row - (global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
L45:
	*kmax = k;
	bindx[*location + 1] = k;
	val[*location] = 1.;
	goto L100;
    }

/*  The bottom boundary (Dirichlet b.c.): */
    if (j + *nyst - 1 == 0 && (i + *nxst - 1 > 0 && i + *nxst - 1 < 
	    global_2.n + 1)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	*kmax = k;
	bindx[*location + 1] = k;
	val[*location] = 1.;
	goto L100;
    }

/*  The top boundary: */
    if (j + *nyst - 1 == global_2.n + 1 && (i + *nxst - 1 > 0 && i + *
	    nxst - 1 < global_2.n + 1)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	*kmax = k;
	bindx[*location + 1] = k;
	val[*location] = 1.;
	goto L100;
    }
/* --------    end of the b.c. equations setting -------- */

/*  Set the eq. for the inner grid points.  */ 
    if (*order == 2) {
        m_2 = *m * *m;
        d1 = *m * *alpha*exp(*m * y); 
        d2 = d1 * d1;
        kn2= d2;
        hkn2 = kn2 * h * h;
	a5 = hkn2 - 4.;
	a2 = 1.;
	a4 = 1.;
	a6 = 1.;
	a8 = 1.;
	a1 = 0.;
        a3 = 0.;
        a7 = 0.;
        a9 = 0.;
	goto L40;
        
    }

/*  definition of the coefficients following Turkel & Harrari 1995, Turkel & Erlangga 2012:  */
    if(*i_scheme == 1 && (*order == 4 || *order == 6)) {
       xph = x + h;
       xmh = x - h;
       yph = y + h;
       ymh = y - h;
       m_2 = *m * *m;
/*     del= 2.; */
/*     del = -1.; */
       del= 0.;   
       d1 = *m * *alpha*exp(*m * yph); 
       d2 = d1 * d1;
       kn1= d2;
       hkn_2 = kn1 * h * h;
/*     if(*order == 4) a1 = hkn_2 * 7. / 360. + 1./6.; */
       if(*order == 4) a1 = hkn_2 * del / 144. + 1./6.;
       if(*order == 6) a1 = hkn_2 * 7. / 360. + 1./6.  + hkn_2 * hkn_2 * del / 720.;

       d1 = *m * *alpha*exp(*m * yph); 
       d2 = d1 * d1;
       kn2= d2;
       hkn_2 = kn2 * h * h;
/*     if(*order == 4) a2 = hkn_2 * 2. / 45. + 2./3. ; */
       if(*order == 4) a2 = hkn_2 * (6. - del) / 72. + 2./3. ;
       if(*order == 6) a2 = hkn_2 * 2. / 45. + 2./3.  + hkn_2 * hkn_2 * (3. - del * 2.) / 720.;

       d1 = *m * *alpha*exp(*m * yph); 
       d2 = d1 * d1;
       kn3= d2;
       hkn_2 = kn3 * h * h;
/*     if(*order == 4) a3 = hkn_2 * 7. / 360. + 1./6.; */
       if(*order == 4) a3 = hkn_2 * del / 144. + 1./6.;
       if(*order == 6) a3 = hkn_2 * 7. / 360. + 1./6.  + hkn_2 * hkn_2 * del / 720.;

       d1 = *m * *alpha*exp(*m * y); 
       d2 = d1 * d1;
       kn4= d2;
       hkn_2 = kn4 * h * h;
/*     if(*order == 4) a4 = hkn_2 * 2. / 45. + 2./3. ; */
       if(*order == 4) a4 = hkn_2 * (6. - del) / 72. + 2./3. ;
       if(*order == 6) a4 = hkn_2 * 2. / 45. + 2./3.  + hkn_2 * hkn_2 * (3. - del * 2.) / 720.;

       d1 = *m * *alpha*exp(*m * y); 
       d2 = d1 * d1;
       kn5= d2;
       hkn_2 = kn5 * h * h;
/*     if(*order == 4) a5 = hkn_2 * 67. / 90. - 10./3.; */
       if(*order == 4) a5 = hkn_2 * (24. + del) / 36. - 10./3.;
       if(*order == 6) a5 = hkn_2 * 67. / 90. - 10./3.  + hkn_2 * hkn_2 * (del - 3.); 

       d1 = *m * *alpha*exp(*m * y); 
       d2 = d1 * d1;
       kn6= d2;
       hkn_2 = kn6 * h * h;
/*     if(*order == 4) a6 = hkn_2 * 2. / 45. + 2./3. ; */
       if(*order == 4) a6 = hkn_2 * (6. - del) / 72. + 2./3. ;
       if(*order == 6) a6 = hkn_2 * 2. / 45. + 2./3.  + hkn_2 * hkn_2 * (3. - del * 2.) / 720.;

       d1 = *m * *alpha*exp(*m * ymh); 
       d2 = d1 * d1;
       kn7= d2;
       hkn_2 = kn7 * h * h;
/*     if(*order == 4) a7 = hkn_2 * 7. / 360. + 1./6.; */
       if(*order == 4) a7 = hkn_2 * del / 144. + 1./6.;
       if(*order == 6) a7 = hkn_2 * 7. / 360. + 1./6.  + hkn_2 * hkn_2 * del / 720.;

       d1 = *m * *alpha*exp(*m * ymh); 
       d2 = d1 * d1;
       kn8= d2;
       hkn_2 = kn8 * h * h;
/*     if(*order == 4) a8 = hkn_2 * 2. / 45. + 2./3. ; */
       if(*order == 4) a8 = hkn_2 * (6. - del) / 72. + 2./3. ;
       if(*order == 6) a8 = hkn_2 * 2. / 45. + 2./3.  + hkn_2 * hkn_2 * (3. - del * 2.) / 720.;

       d1 = *m * *alpha*exp(*m * ymh); 
       d2 = d1 * d1;
       kn9= d2;
       hkn_2 = kn9 * h * h;
/*     if(*order == 4) a9 = hkn_2 * 7. / 360. + 1./6.; */
       if(*order == 4) a9 = hkn_2 * del / 144. + 1./6.;
       if(*order == 6) a9 = hkn_2 * 7. / 360. + 1./6.  + hkn_2 * hkn_2 * del / 720.;
       goto L40;
    }

/*  definition of the coefficients following Turkel, R&D and Tsynkov 2012:  */
    if(*i_scheme == 2 &&  *order == 6) {
       xph = x + h;
       xmh = x - h;
       yph = y + h;
       ymh = y - h;
       h_3 = h*h*h;
       m_2 = *m * *m;

/*     k**2 at the center point   
       kx2 - the first derivative of k**2 with respect to x
       ky2 - the first derivative of k**2 with respect to y */

       m3 =  *m * m_2;
       kx2 = 0.; 
       ky2 = 2.* m3 * *alpha * exp(*m *y)*exp(*m *y);

       d1 = *m * *alpha*exp(*m * yph); 
       d2 = d1 * d1;
       kn1= d2;
       hkn_2 = kn1 * h * h;
       a1 = 1./6. + hkn_2/90. + h_3/120.*(kx2 + ky2);   

       d1 = *m * *alpha*exp(*m * yph); 
       d2 = d1 * d1;
       kn2= d2;
       hkn_2 = kn2 * h * h;
       a2 = 2./3. + hkn_2/90. + h_3*ky2*(2./3. + hkn_2/6.)/20.; 

       d1 = *m * *alpha*exp(*m * yph); 
       d2 = d1 * d1;
       kn3= d2;
       hkn_2 = kn3 * h * h;
       a3 = 1./6. + hkn_2/90. + h_3/120.*(-kx2 + ky2);

       d1 = *m * *alpha*exp(*m * y); 
       d2 = d1 * d1;
       kn4= d2;
       hkn_2 = kn4 * h * h;
       a4 = 2./3. + hkn_2/90. + h_3*kx2*(2./3. + hkn_2/6.)/20.; 

       d1 = *m * *alpha*exp(*m * y); 
       d2 = d1 * d1;
       kn5= d2;
       hkn_2 = kn5 * h * h;
       d3 = 2.* m_2 * exp( *m * y);
       del_k2 =  d3 *d3;
       a5 = -10./3. +41./45.*hkn_2 + h*h_3*(del_k2 - kn5*kn5)/20. ; 

       d1 = *m * *alpha*exp(*m * y); 
       d2 = d1 * d1;
       kn6= d2;
       hkn_2 = kn6 * h * h;
       a6 = 2./3. + hkn_2/90. - h_3*kx2*(2./3. + hkn_2/6.)/20.; 

       d1 = *m * *alpha*exp(*m * ymh); 
       d2 = d1 * d1;
       kn7= d2;
       hkn_2 = kn7 * h * h;
       a7 = 1./6. + hkn_2/90. + h_3/120.*(kx2 - ky2);

       d1 = *m * *alpha*exp(*m * ymh); 
       d2 = d1 * d1;
       kn8= d2;
       hkn_2 = kn8 * h * h;
       a8 = 2./3. + hkn_2/90. - h_3*ky2*(2./3. + hkn_2/6.)/20.; 

       d1 = *m * *alpha*exp(*m * ymh); 
       d2 = d1 * d1;
       kn9= d2;
       hkn_2 = kn9 * h * h;
       a9 = 1./6. + hkn_2/90. - h_3/120.*(kx2 + ky2);
    }
L40: 
    icn = 0;
    sum = 0.;
    k = bindx[*location];
    bindx[k] = *row + 1;
    val[k] = a4;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row - 1;
    val[k] = a6;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row + (global_2.n + 2);
    val[k] = a2;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    if (*order == 2) {
	goto L50;
    }

    bindx[k] = *row + (global_2.n + 2) + 1;
    val[k] = a1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row + (global_2.n + 2) - 1;
    val[k] = a3;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

L50:

    bindx[k] = *row - (global_2.n + 2);
    val[k] = a8;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    if (*order == 2) {
	goto L55;
    }
    bindx[k] = *row - (global_2.n + 2) + 1;
    val[k] = a7;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row - (global_2.n + 2) - 1;
    val[k] = a9;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

L55:

    *kmax = k;
    bindx[*location + 1] = k;
    val[*location] = a5;
L100:
    rhs[*row] = f;

/*  return 0; */
/*  Normalize the equations: */

    d1 = val[*location];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[*row] /= sum1;
    val[*location] /= sum1;
    i1 = icn;
    for (ii = 1; ii <= i1; ++ii) {
	val[k - ii] /= sum1;
    }

    return 0;
} /* create_matrix_row_9pt_example31_case_b */

/* ************************************************************************ */
/* ******************************************************************** */
/* Subroutine */ int create_matrix_row_9pt_example31_case_c(int *row, int *
	location, double *val, int *bindx, int *kmax,double *
	rhs, int *nxst, int *nyst, int *nxloc, int *nyloc, double *alpha, double *beta, 
        double *c2, int *order, int *i_scheme) 
{
    int i1, i_rhs;
    double  d1;
    static double f, h, h2;
    static int i, j, k;
    static double a, b, c, kx;
    static double x, y, a1, a2, a3, a4, a5, a6, a7, a8, a9;
    static double kn1, kn2, kn3, kn4, kn5, kn6, kn7, kn8, kn9;
    static double  xmh, xph, ymh, yph, hkn_2, h_3, kx2, ky2, del_k2; 
    static int ii;
    static double  pi, del;
    static int icn;
    static double sum, hkn2, sum1;
    double f_rhs, fa, f_rhs_x, f_rhs_xx, f_rhs_xxx, f_rhs_xxxx,
                  fa_y, fa_yy, f_rhs_y, f_rhs_yy, f_rhs_yyyy, f_rhs_xxyy;

/* Purpose:   
   Add one row to an MSR matrix corresponding to a 5 / 9 point discrete   
   approximation to the 2D EXAMPLE_31 operator on an (n+2)*(n+2) rectangle.   
   Example 31_case_c = is the 2D example of Turkel, R&D and Tsynkov 2012.

   Order - is the scheme order: 2nd / 4th/ 6th order scheme.
   i_scheme = 1  for the 2nd order scheme, the 4th or the 6th order schemes of
                 Turkel & Harrari 1995, Turkel & Erlangga 2012.
   i_scheme = 2  for the 6th order scheme of Turkel, R&D and Tsynkov 2012.

   Domain: [0 PI]*[0 PI]    

           Uxx + Uyy + K**2*U = f   

   With Dirichlet boundary condition on all sides, obtained from the analytic
   solution.

   The analytic solution is:   
        U = exp(-K(x)/c)*sin(sqrt(alpha**2+beta**2)*y) 
    
        K(x) = alpha - beta*sin(c * x)    for general alpha, beta and c.
   if  c=-2.*alpha ==> f=0
                                                                                */
/* -----------------------------------------------------------------------------*/
/*   Parameters:   
       row          == global row number of the new row to be added.   
       location     == local row where diagonal of the new row will be stored.   
       val,bindx    == (see user's guide). On output, val[] and bindx[]   
                       are appended such that the new row has been added.   
       Order        == scheme order                                             */

    a = *alpha;
    b = *beta;
    c = *c2;
    pi = atan(1.) * 4.;
    h = pi / (double) (global_2.n + 1);
    i = *location % *nxloc;
    j = *location / *nxloc % *nyloc;
    x = (*nxst + i - 1) * h;
    y = (*nyst + j - 1) * h;

    kx = a - b*sin(c*x);
    
    i_rhs=1; /*  right handside is not zero  - true */
    if(  abs(2.*a + c) < 1.e-7) i_rhs= 0;


    if (j + *nyst - 1 == 0 && (i + *nxst - 1 > 0 && i + *nxst - 1 <
         global_2.n + 1)) f = exp(-kx/c)*sin(sqrt(a*a+b*b)*y) ;

    if (j + *nyst - 1 == global_2.n + 1 && (i + *nxst - 1 > 0 && i + *
            nxst - 1 < global_2.n + 1)) f = exp(-kx/c)*sin(sqrt(a*a+b*b)*y);

    if (i + *nxst - 1 == global_2.n + 1) f = exp(-kx/c)*sin(sqrt(a*a+b*b)*y);
    if (i + *nxst - 1 == 0) f = exp(-kx/c)*sin(sqrt(a*a+b*b)*y);


/*  Set the  boundary conditions equations */

/*  On the right side boundary (Dirichlet b.c.): u=0 */

    if (i + *nxst - 1 == global_2.n + 1) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	if (*order == 2) {
	    goto L35;
	}
	if ((*order == 4 || *order == 6) && (j + *nyst - 1 == 0 || j + *nyst 
		- 1 == global_2.n + 1)) {
	    goto L35;
	}
	bindx[k] = *row + (global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

	bindx[k] = *row - (global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
L35:
	*kmax = k;
	bindx[*location + 1] = k;
/*  the diagonal element: */
	val[*location] = 1.;
	goto L100;
    }

/*  On the left side boundary (Dirichlet b.c.): u=0 */

    if (i + *nxst - 1 == 0) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];

	if (*order == 2) {
	    goto L45;
	}
	if ((*order == 4 || *order == 6) && (j + *nyst - 1 == 0 || j + *nyst 
		- 1 == global_2.n + 1)) {
	    goto L45;
	}
	bindx[k] = *row + (global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;

	bindx[k] = *row - (global_2.n + 2);
	val[k] = 0.;
	d1 = val[k];
	sum += d1 * d1;
	++icn;
	++k;
L45:
	*kmax = k;
	bindx[*location + 1] = k;
	val[*location] = 1.;
	goto L100;
    }

/*  The bottom boundary (Dirichlet b.c.): */
    if (j + *nyst - 1 == 0 && (i + *nxst - 1 > 0 && i + *nxst - 1 < 
	    global_2.n + 1)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	*kmax = k;
	bindx[*location + 1] = k;
	val[*location] = 1.;
	goto L100;
    }

/*  The top boundary: */
    if (j + *nyst - 1 == global_2.n + 1 && (i + *nxst - 1 > 0 && i + *
	    nxst - 1 < global_2.n + 1)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
	*kmax = k;
	bindx[*location + 1] = k;
	val[*location] = 1.;
	goto L100;
    }
/* --------    end of the b.c. equations setting -------- */

/*  Set the eq. for the inner grid points.  */ 
    if (*order == 2) {
        kx = a - b*sin(c*x);
        kn2 = kx *kx;
        hkn2 = kn2 * h * h;
	a5 = hkn2 - 4.;
	a2 = 1.;
	a4 = 1.;
	a6 = 1.;
	a8 = 1.;
	a1 = 0.;
        a3 = 0.;
        a7 = 0.;
        a9 = 0.;

        if(i_rhs == 0) f=0.;
        if(i_rhs == 1 ){
          fa = -b*(2.*a + c)*sin(sqrt(a*a + b*b)*y);
          f_rhs = fa*sin(c*x)*exp(-kx/c);
          f =  h*h*f_rhs;
        }
	goto L40;
    }
/*  definition of the coefficients following Turkel & Harrari 1995, Turkel & Erlangga 2012:  */
    if(*i_scheme == 1 && (*order == 4 || *order == 6)) {
       xph = x + h;
       xmh = x - h;
       yph = y + h;
       ymh = y - h;
/*     del= 2.; */
/*     del = -1.; */
       del= 0.;   
       kx = a - b*sin(c*xph);
       kn1 = kx *kx;
       hkn_2 = kn1 * h * h;
/*     if(*order == 4) a1 = hkn_2 * 7. / 360. + 1./6.; */
       if(*order == 4) a1 = hkn_2 * del / 144. + 1./6.;
       if(*order == 6) a1 = hkn_2 * 7. / 360. + 1./6.  + hkn_2 * hkn_2 * del / 720.;

       kx = a - b*sin(c*x);
       kn2 = kx *kx;
       hkn_2 = kn2 * h * h;
/*     if(*order == 4) a2 = hkn_2 * 2. / 45. + 2./3. ; */
       if(*order == 4) a2 = hkn_2 * (6. - del) / 72. + 2./3. ;
       if(*order == 6) a2 = hkn_2 * 2. / 45. + 2./3.  + hkn_2 * hkn_2 * (3. - del * 2.) / 720.;

       kx = a - b*sin(c*xmh);
       kn3 = kx *kx;
       hkn_2 = kn3 * h * h;
/*     if(*order == 4) a3 = hkn_2 * 7. / 360. + 1./6.; */
       if(*order == 4) a3 = hkn_2 * del / 144. + 1./6.;
       if(*order == 6) a3 = hkn_2 * 7. / 360. + 1./6.  + hkn_2 * hkn_2 * del / 720.;

       kx = a - b*sin(c*xph);
       kn4 = kx *kx;
       hkn_2 = kn4 * h * h;
/*     if(*order == 4) a4 = hkn_2 * 2. / 45. + 2./3. ; */
       if(*order == 4) a4 = hkn_2 * (6. - del) / 72. + 2./3. ;
       if(*order == 6) a4 = hkn_2 * 2. / 45. + 2./3.  + hkn_2 * hkn_2 * (3. - del * 2.) / 720.;

       kx = a - b*sin(c*x);
       kn5 = kx *kx;
       hkn_2 = kn5 * h * h;
/*     if(*order == 4) a5 = hkn_2 * 67. / 90. - 10./3.; */
       if(*order == 4) a5 = hkn_2 * (24. + del) / 36. - 10./3.;
       if(*order == 6) a5 = hkn_2 * 67. / 90. - 10./3.  + hkn_2 * hkn_2 * (del - 3.); 

       kx = a - b*sin(c*xmh);
       kn6 = kx *kx;
       hkn_2 = kn6 * h * h;
/*     if(*order == 4) a6 = hkn_2 * 2. / 45. + 2./3. ; */
       if(*order == 4) a6 = hkn_2 * (6. - del) / 72. + 2./3. ;
       if(*order == 6) a6 = hkn_2 * 2. / 45. + 2./3.  + hkn_2 * hkn_2 * (3. - del * 2.) / 720.;

       kx = a - b*sin(c*xph);
       kn7 = kx *kx;
       hkn_2 = kn7 * h * h;
/*     if(*order == 4) a7 = hkn_2 * 7. / 360. + 1./6.; */
       if(*order == 4) a7 = hkn_2 * del / 144. + 1./6.;
       if(*order == 6) a7 = hkn_2 * 7. / 360. + 1./6.  + hkn_2 * hkn_2 * del / 720.;

       kx = a - b*sin(c*x);
       kn8 = kx *kx;
       hkn_2 = kn8 * h * h;
/*     if(*order == 4) a8 = hkn_2 * 2. / 45. + 2./3. ; */
       if(*order == 4) a8 = hkn_2 * (6. - del) / 72. + 2./3. ;
       if(*order == 6) a8 = hkn_2 * 2. / 45. + 2./3.  + hkn_2 * hkn_2 * (3. - del * 2.) / 720.;

       kx = a - b*sin(c*xmh);
       kn9 = kx *kx;
       hkn_2 = kn9 * h * h;
/*     if(*order == 4) a9 = hkn_2 * 7. / 360. + 1./6.; */
       if(*order == 4) a9 = hkn_2 * del / 144. + 1./6.;
       if(*order == 6) a9 = hkn_2 * 7. / 360. + 1./6.  + hkn_2 * hkn_2 * del / 720.;

       if(i_rhs == 0) f=0.;
       if(i_rhs ==1 ){
         fa = -b*(2.*a + c)*sin(sqrt(a*a + b*b)*y);
         f_rhs = fa*sin(c*x)*exp(-kx/c);
         f =  h*h*f_rhs;
       }
       goto L40;
    }

/*  definition of the coefficients following Turkel, R&D and Tsynkov 2012:  */
    if(*i_scheme == 2 &&  *order == 6) {
       xph = x + h;
       xmh = x - h;
       yph = y + h;
       ymh = y - h;
       h_3 = h*h*h;

/*     The derivatives of k**2 are calculated at the center point   
       kx2 - the first derivative of k**2 with respect to x
       ky2 - the first derivative of k**2 with respect to y         */

       kx2 = -2.*a*b*c*cos(c*x) + b*b*c*sin(2.*c*x); 
       ky2 = 0.;

       kx = a - b*sin(c*xph);
       kn1 = kx *kx;
       hkn_2 = kn1 * h * h;
       a1 = 1./6. + hkn_2/90. + h_3/120.*(kx2 + ky2);   

       kx = a - b*sin(c*x);
       kn2 = kx *kx;
       hkn_2 = kn2 * h * h;
       a2 = 2./3. + hkn_2/90. + h_3*ky2*(2./3. + hkn_2/6.)/20.; 

       kx = a - b*sin(c*xmh);
       kn3 = kx *kx;
       hkn_2 = kn3 * h * h;
       a3 = 1./6. + hkn_2/90. + h_3/120.*(-kx2 + ky2);

       kx = a - b*sin(c*xph);
       kn4 = kx *kx;
       hkn_2 = kn4 * h * h;
       a4 = 2./3. + hkn_2/90. + h_3*kx2*(2./3. + hkn_2/6.)/20.; 

       kx = a - b*sin(c*x);
       kn5 = kx *kx;
       hkn_2 = kn5 * h * h;
       del_k2 =   2.*a*b*c*c*sin(c*x) + 2.*b*b*c*c*cos(2.*c*x);
       a5 = -10./3. +41./45.*hkn_2 + h*h_3*(del_k2 - kn5*kn5)/20. ; 

       kx = a - b*sin(c*xmh);
       kn6 = kx *kx;
       hkn_2 = kn6 * h * h;
       a6 = 2./3. + hkn_2/90. - h_3*kx2*(2./3. + hkn_2/6.)/20.; 

       kx = a - b*sin(c*xph);
       kn7 = kx *kx;
       hkn_2 = kn7 * h * h;
       a7 = 1./6. + hkn_2/90. + h_3/120.*(kx2 - ky2);

       kx = a - b*sin(c*x);
       kn8 = kx *kx;
       hkn_2 = kn8 * h * h;
       a8 = 2./3. + hkn_2/90. - h_3*ky2*(2./3. + hkn_2/6.)/20.; 

       kx = a - b*sin(c*xmh);
       kn9 = kx *kx;
       hkn_2 = kn9 * h * h;
       a9 = 1./6. + hkn_2/90. - h_3/120.*(kx2 + ky2);

       if(i_rhs == 0) f=0.;
       if(i_rhs ==1 ){
          kx = a - b*sin(c*x);
          kn5 = kx*kx;
          h2 = h*h;
/* --------------*/
          fa = -b*(2.*a + c)*sin(sqrt(a*a + b*b)*y);

          f_rhs = fa*sin(c*x)*exp(-kx/c);
          f_rhs_x = fa*exp(-kx/c)*(c*cos(c*x)+0.5*b*sin(2.*c*x));
          f_rhs_xx = fa*exp(-kx/c)*(b*c*cos(c*x)*cos(c*x) - c*c*sin(c*x) + b*c*cos(2.*c*x) +
                     0.5*b*b*sin(2.*c*x)*cos(c*x));

          f_rhs_xxx = fa*exp(-kx/c)*(b*b*c*cos(c*x)*cos(c*x)*cos(c*x) - 3.5*b*c*c*sin(2.*c*x) +
                      2.*b*b*c*cos(2.*c*x)*cos(c*x) +0.5*b*b*b*sin(2.*c*x)*cos(c*x)*cos(c*x) - 
                      c*c*c*cos(c*x) - 0.5*b*b*c*sin(2.*c*x)*sin(c*x));

          f_rhs_xxxx =   b*cos(c*x)*f_rhs_xxx + 
                         fa* exp(-kx/c)*(-6.*b*b*c*c*sin(2*c*x)*cos(c*x) -
                         7.*b*c*c*c*cos(2.*c*x) - 3.*b*b*c*c*cos(2.*c*x)*sin(c*x) +
                         b*b*b*c*cos(2.*c*x)*cos(c*x)*cos(c*x) - 0.5*b*b*b*c*sin(2.*c*x)*sin(2.*c*x) +
                         c*c*c*c*sin(c*x));

          fa_y = -b*(2.*a + c)*sqrt(a*a + b*b)*cos(sqrt(a*a + b*b)*y);

          f_rhs_y = fa_y*sin(c*x)*exp(-kx/c);

          f_rhs_yy = -(a*a + b*b)*f_rhs;

          f_rhs_yyyy =  (a*a + b*b)*(a*a + b*b)*f_rhs;

          f_rhs_xxyy = -(a*a + b*b)*f_rhs_xx;

          f =  h2*((1. - kn5*h2/20.)*f_rhs + h2/12.*(f_rhs_xx + f_rhs_yy) + 
               h2*h2/360.*(f_rhs_xxxx + f_rhs_yyyy) + h2*h2/90.*f_rhs_xxyy +
               h2*h2/60.*(kx2 * f_rhs_x + ky2 * f_rhs_y));
       }
    }
L40: 
    icn = 0;
    sum = 0.;
    k = bindx[*location];
    bindx[k] = *row + 1;
    val[k] = a4;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row - 1;
    val[k] = a6;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row + (global_2.n + 2);
    val[k] = a2;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    if (*order == 2) {
	goto L50;
    }

    bindx[k] = *row + (global_2.n + 2) + 1;
    val[k] = a1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row + (global_2.n + 2) - 1;
    val[k] = a3;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

L50:

    bindx[k] = *row - (global_2.n + 2);
    val[k] = a8;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    if (*order == 2) {
	goto L55;
    }
    bindx[k] = *row - (global_2.n + 2) + 1;
    val[k] = a7;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

    bindx[k] = *row - (global_2.n + 2) - 1;
    val[k] = a9;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;

L55:

    *kmax = k;
    bindx[*location + 1] = k;
    val[*location] = a5;
L100:
    rhs[*row] = f;

/*  return 0; */
/*  Normalize the equations: */

    d1 = val[*location];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[*row] /= sum1;
    val[*location] /= sum1;
    i1 = icn;
    for (ii = 1; ii <= i1; ++ii) {
	val[k - ii] /= sum1;
    }

    return 0;
} /* create_matrix_row_9pt_example31_case_c */

/* ******************************************************************** */
/* ******************************************************************** */

/* Subroutine */ int create_matrix_row_7_27_pt_example32_case_a(int *row, int *
	location, double *val, int *bindx, int *kmax, double *
	rhs, int *nxst, int *nyst, int *nzst, int *nxloc, int *nyloc, int *nzloc, 
        double *l_a, double *n_a , int *order, int *i_scheme)

{
    int i1;
    double d1, d2, d3, d4, d5, d6;
    static double ma, na, la, l2, n2;
    static double ac, as, ad;
    static double f, f1;
    static int i, j, kk, ii, k;
    static double x, y, z, pi;
    static int nx, ny, nz, icn;
    static double sum, sum1;
    static double kn2, hkn_2, h, h2, h3, kx2, ky2, kz2, del_k2;
    static double xmh, xph, ymh, yph, zmh, zph;

    static double ac_i_j_k, ac_ip1_j_k, ac_im1_j_k, ac_i_jp1_k, ac_i_jm1_k,
       ac_i_j_kp1, ac_i_j_km1;
    static double ac_ip1_jp1_k, ac_im1_jp1_k, ac_ip1_jm1_k, ac_im1_jm1_k;
    static double ac_ip1_j_kp1, ac_im1_j_kp1, ac_i_jp1_kp1, ac_i_jm1_kp1,
       ac_ip1_jp1_kp1, ac_im1_jp1_kp1, ac_ip1_jm1_kp1, ac_im1_jm1_kp1;
    static double ac_ip1_j_km1, ac_im1_j_km1, ac_i_jp1_km1, ac_i_jm1_km1,
       ac_ip1_jp1_km1, ac_im1_jp1_km1, ac_ip1_jm1_km1, ac_im1_jm1_km1;


/* Purpose:   
   Add one row to an MSR matrix corresponding to a 7 / 27 point discrete
   approximation to the 3D EXAMPLE_32 operator on a (NX+2)*(NY+2)*(NZ+2) cube.
   Example 32_case_a - is the 3D example of Turkel, R&D and Tsynkov 2012,
   but it is not included in the paper.

   Order - is the scheme order: 2nd / 6th order scheme.
   i_scheme = 1  for the 2nd order scheme or the 6th order scheme of Sutmann
                 in the J. of Computational and Applied Mathematics, 2007
                 pp. 15-31, - eq. 17 (applicable to constant K).
   i_scheme = 2  for the 6th order scheme of Turkel, R&D and Tsynkov 2012.  
 
   For this 3D case no 4th order scheme is implemented.
 
   Domain: [0 PI]*[0 PI]*[0 PI] 
 
           Uxx + Uyy + Uzz + K**2*U = f
 
    With Dirichlet boundary condition on all the boundaries. 

    The analytic solution is:
           U(x,y,z)= sin(sin(l*x)*sinh(m*y)*sin(n*z))  
    
    with:  K**2 = l**2*sin(n*z)**2*(sin(l*x)**2 + sinh(m*y)**2) +
                   n**2*sin(l*x)**2*(sinh(m*y)**2 + sin(n*z)**2 )  

    and:   m**2 = l**2 + n**2  so that f=0
                                                                               */
/*-----------------------------------------------------------------------------*/
/*  Parameters:   
       row          == global row number of the new row to be added. 
       location     == local row where diagonal of the new row will be stored. 
       val,bindx    == (see user's guide). On output, val[] and bindx[] 
                       are appended such that the new row has been added.     */

      pi = atan(1.) * 4.;
      h = pi / (double) (global_2.n + 1);
      nx = global_2.n;
      ny = global_2.n;
      nz = global_2.n;
      i = *location  % *nxloc;
      j = *location / *nxloc % *nyloc;
      kk = *location / (*nxloc * *nyloc) % *nzloc;
      x = (*nxst + i - 1) * h;
      y = (*nyst + j - 1) * h;
      z = (*nzst + kk - 1) * h;


/*  Store the  equation coefficients              */
/*----------------------------------------------- */
      f = 0.;
      la = *l_a;
      na = *n_a ;
      ma = sqrt( la * la + na * na);
      f1 = sin(sin(la * x)*sinh(ma * y)*sin(na * z));

/*  Set Dirichlet b.c. on the x=0 plane: */
      if (i + *nxst - 1 == 0 && (j + *nyst - 1 > 0 && j + *nyst - 1 < ny + 1)
	     && (kk + *nzst - 1 >= 0 && kk + *nzst - 1 <= nz + 1)) {
  	icn = 0;
	sum = 0.;
	k = bindx[*location];
        if (*order == 2) {
            goto L20;
        }
        if (j + *nyst - 1 == ny + 1) {
            goto L21;
        }
        bindx[k] = *row + (nx + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L21:
        if (j + *nyst - 1 == 0) {
            goto L22;
        }
        bindx[k] = *row - (nx + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L22:
        if (kk + *nzst - 1 == nz + 1) {
            goto L23;
        }
        bindx[k] = *row + (nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L23:
        if (kk + *nzst - 1 == 0) {
            goto L20;
        }
        bindx[k] = *row - (nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L20:
        f = f1;
        *kmax = k;
        bindx[*location + 1] = k;
        val[*location] = 1.;
        goto L100;
    }

/*  Set Dirichlet b.c. on the x=PI plane */ 
    if (i + *nxst - 1 == nx + 1 && (j + *nyst - 1 > 0 && j + *nyst - 1 < 
	    ny + 1) && (kk + *nzst - 1 >= 0 && kk + *nzst - 1 <= nz + 1)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        if (*order == 2) {
           goto L10;
        }
        if (j + *nyst - 1 == ny + 1) {
           goto L11;
        }
        bindx[k] = *row + (nx + 2);
        val[k] = 0.;
        d1 = val[k]; 
	sum += d1 * d1;
	++icn;
	++k;
L11:
        if (j + *nyst - 1 == 0) {
           goto L12;
        }
        bindx[k] = *row - (nx + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L12:
        if (kk + *nzst - 1 == nz + 1) {
           goto L13;
        }
        bindx[k] = *row + (nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L13:
        if (kk + *nzst - 1 == 0) {
            goto L10;
        }
        bindx[k] = *row - (nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L10:
        f = f1;
        *kmax = k;
        bindx[*location + 1] = k;
        val[*location] = 1.;
        goto L100;
    }

/*  Set Dirichlet b.c. on the y=0 plane: */
    if (j + *nyst - 1 == 0) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        if (*order == 2) {
            goto L18;
        }
        if (i + *nxst - 1 == nx + 1) {
            goto L15;
        }
        bindx[k] = *row + 1;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L15:
        if (i + *nxst - 1 == 0) {
            goto L16;
        }
        bindx[k] = *row - 1;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L16:
        if (kk + *nzst - 1 == nz + 1) {
            goto L17;
        }
        bindx[k] = *row + (nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L17:
        if (kk + *nzst - 1 == 0) {
            goto L18;
        }
        bindx[k] = *row - (nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L18:
        f = f1;
        *kmax = k;
        bindx[*location + 1] = k;
        val[*location] = 1.;
        goto L100;
    }

/*  Set Dirichlet b.c. on the y=PI plane*/            
    if (j + *nyst - 1 == ny + 1) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        if (*order == 2) {
            goto L28;
        }
        if (i + *nxst - 1 == nx + 1) {
            goto L25;
        }
        bindx[k] = *row + 1;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L25:
        if (i + *nxst - 1 == 0) {
            goto L26;
        }
        bindx[k] = *row - 1;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L26:
        if (kk + *nzst - 1 == nz + 1) {
            goto L27;
        }
        bindx[k] = *row + (nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L27:
        if (kk + *nzst - 1 == 0) {
            goto L28;
        }
        bindx[k] = *row - (nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L28:
        f = f1;
        *kmax = k;
        bindx[*location + 1] = k;
        val[*location] = 1.;
        goto L100;
    }

/*  Set Dirichlet b.c. on the z=0. plane */
    if (kk + *nzst - 1 == 0 && (i + *nxst - 1 > 0 && i + *nxst - 1 
	    < nx + 1) && (j + *nyst - 1 > 0 && j + *nyst - 1 < ny + 1)) {
	icn = 0;
	sum = 0.;
        f = f1;
	k = bindx[*location];
        *kmax = k;
        bindx[*location + 1] = k;
        val[*location] = 1.;
        goto L100;
    }
/*  Set Dirichlet b.c. on the z=PI plane */
    if (kk + *nzst - 1 == nz + 1 && (i + *nxst - 1 > 0 && i + *nxst - 1 
	    < nx + 1) && (j + *nyst - 1 > 0 && j + *nyst - 1 < ny + 1)) {
	icn = 0;
	sum = 0.;
        f = f1;
	k = bindx[*location];
        *kmax = k;
        bindx[*location + 1] = k;
        val[*location] = 1.;
        goto L100;
    }
/* --------    end of the b.c. equations setting -------- */
/*  Set the eq. for the inner grid points. */
      h2 = h * h;
      l2 = la * la;
      n2 = na * na;

      d1 = sin(la * x);
      d2 = d1 * d1;
      d3 = sinh(ma * y);
      d4 = d3 * d3;
      d5 = sin(na * z);
      d6 = d5 * d5;
      kn2= l2 * d6 * (d2 + d4 ) + n2 * d2 * (d4 + d6);

      if (*order == 2) {
         ac_i_j_k = (kn2 * h2 - 6.);
         ac_im1_j_k = 1.;
         ac_ip1_j_k = 1.;
         ac_i_jm1_k = 1.;
         ac_i_jp1_k = 1.;
         ac_i_j_kp1 = 1.;
         ac_i_j_km1 = 1.;
         goto L40;
      } /* endif Order=2 */

/*  definition of the coefficients following Sutmann's 2007 paper in JCAM, for   
    constant K, using the L6c scheme - eq. (17).                               */

    if(*i_scheme == 1 &&  *order == 6) {
/*  k plane: */
/* ------------- */
        d2 = kn2 * h * h;
        ac_i_j_k = -64. / 15.  * (1. - d2 * .25 +
                5. / 256. * d2 * d2  - 1./ 1536. * d2 * (d2 * d2)); 
        as = 7./ 15.  * (1. - d2 / 21.);
        ac = .1  * (1. + d2 / 18.);
        ad = 1./ 30. ;

/*  symmetric points */
        ac_i_jm1_k = as;
        ac_i_jp1_k = as;
        ac_im1_j_k = as;
        ac_ip1_j_k = as;
/*  corner points: */
        ac_im1_jm1_k = ac;
        ac_ip1_jm1_k = ac;
        ac_im1_jp1_k = ac;
        ac_ip1_jp1_k = ac;
/*  k-1 plane: */
/* -------------- */
        ac_i_j_km1 = as;
/*  symmetric points : */
        ac_i_jm1_km1 = ac;
        ac_i_jp1_km1 = ac;
        ac_im1_j_km1 = ac;
        ac_ip1_j_km1 = ac;
/*  corner points: */
        ac_im1_jm1_km1 = ad;
        ac_ip1_jm1_km1 = ad;
        ac_im1_jp1_km1 = ad;
        ac_ip1_jp1_km1 = ad;
/*  k+1 plane: */
/* -------------- */
        ac_i_j_kp1 = as;
/*  symmetric points */
        ac_i_jm1_kp1 = ac;
        ac_i_jp1_kp1 = ac;
        ac_im1_j_kp1 = ac;
        ac_ip1_j_kp1 = ac;
/*  corner points: */
        ac_im1_jm1_kp1 = ad;
        ac_ip1_jm1_kp1 = ad;
        ac_im1_jp1_kp1 = ad;
        ac_ip1_jp1_kp1 = ad;
         goto L40;
    } /* endif 6th order following Sutmann 2007 for cobstant K */

/*----------------*/

/*  definition of the coefficients following Turkel, R&D and Tsynkov 2012:  */
    if(*i_scheme == 2 &&  *order == 6) {
       xph = x + h;
       xmh = x - h;
       yph = y + h;
       ymh = y - h;
       zph = z + h;
       zmh = z - h;
       h3 = h*h*h;

/*     The derivatives of k**2 are calculated at the center point
       kx2 - the first derivative of k**2 with respect to x
       ky2 - the first derivative of k**2 with respect to y   
       kz2 - the first derivative of k**2 with respect to z               */

        kx2 = la*sin(2.*la*x)*(na*na*sinh(ma*y)*sinh(ma*y) + ma*ma*sin(na*z)*sin(na*z));
        ky2 = ma*sinh(2.*ma*y)*(na*na*sin(la*x)*sin(la*x) + la*la*sin(na*z)*sin(na*z));
        kz2 = na*sin(2.*na*z)*(la*la*sinh(ma*y)*sinh(ma*y) + ma*ma*sin(la*x)*sin(la*x));

        del_k2 = 2.*la*la*cos(2.*la*x)*(na*na*sinh(ma*y)*sinh(ma*y) + ma*ma*sin(na*z)*sin(na*z)) +
                 2.*na*na*cos(2.*na*z)*(la*la*sinh(ma*y)*sinh(ma*y) + ma*ma*sin(la*x)*sin(la*x)) +
                 2.*ma*ma*cosh(2.*ma*y)*(na*na*sin(la*x)*sin(la*x)  + la*la*sin(na*z)*sin(na*z));

/*     k plane       */
/*    center point - 0 */
        hkn_2 = kn2 * h2;
        ac_i_j_k = -64./15. + (14./15.)*hkn_2 + h2*h2/20. *(del_k2 - kn2*kn2) ;

/*     symmetric points -  ss points*/
        d1 = sin(la * x);
        d2 = d1 * d1;
        d3 = sinh(ma * ymh);
        d4 = d3 * d3;
        d5 = sin(na * z);
        d6 = d5 * d5;
        kn2= l2 * d6 * (d2 + d4 ) + n2 * d2 * (d4 + d6);
        hkn_2 = kn2 * h2;   
        ac_i_jm1_k =  7./15. - hkn_2/90. - ky2 * h3/20.*(1. + hkn_2/6. - 4./6.);   

        d1 = sin(la * x);
        d2 = d1 * d1;
        d3 = sinh(ma * yph);
        d4 = d3 * d3;
        d5 = sin(na * z);
        d6 = d5 * d5;
        kn2= l2 * d6 * (d2 + d4 ) + n2 * d2 * (d4 + d6);
        hkn_2 = kn2 * h2;   
        ac_i_jp1_k =  7./15. - hkn_2/90. + ky2 * h3/20.*(1. + hkn_2/6. - 4./6.);   

        d1 = sin(la * xph);
        d2 = d1 * d1;
        d3 = sinh(ma * y);
        d4 = d3 * d3;
        d5 = sin(na * z);
        d6 = d5 * d5;
        kn2= l2 * d6 * (d2 + d4 ) + n2 * d2 * (d4 + d6);
        hkn_2 = kn2 * h2;   
        ac_ip1_j_k =  7./15. - hkn_2/90. + kx2 * h3/20.*(1. + hkn_2/6. - 4./6.);   

        d1 = sin(la * xmh);
        d2 = d1 * d1;
        d3 = sinh(ma * y);
        d4 = d3 * d3;
        d5 = sin(na * z);
        d6 = d5 * d5;
        kn2= l2 * d6 * (d2 + d4 ) + n2 * d2 * (d4 + d6);
        hkn_2 = kn2 * h2;   
        ac_im1_j_k =  7./15. - hkn_2/90. - kx2 * h3/20.*(1. + hkn_2/6. - 4./6.);   

/*  corner-symmetric points  - cs points: */
        d1 = sin(la * xmh);
        d2 = d1 * d1;
        d3 = sinh(ma * ymh);
        d4 = d3 * d3;
        d5 = sin(na * z);
        d6 = d5 * d5;
        kn2= l2 * d6 * (d2 + d4 ) + n2 * d2 * (d4 + d6);
        hkn_2 = kn2 * h2;   
        ac_im1_jm1_k =  1./10. + hkn_2/90. -  h3/120.*(kx2 + ky2);          

        d1 = sin(la * xph);
        d2 = d1 * d1;
        d3 = sinh(ma * ymh);
        d4 = d3 * d3;
        d5 = sin(na * z);
        d6 = d5 * d5;
        kn2= l2 * d6 * (d2 + d4 ) + n2 * d2 * (d4 + d6);
        hkn_2 = kn2 * h2;   
        ac_ip1_jm1_k =  1./10. + hkn_2/90. + h3/120.*(kx2 - ky2);          

        d1 = sin(la * xmh);
        d2 = d1 * d1;
        d3 = sinh(ma * yph);
        d4 = d3 * d3;
        d5 = sin(na * z);
        d6 = d5 * d5;
        kn2= l2 * d6 * (d2 + d4 ) + n2 * d2 * (d4 + d6);
        hkn_2 = kn2 * h2;   
        ac_im1_jp1_k =  1./10. + hkn_2/90. + h3/120.*(-kx2 + ky2);          

        d1 = sin(la * xph);
        d2 = d1 * d1;
        d3 = sinh(ma * yph);
        d4 = d3 * d3;
        d5 = sin(na * z);
        d6 = d5 * d5;
        kn2= l2 * d6 * (d2 + d4 ) + n2 * d2 * (d4 + d6);
        hkn_2 = kn2 * h2;   
        ac_ip1_jp1_k =  1./10. + hkn_2/90. + h3/120.*(kx2 + ky2); 

/*  k-1 plane: */
/* -------------- */
/*     symmetric points -  ss point*/
        d1 = sin(la * x);
        d2 = d1 * d1;
        d3 = sinh(ma * y);
        d4 = d3 * d3;
        d5 = sin(na * zmh);
        d6 = d5 * d5;
        kn2= l2 * d6 * (d2 + d4 ) + n2 * d2 * (d4 + d6);
        hkn_2 = kn2 * h2;   
        ac_i_j_km1 =  7./15. - hkn_2/90. - kz2 * h3/20.*(1. + hkn_2/6. - 4./6.);

/*  corner-symmetric points  - cs points: */
        d1 = sin(la * x);
        d2 = d1 * d1;
        d3 = sinh(ma * ymh);
        d4 = d3 * d3;
        d5 = sin(na * zmh);
        d6 = d5 * d5;
        kn2= l2 * d6 * (d2 + d4 ) + n2 * d2 * (d4 + d6);
        hkn_2 = kn2 * h2;   
        ac_i_jm1_km1 =  1./10. + hkn_2/90. -  h3/120.*(ky2 + kz2);          

        d1 = sin(la * x);
        d2 = d1 * d1;
        d3 = sinh(ma * yph);
        d4 = d3 * d3;
        d5 = sin(na * zmh);
        d6 = d5 * d5;
        kn2= l2 * d6 * (d2 + d4 ) + n2 * d2 * (d4 + d6);
        hkn_2 = kn2 * h2;   
        ac_i_jp1_km1 =  1./10. + hkn_2/90. +  h3/120.*(ky2 - kz2);          

        d1 = sin(la * xmh);
        d2 = d1 * d1;
        d3 = sinh(ma * y);
        d4 = d3 * d3;
        d5 = sin(na * zmh);
        d6 = d5 * d5;
        kn2= l2 * d6 * (d2 + d4 ) + n2 * d2 * (d4 + d6);
        hkn_2 = kn2 * h2;   
        ac_im1_j_km1 =  1./10. + hkn_2/90. -  h3/120.*(kx2 + kz2);          

        d1 = sin(la * xph);
        d2 = d1 * d1;
        d3 = sinh(ma * y);
        d4 = d3 * d3;
        d5 = sin(na * zmh);
        d6 = d5 * d5;
        kn2= l2 * d6 * (d2 + d4 ) + n2 * d2 * (d4 + d6);
        hkn_2 = kn2 * h2;   
        ac_ip1_j_km1 =  1./10. + hkn_2/90. +  h3/120.*(kx2 - kz2);          

/*  corner points - cc points: */
/*      d1 = sin(la * xmh);
        d2 = d1 * d1;
        d3 = sinh(ma * ymh);
        d4 = d3 * d3;
        d5 = sin(na * zmh);
        d6 = d5 * d5;
        kn2= l2 * d6 * (d2 + d4 ) + n2 * d2 * (d4 + d6);
        hkn_2 = kn2 * h2;    
*/

        ac_im1_jm1_km1 = 1./30.;          
        ac_ip1_jm1_km1 = 1./30.;
        ac_im1_jp1_km1 = 1./30.;
        ac_ip1_jp1_km1 = 1./30.;

/*  k+1 plane: */
/* -------------- */
/*   symmetric points -  ss point*/
        d1 = sin(la * x);
        d2 = d1 * d1;
        d3 = sinh(ma * y);
        d4 = d3 * d3;
        d5 = sin(na * zph);
        d6 = d5 * d5;
        kn2= l2 * d6 * (d2 + d4 ) + n2 * d2 * (d4 + d6);
        hkn_2 = kn2 * h2;   
        ac_i_j_kp1 =  7./15. - hkn_2/90. + kz2 * h3/20.*(1. + hkn_2/6. - 4./6.);

/*  corner-symmetric points  - cs points: */
        d1 = sin(la * x);
        d2 = d1 * d1;
        d3 = sinh(ma * ymh);
        d4 = d3 * d3;
        d5 = sin(na * zph);
        d6 = d5 * d5;
        kn2= l2 * d6 * (d2 + d4 ) + n2 * d2 * (d4 + d6);
        hkn_2 = kn2 * h2;   
        ac_i_jm1_kp1 =  1./10. + hkn_2/90. +  h3/120.*(-ky2 + kz2);          

        d1 = sin(la * x);
        d2 = d1 * d1;
        d3 = sinh(ma * yph);
        d4 = d3 * d3;
        d5 = sin(na * zph);
        d6 = d5 * d5;
        kn2= l2 * d6 * (d2 + d4 ) + n2 * d2 * (d4 + d6);
        hkn_2 = kn2 * h2;   
        ac_i_jp1_kp1 =  1./10. + hkn_2/90. +  h3/120.*(ky2 + kz2);          

        d1 = sin(la * xmh);
        d2 = d1 * d1;
        d3 = sinh(ma * y);
        d4 = d3 * d3;
        d5 = sin(na * zph);
        d6 = d5 * d5;
        kn2= l2 * d6 * (d2 + d4 ) + n2 * d2 * (d4 + d6);
        hkn_2 = kn2 * h2;   
        ac_im1_j_kp1 =  1./10. + hkn_2/90. +  h3/120.*(-kx2 + kz2);          

        d1 = sin(la * xph);
        d2 = d1 * d1;
        d3 = sinh(ma * y);
        d4 = d3 * d3;
        d5 = sin(na * zph);
        d6 = d5 * d5;
        kn2= l2 * d6 * (d2 + d4 ) + n2 * d2 * (d4 + d6);
        hkn_2 = kn2 * h2;   
        ac_ip1_j_kp1 =  1./10. + hkn_2/90. +  h3/120.*(kx2 + kz2);          

/*  corner points - cc points: */
        ac_im1_jm1_kp1 = 1./30.;
        ac_ip1_jm1_kp1 = 1./30.;
        ac_im1_jp1_kp1 = 1./30.;
        ac_ip1_jp1_kp1 = 1./30.;
    }
L40:
    icn = 0;
    sum = 0.;
    k = bindx[*location];
    bindx[k] = *row + 1;
    val[k] =  ac_ip1_j_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 1;
    val[k] = ac_im1_j_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + (nx + 2);
    val[k] = ac_i_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - (nx + 2);
    val[k] = ac_i_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + (nx + 2) * (ny + 2);
    val[k] = ac_i_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - (nx + 2) * (ny + 2);
    val[k] = ac_i_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    if (*order == 2) {
        goto L50;
    }
    bindx[k] = *row + (nx + 2) + 1;
    val[k] = ac_ip1_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + (nx + 2) - 1;
    val[k] = ac_im1_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - (nx + 2) + 1;
    val[k] = ac_ip1_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - (nx + 2) - 1;
    val[k] = ac_im1_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* ----- k+1 plane: */
    bindx[k] = *row + (nx + 2) * (ny + 2) + 1;
    val[k] = ac_ip1_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + (nx + 2) * (ny + 2) - 1;
    val[k] = ac_im1_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + (nx + 2) * (ny + 2) + (nx + 2);
    val[k] = ac_i_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + (nx + 2) * (ny + 2) - (nx + 2);
    val[k] = ac_i_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* -- */
    bindx[k] = *row + (nx + 2) * (ny + 2) + (nx + 2) + 1;
    val[k] = ac_ip1_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + (nx + 2) * (ny + 2) + (nx + 2) - 1;
    val[k] = ac_im1_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + (nx + 2) * (ny + 2) - (nx + 2) + 1;
    val[k] = ac_ip1_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + (nx + 2) * (ny + 2) - (nx + 2) - 1;
    val[k] = ac_im1_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* ----- k-1 plane: */
    bindx[k] = *row - (nx + 2) * (ny + 2) + 1;
    val[k] = ac_ip1_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - (nx + 2) * (ny + 2) - 1;
    val[k] = ac_im1_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - (nx + 2) * (ny + 2) + (nx + 2);
    val[k] = ac_i_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - (nx + 2) * (ny + 2) - (nx + 2);
    val[k] = ac_i_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* -- */
    bindx[k] = *row - (nx + 2) * (ny + 2) + (nx + 2) + 1;
    val[k] = ac_ip1_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - (nx + 2) * (ny + 2) + (nx + 2) - 1;
    val[k] = ac_im1_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - (nx + 2) * (ny + 2) - (nx + 2) + 1;
    val[k] = ac_ip1_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - (nx + 2) * (ny + 2) - (nx + 2) - 1;
    val[k] = ac_im1_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
L50:
    *kmax = k;
    bindx[*location + 1] = k;
    val[*location] = ac_i_j_k;
L100:
    rhs[*row] = f;
/*  return */
/*  normalize the equations: */
    d1 = val[*location];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[*row] /= sum1;
    val[*location] /= sum1;
    i1 = icn;
    for (ii = 1; ii <= i1; ++ii) {
        val[*kmax - ii] /= sum1;
    }
/*  printf("\n%7d %13.10f %13.10f", *row, rhs[*row], val[*location]);
     printf("\n    ");
    for (ii=1; ii<=i1; ++ii) printf(" i=%d  %d", ii,bindx[*kmax - ii]) ;     
    for (ii=1; ii<=i1; ++ii) printf(" %13.10f", val[*kmax - ii]) ;   */
    return 0;
} /* create_matrix_row_7_27pt_example32_case_a */

/* ************************************************************************ */
/* ******************************************************************** */

/* Subroutine */ int create_matrix_row_7_27_pt_example32_case_c(int *row, int *
	location, double *val, int *bindx, int *kmax, double *
	rhs, int *nxst, int *nyst, int *nzst, int *nxloc, int *nyloc, int *nzloc, 
        double *a2, double *b2 , double *c2, double *gam, int *order, int *i_scheme)

{
    int i1, i_rhs;
    double d1, d2;
    static double a, b, c, beta, gama;
    static double ac, as, ad;
    static double f, f1;
    static int i, j, kk, ii, k;
    static double x, y, z, pi;
    static int nx, ny, nz, icn;
    static double sum, sum1;
    static double kn, kn2, hkn_2, h, h2, h3, kx2, ky2, kz2, del_k2;
    static double xmh, xph, ymh, yph, zmh, zph;
    double f_rhs,  psi, kp, kpp, kppp,kpppp, f_rhs_x, f_rhs_xx, f_rhs_xxx, f_rhs_xxxx, 
                   f_rhs_y, f_rhs_yy, f_rhs_yyy, f_rhs_yyyy, 
                   f_rhs_z, f_rhs_zz, f_rhs_zzz, f_rhs_zzzz, f_rhs_xxyy, f_rhs_xxzz, f_rhs_yyzz;

    static double ac_i_j_k,ac_ip1_j_k,ac_im1_j_k,ac_i_jp1_k,ac_i_jm1_k,
       ac_i_j_kp1,ac_i_j_km1;
    static double ac_ip1_jp1_k,ac_im1_jp1_k,ac_ip1_jm1_k,ac_im1_jm1_k;
    static double ac_ip1_j_kp1,ac_im1_j_kp1,ac_i_jp1_kp1,ac_i_jm1_kp1,
       ac_ip1_jp1_kp1,ac_im1_jp1_kp1,ac_ip1_jm1_kp1,ac_im1_jm1_kp1;
    static double ac_ip1_j_km1,ac_im1_j_km1,ac_i_jp1_km1,ac_i_jm1_km1,
       ac_ip1_jp1_km1,ac_im1_jp1_km1,ac_ip1_jm1_km1,ac_im1_jm1_km1;

/* Purpose:   
   Add one row to an MSR matrix corresponding to a 7 / 27 point discrete
   approximation to the 3D EXAMPLE_32 operator on a (NX+2)*(NY+2)*(NZ+2) cube.
   Example 32_case_c = is the 3D example of Turkel, R&D and Tsynkov 2012.

   Order - is the scheme order: 2nd / 6th order scheme.
   i_scheme = 1  for the 2nd order scheme or the 6th order scheme of Sutmann 
                 in the J. of Computational and Applied Mathematics, 2007,
                 pp. 15-31, - eq. 17 (applicable to constant K).
   i_scheme = 2  for the 6th order scheme of Turkel, R&D and Tsynkov 2012.

   For this 3D case no 4th order scheme is implemented.
 
   Domain: [0 PI]*[0 PI]*[0 PI] 
 
           Uxx + Uyy + Uzz + K**2*U = f
 
    With Dirichlet boundary condition on all the boundaries.

    The analytic solution is:
          u(x,y)=exp(-K/c)*sin(beta*y)*sin(gama*z)   for general a, b, c and gama.                              

    and:  K = a - b*sin(c*x)    
    if  c = -2*a ==> f=0                
                                                                                     */
/*-----------------------------------------------------------------------------------*/
/*  Parameters:   
       row          == global row number of the new row to be added.   
       location     == local row where diagonal of the new row will be stored.   
       val,bindx    == (see user's guide). On output, val[] and bindx[]   
                       are appended such that the new row has been added.           */

      pi = atan(1.) * 4.;
      h = pi / (double) (global_2.n + 1);
      nx = global_2.n;
      ny = global_2.n;
      nz = global_2.n;
      i = *location  % *nxloc;
      j = *location / *nxloc % *nyloc;
      kk = *location / (*nxloc * *nyloc) % *nzloc;
      x = (*nxst + i - 1) * h;
      y = (*nyst + j - 1) * h;
      z = (*nzst + kk - 1) * h;


/*  Store the  equation coefficients              */
/*----------------------------------------------- */
      a = *a2;
      b = *b2;
      c = *c2;
      gama = *gam;
      beta = sqrt(a*a + b*b - gama*gama);

      i_rhs=1; /*  right handside is not zero  - true */
      if(  abs(2.*a + c) < 1.e-7) i_rhs= 0;

      kn = a - b*sin(c*x);

      f1 = exp(-kn/c)*sin(beta*y)*sin(gama*z); 

/*  Set Dirichlet b.c. on the x=0 plane: */
      if (i + *nxst - 1 == 0 && (j + *nyst - 1 > 0 && j + *nyst - 1 < ny + 1)
	     && (kk + *nzst - 1 >= 0 && kk + *nzst - 1 <= nz + 1)) {
  	icn = 0;
	sum = 0.;
	k = bindx[*location];
        if (*order == 2) {
            goto L20;
        }
        if (j + *nyst - 1 == ny + 1) {
            goto L21;
        }
        bindx[k] = *row + (nx + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L21:
        if (j + *nyst - 1 == 0) {
            goto L22;
        }
        bindx[k] = *row - (nx + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L22:
        if (kk + *nzst - 1 == nz + 1) {
            goto L23;
        }
        bindx[k] = *row + (nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L23:
        if (kk + *nzst - 1 == 0) {
            goto L20;
        }
        bindx[k] = *row - (nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L20:
        f = f1;
        *kmax = k;
        bindx[*location + 1] = k;
        val[*location] = 1.;
        goto L100;
    }

/*  Set Dirichlet b.c. on the x=PI plane */ 
    if (i + *nxst - 1 == nx + 1 && (j + *nyst - 1 > 0 && j + *nyst - 1 < 
	    ny + 1) && (kk + *nzst - 1 >= 0 && kk + *nzst - 1 <= nz + 1)) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        if (*order == 2) {
           goto L10;
        }
        if (j + *nyst - 1 == ny + 1) {
           goto L11;
        }
        bindx[k] = *row + (nx + 2);
        val[k] = 0.;
        d1 = val[k]; 
	sum += d1 * d1;
	++icn;
	++k;
L11:
        if (j + *nyst - 1 == 0) {
           goto L12;
        }
        bindx[k] = *row - (nx + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L12:
        if (kk + *nzst - 1 == nz + 1) {
           goto L13;
        }
        bindx[k] = *row + (nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L13:
        if (kk + *nzst - 1 == 0) {
            goto L10;
        }
        bindx[k] = *row - (nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L10:
        f = f1;
        *kmax = k;
        bindx[*location + 1] = k;
        val[*location] = 1.;
        goto L100;
    }

/*  Set Dirichlet b.c. on the y=0 plane: */
    if (j + *nyst - 1 == 0) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        if (*order == 2) {
            goto L18;
        }
        if (i + *nxst - 1 == nx + 1) {
            goto L15;
        }
        bindx[k] = *row + 1;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L15:
        if (i + *nxst - 1 == 0) {
            goto L16;
        }
        bindx[k] = *row - 1;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L16:
        if (kk + *nzst - 1 == nz + 1) {
            goto L17;
        }
        bindx[k] = *row + (nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L17:
        if (kk + *nzst - 1 == 0) {
            goto L18;
        }
        bindx[k] = *row - (nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L18:
        f = f1;
        *kmax = k;
        bindx[*location + 1] = k;
        val[*location] = 1.;
        goto L100;
    }

/*  Set Dirichlet b.c. on the y=PI plane*/            
    if (j + *nyst - 1 == ny + 1) {
	icn = 0;
	sum = 0.;
	k = bindx[*location];
        if (*order == 2) {
            goto L28;
        }
        if (i + *nxst - 1 == nx + 1) {
            goto L25;
        }
        bindx[k] = *row + 1;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L25:
        if (i + *nxst - 1 == 0) {
            goto L26;
        }
        bindx[k] = *row - 1;
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L26:
        if (kk + *nzst - 1 == nz + 1) {
            goto L27;
        }
        bindx[k] = *row + (nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L27:
        if (kk + *nzst - 1 == 0) {
            goto L28;
        }
        bindx[k] = *row - (nx + 2) * (ny + 2);
        val[k] = 0.;
        d1 = val[k];
        sum += d1 * d1;
        ++icn;
        ++k;
L28:
        f = f1;
        *kmax = k;
        bindx[*location + 1] = k;
        val[*location] = 1.;
        goto L100;
    }

/*  Set Dirichlet b.c. on the z=0. plane */
    if (kk + *nzst - 1 == 0 && (i + *nxst - 1 > 0 && i + *nxst - 1 
	    < nx + 1) && (j + *nyst - 1 > 0 && j + *nyst - 1 < ny + 1)) {
	icn = 0;
	sum = 0.;
        f = f1;
	k = bindx[*location];
        *kmax = k;
        bindx[*location + 1] = k;
        val[*location] = 1.;
        goto L100;
    }
/*  Set Dirichlet b.c. on the z=PI plane */
    if (kk + *nzst - 1 == nz + 1 && (i + *nxst - 1 > 0 && i + *nxst - 1 
	    < nx + 1) && (j + *nyst - 1 > 0 && j + *nyst - 1 < ny + 1)) {
	icn = 0;
	sum = 0.;
        f = f1;
	k = bindx[*location];
        *kmax = k;
        bindx[*location + 1] = k;
        val[*location] = 1.;
        goto L100;
    }

/* --------    end of the b.c. equations setting -------- */

/*  Set the eq. for the inner grid points. */
      h2 = h * h;
      kn = a - b*sin(c*x);
      kn2 = kn * kn;

      if (*order == 2) {
         ac_i_j_k = (kn2 * h2 - 6.);
         ac_im1_j_k = 1.;
         ac_ip1_j_k = 1.;
         ac_i_jm1_k = 1.;
         ac_i_jp1_k = 1.;
         ac_i_j_kp1 = 1.;
         ac_i_j_km1 = 1.;

        if(i_rhs == 0) f=0.;
        if(i_rhs == 1 ){
          psi = -b*(2*a + c)*exp(-kn/c)*sin(beta*y)*sin(gama*z);   
          f_rhs =  psi*sin(c*x);
          f =  h*h*f_rhs;
        }
         goto L40;
      } /* endif Order=2 */

/*  definition of the coefficients following Sutmann's 2007 paper in JCAM, for   
    constant K, using the L6c scheme - eq. (17).                               */

    if(*i_scheme == 1 &&  *order == 6) {

/*  k plane: */
/* ------------- */
        d2 = kn2 * h * h;
        ac_i_j_k = -64. / 15.  * (1. - d2 * .25 +
                5. / 256. * d2 * d2  - 1./ 1536. * d2 * (d2 * d2)); 
        as = 7./ 15.  * (1. - d2 / 21.);
        ac = .1  * (1. + d2 / 18.);
        ad = 1./ 30. ;

/*  symmetric points */
        ac_i_jm1_k = as;
        ac_i_jp1_k = as;
        ac_im1_j_k = as;
        ac_ip1_j_k = as;
/*  corner points: */
        ac_im1_jm1_k = ac;
        ac_ip1_jm1_k = ac;
        ac_im1_jp1_k = ac;
        ac_ip1_jp1_k = ac;
/*  k-1 plane: */
/* -------------- */
        ac_i_j_km1 = as;
/*  symmetric points : */
        ac_i_jm1_km1 = ac;
        ac_i_jp1_km1 = ac;
        ac_im1_j_km1 = ac;
        ac_ip1_j_km1 = ac;
/*  corner points: */
        ac_im1_jm1_km1 = ad;
        ac_ip1_jm1_km1 = ad;
        ac_im1_jp1_km1 = ad;
        ac_ip1_jp1_km1 = ad;
/*  k+1 plane: */
/* -------------- */
        ac_i_j_kp1 = as;
/*  symmetric points */
        ac_i_jm1_kp1 = ac;
        ac_i_jp1_kp1 = ac;
        ac_im1_j_kp1 = ac;
        ac_ip1_j_kp1 = ac;
/*  corner points: */
        ac_im1_jm1_kp1 = ad;
        ac_ip1_jm1_kp1 = ad;
        ac_im1_jp1_kp1 = ad;
        ac_ip1_jp1_kp1 = ad;

        if(i_rhs == 0) f=0.;
        if(i_rhs == 1 ){
          psi = -b*(2*a + c)*exp(-kn/c)*sin(beta*y)*sin(gama*z);   
          f_rhs =  psi*sin(c*x);
          f =  h*h*f_rhs;
        }
         goto L40;
    } /* endif 6th order following Sutmann 2007 for cobstant K */

/*----------------*/

/*  definition of the coefficients following Turkel, R&D and Tsynkov 2012:  */
    if(*i_scheme == 2 &&  *order == 6) {
       xph = x + h;
       xmh = x - h;
       yph = y + h;
       ymh = y - h;
       zph = z + h;
       zmh = z - h;
       h3 = h*h*h;

/*     The derivatives of k**2 are calculated at the center point
       kx2 - the first derivative of k**2 with reserct to x
       ky2 - the first derivative of k**2 with reserct to y   
       kz2 - the first derivative of k**2 with reserct to z                 */

        kx2 = -2. * a*b*c*cos(c*x) + b*b*c*sin(2.*c*x);
        ky2 = 0.;
        kz2 = 0.;
        del_k2 =  2.*a*b*c*c*sin(c*x) + 2.*b*b*c*c*cos(2.*c*x);

/*     k plane       */
/*    center point - 0 */
        kn = a - b*sin(c*x);
        kn2 = kn * kn;
        hkn_2 = kn2 * h2;
        ac_i_j_k = -64./15. + (14./15.)*hkn_2 + h2*h2/20. *(del_k2 - kn2*kn2) ;

/*     symmetric points -  ss points*/
        kn = a - b*sin(c*x);
        kn2 = kn * kn;
        hkn_2 = kn2 * h2;   
        ac_i_jm1_k =  7./15. - hkn_2/90. - ky2 * h3/20.*(1./3. + hkn_2/6.);   

        kn = a - b*sin(c*x);
        kn2 = kn * kn;
        hkn_2 = kn2 * h2;   
        ac_i_jp1_k =  7./15. - hkn_2/90. + ky2 * h3/20.*(1./3. + hkn_2/6.);   

        kn = a - b*sin(c*xph);
        kn2 = kn * kn;
        hkn_2 = kn2 * h2;   
        ac_ip1_j_k =  7./15. - hkn_2/90. + kx2 * h3/20.*(1./3. + hkn_2/6.);   

        kn = a - b*sin(c*xmh);
        kn2 = kn * kn;
        hkn_2 = kn2 * h2;   
        ac_im1_j_k =  7./15. - hkn_2/90. - kx2 * h3/20.*(1./3. + hkn_2/6.);   

/*  corner-symmetric points  - cs points: */
        kn = a - b*sin(c*xmh);
        kn2 = kn * kn;
        hkn_2 = kn2 * h2;   
        ac_im1_jm1_k =  1./10. + hkn_2/90. -  h3/120.*(kx2 + ky2);          

        kn = a - b*sin(c*xph);
        kn2 = kn * kn;
        hkn_2 = kn2 * h2;   
        ac_ip1_jm1_k =  1./10. + hkn_2/90. + h3/120.*(kx2 - ky2);          

        kn = a - b*sin(c*xmh);
        kn2 = kn * kn;
        hkn_2 = kn2 * h2;   
        ac_im1_jp1_k =  1./10. + hkn_2/90. + h3/120.*(-kx2 + ky2);          

        kn = a - b*sin(c*xph);
        kn2 = kn * kn;
        hkn_2 = kn2 * h2;   
        ac_ip1_jp1_k =  1./10. + hkn_2/90. + h3/120.*(kx2 + ky2); 

/*  k-1 plane: */
/* -------------- */
/*     symmetric points -  ss point*/
        kn = a - b*sin(c*x);
        kn2 = kn * kn;
        hkn_2 = kn2 * h2;   
        ac_i_j_km1 =  7./15. - hkn_2/90. - kz2 * h3/20.*(1./3. + hkn_2/6.);

/*  corner-symmetric points  - cs points: */
        kn = a - b*sin(c*x);
        kn2 = kn * kn;
        hkn_2 = kn2 * h2;   
        ac_i_jm1_km1 =  1./10. + hkn_2/90. -  h3/120.*(ky2 + kz2);          

        kn = a - b*sin(c*x);
        kn2 = kn * kn;
        hkn_2 = kn2 * h2;   
        ac_i_jp1_km1 =  1./10. + hkn_2/90. +  h3/120.*(ky2 - kz2);          

        kn = a - b*sin(c*xmh);
        kn2 = kn * kn;
        hkn_2 = kn2 * h2;   
        ac_im1_j_km1 =  1./10. + hkn_2/90. -  h3/120.*(kx2 + kz2);          

        kn = a - b*sin(c*xph);
        kn2 = kn * kn;
        hkn_2 = kn2 * h2;   
        ac_ip1_j_km1 =  1./10. + hkn_2/90. +  h3/120.*(kx2 - kz2);          

/*  corner points - cc points: */
        ac_im1_jm1_km1 = 1./30.;          
        ac_ip1_jm1_km1 = 1./30.;
        ac_im1_jp1_km1 = 1./30.;
        ac_ip1_jp1_km1 = 1./30.;

/*  k+1 plane: */
/* -------------- */
/*   symmetric points -  ss point*/
        kn = a - b*sin(c*x);
        kn2 = kn * kn;
        hkn_2 = kn2 * h2;   
        ac_i_j_kp1 =  7./15. - hkn_2/90. + kz2 * h3/20.*(1./3. + hkn_2/6.);

/*  corner-symmetric points  - cs points: */
        kn = a - b*sin(c*x);
        kn2 = kn * kn;
        hkn_2 = kn2 * h2;   
        ac_i_jm1_kp1 =  1./10. + hkn_2/90. +  h3/120.*(-ky2 + kz2);          

        kn = a - b*sin(c*x);
        kn2 = kn * kn;
        hkn_2 = kn2 * h2;   
        ac_i_jp1_kp1 =  1./10. + hkn_2/90. +  h3/120.*(ky2 + kz2);          

        kn = a - b*sin(c*xmh);
        kn2 = kn * kn;
        hkn_2 = kn2 * h2;   
        ac_im1_j_kp1 =  1./10. + hkn_2/90. +  h3/120.*(-kx2 + kz2);          

        kn = a - b*sin(c*xph);
        kn2 = kn * kn;
        hkn_2 = kn2 * h2;   
        ac_ip1_j_kp1 =  1./10. + hkn_2/90. +  h3/120.*(kx2 + kz2);          

/*  corner points - cc points: */
        ac_im1_jm1_kp1 = 1./30.;
        ac_ip1_jm1_kp1 = 1./30.;
        ac_im1_jp1_kp1 = 1./30.;
        ac_ip1_jp1_kp1 = 1./30.;

       if(i_rhs == 0) f=0.;
       if(i_rhs ==1 ){
          kn = a - b*sin(c*x);
          kn2 = kn*kn;
          h2 = h*h;
          psi = -b*(2*a + c)*exp(-kn/c)*sin(beta*y)*sin(gama*z);   
          f_rhs =  psi*sin(c*x);
          kp    = -b*c*cos(c*x);
          kpp   =  b*c*c*sin(c*x);
          kppp  =  b*c*c*c*cos(c*x);
          kpppp = -b*c*c*c*c*sin(c*x);
          f_rhs_x = psi*(c*cos(c*x) - (kp/c)*sin(c*x));
          f_rhs_xx = psi*((-c*c -kpp/c + (kp/c)*(kp/c))*sin(c*x) - 2.* kp*cos(c*x));
          f_rhs_xxx =psi*((-kppp/c + 3.*kp*c + 3.*kp*kpp/(c*c) - (kp/c)*(kp/c)*(kp/c))*sin(c*x) +
                    (-c*c*c - 3.*kpp + 3.*kp*kp/c)*cos(c*x)); 
          f_rhs_xxxx= psi*((c*c*c*c - kpppp/c + 6.*kpp*c - 6.*kp*kp + 4.*kp*kppp/(c*c) +
                     3.*kpp*kpp/(c*c) - 6.*kp*kp*kpp/(c*c*c) + (kp/c)*(kp/c)*(kp/c)*(kp/c))*sin(c*x) +
                     (-4.*kppp + 4.*kp*c*c + 12.*kp*kpp/c - 4.*kp*kp*kp/(c*c))*cos(c*x));
          f_rhs_y = -b*(2.*a + c)*beta*sin(c*x)*exp(-kn/c)*cos(beta*y)*sin(gama*z);
          f_rhs_yy = -beta*beta*f_rhs;
          f_rhs_yyy = -beta*beta*f_rhs_y;
          f_rhs_yyyy= beta*beta*beta*beta*f_rhs;
          f_rhs_z =  -b*(2.*a + c)*gama*sin(c*x)*exp(-kn/c)*sin(beta*y)*cos(gama*z);
          f_rhs_zz = -gama*gama*f_rhs;
          f_rhs_zzz = -gama*gama*f_rhs_z;
          f_rhs_zzzz=  gama*gama*gama*gama*f_rhs;
          f_rhs_xxyy = -beta*beta*f_rhs_xx; 
          f_rhs_xxzz = -gama*gama*f_rhs_xx; 
          f_rhs_yyzz = beta*beta*gama*gama*f_rhs;
          f =  h2*((1. - kn2*h2/20.)*f_rhs + h2/12.*(f_rhs_xx + f_rhs_yy + f_rhs_zz) +
               h2*h2/360.*(f_rhs_xxxx + f_rhs_yyyy + f_rhs_zzzz) + h2*h2/90.*(f_rhs_xxyy +
               f_rhs_xxzz + f_rhs_yyzz) +
               h2*h2/60.*(kx2 * f_rhs_x + ky2 * f_rhs_y + kz2*f_rhs_z));
        }
    }
L40:
    icn = 0;
    sum = 0.;
    k = bindx[*location];
    bindx[k] = *row + 1;
    val[k] =  ac_ip1_j_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - 1;
    val[k] = ac_im1_j_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + (nx + 2);
    val[k] = ac_i_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - (nx + 2);
    val[k] = ac_i_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + (nx + 2) * (ny + 2);
    val[k] = ac_i_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - (nx + 2) * (ny + 2);
    val[k] = ac_i_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    if (*order == 2) {
        goto L50;
    }
    bindx[k] = *row + (nx + 2) + 1;
    val[k] = ac_ip1_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + (nx + 2) - 1;
    val[k] = ac_im1_jp1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - (nx + 2) + 1;
    val[k] = ac_ip1_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - (nx + 2) - 1;
    val[k] = ac_im1_jm1_k;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* ----- k+1 plane: */
    bindx[k] = *row + (nx + 2) * (ny + 2) + 1;
    val[k] = ac_ip1_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + (nx + 2) * (ny + 2) - 1;
    val[k] = ac_im1_j_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + (nx + 2) * (ny + 2) + (nx + 2);
    val[k] = ac_i_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + (nx + 2) * (ny + 2) - (nx + 2);
    val[k] = ac_i_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* -- */
    bindx[k] = *row + (nx + 2) * (ny + 2) + (nx + 2) + 1;
    val[k] = ac_ip1_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + (nx + 2) * (ny + 2) + (nx + 2) - 1;
    val[k] = ac_im1_jp1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + (nx + 2) * (ny + 2) - (nx + 2) + 1;
    val[k] = ac_ip1_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row + (nx + 2) * (ny + 2) - (nx + 2) - 1;
    val[k] = ac_im1_jm1_kp1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* ----- k-1 plane: */
    bindx[k] = *row - (nx + 2) * (ny + 2) + 1;
    val[k] = ac_ip1_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - (nx + 2) * (ny + 2) - 1;
    val[k] = ac_im1_j_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - (nx + 2) * (ny + 2) + (nx + 2);
    val[k] = ac_i_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - (nx + 2) * (ny + 2) - (nx + 2);
    val[k] = ac_i_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
/* -- */
    bindx[k] = *row - (nx + 2) * (ny + 2) + (nx + 2) + 1;
    val[k] = ac_ip1_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - (nx + 2) * (ny + 2) + (nx + 2) - 1;
    val[k] = ac_im1_jp1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - (nx + 2) * (ny + 2) - (nx + 2) + 1;
    val[k] = ac_ip1_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
    bindx[k] = *row - (nx + 2) * (ny + 2) - (nx + 2) - 1;
    val[k] = ac_im1_jm1_km1;
    d1 = val[k];
    sum += d1 * d1;
    ++icn;
    ++k;
L50:
    *kmax = k;
    bindx[*location + 1] = k;
    val[*location] = ac_i_j_k;
L100:
    rhs[*row] = f;
/*  return */
/*  normalize the equations: */
    d1 = val[*location];
    sum += d1 * d1;
    sum1 = sqrt(sum);
    rhs[*row] /= sum1;
    val[*location] /= sum1;
    i1 = icn;
    for (ii = 1; ii <= i1; ++ii) {
        val[*kmax - ii] /= sum1;
    }
/*  printf("\n%7d %13.10f %13.10f", *row, rhs[*row], val[*location]);
     printf("\n    ");
    for (ii=1; ii<=i1; ++ii) printf(" i=%d  %d", ii,bindx[*kmax - ii]) ;     
    for (ii=1; ii<=i1; ++ii) printf(" %13.10f", val[*kmax - ii]) ;   */
    return 0;
} /* create_matrix_row_7_27pt_example32_case_c */

/* ************************************************************************ */
/* ******************************************************************** */

/* Subroutine */ int transf_3d(double *q, double *ecoef, int *
	update_inv, int *n_update, int *comm_cart, int *me, 
	int *rank_source_x, int *rank_source_y, int *
	rank_source_z, int *rank_dest_x, int *rank_dest_y, 
	int *rank_dest_z, int *index_s_xm, int *index_s_xp, 
	int *index_s_ym, int *index_s_yp, int *index_s_zm, 
	int *index_s_zp, int *index_r_xm, int *index_r_xp, 
	int *index_r_ym, int *index_r_yp, int *index_r_zm, 
	int *index_r_zp, int *ixp, int *ixm, int *iyp, 
	int *iym, int *izp, int *izm, int *ntype)
{
    int i1;
    static int tagx_min, tagy_min, tagz_min, i, tagx_plus, 
	    tagy_plus, tagz_plus;
    MPI_Status status;

    /* Parameter adjustments */
    --index_r_zp;
    --index_r_zm;
    --index_r_yp;
    --index_r_ym;
    --index_r_xp;
    --index_r_xm;
    --index_s_zp;
    --index_s_zm;
    --index_s_yp;
    --index_s_ym;
    --index_s_xp;
    --index_s_xm;   

    /* Function Body */
    tagx_min = 10;
    tagx_plus = 20;
    tagy_min = 30;
    tagy_plus = 40;
    tagz_min = 50;
    tagz_plus = 60;

/*   Define the vectors to be transfered to the neighboring processors */
    i1 = *ixm;
    for (i = 1; i <= i1; ++i) {
	buf_s_xm[i - 1] = q[index_s_xm[i] - 1];
    }

    i1 = *ixp;
    for (i = 1; i <= i1; ++i) {
	buf_s_xp[i - 1] = q[index_s_xp[i] - 1];
    }

    if (*ntype == 1) {
	goto L10;
    }
    i1 = *iym;
    for (i = 1; i <= i1; ++i) {
	buf_s_ym[i - 1] = q[index_s_ym[i] - 1];
    }

    i1 = *iyp;
    for (i = 1; i <= i1; ++i) {
	buf_s_yp[i - 1] = q[index_s_yp[i] - 1];
    }

    if (*ntype == 2) {
	goto L10;
    }
    i1 = *izm;
    for (i = 1; i <= i1; ++i) {
	buf_s_zm[i - 1] = q[index_s_zm[i] - 1];
    }
/*  if(*me == 0) printf("\n **** buffer to be send ***");  */
    i1 = *izp;
    for (i = 1; i <= i1; ++i) {
	buf_s_zp[i - 1] = q[index_s_zp[i] - 1];
/*     if(*me ==0) printf("\n%8d  %13.7e  %8d", i,  buf_s_zp[i - 1], index_s_zp[i] - 1 );  */
    }
L10:

/* send the correction to the neighbors */

    if (*rank_source_x > -1) {
	MPI_Sendrecv(buf_s_xm,*ixm,MPI_DOUBLE_PRECISION,*rank_source_x,tagx_min, 
		buf_r_xm,*ixm,MPI_DOUBLE_PRECISION,*rank_source_x,tagx_plus, 
		*comm_cart,&status );
    }

    if (*rank_dest_x > -1) {
	MPI_Sendrecv(buf_s_xp,*ixp,MPI_DOUBLE_PRECISION,*rank_dest_x,tagx_plus, 
		buf_r_xp,*ixp,MPI_DOUBLE_PRECISION,*rank_dest_x,tagx_min, 
		*comm_cart,&status );
    }

    if (*ntype == 1) {
	goto L20;
    }
    if (*rank_source_y > -1) {
	MPI_Sendrecv(buf_s_ym,*iym,MPI_DOUBLE_PRECISION,*rank_source_y,tagy_min, 
		buf_r_ym,*iym,MPI_DOUBLE_PRECISION,*rank_source_y,tagy_plus, 
		*comm_cart,&status );
    }

    if (*rank_dest_y > -1) {
	MPI_Sendrecv(buf_s_yp,*iyp,MPI_DOUBLE_PRECISION,*rank_dest_y,tagy_plus, 
		buf_r_yp,*iyp,MPI_DOUBLE_PRECISION,*rank_dest_y,tagy_min, 
		*comm_cart,&status );
    }

    if (*ntype == 2) {
	goto L20;
    }
    if (*rank_source_z > -1) {
	MPI_Sendrecv(buf_s_zm,*izm,MPI_DOUBLE_PRECISION,*rank_source_z,tagz_min, 
		buf_r_zm,*izm,MPI_DOUBLE_PRECISION,*rank_source_z,tagz_plus, 
		*comm_cart,&status );
    }

    if (*rank_dest_z > -1) {
	MPI_Sendrecv(buf_s_zp,*izp,MPI_DOUBLE_PRECISION,*rank_dest_z,tagz_plus, 
		buf_r_zp,*izp,MPI_DOUBLE_PRECISION,*rank_dest_z,tagz_min, 
		*comm_cart,&status );
    }
L20:

/*  Update qnewl() of the processor's inner points due to the   
    neighboring processors */

    i1 = *ixm;
    for (i = 1; i <= i1; ++i) {
	q[index_r_xm[i] - 1] += buf_r_xm[i - 1];
    }

    i1 = *ixp;
    for (i = 1; i <= i1; ++i) {
	q[index_r_xp[i] - 1] += buf_r_xp[i - 1];
    }

    if (*ntype == 1) {
	goto L30;
    }
    i1 = *iym;
    for (i = 1; i <= i1; ++i) {
	q[index_r_ym[i] - 1] += buf_r_ym[i - 1];
    }

    i1 = *iyp;
    for (i = 1; i <= i1; ++i) {
	q[index_r_yp[i] - 1] += buf_r_yp[i - 1];
    }

    if (*ntype == 2) {
	goto L30;
    }
    i1 = *izm;
    for (i = 1; i <= i1; ++i) {
	q[index_r_zm[i] - 1] += buf_r_zm[i - 1];
    }

/*  if(*me ==0) printf("\n ******* received buffer ***");  */
    i1 = *izp;
    for (i = 1; i <= i1; ++i) {
	q[index_r_zp[i] - 1] += buf_r_zp[i - 1];
/*     if(*me ==0) printf("\n%8d  %13.7e %8d", i,  buf_r_zp[i - 1], index_r_zp[i] - 1 );  */
    }
L30:

/*  Get  the new values of the inner boundary points: */

    i1 = *n_update - 1;
    for (i = 0; i <= i1; ++i) {
	q[update_inv[i]] *= ecoef[update_inv[i]];
    }

    i1 = *ixm;
    for (i = 1; i <= i1; ++i) {
	buf_s_xm[i - 1] = q[index_r_xm[i] - 1];
    }
    i1 = *ixp;
    for (i = 1; i <= i1; ++i) {
	buf_s_xp[i - 1] = q[index_r_xp[i] - 1];
    }
    if (*ntype == 1) {
	goto L40;
    }
    i1 = *iym;
    for (i = 1; i <= i1; ++i) {
	buf_s_ym[i - 1] = q[index_r_ym[i] - 1];
    }
    i1 = *iyp;
    for (i = 1; i <= i1; ++i) {
	buf_s_yp[i - 1] = q[index_r_yp[i] - 1];
    }
    if (*ntype == 2) {
	goto L40;
    }
    i1 = *izm;
    for (i = 1; i <= i1; ++i) {
	buf_s_zm[i - 1] = q[index_r_zm[i] - 1];
    }
    i1 = *izp;
 /* if(*me ==0) printf("\n *******  buffer  to be sent ***");  */
    for (i = 1; i <= i1; ++i) {
	buf_s_zp[i - 1] = q[index_r_zp[i] - 1];
/*     if(*me ==0) printf("\n%8d  %13.7e", i,  buf_s_zp[i - 1] );  */
    }
L40:

/* Send the new values of XNEW of the inner points to the neighbors */

    if (*rank_source_x > -1) {
	MPI_Sendrecv(buf_s_xm,*ixm,MPI_DOUBLE_PRECISION,*rank_source_x,tagx_min, 
		buf_r_xm,*ixm,MPI_DOUBLE_PRECISION,*rank_source_x,tagx_plus, 
		*comm_cart,&status );
    }

    if (*rank_dest_x > -1) {
	MPI_Sendrecv(buf_s_xp,*ixp,MPI_DOUBLE_PRECISION,*rank_dest_x,tagx_plus, 
		buf_r_xp,*ixp,MPI_DOUBLE_PRECISION,*rank_dest_x,tagx_min, 
		*comm_cart,&status );
    }

    if (*ntype == 1) {
	goto L50;
    }
    if (*rank_source_y > -1) {
	MPI_Sendrecv(buf_s_ym,*iym,MPI_DOUBLE_PRECISION,*rank_source_y,tagy_min, 
		buf_r_ym,*iym,MPI_DOUBLE_PRECISION,*rank_source_y,tagy_plus, 
		*comm_cart,&status );
    }

    if (*rank_dest_y > -1) {
	MPI_Sendrecv(buf_s_yp,*iyp,MPI_DOUBLE_PRECISION,*rank_dest_y,tagy_plus, 
		buf_r_yp,*iyp,MPI_DOUBLE_PRECISION,*rank_dest_y,tagy_min, 
		*comm_cart,&status );
    }

    if (*ntype == 2) {
	goto L50;
    }
    if (*rank_source_z > -1) {
	MPI_Sendrecv(buf_s_zm,*izm,MPI_DOUBLE_PRECISION,*rank_source_z,tagz_min, 
		buf_r_zm,*izm,MPI_DOUBLE_PRECISION,*rank_source_z,tagz_plus, 
		*comm_cart,&status );
    }

    if (*rank_dest_z > -1) {
	MPI_Sendrecv(buf_s_zp,*izp,MPI_DOUBLE_PRECISION,*rank_dest_z,tagz_plus, 
		buf_r_zp,*izp,MPI_DOUBLE_PRECISION,*rank_dest_z,tagz_min, 
		*comm_cart,&status );
    }
L50:

/*  Update the external points */

    i1 = *ixm;
    for (i = 1; i <= i1; ++i) {
	q[index_s_xm[i] - 1] = buf_r_xm[i - 1];
    }
    i1 = *ixp;
    for (i = 1; i <= i1; ++i) {
	q[index_s_xp[i] - 1] = buf_r_xp[i - 1];
    }
    if (*ntype == 1) {
	goto L60;
    }
    i1 = *iym;
    for (i = 1; i <= i1; ++i) {
	q[index_s_ym[i] - 1] = buf_r_ym[i - 1];
    }
    i1 = *iyp;
    for (i = 1; i <= i1; ++i) {
	q[index_s_yp[i] - 1] = buf_r_yp[i - 1];
    }
    if (*ntype == 2) {
	goto L60;
    }
    i1 = *izm;
    for (i = 1; i <= i1; ++i) {
	q[index_s_zm[i] - 1] = buf_r_zm[i - 1];
    }
    i1 = *izp;
 /* if(*me ==0) printf("\n *******  new received  buffer ***"); */
    for (i = 1; i <= i1; ++i) {
	q[index_s_zp[i] - 1] = buf_r_zp[i - 1];
/*     if(*me ==0) printf("\n%8d  %13.7e  %8d", i,  buf_r_zp[i - 1], index_s_zp[i] - 1 ); */
    }
L60:
    return 0;
} /* transf_3d */

/* ******************************************************************** */
/* ************************************************************************ */

/* Subroutine */ int transf_2d(double *q, double *ecoef, int *
	update_inv, int *n_update, int *comm_cart, int *me, 
	int *rank_source_x, int *rank_source_y, int *
	rank_dest_x, int *rank_dest_y, int *index_s_xm, int 
	*index_s_xp, int *index_s_ym, int *index_s_yp, int *
	index_r_xm, int *index_r_xp, int *index_r_ym, int *
	index_r_yp, int *ixp, int *ixm, int *iyp, int *iym) 
{
    int i1;
    static int tagx_min, tagy_min, i, tagx_plus, tagy_plus;
    MPI_Status status;

    /* Parameter adjustments */
    --index_r_yp;
    --index_r_ym;
    --index_r_xp;
    --index_r_xm;
    --index_s_yp;
    --index_s_ym;
    --index_s_xp;
    --index_s_xm;

    /* Function Body */
    tagx_min = 10;
    tagx_plus = 20;
    tagy_min = 30;
    tagy_plus = 40;

/*  Define the vectors to be transfered to the neighboring processors */
    i1 = *ixm;
    for (i = 1; i <= i1; ++i) {
	buf_s_xm[i - 1] = q[index_s_xm[i] - 1];
    }

    i1 = *ixp;
    for (i = 1; i <= i1; ++i) {
	buf_s_xp[i - 1] = q[index_s_xp[i] - 1];
    }

    i1 = *iym;
    for (i = 1; i <= i1; ++i) {
	buf_s_ym[i - 1] = q[index_s_ym[i] - 1];
    }

    i1 = *iyp;
    for (i = 1; i <= i1; ++i) {
	buf_s_yp[i - 1] = q[index_s_yp[i] - 1];
    }

/* Send the correction to the neighbors */

    if (*rank_source_x > -1) {
	MPI_Sendrecv(buf_s_xm,*ixm,MPI_DOUBLE_PRECISION,*rank_source_x,tagx_min, 
		buf_r_xm,*ixm,MPI_DOUBLE_PRECISION,*rank_source_x,tagx_plus, 
		*comm_cart,&status );
    }

    if (*rank_dest_x > -1) {
	MPI_Sendrecv(buf_s_xp,*ixp,MPI_DOUBLE_PRECISION,*rank_dest_x,tagx_plus, 
		buf_r_xp,*ixp,MPI_DOUBLE_PRECISION,*rank_dest_x,tagx_min, 
		*comm_cart,&status );
    }

    if (*rank_source_y > -1) {
	MPI_Sendrecv(buf_s_ym,*iym,MPI_DOUBLE_PRECISION,*rank_source_y,tagy_min, 
		buf_r_ym,*iym,MPI_DOUBLE_PRECISION,*rank_source_y,tagy_plus, 
		*comm_cart,&status );
    }

    if (*rank_dest_y > -1) {
	MPI_Sendrecv(buf_s_yp,*iyp,MPI_DOUBLE_PRECISION,*rank_dest_y,tagy_plus, 
		buf_r_yp,*iyp,MPI_DOUBLE_PRECISION,*rank_dest_y,tagy_min, 
		*comm_cart,&status );
    }

/*  Update qnewl() of the processor's inner points due to the 
    neighboring processors */

    i1 = *ixm;
    for (i = 1; i <= i1; ++i) {
	q[index_r_xm[i] - 1] += buf_r_xm[i - 1];
    }

    i1 = *ixp;
    for (i = 1; i <= i1; ++i) {
	q[index_r_xp[i] - 1] += buf_r_xp[i - 1];
    }

    i1 = *iym;
    for (i = 1; i <= i1; ++i) {
	q[index_r_ym[i] - 1] += buf_r_ym[i - 1];
    }

    i1 = *iyp;
    for (i = 1; i <= i1; ++i) {
	q[index_r_yp[i] - 1] += buf_r_yp[i - 1];
    }

/*  Get  the new values of the inner boundary points: */

    i1 = *n_update - 1;
    for (i = 0; i <= i1; ++i) {
	q[update_inv[i]] *= ecoef[update_inv[i]];
    }

    i1 = *ixm;
    for (i = 1; i <= i1; ++i) {
	buf_s_xm[i - 1] = q[index_r_xm[i] - 1];
    }
    i1 = *ixp;
    for (i = 1; i <= i1; ++i) {
	buf_s_xp[i - 1] = q[index_r_xp[i] - 1];
    }
    i1 = *iym;
    for (i = 1; i <= i1; ++i) {
	buf_s_ym[i - 1] = q[index_r_ym[i] - 1];
    }
    i1 = *iyp;
    for (i = 1; i <= i1; ++i) {
	buf_s_yp[i - 1] = q[index_r_yp[i] - 1];
    }

    if (*rank_source_x > -1) {
	MPI_Sendrecv(buf_s_xm,*ixm,MPI_DOUBLE_PRECISION,*rank_source_x,tagx_min, 
		buf_r_xm,*ixm,MPI_DOUBLE_PRECISION,*rank_source_x,tagx_plus, 
		*comm_cart,&status );
    }

    if (*rank_dest_x > -1) {
	MPI_Sendrecv(buf_s_xp,*ixp,MPI_DOUBLE_PRECISION,*rank_dest_x,tagx_plus, 
		buf_r_xp,*ixp,MPI_DOUBLE_PRECISION,*rank_dest_x,tagx_min, 
		*comm_cart,&status );
    }

    if (*rank_source_y > -1) {
	MPI_Sendrecv(buf_s_ym,*iym,MPI_DOUBLE_PRECISION,*rank_source_y,tagy_min, 
		buf_r_ym,*iym,MPI_DOUBLE_PRECISION,*rank_source_y,tagy_plus, 
		*comm_cart,&status );
    }

    if (*rank_dest_y > -1) {
	MPI_Sendrecv(buf_s_yp,*iyp,MPI_DOUBLE_PRECISION,*rank_dest_y,tagy_plus, 
		buf_r_yp,*iyp,MPI_DOUBLE_PRECISION,*rank_dest_y,tagy_min, 
		*comm_cart,&status );
    }

/*  Update the external points */

    i1 = *ixm;
    for (i = 1; i <= i1; ++i) {
	q[index_s_xm[i] - 1] = buf_r_xm[i - 1];
    }
    i1 = *ixp;
    for (i = 1; i <= i1; ++i) {
	q[index_s_xp[i] - 1] = buf_r_xp[i - 1];
    }
    i1 = *iym;
    for (i = 1; i <= i1; ++i) {
	q[index_s_ym[i] - 1] = buf_r_ym[i - 1];
    }
    i1 = *iyp;
    for (i = 1; i <= i1; ++i) {
	q[index_s_yp[i] - 1] = buf_r_yp[i - 1];
    }
    return 0;
} /* transf_2d */

/* ******************************************************************** */
/* ************************************************************************ */
